"use strict";
/**
 * Copyright (c) 2018-2019 mol* contributors, licensed under MIT, See LICENSE file for more info.
 *
 * @author David Sehnal <david.sehnal@gmail.com>
 * @author Alexander Rose <alexander.rose@weirdbyte.de>
 */
Object.defineProperty(exports, "__esModule", { value: true });
var linear_algebra_1 = require("../mol-math/linear-algebra");
var util_1 = require("./camera/util");
var rxjs_1 = require("rxjs");
var transition_1 = require("./camera/transition");
// TODO: slab controls that modify near/far planes?
var Camera = /** @class */ (function () {
    function Camera(state, viewport) {
        if (viewport === void 0) { viewport = util_1.Viewport.create(-1, -1, 1, 1); }
        this.updatedViewProjection = new rxjs_1.BehaviorSubject(this);
        this.view = linear_algebra_1.Mat4.identity();
        this.projection = linear_algebra_1.Mat4.identity();
        this.projectionView = linear_algebra_1.Mat4.identity();
        this.inverseProjectionView = linear_algebra_1.Mat4.identity();
        this.state = Camera.createDefaultSnapshot();
        this.viewOffset = {
            enabled: false,
            fullWidth: 1, fullHeight: 1,
            offsetX: 0, offsetY: 0,
            width: 1, height: 1
        };
        this.transition = new transition_1.CameraTransitionManager(this);
        this.prevProjection = linear_algebra_1.Mat4.identity();
        this.prevView = linear_algebra_1.Mat4.identity();
        this.deltaDirection = linear_algebra_1.Vec3.zero();
        this.newPosition = linear_algebra_1.Vec3.zero();
        this.viewport = viewport;
        Camera.copySnapshot(this.state, state);
    }
    Object.defineProperty(Camera.prototype, "position", {
        get: function () { return this.state.position; },
        set: function (v) { linear_algebra_1.Vec3.copy(this.state.position, v); },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(Camera.prototype, "direction", {
        get: function () { return this.state.direction; },
        set: function (v) { linear_algebra_1.Vec3.copy(this.state.direction, v); },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(Camera.prototype, "up", {
        get: function () { return this.state.up; },
        set: function (v) { linear_algebra_1.Vec3.copy(this.state.up, v); },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(Camera.prototype, "target", {
        get: function () { return this.state.target; },
        set: function (v) { linear_algebra_1.Vec3.copy(this.state.target, v); },
        enumerable: true,
        configurable: true
    });
    Camera.prototype.updateMatrices = function () {
        var snapshot = this.state;
        var height = 2 * Math.tan(snapshot.fov / 2) * linear_algebra_1.Vec3.distance(snapshot.position, snapshot.target);
        snapshot.zoom = this.viewport.height / height;
        switch (this.state.mode) {
            case 'orthographic':
                updateOrtho(this);
                break;
            case 'perspective':
                updatePers(this);
                break;
            default: throw new Error('unknown camera mode');
        }
        var changed = !linear_algebra_1.Mat4.areEqual(this.projection, this.prevProjection, 0.000001 /* Value */) || !linear_algebra_1.Mat4.areEqual(this.view, this.prevView, 0.000001 /* Value */);
        linear_algebra_1.Mat4.mul(this.projectionView, this.projection, this.view);
        linear_algebra_1.Mat4.invert(this.inverseProjectionView, this.projectionView);
        if (changed) {
            linear_algebra_1.Mat4.mul(this.projectionView, this.projection, this.view);
            linear_algebra_1.Mat4.invert(this.inverseProjectionView, this.projectionView);
            linear_algebra_1.Mat4.copy(this.prevView, this.view);
            linear_algebra_1.Mat4.copy(this.prevProjection, this.projection);
            this.updatedViewProjection.next(this);
        }
        return changed;
    };
    Camera.prototype.setState = function (snapshot) {
        this.transition.apply(snapshot);
    };
    Camera.prototype.getSnapshot = function () {
        var ret = Camera.createDefaultSnapshot();
        Camera.copySnapshot(ret, this.state);
        return ret;
    };
    Camera.prototype.getFocus = function (target, radius) {
        var fov = this.state.fov;
        var _a = this.viewport, width = _a.width, height = _a.height;
        var aspect = width / height;
        var aspectFactor = (height < width ? 1 : aspect);
        var currentDistance = linear_algebra_1.Vec3.distance(this.state.position, target);
        var targetDistance = Math.abs((radius / aspectFactor) / Math.sin(fov / 2));
        var deltaDistance = Math.abs(currentDistance - targetDistance);
        linear_algebra_1.Vec3.sub(this.deltaDirection, this.state.position, target);
        linear_algebra_1.Vec3.setMagnitude(this.deltaDirection, this.state.direction, deltaDistance);
        if (currentDistance < targetDistance)
            linear_algebra_1.Vec3.negate(this.deltaDirection, this.deltaDirection);
        linear_algebra_1.Vec3.add(this.newPosition, this.state.position, this.deltaDirection);
        return { target: target, position: linear_algebra_1.Vec3.clone(this.newPosition) };
    };
    Camera.prototype.focus = function (target, radius) {
        this.setState(this.getFocus(target, radius));
    };
    // lookAt(target: Vec3) {
    //     cameraLookAt(this.position, this.up, this.direction, target);
    // }
    // translate(v: Vec3) {
    //     Vec3.add(this.position, this.position, v);
    //     cameraLookAt(this.position, this.up, this.direction, this.target);
    // }
    Camera.prototype.project = function (out, point) {
        return util_1.cameraProject(out, point, this.viewport, this.projectionView);
    };
    Camera.prototype.unproject = function (out, point) {
        return util_1.cameraUnproject(out, point, this.viewport, this.inverseProjectionView);
    };
    Camera.prototype.dispose = function () {
        this.updatedViewProjection.complete();
    };
    return Camera;
}());
exports.Camera = Camera;
(function (Camera) {
    function setViewOffset(out, fullWidth, fullHeight, offsetX, offsetY, width, height) {
        out.fullWidth = fullWidth;
        out.fullHeight = fullHeight;
        out.offsetX = offsetX;
        out.offsetY = offsetY;
        out.width = width;
        out.height = height;
    }
    Camera.setViewOffset = setViewOffset;
    function createDefaultSnapshot() {
        return {
            mode: 'perspective',
            position: linear_algebra_1.Vec3.zero(),
            direction: linear_algebra_1.Vec3.create(0, 0, -1),
            up: linear_algebra_1.Vec3.create(0, 1, 0),
            target: linear_algebra_1.Vec3.create(0, 0, 0),
            near: 1,
            far: 10000,
            fogNear: 1,
            fogFar: 10000,
            fov: Math.PI / 4,
            zoom: 1,
        };
    }
    Camera.createDefaultSnapshot = createDefaultSnapshot;
    function copySnapshot(out, source) {
        if (!source)
            return;
        if (typeof source.mode !== 'undefined')
            out.mode = source.mode;
        if (typeof source.position !== 'undefined')
            linear_algebra_1.Vec3.copy(out.position, source.position);
        if (typeof source.direction !== 'undefined')
            linear_algebra_1.Vec3.copy(out.direction, source.direction);
        if (typeof source.up !== 'undefined')
            linear_algebra_1.Vec3.copy(out.up, source.up);
        if (typeof source.target !== 'undefined')
            linear_algebra_1.Vec3.copy(out.target, source.target);
        if (typeof source.near !== 'undefined')
            out.near = source.near;
        if (typeof source.far !== 'undefined')
            out.far = source.far;
        if (typeof source.fogNear !== 'undefined')
            out.fogNear = source.fogNear;
        if (typeof source.fogFar !== 'undefined')
            out.fogFar = source.fogFar;
        if (typeof source.fov !== 'undefined')
            out.fov = source.fov;
        if (typeof source.zoom !== 'undefined')
            out.zoom = source.zoom;
    }
    Camera.copySnapshot = copySnapshot;
})(Camera || (Camera = {}));
exports.Camera = Camera;
var _center = linear_algebra_1.Vec3.zero();
function updateOrtho(camera) {
    var viewport = camera.viewport, _a = camera.state, zoom = _a.zoom, near = _a.near, far = _a.far, viewOffset = camera.viewOffset;
    var fullLeft = -(viewport.width - viewport.x) / 2;
    var fullRight = (viewport.width - viewport.x) / 2;
    var fullTop = (viewport.height - viewport.y) / 2;
    var fullBottom = -(viewport.height - viewport.y) / 2;
    var dx = (fullRight - fullLeft) / (2 * zoom);
    var dy = (fullTop - fullBottom) / (2 * zoom);
    var cx = (fullRight + fullLeft) / 2;
    var cy = (fullTop + fullBottom) / 2;
    var left = cx - dx;
    var right = cx + dx;
    var top = cy + dy;
    var bottom = cy - dy;
    if (viewOffset.enabled) {
        var zoomW = zoom / (viewOffset.width / viewOffset.fullWidth);
        var zoomH = zoom / (viewOffset.height / viewOffset.fullHeight);
        var scaleW = (fullRight - fullLeft) / viewOffset.width;
        var scaleH = (fullTop - fullBottom) / viewOffset.height;
        left += scaleW * (viewOffset.offsetX / zoomW);
        right = left + scaleW * (viewOffset.width / zoomW);
        top -= scaleH * (viewOffset.offsetY / zoomH);
        bottom = top - scaleH * (viewOffset.height / zoomH);
    }
    // build projection matrix
    linear_algebra_1.Mat4.ortho(camera.projection, left, right, top, bottom, near, far);
    // build view matrix
    linear_algebra_1.Vec3.add(_center, camera.position, camera.direction);
    linear_algebra_1.Mat4.lookAt(camera.view, camera.position, _center, camera.up);
}
function updatePers(camera) {
    var aspect = camera.viewport.width / camera.viewport.height;
    var _a = camera.state, fov = _a.fov, near = _a.near, far = _a.far, viewOffset = camera.viewOffset;
    var top = near * Math.tan(0.5 * fov);
    var height = 2 * top;
    var width = aspect * height;
    var left = -0.5 * width;
    if (viewOffset.enabled) {
        left += viewOffset.offsetX * width / viewOffset.fullWidth;
        top -= viewOffset.offsetY * height / viewOffset.fullHeight;
        width *= viewOffset.width / viewOffset.fullWidth;
        height *= viewOffset.height / viewOffset.fullHeight;
    }
    // build projection matrix
    linear_algebra_1.Mat4.perspective(camera.projection, left, left + width, top, top - height, near, far);
    // build view matrix
    linear_algebra_1.Vec3.add(_center, camera.position, camera.direction);
    linear_algebra_1.Mat4.lookAt(camera.view, camera.position, _center, camera.up);
}
