"use strict";
/**
 * Copyright (c) 2018 Mol* contributors, licensed under MIT, See LICENSE file for more info.
 *
 * @author David Sehnal <david.sehnal@gmail.com>
 */
Object.defineProperty(exports, "__esModule", { value: true });
var camera_1 = require("../camera");
var linear_algebra_1 = require("../../mol-math/linear-algebra");
var interpolate_1 = require("../../mol-math/interpolate");
var CameraTransitionManager = /** @class */ (function () {
    function CameraTransitionManager(camera) {
        this.camera = camera;
        this.t = 0;
        this.func = CameraTransitionManager.defaultTransition;
        this.start = 0;
        this.inTransition = false;
        this.durationMs = 0;
        this.source = camera_1.Camera.createDefaultSnapshot();
        this.target = camera_1.Camera.createDefaultSnapshot();
        this.current = camera_1.Camera.createDefaultSnapshot();
    }
    CameraTransitionManager.prototype.apply = function (to, durationMs, transition) {
        if (durationMs === void 0) { durationMs = 0; }
        if (durationMs <= 0 || to.mode !== this.camera.state.mode) {
            this.finish(to);
            return;
        }
        camera_1.Camera.copySnapshot(this.source, this.camera.state);
        camera_1.Camera.copySnapshot(this.target, this.camera.state);
        camera_1.Camera.copySnapshot(this.target, to);
        this.inTransition = true;
        this.func = transition || CameraTransitionManager.defaultTransition;
        this.start = this.t;
        this.durationMs = durationMs;
    };
    CameraTransitionManager.prototype.tick = function (t) {
        this.t = t;
        this.update();
    };
    CameraTransitionManager.prototype.finish = function (to) {
        camera_1.Camera.copySnapshot(this.camera.state, to);
        this.inTransition = false;
    };
    CameraTransitionManager.prototype.update = function () {
        if (!this.inTransition)
            return;
        var normalized = Math.min((this.t - this.start) / this.durationMs, 1);
        if (normalized === 1) {
            this.finish(this.target);
            return;
        }
        this.func(this.current, normalized, this.source, this.target);
        camera_1.Camera.copySnapshot(this.camera.state, this.current);
    };
    return CameraTransitionManager;
}());
exports.CameraTransitionManager = CameraTransitionManager;
(function (CameraTransitionManager) {
    function defaultTransition(out, t, source, target) {
        camera_1.Camera.copySnapshot(out, target);
        // Rotate direction
        var rot = linear_algebra_1.Quat.identity();
        linear_algebra_1.Quat.slerp(rot, rot, linear_algebra_1.Quat.rotationTo(linear_algebra_1.Quat.zero(), source.direction, target.direction), t);
        linear_algebra_1.Vec3.transformQuat(out.direction, source.direction, rot);
        // Rotate up
        linear_algebra_1.Quat.setIdentity(rot);
        linear_algebra_1.Quat.slerp(rot, rot, linear_algebra_1.Quat.rotationTo(linear_algebra_1.Quat.zero(), source.up, target.up), t);
        linear_algebra_1.Vec3.transformQuat(out.up, source.up, rot);
        // Lerp target
        linear_algebra_1.Vec3.lerp(out.target, source.target, target.target, t);
        // Update position
        var dist = -interpolate_1.lerp(linear_algebra_1.Vec3.distance(source.position, source.target), linear_algebra_1.Vec3.distance(target.position, target.target), t);
        linear_algebra_1.Vec3.scale(out.position, out.direction, dist);
        linear_algebra_1.Vec3.add(out.position, out.position, out.target);
        // Lerp other props
        out.zoom = interpolate_1.lerp(source.zoom, target.zoom, t);
        out.fov = interpolate_1.lerp(source.fov, target.fov, t);
        out.near = interpolate_1.lerp(source.near, target.near, t);
        out.far = interpolate_1.lerp(source.far, target.far, t);
        out.fogNear = interpolate_1.lerp(source.fogNear, target.fogNear, t);
        out.fogFar = interpolate_1.lerp(source.fogFar, target.fogFar, t);
    }
    CameraTransitionManager.defaultTransition = defaultTransition;
})(CameraTransitionManager || (CameraTransitionManager = {}));
exports.CameraTransitionManager = CameraTransitionManager;
