"use strict";
/**
 * Copyright (c) 2018-2019 mol* contributors, licensed under MIT, See LICENSE file for more info.
 *
 * @author Alexander Rose <alexander.rose@weirdbyte.de>
 */
var __assign = (this && this.__assign) || function () {
    __assign = Object.assign || function(t) {
        for (var s, i = 1, n = arguments.length; i < n; i++) {
            s = arguments[i];
            for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p))
                t[p] = s[p];
        }
        return t;
    };
    return __assign.apply(this, arguments);
};
Object.defineProperty(exports, "__esModule", { value: true });
var linear_algebra_1 = require("../../mol-math/linear-algebra");
function Viewport() {
    return Viewport.zero();
}
exports.Viewport = Viewport;
(function (Viewport) {
    function zero() {
        return { x: 0, y: 0, width: 0, height: 0 };
    }
    Viewport.zero = zero;
    function create(x, y, width, height) {
        return { x: x, y: y, width: width, height: height };
    }
    Viewport.create = create;
    function clone(viewport) {
        return __assign({}, viewport);
    }
    Viewport.clone = clone;
    function copy(target, source) {
        return Object.assign(target, source);
    }
    Viewport.copy = copy;
    function set(viewport, x, y, width, height) {
        viewport.x = x;
        viewport.y = y;
        viewport.width = width;
        viewport.height = height;
        return viewport;
    }
    Viewport.set = set;
    function toVec4(v4, viewport) {
        v4[0] = viewport.x;
        v4[1] = viewport.y;
        v4[2] = viewport.width;
        v4[3] = viewport.height;
        return v4;
    }
    Viewport.toVec4 = toVec4;
    function equals(a, b) {
        return a.x === b.x && a.y === b.y && a.width === b.width && a.height === b.height;
    }
    Viewport.equals = equals;
})(Viewport || (Viewport = {}));
exports.Viewport = Viewport;
//
var tmpVec3 = linear_algebra_1.Vec3();
/** Modifies the direction & up vectors in place, both are normalized */
function cameraLookAt(position, up, direction, target) {
    linear_algebra_1.Vec3.sub(tmpVec3, target, position);
    linear_algebra_1.Vec3.normalize(tmpVec3, tmpVec3);
    if (!linear_algebra_1.Vec3.isZero(tmpVec3)) {
        // change direction vector to look at target
        var d = linear_algebra_1.Vec3.dot(tmpVec3, up);
        if (Math.abs(d - 1) < 0.000001 /* Value */) { // parallel
            linear_algebra_1.Vec3.scale(up, direction, -1);
        }
        else if (Math.abs(d + 1) < 0.000001 /* Value */) { // anti parallel
            linear_algebra_1.Vec3.copy(up, direction);
        }
        linear_algebra_1.Vec3.copy(direction, tmpVec3);
        // normalize up vector
        linear_algebra_1.Vec3.cross(tmpVec3, direction, up);
        linear_algebra_1.Vec3.normalize(tmpVec3, tmpVec3);
        linear_algebra_1.Vec3.cross(up, tmpVec3, direction);
        linear_algebra_1.Vec3.normalize(up, up);
    }
}
exports.cameraLookAt = cameraLookAt;
var NEAR_RANGE = 0;
var FAR_RANGE = 1;
var tmpVec4 = linear_algebra_1.Vec4();
/** Transform point into 2D window coordinates. */
function cameraProject(out, point, viewport, projectionView) {
    var vX = viewport.x, vY = viewport.y, vWidth = viewport.width, vHeight = viewport.height;
    // clip space -> NDC -> window coordinates, implicit 1.0 for w component
    linear_algebra_1.Vec4.set(tmpVec4, point[0], point[1], point[2], 1.0);
    // transform into clip space
    linear_algebra_1.Vec4.transformMat4(tmpVec4, tmpVec4, projectionView);
    // transform into NDC
    var w = tmpVec4[3];
    if (w !== 0) {
        tmpVec4[0] /= w;
        tmpVec4[1] /= w;
        tmpVec4[2] /= w;
    }
    // transform into window coordinates, set fourth component is (1/clip.w) as in gl_FragCoord.w
    out[0] = vX + vWidth / 2 * tmpVec4[0] + (0 + vWidth / 2);
    out[1] = vY + vHeight / 2 * tmpVec4[1] + (0 + vHeight / 2);
    out[2] = (FAR_RANGE - NEAR_RANGE) / 2 * tmpVec4[2] + (FAR_RANGE + NEAR_RANGE) / 2;
    out[3] = w === 0 ? 0 : 1 / w;
    return out;
}
exports.cameraProject = cameraProject;
/**
 * Transform point from screen space to 3D coordinates.
 * The point must have x and y set to 2D window coordinates and z between 0 (near) and 1 (far).
 */
function cameraUnproject(out, point, viewport, inverseProjectionView) {
    var vX = viewport.x, vY = viewport.y, vWidth = viewport.width, vHeight = viewport.height;
    var x = point[0] - vX;
    var y = (vHeight - point[1] - 1) - vY;
    var z = point[2];
    out[0] = (2 * x) / vWidth - 1;
    out[1] = (2 * y) / vHeight - 1;
    out[2] = 2 * z - 1;
    return linear_algebra_1.Vec3.transformMat4(out, out, inverseProjectionView);
}
exports.cameraUnproject = cameraUnproject;
