"use strict";
/**
 * Copyright (c) 2018-2019 mol* contributors, licensed under MIT, See LICENSE file for more info.
 *
 * @author Alexander Rose <alexander.rose@weirdbyte.de>
 */
var __assign = (this && this.__assign) || function () {
    __assign = Object.assign || function(t) {
        for (var s, i = 1, n = arguments.length; i < n; i++) {
            s = arguments[i];
            for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p))
                t[p] = s[p];
        }
        return t;
    };
    return __assign.apply(this, arguments);
};
Object.defineProperty(exports, "__esModule", { value: true });
var rxjs_1 = require("rxjs");
var now_1 = require("../mol-util/now");
var linear_algebra_1 = require("../mol-math/linear-algebra");
var input_observer_1 = require("../mol-util/input/input-observer");
var renderer_1 = require("../mol-gl/renderer");
var trackball_1 = require("./controls/trackball");
var util_1 = require("./camera/util");
var context_1 = require("../mol-gl/webgl/context");
var representation_1 = require("../mol-repr/representation");
var scene_1 = require("../mol-gl/scene");
var loci_1 = require("../mol-model/loci");
var camera_1 = require("./camera");
var param_definition_1 = require("../mol-util/param-definition");
var bounding_sphere_helper_1 = require("./helper/bounding-sphere-helper");
var set_1 = require("../mol-util/set");
var interaction_events_1 = require("./helper/interaction-events");
var postprocessing_1 = require("./passes/postprocessing");
var multi_sample_1 = require("./passes/multi-sample");
var util_2 = require("../mol-gl/compute/util");
var draw_1 = require("./passes/draw");
var pick_1 = require("./passes/pick");
exports.Canvas3DParams = {
    // TODO: FPS cap?
    // maxFps: PD.Numeric(30),
    cameraMode: param_definition_1.ParamDefinition.Select('perspective', [['perspective', 'Perspective'], ['orthographic', 'Orthographic']]),
    cameraClipDistance: param_definition_1.ParamDefinition.Numeric(0, { min: 0.0, max: 50.0, step: 0.1 }, { description: 'The distance between camera and scene at which to clip regardless of near clipping plane.' }),
    clip: param_definition_1.ParamDefinition.Interval([1, 100], { min: 1, max: 100, step: 1 }),
    fog: param_definition_1.ParamDefinition.Interval([50, 100], { min: 1, max: 100, step: 1 }),
    multiSample: param_definition_1.ParamDefinition.Group(multi_sample_1.MultiSampleParams),
    postprocessing: param_definition_1.ParamDefinition.Group(postprocessing_1.PostprocessingParams),
    renderer: param_definition_1.ParamDefinition.Group(renderer_1.RendererParams),
    trackball: param_definition_1.ParamDefinition.Group(trackball_1.TrackballControlsParams),
    debug: param_definition_1.ParamDefinition.Group(bounding_sphere_helper_1.DebugHelperParams)
};
var requestAnimationFrame = typeof window !== 'undefined' ? window.requestAnimationFrame : function (f) { return setImmediate(function () { return f(Date.now()); }); };
var Canvas3D;
(function (Canvas3D) {
    function fromCanvas(canvas, props) {
        if (props === void 0) { props = {}; }
        var gl = context_1.getGLContext(canvas, {
            alpha: false,
            antialias: true,
            depth: true,
            preserveDrawingBuffer: true
        });
        if (gl === null)
            throw new Error('Could not create a WebGL rendering context');
        var input = input_observer_1.default.fromElement(canvas);
        return Canvas3D.create(gl, input, props);
    }
    Canvas3D.fromCanvas = fromCanvas;
    function create(gl, input, props) {
        if (props === void 0) { props = {}; }
        var p = __assign({}, param_definition_1.ParamDefinition.getDefaultValues(exports.Canvas3DParams), props);
        var reprRenderObjects = new Map();
        var reprUpdatedSubscriptions = new Map();
        var reprCount = new rxjs_1.BehaviorSubject(0);
        var startTime = now_1.now();
        var didDraw = new rxjs_1.BehaviorSubject(0);
        var camera = new camera_1.Camera({
            near: 0.1,
            far: 10000,
            position: linear_algebra_1.Vec3.create(0, 0, 10),
            mode: p.cameraMode
        });
        var webgl = context_1.createContext(gl);
        var width = gl.drawingBufferWidth;
        var height = gl.drawingBufferHeight;
        var scene = scene_1.default.create(webgl);
        var controls = trackball_1.TrackballControls.create(input, camera, p.trackball);
        var renderer = renderer_1.default.create(webgl, camera, p.renderer);
        var debugHelper = new bounding_sphere_helper_1.BoundingSphereHelper(webgl, scene, p.debug);
        var interactionHelper = new interaction_events_1.Canvas3dInteractionHelper(identify, getLoci, input);
        var drawPass = new draw_1.DrawPass(webgl, renderer, scene, debugHelper);
        var pickPass = new pick_1.PickPass(webgl, renderer, scene, 0.5);
        var postprocessing = new postprocessing_1.PostprocessingPass(webgl, camera, drawPass, p.postprocessing);
        var multiSample = new multi_sample_1.MultiSamplePass(webgl, camera, drawPass, postprocessing, p.multiSample);
        var isUpdating = false;
        var drawPending = false;
        function getLoci(pickingId) {
            var loci = loci_1.EmptyLoci;
            var repr = representation_1.Representation.Empty;
            reprRenderObjects.forEach(function (_, _repr) {
                var _loci = _repr.getLoci(pickingId);
                if (!loci_1.isEmptyLoci(_loci)) {
                    if (!loci_1.isEmptyLoci(loci))
                        console.warn('found another loci');
                    loci = _loci;
                    repr = _repr;
                }
            });
            return { loci: loci, repr: repr };
        }
        function mark(reprLoci, action) {
            var repr = reprLoci.repr, loci = reprLoci.loci;
            var changed = false;
            if (repr) {
                changed = repr.mark(loci, action);
            }
            else {
                reprRenderObjects.forEach(function (_, _repr) { changed = _repr.mark(loci, action) || changed; });
            }
            if (changed) {
                scene.update(void 0, true);
                var prevPickDirty = pickPass.pickDirty;
                draw(true);
                pickPass.pickDirty = prevPickDirty; // marking does not change picking buffers
            }
        }
        var currentNear = -1, currentFar = -1, currentFogNear = -1, currentFogFar = -1;
        function setClipping() {
            var cDist = linear_algebra_1.Vec3.distance(camera.state.position, camera.state.target);
            var bRadius = Math.max(10, scene.boundingSphere.radius);
            var nearFactor = (50 - p.clip[0]) / 50;
            var farFactor = -(50 - p.clip[1]) / 50;
            var near = cDist - (bRadius * nearFactor);
            var far = cDist + (bRadius * farFactor);
            var fogNearFactor = (50 - p.fog[0]) / 50;
            var fogFarFactor = -(50 - p.fog[1]) / 50;
            var fogNear = cDist - (bRadius * fogNearFactor);
            var fogFar = cDist + (bRadius * fogFarFactor);
            if (camera.state.mode === 'perspective') {
                near = Math.max(1, p.cameraClipDistance, near);
                far = Math.max(1, far);
                fogNear = Math.max(1, fogNear);
                fogFar = Math.max(1, fogFar);
            }
            else if (camera.state.mode === 'orthographic') {
                if (p.cameraClipDistance > 0) {
                    near = Math.max(p.cameraClipDistance, near);
                }
            }
            if (near !== currentNear || far !== currentFar || fogNear !== currentFogNear || fogFar !== currentFogFar) {
                camera.setState({ near: near, far: far, fogNear: fogNear, fogFar: fogFar });
                currentNear = near, currentFar = far, currentFogNear = fogNear, currentFogFar = fogFar;
            }
        }
        function render(variant, force) {
            if (isUpdating)
                return false;
            var didRender = false;
            controls.update(currentTime);
            // TODO: is this a good fix? Also, setClipping does not work if the user has manually set a clipping plane.
            if (!camera.transition.inTransition)
                setClipping();
            var cameraChanged = camera.updateMatrices();
            multiSample.update(force || cameraChanged, currentTime);
            if (force || cameraChanged || multiSample.enabled) {
                switch (variant) {
                    case 'pick':
                        pickPass.render();
                        break;
                    case 'draw':
                        renderer.setViewport(0, 0, width, height);
                        if (multiSample.enabled) {
                            multiSample.render();
                        }
                        else {
                            drawPass.render(!postprocessing.enabled);
                            if (postprocessing.enabled)
                                postprocessing.render(true);
                        }
                        pickPass.pickDirty = true;
                        break;
                }
                didRender = true;
            }
            return didRender && cameraChanged;
        }
        var forceNextDraw = false;
        var currentTime = 0;
        function draw(force) {
            if (render('draw', !!force || forceNextDraw)) {
                didDraw.next(now_1.now() - startTime);
            }
            forceNextDraw = false;
            drawPending = false;
        }
        function requestDraw(force) {
            if (drawPending)
                return;
            drawPending = true;
            forceNextDraw = !!force;
        }
        function animate() {
            currentTime = now_1.now();
            camera.transition.tick(currentTime);
            draw(false);
            if (!camera.transition.inTransition)
                interactionHelper.tick(currentTime);
            requestAnimationFrame(animate);
        }
        function identify(x, y) {
            return pickPass.identify(x, y);
        }
        function add(repr) {
            isUpdating = true;
            var oldRO = reprRenderObjects.get(repr);
            var newRO = new Set();
            repr.renderObjects.forEach(function (o) { return newRO.add(o); });
            if (oldRO) {
                if (!set_1.SetUtils.areEqual(newRO, oldRO)) {
                    for (var _i = 0, _a = Array.from(newRO); _i < _a.length; _i++) {
                        var o = _a[_i];
                        if (!oldRO.has(o))
                            scene.add(o);
                    }
                    for (var _b = 0, _c = Array.from(oldRO); _b < _c.length; _b++) {
                        var o = _c[_b];
                        if (!newRO.has(o))
                            scene.remove(o);
                    }
                }
            }
            else {
                repr.renderObjects.forEach(function (o) { return scene.add(o); });
            }
            reprRenderObjects.set(repr, newRO);
            scene.update(repr.renderObjects, false);
            if (debugHelper.isEnabled)
                debugHelper.update();
            isUpdating = false;
            requestDraw(true);
            reprCount.next(reprRenderObjects.size);
        }
        handleResize();
        return {
            webgl: webgl,
            add: function (repr) {
                add(repr);
                reprUpdatedSubscriptions.set(repr, repr.updated.subscribe(function (_) {
                    if (!repr.state.syncManually)
                        add(repr);
                }));
            },
            remove: function (repr) {
                var updatedSubscription = reprUpdatedSubscriptions.get(repr);
                if (updatedSubscription) {
                    updatedSubscription.unsubscribe();
                }
                var renderObjects = reprRenderObjects.get(repr);
                if (renderObjects) {
                    isUpdating = true;
                    renderObjects.forEach(function (o) { return scene.remove(o); });
                    reprRenderObjects.delete(repr);
                    scene.update(void 0, false);
                    if (debugHelper.isEnabled)
                        debugHelper.update();
                    isUpdating = false;
                    requestDraw(true);
                    reprCount.next(reprRenderObjects.size);
                }
            },
            update: function (repr, keepSphere) {
                if (repr) {
                    if (!reprRenderObjects.has(repr))
                        return;
                    scene.update(repr.renderObjects, !!keepSphere);
                }
                else {
                    scene.update(void 0, !!keepSphere);
                }
            },
            clear: function () {
                reprRenderObjects.clear();
                scene.clear();
                debugHelper.clear();
            },
            // draw,
            requestDraw: requestDraw,
            animate: animate,
            identify: identify,
            mark: mark,
            getLoci: getLoci,
            handleResize: handleResize,
            resetCamera: function () {
                camera.focus(scene.boundingSphere.center, scene.boundingSphere.radius);
                requestDraw(true);
            },
            camera: camera,
            downloadScreenshot: function () {
                // TODO
            },
            getPixelData: function (variant) {
                switch (variant) {
                    case 'color': return webgl.getDrawingBufferPixelData();
                    case 'pickObject': return pickPass.objectPickTarget.getPixelData();
                    case 'pickInstance': return pickPass.instancePickTarget.getPixelData();
                    case 'pickGroup': return pickPass.groupPickTarget.getPixelData();
                    case 'depth': return util_2.readTexture(webgl, drawPass.depthTexture);
                }
            },
            didDraw: didDraw,
            setProps: function (props) {
                if (props.cameraMode !== undefined && props.cameraMode !== camera.state.mode) {
                    camera.setState({ mode: props.cameraMode });
                }
                if (props.cameraClipDistance !== undefined)
                    p.cameraClipDistance = props.cameraClipDistance;
                if (props.clip !== undefined)
                    p.clip = [props.clip[0], props.clip[1]];
                if (props.fog !== undefined)
                    p.fog = [props.fog[0], props.fog[1]];
                if (props.postprocessing)
                    postprocessing.setProps(props.postprocessing);
                if (props.multiSample)
                    multiSample.setProps(props.multiSample);
                if (props.renderer)
                    renderer.setProps(props.renderer);
                if (props.trackball)
                    controls.setProps(props.trackball);
                if (props.debug)
                    debugHelper.setProps(props.debug);
                requestDraw(true);
            },
            get props() {
                return {
                    cameraMode: camera.state.mode,
                    cameraClipDistance: p.cameraClipDistance,
                    clip: p.clip,
                    fog: p.fog,
                    postprocessing: __assign({}, postprocessing.props),
                    multiSample: __assign({}, multiSample.props),
                    renderer: __assign({}, renderer.props),
                    trackball: __assign({}, controls.props),
                    debug: __assign({}, debugHelper.props)
                };
            },
            get input() {
                return input;
            },
            get stats() {
                return renderer.stats;
            },
            get interaction() {
                return interactionHelper.events;
            },
            dispose: function () {
                scene.clear();
                debugHelper.clear();
                input.dispose();
                controls.dispose();
                renderer.dispose();
                camera.dispose();
                interactionHelper.dispose();
            }
        };
        function handleResize() {
            width = gl.drawingBufferWidth;
            height = gl.drawingBufferHeight;
            renderer.setViewport(0, 0, width, height);
            util_1.Viewport.set(camera.viewport, 0, 0, width, height);
            util_1.Viewport.set(controls.viewport, 0, 0, width, height);
            drawPass.setSize(width, height);
            pickPass.setSize(width, height);
            postprocessing.setSize(width, height);
            multiSample.setSize(width, height);
            requestDraw(true);
        }
    }
    Canvas3D.create = create;
})(Canvas3D || (Canvas3D = {}));
exports.Canvas3D = Canvas3D;
