"use strict";
/**
 * Copyright (c) 2018-2019 mol* contributors, licensed under MIT, See LICENSE file for more info.
 *
 * @author Alexander Rose <alexander.rose@weirdbyte.de>
 * @author David Sehnal <david.sehnal@gmail.com>
 *
 * This code has been modified from https://github.com/mrdoob/three.js/,
 * copyright (c) 2010-2018 three.js authors. MIT License
 */
var __assign = (this && this.__assign) || function () {
    __assign = Object.assign || function(t) {
        for (var s, i = 1, n = arguments.length; i < n; i++) {
            s = arguments[i];
            for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p))
                t[p] = s[p];
        }
        return t;
    };
    return __assign.apply(this, arguments);
};
Object.defineProperty(exports, "__esModule", { value: true });
var linear_algebra_1 = require("../../mol-math/linear-algebra");
var util_1 = require("../camera/util");
var param_definition_1 = require("../../mol-util/param-definition");
exports.TrackballControlsParams = {
    noScroll: param_definition_1.ParamDefinition.Boolean(true, { isHidden: true }),
    rotateSpeed: param_definition_1.ParamDefinition.Numeric(5.0, { min: 0.1, max: 10, step: 0.1 }),
    zoomSpeed: param_definition_1.ParamDefinition.Numeric(6.0, { min: 0.1, max: 10, step: 0.1 }),
    panSpeed: param_definition_1.ParamDefinition.Numeric(0.8, { min: 0.1, max: 5, step: 0.1 }),
    spin: param_definition_1.ParamDefinition.Boolean(false),
    spinSpeed: param_definition_1.ParamDefinition.Numeric(1, { min: -100, max: 100, step: 1 }),
    staticMoving: param_definition_1.ParamDefinition.Boolean(true, { isHidden: true }),
    dynamicDampingFactor: param_definition_1.ParamDefinition.Numeric(0.2, {}, { isHidden: true }),
    minDistance: param_definition_1.ParamDefinition.Numeric(0.01, {}, { isHidden: true }),
    maxDistance: param_definition_1.ParamDefinition.Numeric(1e150, {}, { isHidden: true })
};
var TrackballControls;
(function (TrackballControls) {
    function create(input, object, props) {
        if (props === void 0) { props = {}; }
        var p = __assign({}, param_definition_1.ParamDefinition.getDefaultValues(exports.TrackballControlsParams), props);
        var viewport = { x: 0, y: 0, width: 0, height: 0 };
        var target = object.target;
        var disposed = false;
        var dragSub = input.drag.subscribe(onDrag);
        var interactionEndSub = input.interactionEnd.subscribe(onInteractionEnd);
        var wheelSub = input.wheel.subscribe(onWheel);
        var pinchSub = input.pinch.subscribe(onPinch);
        var _isInteracting = false;
        // For internal use
        var lastPosition = linear_algebra_1.Vec3.zero();
        var _eye = linear_algebra_1.Vec3.zero();
        var _movePrev = linear_algebra_1.Vec2.zero();
        var _moveCurr = linear_algebra_1.Vec2.zero();
        var _lastAxis = linear_algebra_1.Vec3.zero();
        var _lastAngle = 0;
        var _zoomStart = linear_algebra_1.Vec2.zero();
        var _zoomEnd = linear_algebra_1.Vec2.zero();
        var _panStart = linear_algebra_1.Vec2.zero();
        var _panEnd = linear_algebra_1.Vec2.zero();
        // Initial values for reseting
        var target0 = linear_algebra_1.Vec3.clone(target);
        var position0 = linear_algebra_1.Vec3.clone(object.position);
        var up0 = linear_algebra_1.Vec3.clone(object.up);
        var mouseOnScreenVec2 = linear_algebra_1.Vec2.zero();
        function getMouseOnScreen(pageX, pageY) {
            linear_algebra_1.Vec2.set(mouseOnScreenVec2, (pageX - viewport.x) / viewport.width, (pageY - viewport.y) / viewport.height);
            return mouseOnScreenVec2;
        }
        var mouseOnCircleVec2 = linear_algebra_1.Vec2.zero();
        function getMouseOnCircle(pageX, pageY) {
            linear_algebra_1.Vec2.set(mouseOnCircleVec2, ((pageX - viewport.width * 0.5 - viewport.x) / (viewport.width * 0.5)), ((viewport.height + 2 * (viewport.y - pageY)) / viewport.width) // screen.width intentional
            );
            return mouseOnCircleVec2;
        }
        var rotAxis = linear_algebra_1.Vec3.zero();
        var rotQuat = linear_algebra_1.Quat.zero();
        var rotEyeDir = linear_algebra_1.Vec3.zero();
        var rotObjUpDir = linear_algebra_1.Vec3.zero();
        var rotObjSideDir = linear_algebra_1.Vec3.zero();
        var rotMoveDir = linear_algebra_1.Vec3.zero();
        function rotateCamera() {
            linear_algebra_1.Vec3.set(rotMoveDir, _moveCurr[0] - _movePrev[0], _moveCurr[1] - _movePrev[1], 0);
            var angle = linear_algebra_1.Vec3.magnitude(rotMoveDir);
            if (angle) {
                linear_algebra_1.Vec3.copy(_eye, object.position);
                linear_algebra_1.Vec3.sub(_eye, _eye, target);
                linear_algebra_1.Vec3.normalize(rotEyeDir, linear_algebra_1.Vec3.copy(rotEyeDir, _eye));
                linear_algebra_1.Vec3.normalize(rotObjUpDir, linear_algebra_1.Vec3.copy(rotObjUpDir, object.up));
                linear_algebra_1.Vec3.normalize(rotObjSideDir, linear_algebra_1.Vec3.cross(rotObjSideDir, rotObjUpDir, rotEyeDir));
                linear_algebra_1.Vec3.setMagnitude(rotObjUpDir, rotObjUpDir, _moveCurr[1] - _movePrev[1]);
                linear_algebra_1.Vec3.setMagnitude(rotObjSideDir, rotObjSideDir, _moveCurr[0] - _movePrev[0]);
                linear_algebra_1.Vec3.add(rotMoveDir, linear_algebra_1.Vec3.copy(rotMoveDir, rotObjUpDir), rotObjSideDir);
                linear_algebra_1.Vec3.normalize(rotAxis, linear_algebra_1.Vec3.cross(rotAxis, rotMoveDir, _eye));
                angle *= p.rotateSpeed;
                linear_algebra_1.Quat.setAxisAngle(rotQuat, rotAxis, angle);
                linear_algebra_1.Vec3.transformQuat(_eye, _eye, rotQuat);
                linear_algebra_1.Vec3.transformQuat(object.up, object.up, rotQuat);
                linear_algebra_1.Vec3.copy(_lastAxis, rotAxis);
                _lastAngle = angle;
            }
            else if (!p.staticMoving && _lastAngle) {
                _lastAngle *= Math.sqrt(1.0 - p.dynamicDampingFactor);
                linear_algebra_1.Vec3.sub(_eye, linear_algebra_1.Vec3.copy(_eye, object.position), target);
                linear_algebra_1.Quat.setAxisAngle(rotQuat, _lastAxis, _lastAngle);
                linear_algebra_1.Vec3.transformQuat(_eye, _eye, rotQuat);
                linear_algebra_1.Vec3.transformQuat(object.up, object.up, rotQuat);
            }
            linear_algebra_1.Vec2.copy(_movePrev, _moveCurr);
        }
        function zoomCamera() {
            var factor = 1.0 + (_zoomEnd[1] - _zoomStart[1]) * p.zoomSpeed;
            if (factor !== 1.0 && factor > 0.0) {
                linear_algebra_1.Vec3.scale(_eye, _eye, factor);
            }
            if (p.staticMoving) {
                linear_algebra_1.Vec2.copy(_zoomStart, _zoomEnd);
            }
            else {
                _zoomStart[1] += (_zoomEnd[1] - _zoomStart[1]) * p.dynamicDampingFactor;
            }
        }
        var panMouseChange = linear_algebra_1.Vec2.zero();
        var panObjUp = linear_algebra_1.Vec3.zero();
        var panOffset = linear_algebra_1.Vec3.zero();
        function panCamera() {
            linear_algebra_1.Vec2.sub(panMouseChange, linear_algebra_1.Vec2.copy(panMouseChange, _panEnd), _panStart);
            if (linear_algebra_1.Vec2.squaredMagnitude(panMouseChange)) {
                linear_algebra_1.Vec2.scale(panMouseChange, panMouseChange, linear_algebra_1.Vec3.magnitude(_eye) * p.panSpeed);
                linear_algebra_1.Vec3.cross(panOffset, linear_algebra_1.Vec3.copy(panOffset, _eye), object.up);
                linear_algebra_1.Vec3.setMagnitude(panOffset, panOffset, panMouseChange[0]);
                linear_algebra_1.Vec3.setMagnitude(panObjUp, object.up, panMouseChange[1]);
                linear_algebra_1.Vec3.add(panOffset, panOffset, panObjUp);
                linear_algebra_1.Vec3.add(object.position, object.position, panOffset);
                linear_algebra_1.Vec3.add(target, target, panOffset);
                if (p.staticMoving) {
                    linear_algebra_1.Vec2.copy(_panStart, _panEnd);
                }
                else {
                    linear_algebra_1.Vec2.sub(panMouseChange, _panEnd, _panStart);
                    linear_algebra_1.Vec2.scale(panMouseChange, panMouseChange, p.dynamicDampingFactor);
                    linear_algebra_1.Vec2.add(_panStart, _panStart, panMouseChange);
                }
            }
        }
        /** Ensure the distance between object and target is within the min/max distance */
        function checkDistances() {
            if (linear_algebra_1.Vec3.squaredMagnitude(_eye) > p.maxDistance * p.maxDistance) {
                linear_algebra_1.Vec3.setMagnitude(_eye, _eye, p.maxDistance);
                linear_algebra_1.Vec3.add(object.position, target, _eye);
                linear_algebra_1.Vec2.copy(_zoomStart, _zoomEnd);
            }
            if (linear_algebra_1.Vec3.squaredMagnitude(_eye) < p.minDistance * p.minDistance) {
                linear_algebra_1.Vec3.setMagnitude(_eye, _eye, p.minDistance);
                linear_algebra_1.Vec3.add(object.position, target, _eye);
                linear_algebra_1.Vec2.copy(_zoomStart, _zoomEnd);
            }
        }
        var lastUpdated = -1;
        /** Update the object's position, direction and up vectors */
        function update(t) {
            if (lastUpdated === t)
                return;
            if (p.spin)
                spin(t - lastUpdated);
            linear_algebra_1.Vec3.sub(_eye, object.position, target);
            rotateCamera();
            zoomCamera();
            panCamera();
            linear_algebra_1.Vec3.add(object.position, target, _eye);
            checkDistances();
            util_1.cameraLookAt(object.position, object.up, object.direction, target);
            if (linear_algebra_1.Vec3.squaredDistance(lastPosition, object.position) > 0.000001 /* Value */) {
                linear_algebra_1.Vec3.copy(lastPosition, object.position);
            }
            lastUpdated = t;
        }
        /** Reset object's vectors and the target vector to their initial values */
        function reset() {
            linear_algebra_1.Vec3.copy(target, target0);
            linear_algebra_1.Vec3.copy(object.position, position0);
            linear_algebra_1.Vec3.copy(object.up, up0);
            linear_algebra_1.Vec3.sub(_eye, object.position, target);
            util_1.cameraLookAt(object.position, object.up, object.direction, target);
            linear_algebra_1.Vec3.copy(lastPosition, object.position);
        }
        // listeners
        function onDrag(_a) {
            var pageX = _a.pageX, pageY = _a.pageY, buttons = _a.buttons, isStart = _a.isStart;
            _isInteracting = true;
            if (isStart) {
                if (buttons === 1 /* Primary */) {
                    linear_algebra_1.Vec2.copy(_moveCurr, getMouseOnCircle(pageX, pageY));
                    linear_algebra_1.Vec2.copy(_movePrev, _moveCurr);
                }
                else if (buttons === 4 /* Auxilary */) {
                    linear_algebra_1.Vec2.copy(_zoomStart, getMouseOnScreen(pageX, pageY));
                    linear_algebra_1.Vec2.copy(_zoomEnd, _zoomStart);
                }
                else if (buttons === 2 /* Secondary */) {
                    linear_algebra_1.Vec2.copy(_panStart, getMouseOnScreen(pageX, pageY));
                    linear_algebra_1.Vec2.copy(_panEnd, _panStart);
                }
            }
            if (buttons === 1 /* Primary */) {
                linear_algebra_1.Vec2.copy(_movePrev, _moveCurr);
                linear_algebra_1.Vec2.copy(_moveCurr, getMouseOnCircle(pageX, pageY));
            }
            else if (buttons === 4 /* Auxilary */) {
                linear_algebra_1.Vec2.copy(_zoomEnd, getMouseOnScreen(pageX, pageY));
            }
            else if (buttons === 2 /* Secondary */) {
                linear_algebra_1.Vec2.copy(_panEnd, getMouseOnScreen(pageX, pageY));
            }
        }
        function onInteractionEnd() {
            _isInteracting = false;
        }
        function onWheel(_a) {
            var dy = _a.dy;
            _zoomStart[1] -= dy * 0.0001;
        }
        function onPinch(_a) {
            var fraction = _a.fraction;
            _isInteracting = true;
            _zoomStart[1] -= (fraction - 1) * 0.1;
        }
        function dispose() {
            if (disposed)
                return;
            disposed = true;
            dragSub.unsubscribe();
            wheelSub.unsubscribe();
            pinchSub.unsubscribe();
            interactionEndSub.unsubscribe();
        }
        var _spinSpeed = linear_algebra_1.Vec2.create(0.005, 0);
        function spin(deltaT) {
            var frameSpeed = (p.spinSpeed || 0) / 1000;
            _spinSpeed[0] = 60 * Math.min(Math.abs(deltaT), 1000 / 8) / 1000 * frameSpeed;
            if (!_isInteracting)
                linear_algebra_1.Vec2.add(_moveCurr, _movePrev, _spinSpeed);
        }
        // force an update at start
        update(0);
        return {
            viewport: viewport,
            get props() { return p; },
            setProps: function (props) {
                Object.assign(p, props);
            },
            update: update,
            reset: reset,
            dispose: dispose
        };
    }
    TrackballControls.create = create;
})(TrackballControls || (TrackballControls = {}));
exports.TrackballControls = TrackballControls;
