"use strict";
/**
 * Copyright (c) 2019 mol* contributors, licensed under MIT, See LICENSE file for more info.
 *
 * @author Alexander Rose <alexander.rose@weirdbyte.de>
 */
var __assign = (this && this.__assign) || function () {
    __assign = Object.assign || function(t) {
        for (var s, i = 1, n = arguments.length; i < n; i++) {
            s = arguments[i];
            for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p))
                t[p] = s[p];
        }
        return t;
    };
    return __assign.apply(this, arguments);
};
Object.defineProperty(exports, "__esModule", { value: true });
var util_1 = require("../../mol-gl/compute/util");
var schema_1 = require("../../mol-gl/renderable/schema");
var mol_util_1 = require("../../mol-util");
var linear_algebra_1 = require("../../mol-math/linear-algebra");
var shader_code_1 = require("../../mol-gl/shader-code");
var render_item_1 = require("../../mol-gl/webgl/render-item");
var renderable_1 = require("../../mol-gl/renderable");
var param_definition_1 = require("../../mol-util/param-definition");
var render_target_1 = require("../../mol-gl/webgl/render-target");
var camera_1 = require("../../mol-canvas3d/camera");
var quad_vert_1 = require("../../mol-gl/shader/quad.vert");
var compose_frag_1 = require("../../mol-gl/shader/compose.frag");
var ComposeSchema = __assign({}, util_1.QuadSchema, { tColor: schema_1.TextureSpec('texture', 'rgba', 'ubyte', 'nearest'), uTexSize: schema_1.UniformSpec('v2'), uWeight: schema_1.UniformSpec('f') });
function getComposeRenderable(ctx, colorTexture) {
    var values = __assign({}, util_1.QuadValues, { tColor: mol_util_1.ValueCell.create(colorTexture), uTexSize: mol_util_1.ValueCell.create(linear_algebra_1.Vec2.create(colorTexture.width, colorTexture.height)), uWeight: mol_util_1.ValueCell.create(1.0) });
    var schema = __assign({}, ComposeSchema);
    var shaderCode = shader_code_1.ShaderCode(quad_vert_1.default, compose_frag_1.default);
    var renderItem = render_item_1.createComputeRenderItem(ctx, 'triangles', shaderCode, schema, values);
    return renderable_1.createComputeRenderable(renderItem, values);
}
exports.MultiSampleParams = {
    mode: param_definition_1.ParamDefinition.Select('off', [['off', 'Off'], ['on', 'On'], ['temporal', 'Temporal']]),
    sampleLevel: param_definition_1.ParamDefinition.Numeric(2, { min: 0, max: 5, step: 1 }),
};
var MultiSamplePass = /** @class */ (function () {
    function MultiSamplePass(webgl, camera, drawPass, postprocessing, props) {
        this.webgl = webgl;
        this.camera = camera;
        this.drawPass = drawPass;
        this.postprocessing = postprocessing;
        this.sampleIndex = -1;
        this.currentTime = 0;
        this.lastRenderTime = 0;
        var gl = webgl.gl;
        this.composeTarget = render_target_1.createRenderTarget(webgl, gl.drawingBufferWidth, gl.drawingBufferHeight);
        this.holdTarget = render_target_1.createRenderTarget(webgl, gl.drawingBufferWidth, gl.drawingBufferHeight);
        this.compose = getComposeRenderable(webgl, drawPass.colorTarget.texture);
        this.props = __assign({}, param_definition_1.ParamDefinition.getDefaultValues(exports.MultiSampleParams), props);
    }
    Object.defineProperty(MultiSamplePass.prototype, "enabled", {
        get: function () {
            if (this.props.mode === 'temporal') {
                if (this.currentTime - this.lastRenderTime > 200) {
                    return this.sampleIndex !== -1;
                }
                else {
                    this.sampleIndex = 0;
                    return false;
                }
            }
            else if (this.props.mode === 'on') {
                return true;
            }
            else {
                return false;
            }
        },
        enumerable: true,
        configurable: true
    });
    MultiSamplePass.prototype.update = function (changed, currentTime) {
        if (changed)
            this.lastRenderTime = currentTime;
        this.currentTime = currentTime;
    };
    MultiSamplePass.prototype.setSize = function (width, height) {
        this.composeTarget.setSize(width, height);
        this.holdTarget.setSize(width, height);
        mol_util_1.ValueCell.update(this.compose.values.uTexSize, linear_algebra_1.Vec2.set(this.compose.values.uTexSize.ref.value, width, height));
    };
    MultiSamplePass.prototype.setProps = function (props) {
        if (props.mode !== undefined)
            this.props.mode = props.mode;
        if (props.sampleLevel !== undefined)
            this.props.sampleLevel = props.sampleLevel;
    };
    MultiSamplePass.prototype.render = function () {
        if (this.props.mode === 'temporal') {
            this.renderTemporalMultiSample();
        }
        else {
            this.renderMultiSample();
        }
    };
    MultiSamplePass.prototype.renderMultiSample = function () {
        var _a = this, camera = _a.camera, compose = _a.compose, composeTarget = _a.composeTarget, drawPass = _a.drawPass, postprocessing = _a.postprocessing, webgl = _a.webgl;
        var gl = webgl.gl, state = webgl.state;
        // based on the Multisample Anti-Aliasing Render Pass
        // contributed to three.js by bhouston / http://clara.io/
        //
        // This manual approach to MSAA re-renders the scene once for
        // each sample with camera jitter and accumulates the results.
        var offsetList = JitterVectors[Math.max(0, Math.min(this.props.sampleLevel, 5))];
        var baseSampleWeight = 1.0 / offsetList.length;
        var roundingRange = 1 / 32;
        camera.viewOffset.enabled = true;
        mol_util_1.ValueCell.update(compose.values.tColor, postprocessing.enabled ? postprocessing.target.texture : drawPass.colorTarget.texture);
        compose.update();
        var _b = drawPass.colorTarget, width = _b.width, height = _b.height;
        // render the scene multiple times, each slightly jitter offset
        // from the last and accumulate the results.
        for (var i = 0; i < offsetList.length; ++i) {
            var offset = offsetList[i];
            camera_1.Camera.setViewOffset(camera.viewOffset, width, height, offset[0], offset[1], width, height);
            camera.updateMatrices();
            // the theory is that equal weights for each sample lead to an accumulation of rounding
            // errors. The following equation varies the sampleWeight per sample so that it is uniformly
            // distributed across a range of values whose rounding errors cancel each other out.
            var uniformCenteredDistribution = -0.5 + (i + 0.5) / offsetList.length;
            var sampleWeight = baseSampleWeight + roundingRange * uniformCenteredDistribution;
            mol_util_1.ValueCell.update(compose.values.uWeight, sampleWeight);
            // render scene and optionally postprocess
            drawPass.render(false);
            if (postprocessing.enabled)
                postprocessing.render(false);
            // compose rendered scene with compose target
            composeTarget.bind();
            gl.viewport(0, 0, width, height);
            state.enable(gl.BLEND);
            state.blendEquationSeparate(gl.FUNC_ADD, gl.FUNC_ADD);
            state.blendFuncSeparate(gl.ONE, gl.ONE, gl.ONE, gl.ONE);
            state.disable(gl.DEPTH_TEST);
            state.disable(gl.SCISSOR_TEST);
            state.depthMask(false);
            if (i === 0) {
                state.clearColor(0, 0, 0, 0);
                gl.clear(gl.COLOR_BUFFER_BIT);
            }
            compose.render();
        }
        mol_util_1.ValueCell.update(compose.values.uWeight, 1.0);
        mol_util_1.ValueCell.update(compose.values.tColor, composeTarget.texture);
        compose.update();
        webgl.unbindFramebuffer();
        gl.viewport(0, 0, width, height);
        state.disable(gl.BLEND);
        compose.render();
        camera.viewOffset.enabled = false;
        camera.updateMatrices();
    };
    MultiSamplePass.prototype.renderTemporalMultiSample = function () {
        var _a = this, camera = _a.camera, compose = _a.compose, composeTarget = _a.composeTarget, holdTarget = _a.holdTarget, postprocessing = _a.postprocessing, drawPass = _a.drawPass, webgl = _a.webgl;
        var gl = webgl.gl, state = webgl.state;
        // based on the Multisample Anti-Aliasing Render Pass
        // contributed to three.js by bhouston / http://clara.io/
        //
        // This manual approach to MSAA re-renders the scene once for
        // each sample with camera jitter and accumulates the results.
        var offsetList = JitterVectors[Math.max(0, Math.min(this.props.sampleLevel, 5))];
        if (this.sampleIndex === -1)
            return;
        if (this.sampleIndex >= offsetList.length) {
            this.sampleIndex = -1;
            return;
        }
        var i = this.sampleIndex;
        if (i === 0) {
            drawPass.render(false);
            if (postprocessing.enabled)
                postprocessing.render(false);
            mol_util_1.ValueCell.update(compose.values.uWeight, 1.0);
            mol_util_1.ValueCell.update(compose.values.tColor, postprocessing.enabled ? postprocessing.target.texture : drawPass.colorTarget.texture);
            compose.update();
            holdTarget.bind();
            state.disable(gl.BLEND);
            compose.render();
        }
        var sampleWeight = 1.0 / offsetList.length;
        camera.viewOffset.enabled = true;
        mol_util_1.ValueCell.update(compose.values.tColor, postprocessing.enabled ? postprocessing.target.texture : drawPass.colorTarget.texture);
        mol_util_1.ValueCell.update(compose.values.uWeight, sampleWeight);
        compose.update();
        var _b = drawPass.colorTarget, width = _b.width, height = _b.height;
        // render the scene multiple times, each slightly jitter offset
        // from the last and accumulate the results.
        var numSamplesPerFrame = Math.pow(2, this.props.sampleLevel);
        for (var i_1 = 0; i_1 < numSamplesPerFrame; ++i_1) {
            var offset = offsetList[this.sampleIndex];
            camera_1.Camera.setViewOffset(camera.viewOffset, width, height, offset[0], offset[1], width, height);
            camera.updateMatrices();
            // render scene and optionally postprocess
            drawPass.render(false);
            if (postprocessing.enabled)
                postprocessing.render(false);
            // compose rendered scene with compose target
            composeTarget.bind();
            state.enable(gl.BLEND);
            state.blendEquationSeparate(gl.FUNC_ADD, gl.FUNC_ADD);
            state.blendFuncSeparate(gl.ONE, gl.ONE, gl.ONE, gl.ONE);
            state.disable(gl.DEPTH_TEST);
            state.disable(gl.SCISSOR_TEST);
            state.depthMask(false);
            if (this.sampleIndex === 0) {
                state.clearColor(0, 0, 0, 0);
                gl.clear(gl.COLOR_BUFFER_BIT);
            }
            compose.render();
            this.sampleIndex += 1;
            if (this.sampleIndex >= offsetList.length)
                break;
        }
        var accumulationWeight = this.sampleIndex * sampleWeight;
        if (accumulationWeight > 0) {
            mol_util_1.ValueCell.update(compose.values.uWeight, 1.0);
            mol_util_1.ValueCell.update(compose.values.tColor, composeTarget.texture);
            compose.update();
            webgl.unbindFramebuffer();
            gl.viewport(0, 0, width, height);
            state.disable(gl.BLEND);
            compose.render();
        }
        if (accumulationWeight < 1.0) {
            mol_util_1.ValueCell.update(compose.values.uWeight, 1.0 - accumulationWeight);
            mol_util_1.ValueCell.update(compose.values.tColor, holdTarget.texture);
            compose.update();
            webgl.unbindFramebuffer();
            gl.viewport(0, 0, width, height);
            if (accumulationWeight === 0)
                state.disable(gl.BLEND);
            else
                state.enable(gl.BLEND);
            compose.render();
        }
        camera.viewOffset.enabled = false;
        camera.updateMatrices();
        if (this.sampleIndex >= offsetList.length)
            this.sampleIndex = -1;
    };
    return MultiSamplePass;
}());
exports.MultiSamplePass = MultiSamplePass;
var JitterVectors = [
    [
        [0, 0]
    ],
    [
        [4, 4], [-4, -4]
    ],
    [
        [-2, -6], [6, -2], [-6, 2], [2, 6]
    ],
    [
        [1, -3], [-1, 3], [5, 1], [-3, -5],
        [-5, 5], [-7, -1], [3, 7], [7, -7]
    ],
    [
        [1, 1], [-1, -3], [-3, 2], [4, -1],
        [-5, -2], [2, 5], [5, 3], [3, -5],
        [-2, 6], [0, -7], [-4, -6], [-6, 4],
        [-8, 0], [7, -4], [6, 7], [-7, -8]
    ],
    [
        [-4, -7], [-7, -5], [-3, -5], [-5, -4],
        [-1, -4], [-2, -2], [-6, -1], [-4, 0],
        [-7, 1], [-1, 2], [-6, 3], [-3, 3],
        [-7, 6], [-3, 6], [-5, 7], [-1, 7],
        [5, -7], [1, -6], [6, -5], [4, -4],
        [2, -3], [7, -2], [1, -1], [4, -1],
        [2, 1], [6, 2], [0, 4], [4, 4],
        [2, 5], [7, 5], [5, 6], [3, 7]
    ]
];
JitterVectors.forEach(function (offsetList) {
    offsetList.forEach(function (offset) {
        // 0.0625 = 1 / 16
        offset[0] *= 0.0625;
        offset[1] *= 0.0625;
    });
});
