"use strict";
/**
 * Copyright (c) 2018 mol* contributors, licensed under MIT, See LICENSE file for more info.
 *
 * @author Alexander Rose <alexander.rose@weirdbyte.de>
 */
Object.defineProperty(exports, "__esModule", { value: true });
var mol_util_1 = require("../../mol-util");
var util_1 = require("../../mol-gl/renderable/util");
var color_1 = require("../../mol-util/color");
var linear_algebra_1 = require("../../mol-math/linear-algebra");
var location_1 = require("../../mol-model/location");
var geometry_1 = require("./geometry");
function createColors(locationIt, colorTheme, colorData) {
    switch (geometry_1.Geometry.getGranularity(locationIt, colorTheme.granularity)) {
        case 'uniform': return createUniformColor(locationIt, colorTheme.color, colorData);
        case 'group': return createGroupColor(locationIt, colorTheme.color, colorData);
        case 'groupInstance': return createGroupInstanceColor(locationIt, colorTheme.color, colorData);
        case 'instance': return createInstanceColor(locationIt, colorTheme.color, colorData);
    }
}
exports.createColors = createColors;
function createValueColor(value, colorData) {
    if (colorData) {
        mol_util_1.ValueCell.update(colorData.uColor, color_1.Color.toVec3Normalized(colorData.uColor.ref.value, value));
        if (colorData.dColorType.ref.value !== 'uniform') {
            mol_util_1.ValueCell.update(colorData.dColorType, 'uniform');
        }
        return colorData;
    }
    else {
        return {
            uColor: mol_util_1.ValueCell.create(color_1.Color.toVec3Normalized(linear_algebra_1.Vec3(), value)),
            tColor: mol_util_1.ValueCell.create({ array: new Uint8Array(3), width: 1, height: 1 }),
            uColorTexDim: mol_util_1.ValueCell.create(linear_algebra_1.Vec2.create(1, 1)),
            dColorType: mol_util_1.ValueCell.create('uniform'),
        };
    }
}
exports.createValueColor = createValueColor;
/** Creates color uniform */
function createUniformColor(locationIt, color, colorData) {
    return createValueColor(color(location_1.NullLocation, false), colorData);
}
exports.createUniformColor = createUniformColor;
function createTextureColor(colors, type, colorData) {
    if (colorData) {
        mol_util_1.ValueCell.update(colorData.tColor, colors);
        mol_util_1.ValueCell.update(colorData.uColorTexDim, linear_algebra_1.Vec2.create(colors.width, colors.height));
        if (colorData.dColorType.ref.value !== type) {
            mol_util_1.ValueCell.update(colorData.dColorType, type);
        }
        return colorData;
    }
    else {
        return {
            uColor: mol_util_1.ValueCell.create(linear_algebra_1.Vec3.zero()),
            tColor: mol_util_1.ValueCell.create(colors),
            uColorTexDim: mol_util_1.ValueCell.create(linear_algebra_1.Vec2.create(colors.width, colors.height)),
            dColorType: mol_util_1.ValueCell.create(type),
        };
    }
}
exports.createTextureColor = createTextureColor;
/** Creates color texture with color for each instance/unit */
function createInstanceColor(locationIt, color, colorData) {
    var instanceCount = locationIt.instanceCount;
    var colors = util_1.createTextureImage(Math.max(1, instanceCount), 3, Uint8Array, colorData && colorData.tColor.ref.value.array);
    locationIt.reset();
    while (locationIt.hasNext) {
        var _a = locationIt.move(), location_2 = _a.location, isSecondary = _a.isSecondary, instanceIndex = _a.instanceIndex;
        color_1.Color.toArray(color(location_2, isSecondary), colors.array, instanceIndex * 3);
        locationIt.skipInstance();
    }
    return createTextureColor(colors, 'instance', colorData);
}
exports.createInstanceColor = createInstanceColor;
/** Creates color texture with color for each group (i.e. shared across instances/units) */
function createGroupColor(locationIt, color, colorData) {
    var groupCount = locationIt.groupCount;
    var colors = util_1.createTextureImage(Math.max(1, groupCount), 3, Uint8Array, colorData && colorData.tColor.ref.value.array);
    locationIt.reset();
    while (locationIt.hasNext && !locationIt.isNextNewInstance) {
        var _a = locationIt.move(), location_3 = _a.location, isSecondary = _a.isSecondary, groupIndex = _a.groupIndex;
        color_1.Color.toArray(color(location_3, isSecondary), colors.array, groupIndex * 3);
    }
    return createTextureColor(colors, 'group', colorData);
}
exports.createGroupColor = createGroupColor;
/** Creates color texture with color for each group in each instance (i.e. for each unit) */
function createGroupInstanceColor(locationIt, color, colorData) {
    var groupCount = locationIt.groupCount, instanceCount = locationIt.instanceCount;
    var count = instanceCount * groupCount;
    var colors = util_1.createTextureImage(Math.max(1, count), 3, Uint8Array, colorData && colorData.tColor.ref.value.array);
    locationIt.reset();
    while (locationIt.hasNext) {
        var _a = locationIt.move(), location_4 = _a.location, isSecondary = _a.isSecondary, index = _a.index;
        color_1.Color.toArray(color(location_4, isSecondary), colors.array, index * 3);
    }
    return createTextureColor(colors, 'groupInstance', colorData);
}
exports.createGroupInstanceColor = createGroupInstanceColor;
