"use strict";
/**
 * Copyright (c) 2018-2019 mol* contributors, licensed under MIT, See LICENSE file for more info.
 *
 * @author Alexander Rose <alexander.rose@weirdbyte.de>
 */
Object.defineProperty(exports, "__esModule", { value: true });
var value_cell_1 = require("../../../mol-util/value-cell");
var linear_algebra_1 = require("../../../mol-math/linear-algebra");
var util_1 = require("../../../mol-data/util");
var util_2 = require("../../util");
var sphere_1 = require("./builder/sphere");
var cylinder_1 = require("./builder/cylinder");
var tmpV = linear_algebra_1.Vec3.zero();
var tmpMat3 = linear_algebra_1.Mat3.zero();
var tmpVecA = linear_algebra_1.Vec3.zero();
var tmpVecB = linear_algebra_1.Vec3.zero();
var tmpVecC = linear_algebra_1.Vec3.zero();
var tmpVecD = linear_algebra_1.Vec3.zero();
var MeshBuilder;
(function (MeshBuilder) {
    function createState(initialCount, chunkSize, mesh) {
        if (initialCount === void 0) { initialCount = 2048; }
        if (chunkSize === void 0) { chunkSize = 1024; }
        return {
            currentGroup: -1,
            vertices: util_1.ChunkedArray.create(Float32Array, 3, chunkSize, mesh ? mesh.vertexBuffer.ref.value : initialCount),
            normals: util_1.ChunkedArray.create(Float32Array, 3, chunkSize, mesh ? mesh.normalBuffer.ref.value : initialCount),
            indices: util_1.ChunkedArray.create(Uint32Array, 3, chunkSize * 3, mesh ? mesh.indexBuffer.ref.value : initialCount * 3),
            groups: util_1.ChunkedArray.create(Float32Array, 1, chunkSize, mesh ? mesh.groupBuffer.ref.value : initialCount),
            mesh: mesh
        };
    }
    MeshBuilder.createState = createState;
    function addTriangle(state, a, b, c) {
        var vertices = state.vertices, normals = state.normals, indices = state.indices, groups = state.groups, currentGroup = state.currentGroup;
        var offset = vertices.elementCount;
        // positions
        util_1.ChunkedArray.add3(vertices, a[0], a[1], a[2]);
        util_1.ChunkedArray.add3(vertices, b[0], b[1], b[2]);
        util_1.ChunkedArray.add3(vertices, c[0], c[1], c[2]);
        linear_algebra_1.Vec3.triangleNormal(tmpV, a, b, c);
        for (var i = 0; i < 3; ++i) {
            util_1.ChunkedArray.add3(normals, tmpV[0], tmpV[1], tmpV[2]); // normal
            util_1.ChunkedArray.add(groups, currentGroup); // group
        }
        util_1.ChunkedArray.add3(indices, offset, offset + 1, offset + 2);
    }
    MeshBuilder.addTriangle = addTriangle;
    function addTriangleStrip(state, vertices, indices) {
        linear_algebra_1.Vec3.fromArray(tmpVecC, vertices, indices[0] * 3);
        linear_algebra_1.Vec3.fromArray(tmpVecD, vertices, indices[1] * 3);
        for (var i = 2, il = indices.length; i < il; i += 2) {
            linear_algebra_1.Vec3.copy(tmpVecA, tmpVecC);
            linear_algebra_1.Vec3.copy(tmpVecB, tmpVecD);
            linear_algebra_1.Vec3.fromArray(tmpVecC, vertices, indices[i] * 3);
            linear_algebra_1.Vec3.fromArray(tmpVecD, vertices, indices[i + 1] * 3);
            addTriangle(state, tmpVecA, tmpVecB, tmpVecC);
            addTriangle(state, tmpVecB, tmpVecD, tmpVecC);
        }
    }
    MeshBuilder.addTriangleStrip = addTriangleStrip;
    function addTriangleFan(state, vertices, indices) {
        linear_algebra_1.Vec3.fromArray(tmpVecA, vertices, indices[0] * 3);
        for (var i = 2, il = indices.length; i < il; ++i) {
            linear_algebra_1.Vec3.fromArray(tmpVecB, vertices, indices[i - 1] * 3);
            linear_algebra_1.Vec3.fromArray(tmpVecC, vertices, indices[i] * 3);
            addTriangle(state, tmpVecA, tmpVecC, tmpVecB);
        }
    }
    MeshBuilder.addTriangleFan = addTriangleFan;
    function addPrimitive(state, t, primitive) {
        var va = primitive.vertices, na = primitive.normals, ia = primitive.indices;
        var vertices = state.vertices, normals = state.normals, indices = state.indices, groups = state.groups, currentGroup = state.currentGroup;
        var offset = vertices.elementCount;
        var n = util_2.getNormalMatrix(tmpMat3, t);
        for (var i = 0, il = va.length; i < il; i += 3) {
            // position
            linear_algebra_1.Vec3.transformMat4(tmpV, linear_algebra_1.Vec3.fromArray(tmpV, va, i), t);
            util_1.ChunkedArray.add3(vertices, tmpV[0], tmpV[1], tmpV[2]);
            // normal
            linear_algebra_1.Vec3.transformMat3(tmpV, linear_algebra_1.Vec3.fromArray(tmpV, na, i), n);
            util_1.ChunkedArray.add3(normals, tmpV[0], tmpV[1], tmpV[2]);
            // group
            util_1.ChunkedArray.add(groups, currentGroup);
        }
        for (var i = 0, il = ia.length; i < il; i += 3) {
            util_1.ChunkedArray.add3(indices, ia[i] + offset, ia[i + 1] + offset, ia[i + 2] + offset);
        }
    }
    MeshBuilder.addPrimitive = addPrimitive;
    function addCage(state, t, cage, radius, detail) {
        var va = cage.vertices, ea = cage.edges;
        var cylinderProps = { radiusTop: radius, radiusBottom: radius };
        for (var i = 0, il = ea.length; i < il; i += 2) {
            linear_algebra_1.Vec3.fromArray(tmpVecA, va, ea[i] * 3);
            linear_algebra_1.Vec3.fromArray(tmpVecB, va, ea[i + 1] * 3);
            linear_algebra_1.Vec3.transformMat4(tmpVecA, tmpVecA, t);
            linear_algebra_1.Vec3.transformMat4(tmpVecB, tmpVecB, t);
            sphere_1.addSphere(state, tmpVecA, radius, detail);
            sphere_1.addSphere(state, tmpVecB, radius, detail);
            cylinder_1.addCylinder(state, tmpVecA, tmpVecB, 1, cylinderProps);
        }
    }
    MeshBuilder.addCage = addCage;
    function getMesh(state) {
        var vertices = state.vertices, normals = state.normals, indices = state.indices, groups = state.groups, mesh = state.mesh;
        var vb = util_1.ChunkedArray.compact(vertices, true);
        var ib = util_1.ChunkedArray.compact(indices, true);
        var nb = util_1.ChunkedArray.compact(normals, true);
        var gb = util_1.ChunkedArray.compact(groups, true);
        return {
            kind: 'mesh',
            vertexCount: state.vertices.elementCount,
            triangleCount: state.indices.elementCount,
            vertexBuffer: mesh ? value_cell_1.ValueCell.update(mesh.vertexBuffer, vb) : value_cell_1.ValueCell.create(vb),
            indexBuffer: mesh ? value_cell_1.ValueCell.update(mesh.indexBuffer, ib) : value_cell_1.ValueCell.create(ib),
            normalBuffer: mesh ? value_cell_1.ValueCell.update(mesh.normalBuffer, nb) : value_cell_1.ValueCell.create(nb),
            groupBuffer: mesh ? value_cell_1.ValueCell.update(mesh.groupBuffer, gb) : value_cell_1.ValueCell.create(gb),
            normalsComputed: true,
        };
    }
    MeshBuilder.getMesh = getMesh;
})(MeshBuilder = exports.MeshBuilder || (exports.MeshBuilder = {}));
