/**
 * Copyright (c) 2018-2019 mol* contributors, licensed under MIT, See LICENSE file for more info.
 *
 * @author Alexander Rose <alexander.rose@weirdbyte.de>
 */
import { ValueCell } from '../../mol-util';
import { AttributeItemSize, ElementsKind, AttributeValues, AttributeKind } from '../webgl/buffer';
import { UniformKind, UniformValues } from '../webgl/uniform';
import { DefineKind, DefineValues } from '../shader-code';
import { Vec2, Vec3, Vec4, Mat3, Mat4 } from '../../mol-math/linear-algebra';
import { TextureImage, TextureVolume } from './util';
import { TextureValues, TextureType, TextureFormat, TextureFilter, TextureKind, Texture } from '../webgl/texture';
import { Sphere3D } from '../../mol-math/geometry';
export declare type ValueKindType = {
    'number': number;
    'string': string;
    'boolean': string;
    'any': any;
    'm4': Mat4;
    'float32': Float32Array;
    'sphere': Sphere3D;
};
export declare type ValueKind = keyof ValueKindType;
export declare type KindValue = {
    'f': number;
    'i': number;
    'v2': Vec2;
    'v3': Vec3;
    'v4': Vec4;
    'm3': Mat3;
    'm4': Mat4;
    't': number;
    'uint8': Uint8Array;
    'int8': Int8Array;
    'uint16': Uint16Array;
    'int16': Int16Array;
    'uint32': Uint32Array;
    'int32': Int32Array;
    'float32': Float32Array;
    'image-uint8': TextureImage<Uint8Array>;
    'image-float32': TextureImage<Float32Array>;
    'image-depth': TextureImage<Uint8Array>;
    'volume-uint8': TextureVolume<Uint8Array>;
    'volume-float32': TextureVolume<Float32Array>;
    'texture': Texture;
    'texture2d': Texture;
    'texture3d': Texture;
    'number': number;
    'string': string;
    'boolean': boolean;
    'any': any;
    'sphere': Sphere3D;
};
export declare type Values<S extends RenderableSchema> = {
    [k in keyof S]: ValueCell<KindValue[S[k]['kind']]>;
};
export declare function splitValues(schema: RenderableSchema, values: RenderableValues): {
    attributeValues: AttributeValues;
    defineValues: DefineValues;
    textureValues: TextureValues;
    uniformValues: UniformValues;
    materialUniformValues: UniformValues;
};
export declare type Versions<T extends RenderableValues> = {
    [k in keyof T]: number;
};
export declare function getValueVersions<T extends RenderableValues>(values: T): Versions<T>;
export declare type AttributeSpec<K extends AttributeKind> = {
    type: 'attribute';
    kind: K;
    itemSize: AttributeItemSize;
    divisor: number;
};
export declare function AttributeSpec<K extends AttributeKind>(kind: K, itemSize: AttributeItemSize, divisor: number): AttributeSpec<K>;
export declare type UniformSpec<K extends UniformKind> = {
    type: 'uniform';
    kind: K;
    isMaterial: boolean;
};
export declare function UniformSpec<K extends UniformKind>(kind: K, isMaterial?: boolean): UniformSpec<K>;
export declare type TextureSpec<K extends TextureKind> = {
    type: 'texture';
    kind: K;
    format: TextureFormat;
    dataType: TextureType;
    filter: TextureFilter;
};
export declare function TextureSpec<K extends TextureKind>(kind: K, format: TextureFormat, dataType: TextureType, filter: TextureFilter): TextureSpec<K>;
export declare type ElementsSpec<K extends ElementsKind> = {
    type: 'elements';
    kind: K;
};
export declare function ElementsSpec<K extends ElementsKind>(kind: K): ElementsSpec<K>;
export declare type DefineSpec<K extends DefineKind> = {
    type: 'define';
    kind: K;
    options?: string[];
};
export declare function DefineSpec<K extends DefineKind>(kind: K, options?: string[]): DefineSpec<K>;
export declare type ValueSpec<K extends ValueKind> = {
    type: 'value';
    kind: K;
};
export declare function ValueSpec<K extends ValueKind>(kind: K): ValueSpec<K>;
export declare type RenderableSchema = {
    [k: string]: (AttributeSpec<AttributeKind> | UniformSpec<UniformKind> | TextureSpec<TextureKind> | ValueSpec<ValueKind> | DefineSpec<DefineKind> | ElementsSpec<ElementsKind>);
};
export declare type RenderableValues = {
    [k: string]: ValueCell<any>;
};
export declare const GlobalUniformSchema: {
    uModel: UniformSpec<"m4">;
    uView: UniformSpec<"m4">;
    uInvView: UniformSpec<"m4">;
    uModelView: UniformSpec<"m4">;
    uInvModelView: UniformSpec<"m4">;
    uProjection: UniformSpec<"m4">;
    uInvProjection: UniformSpec<"m4">;
    uModelViewProjection: UniformSpec<"m4">;
    uInvModelViewProjection: UniformSpec<"m4">;
    uLightIntensity: UniformSpec<"f">;
    uAmbientIntensity: UniformSpec<"f">;
    uMetalness: UniformSpec<"f">;
    uRoughness: UniformSpec<"f">;
    uReflectivity: UniformSpec<"f">;
    uIsOrtho: UniformSpec<"f">;
    uPixelRatio: UniformSpec<"f">;
    uViewportHeight: UniformSpec<"f">;
    uViewport: UniformSpec<"v4">;
    uCameraPosition: UniformSpec<"v3">;
    uNear: UniformSpec<"f">;
    uFar: UniformSpec<"f">;
    uFogNear: UniformSpec<"f">;
    uFogFar: UniformSpec<"f">;
    uFogColor: UniformSpec<"v3">;
    uPickingAlphaThreshold: UniformSpec<"f">;
};
export declare type GlobalUniformSchema = typeof GlobalUniformSchema;
export declare type GlobalUniformValues = Values<GlobalUniformSchema>;
export declare const InternalSchema: {
    uObjectId: UniformSpec<"i">;
    uPickable: UniformSpec<"i">;
};
export declare type InternalSchema = typeof InternalSchema;
export declare type InternalValues = {
    [k in keyof InternalSchema]: ValueCell<any>;
};
export declare const ColorSchema: {
    uColor: UniformSpec<"v3">;
    uColorTexDim: UniformSpec<"v2">;
    tColor: TextureSpec<"image-uint8">;
    dColorType: DefineSpec<"string">;
};
export declare type ColorSchema = typeof ColorSchema;
export declare type ColorValues = Values<ColorSchema>;
export declare const SizeSchema: {
    uSize: UniformSpec<"f">;
    uSizeTexDim: UniformSpec<"v2">;
    tSize: TextureSpec<"image-uint8">;
    dSizeType: DefineSpec<"string">;
    uSizeFactor: UniformSpec<"f">;
};
export declare type SizeSchema = typeof SizeSchema;
export declare type SizeValues = Values<SizeSchema>;
export declare const MarkerSchema: {
    uMarkerTexDim: UniformSpec<"v2">;
    tMarker: TextureSpec<"image-uint8">;
};
export declare type MarkerSchema = typeof MarkerSchema;
export declare type MarkerValues = Values<MarkerSchema>;
export declare const OverpaintSchema: {
    uOverpaintTexDim: UniformSpec<"v2">;
    tOverpaint: TextureSpec<"image-uint8">;
    dOverpaint: DefineSpec<"boolean">;
};
export declare type OverpaintSchema = typeof OverpaintSchema;
export declare type OverpaintValues = Values<OverpaintSchema>;
export declare const TransparencySchema: {
    uTransparencyTexDim: UniformSpec<"v2">;
    tTransparency: TextureSpec<"image-uint8">;
    dTransparency: DefineSpec<"boolean">;
    dTransparencyVariant: DefineSpec<"string">;
};
export declare type TransparencySchema = typeof TransparencySchema;
export declare type TransparencyValues = Values<TransparencySchema>;
export declare const BaseSchema: {
    aInstance: AttributeSpec<"float32">;
    aGroup: AttributeSpec<"float32">;
    /**
     * final per-instance transform calculated for instance `i` as
     * `aTransform[i] = matrix * transform[i] * extraTransform[i]`
     */
    aTransform: AttributeSpec<"float32">;
    /**
     * final alpha, calculated as `values.alpha * state.alpha`
     */
    uAlpha: UniformSpec<"f">;
    uInstanceCount: UniformSpec<"i">;
    uGroupCount: UniformSpec<"i">;
    uHighlightColor: UniformSpec<"v3">;
    uSelectColor: UniformSpec<"v3">;
    drawCount: ValueSpec<"number">;
    instanceCount: ValueSpec<"number">;
    /** base alpha, see uAlpha  */
    alpha: ValueSpec<"number">;
    /** global transform, see aTransform */
    matrix: ValueSpec<"m4">;
    /** base per-instance transform, see aTransform */
    transform: ValueSpec<"float32">;
    /** additional per-instance transform, see aTransform */
    extraTransform: ValueSpec<"float32">;
    /** bounding sphere taking aTransform into account */
    boundingSphere: ValueSpec<"sphere">;
    /** bounding sphere NOT taking aTransform into account */
    invariantBoundingSphere: ValueSpec<"sphere">;
    dUseFog: DefineSpec<"boolean">;
    uTransparencyTexDim: UniformSpec<"v2">;
    tTransparency: TextureSpec<"image-uint8">;
    dTransparency: DefineSpec<"boolean">;
    dTransparencyVariant: DefineSpec<"string">;
    uOverpaintTexDim: UniformSpec<"v2">;
    tOverpaint: TextureSpec<"image-uint8">;
    dOverpaint: DefineSpec<"boolean">;
    uMarkerTexDim: UniformSpec<"v2">;
    tMarker: TextureSpec<"image-uint8">;
    uColor: UniformSpec<"v3">;
    uColorTexDim: UniformSpec<"v2">;
    tColor: TextureSpec<"image-uint8">;
    dColorType: DefineSpec<"string">;
};
export declare type BaseSchema = typeof BaseSchema;
export declare type BaseValues = Values<BaseSchema>;
