"use strict";
/**
 * Copyright (c) 2018 mol* contributors, licensed under MIT, See LICENSE file for more info.
 *
 * @author Alexander Rose <alexander.rose@weirdbyte.de>
 */
Object.defineProperty(exports, "__esModule", { value: true });
var geometry_1 = require("../../mol-math/geometry");
var linear_algebra_1 = require("../../mol-math/linear-algebra");
var boundary_helper_1 = require("../../mol-math/geometry/boundary-helper");
function calculateTextureInfo(n, itemSize) {
    var sqN = Math.sqrt(n);
    var width = Math.ceil(sqN);
    width = width + (itemSize - (width % itemSize)) % itemSize;
    var height = width > 0 ? Math.ceil(n / width) : 0;
    return { width: width, height: height, length: width * height * itemSize };
}
exports.calculateTextureInfo = calculateTextureInfo;
function createTextureImage(n, itemSize, arrayCtor, array) {
    var _a = calculateTextureInfo(n, itemSize), length = _a.length, width = _a.width, height = _a.height;
    array = array && array.length >= length ? array : new arrayCtor(length);
    return { array: array, width: width, height: height };
}
exports.createTextureImage = createTextureImage;
function printTextureImage(textureImage, scale) {
    if (scale === void 0) { scale = 1; }
    var array = textureImage.array, width = textureImage.width, height = textureImage.height;
    var itemSize = array.length / (width * height);
    var data = new Uint8ClampedArray(width * height * 4);
    if (itemSize === 1) {
        for (var y = 0; y < height; ++y) {
            for (var x = 0; x < width; ++x) {
                data[(y * width + x) * 4 + 3] = array[y * width + x];
            }
        }
    }
    else if (itemSize === 4) {
        data.set(array);
    }
    else {
        console.warn("itemSize '" + itemSize + "' not supported");
    }
    return printImageData(new ImageData(data, width, height), scale);
}
exports.printTextureImage = printTextureImage;
function printImageData(imageData, scale, pixelated) {
    if (scale === void 0) { scale = 1; }
    if (pixelated === void 0) { pixelated = false; }
    var canvas = document.createElement('canvas');
    canvas.width = imageData.width;
    canvas.height = imageData.height;
    var ctx = canvas.getContext('2d');
    if (!ctx)
        throw new Error('Could not create canvas 2d context');
    ctx.putImageData(imageData, 0, 0);
    canvas.toBlob(function (imgBlob) {
        var objectURL = window.URL.createObjectURL(imgBlob);
        var img = document.createElement('img');
        img.src = objectURL;
        img.style.width = imageData.width * scale + 'px';
        img.style.height = imageData.height * scale + 'px';
        if (pixelated)
            img.style.imageRendering = 'pixelated'; // supported only in Chrome
        img.style.position = 'absolute';
        img.style.top = '0px';
        img.style.left = '0px';
        img.style.border = 'solid grey';
        document.body.appendChild(img);
    }, 'image/png');
}
exports.printImageData = printImageData;
//
var v = linear_algebra_1.Vec3.zero();
var boundaryHelper = new boundary_helper_1.BoundaryHelper();
function calculateInvariantBoundingSphere(position, positionCount) {
    boundaryHelper.reset(0);
    for (var i = 0, _i = positionCount * 3; i < _i; i += 3) {
        linear_algebra_1.Vec3.fromArray(v, position, i);
        boundaryHelper.boundaryStep(v, 0);
    }
    boundaryHelper.finishBoundaryStep();
    for (var i = 0, _i = positionCount * 3; i < _i; i += 3) {
        linear_algebra_1.Vec3.fromArray(v, position, i);
        boundaryHelper.extendStep(v, 0);
    }
    return boundaryHelper.getSphere();
}
exports.calculateInvariantBoundingSphere = calculateInvariantBoundingSphere;
function calculateTransformBoundingSphere(invariantBoundingSphere, transform, transformCount) {
    var center = invariantBoundingSphere.center, radius = invariantBoundingSphere.radius;
    boundaryHelper.reset(0);
    for (var i = 0, _i = transformCount; i < _i; ++i) {
        linear_algebra_1.Vec3.transformMat4Offset(v, center, transform, 0, 0, i * 16);
        boundaryHelper.boundaryStep(v, radius);
    }
    boundaryHelper.finishBoundaryStep();
    for (var i = 0, _i = transformCount; i < _i; ++i) {
        linear_algebra_1.Vec3.transformMat4Offset(v, center, transform, 0, 0, i * 16);
        boundaryHelper.extendStep(v, radius);
    }
    return boundaryHelper.getSphere();
}
exports.calculateTransformBoundingSphere = calculateTransformBoundingSphere;
function calculateBoundingSphere(position, positionCount, transform, transformCount, padding) {
    if (padding === void 0) { padding = 0; }
    var invariantBoundingSphere = calculateInvariantBoundingSphere(position, positionCount);
    var boundingSphere = calculateTransformBoundingSphere(invariantBoundingSphere, transform, transformCount);
    geometry_1.Sphere3D.expand(boundingSphere, boundingSphere, padding);
    geometry_1.Sphere3D.expand(invariantBoundingSphere, invariantBoundingSphere, padding);
    return { boundingSphere: boundingSphere, invariantBoundingSphere: invariantBoundingSphere };
}
exports.calculateBoundingSphere = calculateBoundingSphere;
