"use strict";
/**
 * Copyright (c) 2018-2019 mol* contributors, licensed under MIT, See LICENSE file for more info.
 *
 * @author Alexander Rose <alexander.rose@weirdbyte.de>
 */
var __assign = (this && this.__assign) || function () {
    __assign = Object.assign || function(t) {
        for (var s, i = 1, n = arguments.length; i < n; i++) {
            s = arguments[i];
            for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p))
                t[p] = s[p];
        }
        return t;
    };
    return __assign.apply(this, arguments);
};
Object.defineProperty(exports, "__esModule", { value: true });
var util_1 = require("../mol-canvas3d/camera/util");
var linear_algebra_1 = require("../mol-math/linear-algebra");
var color_1 = require("../mol-util/color");
var mol_util_1 = require("../mol-util");
var param_definition_1 = require("../mol-util/param-definition");
var object_1 = require("../mol-util/object");
exports.RendererParams = {
    backgroundColor: param_definition_1.ParamDefinition.Color(color_1.Color(0x000000)),
    pickingAlphaThreshold: param_definition_1.ParamDefinition.Numeric(0.5, { min: 0.0, max: 1.0, step: 0.01 }, { description: 'The minimum opacity value needed for an object to be pickable.' }),
    lightIntensity: param_definition_1.ParamDefinition.Numeric(0.6, { min: 0.0, max: 1.0, step: 0.01 }),
    ambientIntensity: param_definition_1.ParamDefinition.Numeric(0.4, { min: 0.0, max: 1.0, step: 0.01 }),
    metalness: param_definition_1.ParamDefinition.Numeric(0.0, { min: 0.0, max: 1.0, step: 0.01 }),
    roughness: param_definition_1.ParamDefinition.Numeric(0.4, { min: 0.0, max: 1.0, step: 0.01 }),
    reflectivity: param_definition_1.ParamDefinition.Numeric(0.5, { min: 0.0, max: 1.0, step: 0.01 }),
};
var Renderer;
(function (Renderer) {
    function create(ctx, camera, props) {
        if (props === void 0) { props = {}; }
        var gl = ctx.gl, state = ctx.state, stats = ctx.stats;
        var p = object_1.deepClone(__assign({}, param_definition_1.ParamDefinition.getDefaultValues(exports.RendererParams), props));
        var viewport = util_1.Viewport();
        var bgColor = color_1.Color.toVec3Normalized(linear_algebra_1.Vec3(), p.backgroundColor);
        var view = linear_algebra_1.Mat4.clone(camera.view);
        var invView = linear_algebra_1.Mat4.invert(linear_algebra_1.Mat4.identity(), view);
        var modelView = linear_algebra_1.Mat4.clone(camera.view);
        var invModelView = linear_algebra_1.Mat4.invert(linear_algebra_1.Mat4.identity(), modelView);
        var invProjection = linear_algebra_1.Mat4.invert(linear_algebra_1.Mat4.identity(), camera.projection);
        var modelViewProjection = linear_algebra_1.Mat4.mul(linear_algebra_1.Mat4.identity(), modelView, camera.projection);
        var invModelViewProjection = linear_algebra_1.Mat4.invert(linear_algebra_1.Mat4.identity(), modelViewProjection);
        var globalUniforms = {
            uModel: mol_util_1.ValueCell.create(linear_algebra_1.Mat4.identity()),
            uView: mol_util_1.ValueCell.create(camera.view),
            uInvView: mol_util_1.ValueCell.create(invView),
            uModelView: mol_util_1.ValueCell.create(modelView),
            uInvModelView: mol_util_1.ValueCell.create(invModelView),
            uInvProjection: mol_util_1.ValueCell.create(invProjection),
            uProjection: mol_util_1.ValueCell.create(linear_algebra_1.Mat4.clone(camera.projection)),
            uModelViewProjection: mol_util_1.ValueCell.create(modelViewProjection),
            uInvModelViewProjection: mol_util_1.ValueCell.create(invModelViewProjection),
            uIsOrtho: mol_util_1.ValueCell.create(camera.state.mode === 'orthographic' ? 1 : 0),
            uPixelRatio: mol_util_1.ValueCell.create(ctx.pixelRatio),
            uViewportHeight: mol_util_1.ValueCell.create(viewport.height),
            uViewport: mol_util_1.ValueCell.create(util_1.Viewport.toVec4(linear_algebra_1.Vec4(), viewport)),
            uLightIntensity: mol_util_1.ValueCell.create(p.lightIntensity),
            uAmbientIntensity: mol_util_1.ValueCell.create(p.ambientIntensity),
            uMetalness: mol_util_1.ValueCell.create(p.metalness),
            uRoughness: mol_util_1.ValueCell.create(p.roughness),
            uReflectivity: mol_util_1.ValueCell.create(p.reflectivity),
            uCameraPosition: mol_util_1.ValueCell.create(linear_algebra_1.Vec3.clone(camera.state.position)),
            uNear: mol_util_1.ValueCell.create(camera.state.near),
            uFar: mol_util_1.ValueCell.create(camera.state.far),
            uFogNear: mol_util_1.ValueCell.create(camera.state.fogNear),
            uFogFar: mol_util_1.ValueCell.create(camera.state.fogFar),
            uFogColor: mol_util_1.ValueCell.create(bgColor),
            uPickingAlphaThreshold: mol_util_1.ValueCell.create(p.pickingAlphaThreshold),
        };
        var globalUniformList = Object.entries(globalUniforms);
        var globalUniformsNeedUpdate = true;
        var renderObject = function (r, variant) {
            var program = r.getProgram(variant);
            if (r.state.visible) {
                if (state.currentProgramId !== program.id) {
                    // console.log('new program')
                    globalUniformsNeedUpdate = true;
                    program.use();
                }
                if (globalUniformsNeedUpdate) {
                    // console.log('globalUniformsNeedUpdate')
                    program.setUniforms(globalUniformList);
                    globalUniformsNeedUpdate = false;
                }
                if (r.values.dDoubleSided) {
                    if (r.values.dDoubleSided.ref.value) {
                        state.disable(gl.CULL_FACE);
                    }
                    else {
                        state.enable(gl.CULL_FACE);
                    }
                }
                else {
                    // webgl default
                    state.disable(gl.CULL_FACE);
                }
                if (r.values.dFlipSided) {
                    if (r.values.dFlipSided.ref.value) {
                        state.frontFace(gl.CW);
                        state.cullFace(gl.FRONT);
                    }
                    else {
                        state.frontFace(gl.CCW);
                        state.cullFace(gl.BACK);
                    }
                }
                else {
                    // webgl default
                    state.frontFace(gl.CCW);
                    state.cullFace(gl.BACK);
                }
                r.render(variant);
            }
        };
        var render = function (scene, variant, clear) {
            mol_util_1.ValueCell.update(globalUniforms.uModel, scene.view);
            mol_util_1.ValueCell.update(globalUniforms.uView, camera.view);
            mol_util_1.ValueCell.update(globalUniforms.uInvView, linear_algebra_1.Mat4.invert(invView, camera.view));
            mol_util_1.ValueCell.update(globalUniforms.uModelView, linear_algebra_1.Mat4.mul(modelView, scene.view, camera.view));
            mol_util_1.ValueCell.update(globalUniforms.uInvModelView, linear_algebra_1.Mat4.invert(invModelView, modelView));
            mol_util_1.ValueCell.update(globalUniforms.uProjection, camera.projection);
            mol_util_1.ValueCell.update(globalUniforms.uInvProjection, linear_algebra_1.Mat4.invert(invProjection, camera.projection));
            mol_util_1.ValueCell.update(globalUniforms.uModelViewProjection, linear_algebra_1.Mat4.mul(modelViewProjection, modelView, camera.projection));
            mol_util_1.ValueCell.update(globalUniforms.uInvModelViewProjection, linear_algebra_1.Mat4.invert(invModelViewProjection, modelViewProjection));
            mol_util_1.ValueCell.update(globalUniforms.uIsOrtho, camera.state.mode === 'orthographic' ? 1 : 0);
            mol_util_1.ValueCell.update(globalUniforms.uCameraPosition, camera.state.position);
            mol_util_1.ValueCell.update(globalUniforms.uFar, camera.state.far);
            mol_util_1.ValueCell.update(globalUniforms.uNear, camera.state.near);
            mol_util_1.ValueCell.update(globalUniforms.uFogFar, camera.state.fogFar);
            mol_util_1.ValueCell.update(globalUniforms.uFogNear, camera.state.fogNear);
            globalUniformsNeedUpdate = true;
            state.currentRenderItemId = -1;
            var renderables = scene.renderables;
            state.disable(gl.SCISSOR_TEST);
            state.disable(gl.BLEND);
            state.depthMask(true);
            state.colorMask(true, true, true, true);
            state.enable(gl.DEPTH_TEST);
            if (clear) {
                if (variant === 'color') {
                    state.clearColor(bgColor[0], bgColor[1], bgColor[2], 1.0);
                }
                else {
                    state.clearColor(1, 1, 1, 1);
                }
                gl.clear(gl.COLOR_BUFFER_BIT | gl.DEPTH_BUFFER_BIT);
            }
            if (variant === 'color') {
                for (var i = 0, il = renderables.length; i < il; ++i) {
                    var r = renderables[i];
                    if (r.state.opaque)
                        renderObject(r, variant);
                }
                state.blendFunc(gl.SRC_ALPHA, gl.ONE_MINUS_SRC_ALPHA);
                state.enable(gl.BLEND);
                for (var i = 0, il = renderables.length; i < il; ++i) {
                    var r = renderables[i];
                    state.depthMask(r.values.uAlpha.ref.value === 1.0);
                    if (!r.state.opaque)
                        renderObject(r, variant);
                }
            }
            else { // picking & depth
                for (var i = 0, il = renderables.length; i < il; ++i) {
                    renderObject(renderables[i], variant);
                }
            }
            gl.finish();
        };
        return {
            clear: function () {
                state.depthMask(true);
                state.colorMask(true, true, true, true);
                state.clearColor(bgColor[0], bgColor[1], bgColor[2], 1.0);
                gl.clear(gl.COLOR_BUFFER_BIT | gl.DEPTH_BUFFER_BIT);
            },
            render: render,
            setProps: function (props) {
                if (props.pickingAlphaThreshold !== undefined && props.pickingAlphaThreshold !== p.pickingAlphaThreshold) {
                    p.pickingAlphaThreshold = props.pickingAlphaThreshold;
                    mol_util_1.ValueCell.update(globalUniforms.uPickingAlphaThreshold, p.pickingAlphaThreshold);
                }
                if (props.backgroundColor !== undefined && props.backgroundColor !== p.backgroundColor) {
                    p.backgroundColor = props.backgroundColor;
                    color_1.Color.toVec3Normalized(bgColor, p.backgroundColor);
                    mol_util_1.ValueCell.update(globalUniforms.uFogColor, linear_algebra_1.Vec3.copy(globalUniforms.uFogColor.ref.value, bgColor));
                }
                if (props.lightIntensity !== undefined && props.lightIntensity !== p.lightIntensity) {
                    p.lightIntensity = props.lightIntensity;
                    mol_util_1.ValueCell.update(globalUniforms.uLightIntensity, p.lightIntensity);
                }
                if (props.ambientIntensity !== undefined && props.ambientIntensity !== p.ambientIntensity) {
                    p.ambientIntensity = props.ambientIntensity;
                    mol_util_1.ValueCell.update(globalUniforms.uAmbientIntensity, p.ambientIntensity);
                }
                if (props.metalness !== undefined && props.metalness !== p.metalness) {
                    p.metalness = props.metalness;
                    mol_util_1.ValueCell.update(globalUniforms.uMetalness, p.metalness);
                }
                if (props.roughness !== undefined && props.roughness !== p.roughness) {
                    p.roughness = props.roughness;
                    mol_util_1.ValueCell.update(globalUniforms.uRoughness, p.roughness);
                }
                if (props.reflectivity !== undefined && props.reflectivity !== p.reflectivity) {
                    p.reflectivity = props.reflectivity;
                    mol_util_1.ValueCell.update(globalUniforms.uReflectivity, p.reflectivity);
                }
            },
            setViewport: function (x, y, width, height) {
                gl.viewport(x, y, width, height);
                if (x !== viewport.x || y !== viewport.y || width !== viewport.width || height !== viewport.height) {
                    util_1.Viewport.set(viewport, x, y, width, height);
                    mol_util_1.ValueCell.update(globalUniforms.uViewportHeight, height);
                    mol_util_1.ValueCell.update(globalUniforms.uViewport, linear_algebra_1.Vec4.set(globalUniforms.uViewport.ref.value, x, y, width, height));
                }
            },
            get props() {
                return p;
            },
            get stats() {
                return {
                    programCount: ctx.programCache.count,
                    shaderCount: ctx.shaderCache.count,
                    bufferCount: stats.bufferCount,
                    framebufferCount: stats.framebufferCount,
                    renderbufferCount: stats.renderbufferCount,
                    textureCount: stats.textureCount,
                    vaoCount: stats.vaoCount,
                    drawCount: stats.drawCount,
                    instanceCount: stats.instanceCount,
                    instancedDrawCount: stats.instancedDrawCount,
                };
            },
            dispose: function () {
                // TODO
            }
        };
    }
    Renderer.create = create;
})(Renderer || (Renderer = {}));
exports.default = Renderer;
