"use strict";
/**
 * Copyright (c) 2018 mol* contributors, licensed under MIT, See LICENSE file for more info.
 *
 * @author Alexander Rose <alexander.rose@weirdbyte.de>
 */
Object.defineProperty(exports, "__esModule", { value: true });
var render_object_1 = require("./render-object");
var object3d_1 = require("./object3d");
var geometry_1 = require("../mol-math/geometry");
var linear_algebra_1 = require("../mol-math/linear-algebra");
var boundary_helper_1 = require("../mol-math/geometry/boundary-helper");
var boundaryHelper = new boundary_helper_1.BoundaryHelper();
function calculateBoundingSphere(renderables, boundingSphere) {
    boundaryHelper.reset(0.1);
    for (var i = 0, il = renderables.length; i < il; ++i) {
        var boundingSphere_1 = renderables[i].values.boundingSphere.ref.value;
        if (!boundingSphere_1.radius)
            continue;
        boundaryHelper.boundaryStep(boundingSphere_1.center, boundingSphere_1.radius);
    }
    boundaryHelper.finishBoundaryStep();
    for (var i = 0, il = renderables.length; i < il; ++i) {
        var boundingSphere_2 = renderables[i].values.boundingSphere.ref.value;
        if (!boundingSphere_2.radius)
            continue;
        boundaryHelper.extendStep(boundingSphere_2.center, boundingSphere_2.radius);
    }
    linear_algebra_1.Vec3.copy(boundingSphere.center, boundaryHelper.center);
    boundingSphere.radius = boundaryHelper.radius;
    return boundingSphere;
}
function renderableSort(a, b) {
    var drawProgramIdA = a.getProgram('color').id;
    var drawProgramIdB = b.getProgram('color').id;
    var materialIdA = a.materialId;
    var materialIdB = b.materialId;
    var zA = a.values.boundingSphere.ref.value.center[2];
    var zB = b.values.boundingSphere.ref.value.center[2];
    if (drawProgramIdA !== drawProgramIdB) {
        return drawProgramIdA - drawProgramIdB; // sort by program id to minimize gl state changes
    }
    else if (materialIdA !== materialIdB) {
        return materialIdA - materialIdB; // sort by material id to minimize gl state changes
    }
    else if (zA !== zB) {
        return a.state.opaque
            ? zA - zB // when opaque, draw closer elements first to minimize overdraw
            : zB - zA; // when transparent, draw elements last to maximize partial visibility
    }
    else {
        return a.id - b.id;
    }
}
var Scene;
(function (Scene) {
    function create(ctx) {
        var renderableMap = new Map();
        var renderables = [];
        var boundingSphere = geometry_1.Sphere3D.zero();
        var boundingSphereDirty = true;
        var object3d = object3d_1.Object3D.create();
        return {
            get view() { return object3d.view; },
            get position() { return object3d.position; },
            get direction() { return object3d.direction; },
            get up() { return object3d.up; },
            update: function (objects, keepBoundingSphere) {
                object3d_1.Object3D.update(object3d);
                if (objects) {
                    for (var i = 0, il = objects.length; i < il; ++i) {
                        var o = renderableMap.get(objects[i]);
                        if (!o)
                            continue;
                        o.update();
                    }
                }
                else {
                    for (var i = 0, il = renderables.length; i < il; ++i) {
                        renderables[i].update();
                    }
                }
                if (!keepBoundingSphere)
                    boundingSphereDirty = true;
            },
            add: function (o) {
                if (!renderableMap.has(o)) {
                    var renderable = render_object_1.createRenderable(ctx, o);
                    renderables.push(renderable);
                    renderables.sort(renderableSort);
                    renderableMap.set(o, renderable);
                    boundingSphereDirty = true;
                    return renderable;
                }
                else {
                    console.warn("RenderObject with id '" + o.id + "' already present");
                    return renderableMap.get(o);
                }
            },
            remove: function (o) {
                var renderable = renderableMap.get(o);
                if (renderable) {
                    renderable.dispose();
                    renderables.splice(renderables.indexOf(renderable), 1);
                    renderables.sort(renderableSort);
                    renderableMap.delete(o);
                    boundingSphereDirty = true;
                }
            },
            has: function (o) {
                return renderableMap.has(o);
            },
            clear: function () {
                for (var i = 0, il = renderables.length; i < il; ++i) {
                    renderables[i].dispose();
                }
                renderables.length = 0;
                renderableMap.clear();
                boundingSphereDirty = true;
            },
            forEach: function (callbackFn) {
                renderableMap.forEach(callbackFn);
            },
            get count() {
                return renderables.length;
            },
            renderables: renderables,
            get boundingSphere() {
                if (boundingSphereDirty)
                    calculateBoundingSphere(renderables, boundingSphere);
                boundingSphereDirty = false;
                return boundingSphere;
            }
        };
    }
    Scene.create = create;
})(Scene || (Scene = {}));
exports.default = Scene;
