"use strict";
/**
 * Copyright (c) 2018-2019 mol* contributors, licensed under MIT, See LICENSE file for more info.
 *
 * @author Alexander Rose <alexander.rose@weirdbyte.de>
 */
var __assign = (this && this.__assign) || function () {
    __assign = Object.assign || function(t) {
        for (var s, i = 1, n = arguments.length; i < n; i++) {
            s = arguments[i];
            for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p))
                t[p] = s[p];
        }
        return t;
    };
    return __assign.apply(this, arguments);
};
Object.defineProperty(exports, "__esModule", { value: true });
var id_factory_1 = require("../../mol-util/id-factory");
var getNextBufferId = id_factory_1.idFactory();
function getUsageHint(ctx, usageHint) {
    var gl = ctx.gl;
    switch (usageHint) {
        case 'static': return gl.STATIC_DRAW;
        case 'dynamic': return gl.DYNAMIC_DRAW;
        case 'stream': return gl.STREAM_DRAW;
    }
}
exports.getUsageHint = getUsageHint;
function getDataType(ctx, dataType) {
    var gl = ctx.gl;
    switch (dataType) {
        case 'uint8': return gl.UNSIGNED_BYTE;
        case 'int8': return gl.BYTE;
        case 'uint16': return gl.UNSIGNED_SHORT;
        case 'int16': return gl.SHORT;
        case 'uint32': return gl.UNSIGNED_INT;
        case 'int32': return gl.INT;
        case 'float32': return gl.FLOAT;
    }
}
exports.getDataType = getDataType;
function dataTypeFromArray(ctx, array) {
    var gl = ctx.gl;
    if (array instanceof Uint8Array) {
        return gl.UNSIGNED_BYTE;
    }
    else if (array instanceof Int8Array) {
        return gl.BYTE;
    }
    else if (array instanceof Uint16Array) {
        return gl.UNSIGNED_SHORT;
    }
    else if (array instanceof Int16Array) {
        return gl.SHORT;
    }
    else if (array instanceof Uint32Array) {
        return gl.UNSIGNED_INT;
    }
    else if (array instanceof Int32Array) {
        return gl.INT;
    }
    else if (array instanceof Float32Array) {
        return gl.FLOAT;
    }
    else {
        throw new Error('Should nevver happen');
    }
}
function getBufferType(ctx, bufferType) {
    var gl = ctx.gl;
    switch (bufferType) {
        case 'attribute': return gl.ARRAY_BUFFER;
        case 'elements': return gl.ELEMENT_ARRAY_BUFFER;
        case 'uniform': return gl.UNIFORM_BUFFER;
    }
}
exports.getBufferType = getBufferType;
function createBuffer(ctx, array, usageHint, bufferType) {
    var gl = ctx.gl, stats = ctx.stats;
    var _buffer = gl.createBuffer();
    if (_buffer === null) {
        throw new Error('Could not create WebGL buffer');
    }
    var _usageHint = getUsageHint(ctx, usageHint);
    var _bufferType = getBufferType(ctx, bufferType);
    var _dataType = dataTypeFromArray(ctx, array);
    var _bpe = array.BYTES_PER_ELEMENT;
    var _length = array.length;
    function updateData(array) {
        gl.bindBuffer(_bufferType, _buffer);
        gl.bufferData(_bufferType, array, _usageHint);
    }
    updateData(array);
    var destroyed = false;
    stats.bufferCount += 1;
    return {
        id: getNextBufferId(),
        _buffer: _buffer,
        _usageHint: _usageHint,
        _bufferType: _bufferType,
        _dataType: _dataType,
        _bpe: _bpe,
        length: _length,
        updateData: updateData,
        updateSubData: function (array, offset, count) {
            gl.bindBuffer(_bufferType, _buffer);
            gl.bufferSubData(_bufferType, offset * _bpe, array.subarray(offset, offset + count));
        },
        destroy: function () {
            if (destroyed)
                return;
            gl.deleteBuffer(_buffer);
            destroyed = true;
            stats.bufferCount -= 1;
        }
    };
}
exports.createBuffer = createBuffer;
function getAttribType(gl, kind, itemSize) {
    switch (kind) {
        case 'int32':
            switch (itemSize) {
                case 1: return gl.INT;
                case 2: return gl.INT_VEC2;
                case 3: return gl.INT_VEC3;
                case 4: return gl.INT_VEC4;
            }
            break;
        case 'float32':
            switch (itemSize) {
                case 1: return gl.FLOAT;
                case 2: return gl.FLOAT_VEC2;
                case 3: return gl.FLOAT_VEC3;
                case 4: return gl.FLOAT_VEC4;
                case 16: return gl.FLOAT_MAT4;
            }
            break;
    }
    throw new Error("unknown attribute type for kind '" + kind + "' and itemSize '" + itemSize + "'");
}
exports.getAttribType = getAttribType;
function createAttributeBuffer(ctx, array, itemSize, divisor, usageHint) {
    if (usageHint === void 0) { usageHint = 'dynamic'; }
    var gl = ctx.gl;
    var instancedArrays = ctx.extensions.instancedArrays;
    var buffer = createBuffer(ctx, array, usageHint, 'attribute');
    var _buffer = buffer._buffer, _bufferType = buffer._bufferType, _dataType = buffer._dataType, _bpe = buffer._bpe;
    return __assign({}, buffer, { bind: function (location) {
            gl.bindBuffer(_bufferType, _buffer);
            if (itemSize === 16) {
                for (var i = 0; i < 4; ++i) {
                    gl.enableVertexAttribArray(location + i);
                    gl.vertexAttribPointer(location + i, 4, _dataType, false, 4 * 4 * _bpe, i * 4 * _bpe);
                    instancedArrays.vertexAttribDivisor(location + i, divisor);
                }
            }
            else {
                gl.enableVertexAttribArray(location);
                gl.vertexAttribPointer(location, itemSize, _dataType, false, 0, 0);
                instancedArrays.vertexAttribDivisor(location, divisor);
            }
        } });
}
exports.createAttributeBuffer = createAttributeBuffer;
function createAttributeBuffers(ctx, schema, values) {
    var buffers = [];
    Object.keys(schema).forEach(function (k) {
        var spec = schema[k];
        if (spec.type === 'attribute') {
            buffers[buffers.length] = [k, createAttributeBuffer(ctx, values[k].ref.value, spec.itemSize, spec.divisor)];
        }
    });
    return buffers;
}
exports.createAttributeBuffers = createAttributeBuffers;
function createElementsBuffer(ctx, array, usageHint) {
    if (usageHint === void 0) { usageHint = 'static'; }
    var gl = ctx.gl;
    var buffer = createBuffer(ctx, array, usageHint, 'elements');
    var _buffer = buffer._buffer;
    return __assign({}, buffer, { bind: function () {
            gl.bindBuffer(gl.ELEMENT_ARRAY_BUFFER, _buffer);
        } });
}
exports.createElementsBuffer = createElementsBuffer;
