"use strict";
/**
 * Copyright (c) 2018-2019 mol* contributors, licensed under MIT, See LICENSE file for more info.
 *
 * @author Alexander Rose <alexander.rose@weirdbyte.de>
 */
Object.defineProperty(exports, "__esModule", { value: true });
var shader_code_1 = require("../shader-code");
var uniform_1 = require("./uniform");
var buffer_1 = require("./buffer");
var reference_cache_1 = require("../../mol-util/reference-cache");
var id_factory_1 = require("../../mol-util/id-factory");
var util_1 = require("../../mol-data/util");
var debug_1 = require("../../mol-util/debug");
var getNextProgramId = id_factory_1.idFactory();
function getLocations(gl, program, schema) {
    var locations = {};
    Object.keys(schema).forEach(function (k) {
        var spec = schema[k];
        if (spec.type === 'attribute') {
            var loc = gl.getAttribLocation(program, k);
            // if (loc === -1) console.info(`Could not get attribute location for '${k}'`)
            locations[k] = loc;
        }
        else if (spec.type === 'uniform' || spec.type === 'texture') {
            var loc = gl.getUniformLocation(program, k);
            // if (loc === null) console.info(`Could not get uniform location for '${k}'`)
            locations[k] = loc;
        }
    });
    return locations;
}
function checkActiveAttributes(gl, program, schema) {
    var attribCount = gl.getProgramParameter(program, gl.ACTIVE_ATTRIBUTES);
    for (var i = 0; i < attribCount; ++i) {
        var info = gl.getActiveAttrib(program, i);
        if (info) {
            var name_1 = info.name, type = info.type;
            if (name_1.startsWith('__activeAttribute')) {
                // name assigned by `gl.shim.ts`, ignore for checks
                continue;
            }
            var spec = schema[name_1];
            if (spec === undefined) {
                throw new Error("missing 'uniform' or 'texture' with name '" + name_1 + "' in schema");
            }
            if (spec.type !== 'attribute') {
                throw new Error("'" + name_1 + "' must be of type 'attribute' but is '" + spec.type + "'");
            }
            var attribType = buffer_1.getAttribType(gl, spec.kind, spec.itemSize);
            if (attribType !== type) {
                throw new Error("unexpected attribute type for " + name_1);
            }
        }
    }
}
function checkActiveUniforms(gl, program, schema) {
    var attribCount = gl.getProgramParameter(program, gl.ACTIVE_UNIFORMS);
    for (var i = 0; i < attribCount; ++i) {
        var info = gl.getActiveUniform(program, i);
        if (info) {
            var name_2 = info.name, type = info.type;
            if (name_2.startsWith('__activeUniform')) {
                // name assigned by `gl.shim.ts`, ignore for checks
                continue;
            }
            var spec = schema[name_2];
            if (spec === undefined) {
                throw new Error("missing 'uniform' or 'texture' with name '" + name_2 + "' in schema");
            }
            if (spec.type === 'uniform') {
                var uniformType = uniform_1.getUniformType(gl, spec.kind);
                if (uniformType !== type) {
                    throw new Error("unexpected uniform type for " + name_2);
                }
            }
            else if (spec.type === 'texture') {
                if (spec.kind === 'image-float32' || spec.kind === 'image-uint8') {
                    if (type !== gl.SAMPLER_2D) {
                        throw new Error("unexpected sampler type for '" + name_2 + "'");
                    }
                }
                else if (spec.kind === 'volume-float32' || spec.kind === 'volume-uint8') {
                    if (type !== gl.SAMPLER_3D) {
                        throw new Error("unexpected sampler type for '" + name_2 + "'");
                    }
                }
                else {
                    // TODO
                }
            }
            else {
                throw new Error("'" + name_2 + "' must be of type 'uniform' or 'texture' but is '" + spec.type + "'");
            }
        }
    }
}
function createProgram(gl, state, extensions, shaderCache, props) {
    var defineValues = props.defineValues, _shaderCode = props.shaderCode, schema = props.schema;
    var program = gl.createProgram();
    if (program === null) {
        throw new Error('Could not create WebGL program');
    }
    var programId = getNextProgramId();
    var shaderCode = shader_code_1.addShaderDefines(gl, extensions, defineValues, _shaderCode);
    var vertShaderRef = shaderCache.get({ type: 'vert', source: shaderCode.vert });
    var fragShaderRef = shaderCache.get({ type: 'frag', source: shaderCode.frag });
    vertShaderRef.value.attach(program);
    fragShaderRef.value.attach(program);
    gl.linkProgram(program);
    if (debug_1.isDebugMode) {
        // no-op in FF on Mac, see https://bugzilla.mozilla.org/show_bug.cgi?id=1284425
        // gl.validateProgram(program)
        if (!gl.getProgramParameter(program, gl.LINK_STATUS)) {
            throw new Error("Could not compile WebGL program. \n\n" + gl.getProgramInfoLog(program));
        }
    }
    var locations = getLocations(gl, program, schema);
    var uniformSetters = uniform_1.getUniformSetters(schema);
    if (debug_1.isDebugMode) {
        checkActiveAttributes(gl, program, schema);
        checkActiveUniforms(gl, program, schema);
    }
    var destroyed = false;
    return {
        id: programId,
        use: function () {
            // console.log('use', programId)
            state.currentProgramId = programId;
            gl.useProgram(program);
        },
        setUniforms: function (uniformValues) {
            for (var i = 0, il = uniformValues.length; i < il; ++i) {
                var _a = uniformValues[i], k = _a[0], v = _a[1];
                if (v) {
                    var l = locations[k];
                    if (l !== null)
                        uniformSetters[k](gl, l, v.ref.value);
                }
            }
        },
        bindAttributes: function (attribueBuffers) {
            for (var i = 0, il = attribueBuffers.length; i < il; ++i) {
                var _a = attribueBuffers[i], k = _a[0], buffer = _a[1];
                var l = locations[k];
                if (l !== -1)
                    buffer.bind(l);
            }
        },
        bindTextures: function (textures) {
            for (var i = 0, il = textures.length; i < il; ++i) {
                var _a = textures[i], k = _a[0], texture = _a[1];
                var l = locations[k];
                if (l !== null) {
                    // TODO if the order and count of textures in a material can be made invariant
                    //      bind needs to be called only when the material changes
                    texture.bind(i);
                    uniformSetters[k](gl, l, i);
                }
            }
        },
        destroy: function () {
            if (destroyed)
                return;
            vertShaderRef.free();
            fragShaderRef.free();
            gl.deleteProgram(program);
            destroyed = true;
        }
    };
}
exports.createProgram = createProgram;
function defineValueHash(v) {
    return typeof v === 'boolean' ? (v ? 1 : 0) :
        typeof v === 'number' ? v : util_1.hashString(v);
}
function createProgramCache(gl, state, extensions, shaderCache) {
    return reference_cache_1.createReferenceCache(function (props) {
        var array = [props.shaderCode.id];
        Object.keys(props.defineValues).forEach(function (k) {
            var v = props.defineValues[k].ref.value;
            array.push(util_1.hashString(k), defineValueHash(v));
        });
        return util_1.hashFnv32a(array).toString();
    }, function (props) { return createProgram(gl, state, extensions, shaderCache, props); }, function (program) { program.destroy(); });
}
exports.createProgramCache = createProgramCache;
