"use strict";
/**
 * Copyright (c) 2018-2019 mol* contributors, licensed under MIT, See LICENSE file for more info.
 *
 * @author Alexander Rose <alexander.rose@weirdbyte.de>
 */
var __assign = (this && this.__assign) || function () {
    __assign = Object.assign || function(t) {
        for (var s, i = 1, n = arguments.length; i < n; i++) {
            s = arguments[i];
            for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p))
                t[p] = s[p];
        }
        return t;
    };
    return __assign.apply(this, arguments);
};
Object.defineProperty(exports, "__esModule", { value: true });
var buffer_1 = require("./buffer");
var texture_1 = require("./texture");
var context_1 = require("./context");
var schema_1 = require("../renderable/schema");
var id_factory_1 = require("../../mol-util/id-factory");
var vertex_array_1 = require("./vertex-array");
var mol_util_1 = require("../../mol-util");
var framebuffer_1 = require("./framebuffer");
var debug_1 = require("../../mol-util/debug");
var getNextRenderItemId = id_factory_1.idFactory();
function getDrawMode(ctx, drawMode) {
    var gl = ctx.gl;
    switch (drawMode) {
        case 'points': return gl.POINTS;
        case 'lines': return gl.LINES;
        case 'line-strip': return gl.LINE_STRIP;
        case 'line-loop': return gl.LINE_LOOP;
        case 'triangles': return gl.TRIANGLES;
        case 'triangle-strip': return gl.TRIANGLE_STRIP;
        case 'triangle-fan': return gl.TRIANGLE_FAN;
    }
}
exports.getDrawMode = getDrawMode;
//
var GraphicsRenderVariantDefines = {
    'color': {},
    'pickObject': { dColorType: mol_util_1.ValueCell.create('objectPicking') },
    'pickInstance': { dColorType: mol_util_1.ValueCell.create('instancePicking') },
    'pickGroup': { dColorType: mol_util_1.ValueCell.create('groupPicking') },
    'depth': { dColorType: mol_util_1.ValueCell.create('depth') }
};
var ComputeRenderVariantDefines = {
    'compute': {},
};
function createValueChanges() {
    return {
        attributes: false,
        defines: false,
        elements: false,
        textures: false,
    };
}
function resetValueChanges(valueChanges) {
    valueChanges.attributes = false;
    valueChanges.defines = false;
    valueChanges.elements = false;
    valueChanges.textures = false;
}
function createGraphicsRenderItem(ctx, drawMode, shaderCode, schema, values, materialId) {
    return createRenderItem(ctx, drawMode, shaderCode, schema, values, materialId, GraphicsRenderVariantDefines);
}
exports.createGraphicsRenderItem = createGraphicsRenderItem;
function createComputeRenderItem(ctx, drawMode, shaderCode, schema, values, materialId) {
    if (materialId === void 0) { materialId = -1; }
    return createRenderItem(ctx, drawMode, shaderCode, schema, values, materialId, ComputeRenderVariantDefines);
}
exports.createComputeRenderItem = createComputeRenderItem;
/**
 * Creates a render item
 *
 * - assumes that `values.drawCount` and `values.instanceCount` exist
 */
function createRenderItem(ctx, drawMode, shaderCode, schema, values, materialId, renderVariantDefines) {
    var id = getNextRenderItemId();
    var stats = ctx.stats, state = ctx.state, programCache = ctx.programCache;
    var _a = ctx.extensions, instancedArrays = _a.instancedArrays, vertexArrayObject = _a.vertexArrayObject;
    var _b = schema_1.splitValues(schema, values), attributeValues = _b.attributeValues, defineValues = _b.defineValues, textureValues = _b.textureValues, uniformValues = _b.uniformValues, materialUniformValues = _b.materialUniformValues;
    var uniformValueEntries = Object.entries(uniformValues);
    var materialUniformValueEntries = Object.entries(materialUniformValues);
    var defineValueEntries = Object.entries(defineValues);
    var versions = schema_1.getValueVersions(values);
    var glDrawMode = getDrawMode(ctx, drawMode);
    var programs = {};
    Object.keys(renderVariantDefines).forEach(function (k) {
        var variantDefineValues = renderVariantDefines[k];
        programs[k] = programCache.get({
            defineValues: __assign({}, defineValues, variantDefineValues),
            shaderCode: shaderCode,
            schema: schema
        });
    });
    var textures = texture_1.createTextures(ctx, schema, textureValues);
    var attributeBuffers = buffer_1.createAttributeBuffers(ctx, schema, attributeValues);
    var elementsBuffer;
    var elements = values.elements;
    if (elements && elements.ref.value) {
        elementsBuffer = buffer_1.createElementsBuffer(ctx, elements.ref.value);
    }
    var vertexArrays = {};
    Object.keys(renderVariantDefines).forEach(function (k) {
        vertexArrays[k] = vertex_array_1.createVertexArray(ctx, programs[k].value, attributeBuffers, elementsBuffer);
    });
    var drawCount = values.drawCount.ref.value;
    var instanceCount = values.instanceCount.ref.value;
    stats.drawCount += drawCount;
    stats.instanceCount += instanceCount;
    stats.instancedDrawCount += instanceCount * drawCount;
    var valueChanges = createValueChanges();
    var destroyed = false;
    var currentProgramId = -1;
    return {
        id: id,
        materialId: materialId,
        getProgram: function (variant) { return programs[variant].value; },
        render: function (variant) {
            if (drawCount === 0 || instanceCount === 0)
                return;
            var program = programs[variant].value;
            if (program.id === currentProgramId && state.currentRenderItemId === id) {
                program.setUniforms(uniformValueEntries);
                program.bindTextures(textures);
            }
            else {
                var vertexArray = vertexArrays[variant];
                if (program.id !== state.currentProgramId || program.id !== currentProgramId ||
                    materialId === -1 || materialId !== state.currentMaterialId) {
                    // console.log('program.id changed or materialId changed/-1', materialId)
                    if (program.id !== state.currentProgramId)
                        program.use();
                    program.setUniforms(materialUniformValueEntries);
                    state.currentMaterialId = materialId;
                    currentProgramId = program.id;
                }
                program.setUniforms(uniformValueEntries);
                program.bindTextures(textures);
                if (vertexArrayObject && vertexArray) {
                    vertexArrayObject.bindVertexArray(vertexArray);
                    // need to bind elements buffer explicitly since it is not always recorded in the VAO
                    if (elementsBuffer)
                        elementsBuffer.bind();
                }
                else {
                    if (elementsBuffer)
                        elementsBuffer.bind();
                    program.bindAttributes(attributeBuffers);
                }
                state.currentRenderItemId = id;
            }
            if (debug_1.isDebugMode) {
                framebuffer_1.checkFramebufferStatus(ctx.gl);
            }
            if (elementsBuffer) {
                instancedArrays.drawElementsInstanced(glDrawMode, drawCount, elementsBuffer._dataType, 0, instanceCount);
            }
            else {
                instancedArrays.drawArraysInstanced(glDrawMode, 0, drawCount, instanceCount);
            }
            if (debug_1.isDebugMode) {
                try {
                    context_1.checkError(ctx.gl);
                }
                catch (e) {
                    // console.log('shaderCode', shaderCode)
                    // console.log('schema', schema)
                    // console.log('attributeBuffers', attributeBuffers)
                    throw new Error("Error rendering item id " + id + ": '" + e + "'");
                }
            }
        },
        update: function () {
            resetValueChanges(valueChanges);
            for (var i = 0, il = defineValueEntries.length; i < il; ++i) {
                var _a = defineValueEntries[i], k = _a[0], value = _a[1];
                if (value.ref.version !== versions[k]) {
                    // console.log('define version changed', k)
                    valueChanges.defines = true;
                    versions[k] = value.ref.version;
                }
            }
            if (valueChanges.defines) {
                // console.log('some defines changed, need to rebuild programs')
                Object.keys(renderVariantDefines).forEach(function (k) {
                    var variantDefineValues = renderVariantDefines[k];
                    programs[k].free();
                    programs[k] = programCache.get({
                        defineValues: __assign({}, defineValues, variantDefineValues),
                        shaderCode: shaderCode,
                        schema: schema
                    });
                });
            }
            if (values.drawCount.ref.version !== versions.drawCount) {
                // console.log('drawCount version changed')
                stats.drawCount += values.drawCount.ref.value - drawCount;
                stats.instancedDrawCount += instanceCount * values.drawCount.ref.value - instanceCount * drawCount;
                drawCount = values.drawCount.ref.value;
                versions.drawCount = values.drawCount.ref.version;
            }
            if (values.instanceCount.ref.version !== versions.instanceCount) {
                // console.log('instanceCount version changed')
                stats.instanceCount += values.instanceCount.ref.value - instanceCount;
                stats.instancedDrawCount += values.instanceCount.ref.value * drawCount - instanceCount * drawCount;
                instanceCount = values.instanceCount.ref.value;
                versions.instanceCount = values.instanceCount.ref.version;
            }
            for (var i = 0, il = attributeBuffers.length; i < il; ++i) {
                var _b = attributeBuffers[i], k = _b[0], buffer = _b[1];
                var value = attributeValues[k];
                if (value.ref.version !== versions[k]) {
                    if (buffer.length >= value.ref.value.length) {
                        // console.log('attribute array large enough to update', k, value.ref.id, value.ref.version)
                        buffer.updateData(value.ref.value);
                    }
                    else {
                        // console.log('attribute array to small, need to create new attribute', k, value.ref.id, value.ref.version)
                        buffer.destroy();
                        var _c = schema[k], itemSize = _c.itemSize, divisor = _c.divisor;
                        attributeBuffers[i][1] = buffer_1.createAttributeBuffer(ctx, value.ref.value, itemSize, divisor);
                        valueChanges.attributes = true;
                    }
                    versions[k] = value.ref.version;
                }
            }
            if (elementsBuffer && values.elements.ref.version !== versions.elements) {
                if (elementsBuffer.length >= values.elements.ref.value.length) {
                    // console.log('elements array large enough to update', values.elements.ref.id, values.elements.ref.version)
                    elementsBuffer.updateData(values.elements.ref.value);
                }
                else {
                    // console.log('elements array to small, need to create new elements', values.elements.ref.id, values.elements.ref.version)
                    elementsBuffer.destroy();
                    elementsBuffer = buffer_1.createElementsBuffer(ctx, values.elements.ref.value);
                    valueChanges.elements = true;
                }
                versions.elements = values.elements.ref.version;
            }
            if (valueChanges.attributes || valueChanges.defines || valueChanges.elements) {
                // console.log('program/defines or buffers changed, update vaos')
                var vertexArrayObject_1 = ctx.extensions.vertexArrayObject;
                if (vertexArrayObject_1) {
                    Object.keys(renderVariantDefines).forEach(function (k) {
                        vertexArrayObject_1.bindVertexArray(vertexArrays[k]);
                        if (elementsBuffer && (valueChanges.defines || valueChanges.elements)) {
                            elementsBuffer.bind();
                        }
                        if (valueChanges.attributes || valueChanges.defines) {
                            programs[k].value.bindAttributes(attributeBuffers);
                        }
                        vertexArrayObject_1.bindVertexArray(null);
                    });
                }
            }
            for (var i = 0, il = textures.length; i < il; ++i) {
                var _d = textures[i], k = _d[0], texture = _d[1];
                var value = textureValues[k];
                if (value.ref.version !== versions[k]) {
                    // update of textures with kind 'texture' is done externally
                    if (schema[k].kind !== 'texture') {
                        // console.log('texture version changed, uploading image', k)
                        texture.load(value.ref.value);
                        versions[k] = value.ref.version;
                        valueChanges.textures = true;
                    }
                    else {
                        textures[i][1] = value.ref.value;
                    }
                }
            }
            return valueChanges;
        },
        destroy: function () {
            if (!destroyed) {
                Object.keys(renderVariantDefines).forEach(function (k) {
                    programs[k].free();
                    vertex_array_1.deleteVertexArray(ctx, vertexArrays[k]);
                });
                textures.forEach(function (_a) {
                    var k = _a[0], texture = _a[1];
                    // lifetime of textures with kind 'texture' is defined externally
                    if (schema[k].kind !== 'texture') {
                        texture.destroy();
                    }
                });
                attributeBuffers.forEach(function (_a) {
                    var _ = _a[0], buffer = _a[1];
                    return buffer.destroy();
                });
                if (elementsBuffer)
                    elementsBuffer.destroy();
                destroyed = true;
            }
        }
    };
}
exports.createRenderItem = createRenderItem;
