"use strict";
/**
 * Copyright (c) 2018-2019 mol* contributors, licensed under MIT, See LICENSE file for more info.
 *
 * @author Alexander Rose <alexander.rose@weirdbyte.de>
 */
Object.defineProperty(exports, "__esModule", { value: true });
var id_factory_1 = require("../../mol-util/id-factory");
var compat_1 = require("./compat");
var getNextTextureId = id_factory_1.idFactory();
function getTarget(ctx, kind) {
    var gl = ctx.gl;
    switch (kind) {
        case 'image-uint8': return gl.TEXTURE_2D;
        case 'image-float32': return gl.TEXTURE_2D;
        case 'image-depth': return gl.TEXTURE_2D;
    }
    if (compat_1.isWebGL2(gl)) {
        switch (kind) {
            case 'volume-uint8': return gl.TEXTURE_3D;
            case 'volume-float32': return gl.TEXTURE_3D;
        }
    }
    throw new Error("unknown texture kind '" + kind + "'");
}
exports.getTarget = getTarget;
function getFormat(ctx, format, type) {
    var gl = ctx.gl;
    switch (format) {
        case 'alpha':
            if (compat_1.isWebGL2(gl) && type === 'float')
                return gl.RED;
            else
                return gl.ALPHA;
        case 'rgb': return gl.RGB;
        case 'rgba': return gl.RGBA;
        case 'depth': return gl.DEPTH_COMPONENT;
    }
}
exports.getFormat = getFormat;
function getInternalFormat(ctx, format, type) {
    var gl = ctx.gl;
    if (compat_1.isWebGL2(gl)) {
        switch (format) {
            case 'alpha':
                switch (type) {
                    case 'ubyte': return gl.ALPHA;
                    case 'float': return gl.R32F;
                }
            case 'rgb':
                switch (type) {
                    case 'ubyte': return gl.RGB;
                    case 'float': return gl.RGB32F;
                }
            case 'rgba':
                switch (type) {
                    case 'ubyte': return gl.RGBA;
                    case 'float': return gl.RGBA32F;
                }
            case 'depth':
                return gl.DEPTH_COMPONENT16;
        }
    }
    return getFormat(ctx, format, type);
}
exports.getInternalFormat = getInternalFormat;
function getType(ctx, type) {
    var gl = ctx.gl;
    switch (type) {
        case 'ubyte': return gl.UNSIGNED_BYTE;
        case 'ushort': return gl.UNSIGNED_SHORT;
        case 'float': return gl.FLOAT;
    }
}
exports.getType = getType;
function getFilter(ctx, type) {
    var gl = ctx.gl;
    switch (type) {
        case 'nearest': return gl.NEAREST;
        case 'linear': return gl.LINEAR;
    }
}
exports.getFilter = getFilter;
function getAttachment(ctx, attachment) {
    var gl = ctx.gl, extensions = ctx.extensions;
    switch (attachment) {
        case 'depth': return gl.DEPTH_ATTACHMENT;
        case 'stencil': return gl.STENCIL_ATTACHMENT;
        case 'color0':
        case 0: return gl.COLOR_ATTACHMENT0;
    }
    if (extensions.drawBuffers) {
        switch (attachment) {
            case 'color1':
            case 1: return extensions.drawBuffers.COLOR_ATTACHMENT1;
            case 'color2':
            case 2: return extensions.drawBuffers.COLOR_ATTACHMENT2;
            case 'color3':
            case 3: return extensions.drawBuffers.COLOR_ATTACHMENT3;
            case 'color4':
            case 4: return extensions.drawBuffers.COLOR_ATTACHMENT4;
            case 'color5':
            case 5: return extensions.drawBuffers.COLOR_ATTACHMENT5;
            case 'color6':
            case 6: return extensions.drawBuffers.COLOR_ATTACHMENT6;
            case 'color7':
            case 7: return extensions.drawBuffers.COLOR_ATTACHMENT7;
        }
    }
    throw new Error('unknown texture attachment');
}
exports.getAttachment = getAttachment;
function createTexture(ctx, kind, _format, _type, _filter) {
    var id = getNextTextureId();
    var gl = ctx.gl, stats = ctx.stats;
    var texture = gl.createTexture();
    if (texture === null) {
        throw new Error('Could not create WebGL texture');
    }
    // check texture kind and type compatability
    if ((kind.endsWith('float32') && _type !== 'float') ||
        (kind.endsWith('uint8') && _type !== 'ubyte') ||
        (kind.endsWith('depth') && _type !== 'ushort')) {
        throw new Error("texture kind '" + kind + "' and type '" + _type + "' are incompatible");
    }
    var target = getTarget(ctx, kind);
    var filter = getFilter(ctx, _filter);
    var format = getFormat(ctx, _format, _type);
    var internalFormat = getInternalFormat(ctx, _format, _type);
    var type = getType(ctx, _type);
    gl.bindTexture(target, texture);
    gl.texParameteri(target, gl.TEXTURE_MAG_FILTER, filter);
    gl.texParameteri(target, gl.TEXTURE_MIN_FILTER, filter);
    // clamp-to-edge needed for non-power-of-two textures in webgl
    gl.texParameteri(target, gl.TEXTURE_WRAP_S, gl.CLAMP_TO_EDGE);
    gl.texParameteri(target, gl.TEXTURE_WRAP_T, gl.CLAMP_TO_EDGE);
    gl.bindTexture(target, null);
    var width = 0, height = 0, depth = 0;
    var destroyed = false;
    stats.textureCount += 1;
    return {
        id: id,
        target: target,
        format: format,
        internalFormat: internalFormat,
        type: type,
        get width() { return width; },
        get height() { return height; },
        get depth() { return depth; },
        define: function (_width, _height, _depth) {
            width = _width, height = _height, depth = _depth || 0;
            gl.bindTexture(target, texture);
            if (target === gl.TEXTURE_2D) {
                gl.texImage2D(target, 0, internalFormat, width, height, 0, format, type, null);
            }
            else if (compat_1.isWebGL2(gl) && target === gl.TEXTURE_3D && depth !== undefined) {
                gl.texImage3D(target, 0, internalFormat, width, height, depth, 0, format, type, null);
            }
            else {
                throw new Error('unknown texture target');
            }
        },
        load: function (data) {
            gl.bindTexture(target, texture);
            // unpack alignment of 1 since we use textures only for data
            gl.pixelStorei(gl.UNPACK_ALIGNMENT, 1);
            gl.pixelStorei(gl.UNPACK_COLORSPACE_CONVERSION_WEBGL, gl.NONE);
            gl.pixelStorei(gl.UNPACK_PREMULTIPLY_ALPHA_WEBGL, 0);
            if (target === gl.TEXTURE_2D) {
                var _a = data, array = _a.array, _width = _a.width, _height = _a.height;
                width = _width, height = _height;
                gl.texImage2D(target, 0, internalFormat, width, height, 0, format, type, array);
            }
            else if (compat_1.isWebGL2(gl) && target === gl.TEXTURE_3D) {
                var _b = data, array = _b.array, _width = _b.width, _height = _b.height, _depth = _b.depth;
                width = _width, height = _height, depth = _depth;
                gl.texImage3D(target, 0, internalFormat, width, height, depth, 0, format, type, array);
            }
            else {
                throw new Error('unknown texture target');
            }
            gl.bindTexture(target, null);
        },
        bind: function (id) {
            gl.activeTexture(gl.TEXTURE0 + id);
            gl.bindTexture(target, texture);
        },
        unbind: function (id) {
            gl.activeTexture(gl.TEXTURE0 + id);
            gl.bindTexture(target, null);
        },
        attachFramebuffer: function (framebuffer, attachment, layer) {
            framebuffer.bind();
            if (target === gl.TEXTURE_2D) {
                gl.framebufferTexture2D(gl.FRAMEBUFFER, getAttachment(ctx, attachment), gl.TEXTURE_2D, texture, 0);
            }
            else if (compat_1.isWebGL2(gl) && target === gl.TEXTURE_3D) {
                if (layer === undefined)
                    throw new Error('need `layer` to attach 3D texture');
                gl.framebufferTextureLayer(gl.FRAMEBUFFER, getAttachment(ctx, attachment), texture, 0, layer);
            }
            else {
                throw new Error('unknown texture target');
            }
        },
        detachFramebuffer: function (framebuffer, attachment) {
            framebuffer.bind();
            if (target === gl.TEXTURE_2D) {
                gl.framebufferTexture2D(gl.FRAMEBUFFER, getAttachment(ctx, attachment), gl.TEXTURE_2D, null, 0);
            }
            else if (compat_1.isWebGL2(gl) && target === gl.TEXTURE_3D) {
                gl.framebufferTextureLayer(gl.FRAMEBUFFER, getAttachment(ctx, attachment), null, 0, 0);
            }
            else {
                throw new Error('unknown texture target');
            }
        },
        destroy: function () {
            if (destroyed)
                return;
            gl.deleteTexture(texture);
            destroyed = true;
            stats.textureCount -= 1;
        }
    };
}
exports.createTexture = createTexture;
function createTextures(ctx, schema, values) {
    var textures = [];
    Object.keys(schema).forEach(function (k) {
        var spec = schema[k];
        if (spec.type === 'texture') {
            if (spec.kind === 'texture') {
                textures[textures.length] = [k, values[k].ref.value];
            }
            else {
                var texture = createTexture(ctx, spec.kind, spec.format, spec.dataType, spec.filter);
                texture.load(values[k].ref.value);
                textures[textures.length] = [k, texture];
            }
        }
    });
    return textures;
}
exports.createTextures = createTextures;
