"use strict";
/**
 * Copyright (c) 2017-2018 mol* contributors, licensed under MIT, See LICENSE file for more info.
 *
 * @author David Sehnal <david.sehnal@gmail.com>
 * @author Alexander Rose <alexander.rose@weirdbyte.de>
 */
Object.defineProperty(exports, "__esModule", { value: true });
var db_1 = require("../../../mol-data/db");
var linear_algebra_1 = require("../../../mol-math/linear-algebra");
var mol_util_1 = require("../../../mol-util");
var Data = require("./data-model");
function toDatabaseCollection(schema, file) {
    var dbc = {};
    for (var _i = 0, _a = file.blocks; _i < _a.length; _i++) {
        var data = _a[_i];
        dbc[data.header] = toDatabase(schema, data);
    }
    return dbc;
}
exports.toDatabaseCollection = toDatabaseCollection;
function toDatabase(schema, frame) {
    return createDatabase(schema, frame);
}
exports.toDatabase = toDatabase;
function toTable(schema, category) {
    return new CategoryTable(category, schema, true);
}
exports.toTable = toTable;
function getColumnCtor(t) {
    switch (t.valueType) {
        case 'str': return function (f, c, k) { return createColumn(t, f, f.str, f.toStringArray); };
        case 'int': return function (f, c, k) { return createColumn(t, f, f.int, f.toIntArray); };
        case 'float': return function (f, c, k) { return createColumn(t, f, f.float, f.toFloatArray); };
        case 'list': throw new Error('Use createListColumn instead.');
        case 'tensor': throw new Error('Use createTensorColumn instead.');
    }
}
function createColumn(schema, field, value, toArray) {
    return {
        schema: schema,
        __array: field.__array,
        isDefined: field.isDefined,
        rowCount: field.rowCount,
        value: value,
        valueKind: field.valueKind,
        areValuesEqual: field.areValuesEqual,
        toArray: toArray
    };
}
function createListColumn(schema, category, key) {
    var separator = schema.separator;
    var itemParse = schema.itemParse;
    var f = category.getField(key);
    var value = f ? function (row) { return f.str(row).split(separator).map(function (x) { return itemParse(x.trim()); }).filter(function (x) { return !!x; }); } : function (row) { return []; };
    var toArray = function (params) { return db_1.ColumnHelpers.createAndFillArray(category.rowCount, value, params); };
    return {
        schema: schema,
        __array: void 0,
        isDefined: !!f,
        rowCount: category.rowCount,
        value: value,
        valueKind: f ? f.valueKind : function () { return 1 /* NotPresent */; },
        areValuesEqual: function (rowA, rowB) { return mol_util_1.arrayEqual(value(rowA), value(rowB)); },
        toArray: toArray
    };
}
function createTensorColumn(schema, category, key) {
    var space = schema.space;
    var zeroOffset = category.fieldNames.indexOf(key + "[0]") >= 0;
    var fst = zeroOffset ? 0 : 1;
    var firstFieldName;
    switch (space.rank) {
        case 1:
            firstFieldName = key + "[" + fst + "]";
            break;
        case 2:
            firstFieldName = key + "[" + fst + "][" + fst + "]";
            break;
        case 3:
            firstFieldName = key + "[" + fst + "][" + fst + "][" + fst + "]";
            break;
        default: throw new Error('Tensors with rank > 3 or rank 0 are currently not supported.');
    }
    var first = category.getField(firstFieldName) || db_1.Column.Undefined(category.rowCount, schema);
    var value = function (row) { return Data.getTensor(category, key, space, row, zeroOffset); };
    var toArray = function (params) { return db_1.ColumnHelpers.createAndFillArray(category.rowCount, value, params); };
    return {
        schema: schema,
        __array: void 0,
        isDefined: first.isDefined,
        rowCount: category.rowCount,
        value: value,
        valueKind: first.valueKind,
        areValuesEqual: function (rowA, rowB) { return linear_algebra_1.Tensor.areEqualExact(value(rowA), value(rowB)); },
        toArray: toArray
    };
}
var CategoryTable = /** @class */ (function () {
    function CategoryTable(category, schema, _isDefined) {
        this._isDefined = _isDefined;
        var fieldKeys = Object.keys(schema);
        this._rowCount = category.rowCount;
        this._columns = fieldKeys;
        this._schema = schema;
        var cache = Object.create(null);
        var _loop_1 = function (k) {
            Object.defineProperty(this_1, k, {
                get: function () {
                    if (cache[k])
                        return cache[k];
                    var fType = schema[k];
                    if (fType.valueType === 'list') {
                        cache[k] = createListColumn(fType, category, k);
                    }
                    else if (fType.valueType === 'tensor') {
                        cache[k] = createTensorColumn(fType, category, k);
                    }
                    else {
                        var ctor = getColumnCtor(fType);
                        var field = category.getField(k);
                        cache[k] = !!field ? ctor(field, category, k) : db_1.Column.Undefined(category.rowCount, fType);
                    }
                    return cache[k];
                },
                enumerable: true,
                configurable: false
            });
        };
        var this_1 = this;
        for (var _i = 0, fieldKeys_1 = fieldKeys; _i < fieldKeys_1.length; _i++) {
            var k = fieldKeys_1[_i];
            _loop_1(k);
        }
    }
    return CategoryTable;
}());
function createDatabase(schema, frame) {
    var tables = Object.create(null);
    for (var _i = 0, _a = Object.keys(schema); _i < _a.length; _i++) {
        var k = _a[_i];
        tables[k] = createTable(k, schema[k], frame);
    }
    return db_1.Database.ofTables(frame.header, schema, tables);
}
function createTable(key, schema, frame) {
    var cat = frame.categories[key];
    return new CategoryTable(cat || Data.CifCategory.empty(key), schema, !!cat);
}
