"use strict";
/**
 * Copyright (c) 2018 mol* contributors, licensed under MIT, See LICENSE file for more info.
 *
 * @author David Sehnal <david.sehnal@gmail.com>
 */
Object.defineProperty(exports, "__esModule", { value: true });
var _3d_1 = require("../../mol-math/linear-algebra/3d");
/**
 * Usage:
 *
 * 1. .reset(tolerance); tolerance plays part in the "extend" step
 * 2. for each point/sphere call boundaryStep()
 * 3. .finishBoundaryStep
 * 4. for each point/sphere call extendStep
 * 5. use .center/.radius or call getSphere/getBox
 */
var BoundaryHelper = /** @class */ (function () {
    function BoundaryHelper() {
        this.count = 0;
        this.extremes = [_3d_1.Vec3.zero(), _3d_1.Vec3.zero(), _3d_1.Vec3.zero(), _3d_1.Vec3.zero(), _3d_1.Vec3.zero(), _3d_1.Vec3.zero()];
        this.u = _3d_1.Vec3.zero();
        this.v = _3d_1.Vec3.zero();
        this.tolerance = 0;
        this.center = _3d_1.Vec3.zero();
        this.radius = 0;
        this.reset(0);
    }
    BoundaryHelper.prototype.reset = function (tolerance) {
        _3d_1.Vec3.set(this.center, 0, 0, 0);
        for (var i = 0; i < 6; i++) {
            var e = i % 2 === 0 ? Number.MAX_VALUE : -Number.MAX_VALUE;
            _3d_1.Vec3.set(this.extremes[i], e, e, e);
        }
        this.radius = 0;
        this.count = 0;
        this.tolerance = tolerance;
    };
    BoundaryHelper.prototype.boundaryStep = function (p, r) {
        updateExtremeMin(0, this.extremes[0], p, r);
        updateExtremeMax(0, this.extremes[1], p, r);
        updateExtremeMin(1, this.extremes[2], p, r);
        updateExtremeMax(1, this.extremes[3], p, r);
        updateExtremeMin(2, this.extremes[4], p, r);
        updateExtremeMax(2, this.extremes[5], p, r);
        this.count++;
    };
    BoundaryHelper.prototype.finishBoundaryStep = function () {
        if (this.count === 0)
            return;
        var maxSpan = 0, mI = 0, mJ = 0;
        for (var i = 0; i < 5; i++) {
            for (var j = i + 1; j < 6; j++) {
                var d = _3d_1.Vec3.squaredDistance(this.extremes[i], this.extremes[j]);
                if (d > maxSpan) {
                    maxSpan = d;
                    mI = i;
                    mJ = j;
                }
            }
        }
        _3d_1.Vec3.add(this.center, this.extremes[mI], this.extremes[mJ]);
        _3d_1.Vec3.scale(this.center, this.center, 0.5);
        this.radius = _3d_1.Vec3.distance(this.center, this.extremes[mI]);
    };
    BoundaryHelper.prototype.extendStep = function (p, r) {
        var d = _3d_1.Vec3.distance(p, this.center);
        if ((1 + this.tolerance) * this.radius >= r + d)
            return;
        _3d_1.Vec3.sub(this.u, p, this.center);
        _3d_1.Vec3.normalize(this.u, this.u);
        _3d_1.Vec3.scale(this.v, this.u, -this.radius);
        _3d_1.Vec3.add(this.v, this.v, this.center);
        _3d_1.Vec3.scale(this.u, this.u, r + d);
        _3d_1.Vec3.add(this.u, this.u, this.center);
        _3d_1.Vec3.add(this.center, this.u, this.v);
        _3d_1.Vec3.scale(this.center, this.center, 0.5);
        this.radius = 0.5 * (r + d + this.radius);
    };
    BoundaryHelper.prototype.getBox = function () {
        _3d_1.Vec3.copy(this.u, this.extremes[0]);
        _3d_1.Vec3.copy(this.v, this.extremes[0]);
        for (var i = 1; i < 6; i++) {
            _3d_1.Vec3.min(this.u, this.u, this.extremes[i]);
            _3d_1.Vec3.max(this.v, this.v, this.extremes[i]);
        }
        return { min: _3d_1.Vec3.clone(this.u), max: _3d_1.Vec3.clone(this.v) };
    };
    BoundaryHelper.prototype.getSphere = function () {
        return { center: _3d_1.Vec3.clone(this.center), radius: this.radius };
    };
    return BoundaryHelper;
}());
exports.BoundaryHelper = BoundaryHelper;
function updateExtremeMin(d, e, center, r) {
    if (center[d] - r < e[d]) {
        _3d_1.Vec3.copy(e, center);
        e[d] -= r;
    }
}
function updateExtremeMax(d, e, center, r) {
    if (center[d] + r > e[d]) {
        _3d_1.Vec3.copy(e, center);
        e[d] += r;
    }
}
