"use strict";
/**
 * Copyright (c) 2017-2019 mol* contributors, licensed under MIT, See LICENSE file for more info.
 *
 * @author Alexander Rose <alexander.rose@weirdbyte.de>
 * @author Michael Krone <michael.krone@uni-tuebingen.de>
 */
var __assign = (this && this.__assign) || function () {
    __assign = Object.assign || function(t) {
        for (var s, i = 1, n = arguments.length; i < n; i++) {
            s = arguments[i];
            for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p))
                t[p] = s[p];
        }
        return t;
    };
    return __assign.apply(this, arguments);
};
Object.defineProperty(exports, "__esModule", { value: true });
var geometry_1 = require("../../geometry");
var int_1 = require("../../../mol-data/int");
var linear_algebra_1 = require("../../linear-algebra");
var mol_util_1 = require("../../../mol-util");
var renderable_1 = require("../../../mol-gl/renderable");
var texture_1 = require("../../../mol-gl/webgl/texture");
var float_packing_1 = require("../../../mol-util/float-packing");
var shader_code_1 = require("../../../mol-gl/shader-code");
var render_item_1 = require("../../../mol-gl/webgl/render-item");
var schema_1 = require("../../../mol-gl/renderable/schema");
var gaussian_density_vert_1 = require("../../../mol-gl/shader/gaussian-density.vert");
var gaussian_density_frag_1 = require("../../../mol-gl/shader/gaussian-density.frag");
exports.GaussianDensitySchema = {
    drawCount: schema_1.ValueSpec('number'),
    instanceCount: schema_1.ValueSpec('number'),
    aRadius: schema_1.AttributeSpec('float32', 1, 0),
    aPosition: schema_1.AttributeSpec('float32', 3, 0),
    aGroup: schema_1.AttributeSpec('float32', 1, 0),
    uCurrentSlice: schema_1.UniformSpec('f'),
    uCurrentX: schema_1.UniformSpec('f'),
    uCurrentY: schema_1.UniformSpec('f'),
    uBboxMin: schema_1.UniformSpec('v3', true),
    uBboxSize: schema_1.UniformSpec('v3', true),
    uGridDim: schema_1.UniformSpec('v3', true),
    uGridTexDim: schema_1.UniformSpec('v3', true),
    uGridTexScale: schema_1.UniformSpec('v2', true),
    uAlpha: schema_1.UniformSpec('f', true),
    uResolution: schema_1.UniformSpec('f', true),
    tMinDistanceTex: schema_1.TextureSpec('texture', 'rgba', 'float', 'nearest'),
    dGridTexType: schema_1.DefineSpec('string', ['2d', '3d']),
    dCalcType: schema_1.DefineSpec('string', ['density', 'minDistance', 'groupId']),
};
exports.GaussianDensityShaderCode = shader_code_1.ShaderCode(gaussian_density_vert_1.default, gaussian_density_frag_1.default, { standardDerivatives: false, fragDepth: false });
/** name for shared framebuffer used for gpu gaussian surface operations */
var FramebufferName = 'gaussian-density';
function GaussianDensityGPU(position, box, radius, props, webgl) {
    // always use texture2d when the gaussian density needs to be downloaded from the GPU,
    // it's faster than texture3d
    // console.time('GaussianDensityTexture2d')
    var _a = calcGaussianDensityTexture2d(webgl, position, box, radius, props), scale = _a.scale, bbox = _a.bbox, texture = _a.texture, gridDim = _a.gridDim, gridTexDim = _a.gridTexDim;
    // webgl.waitForGpuCommandsCompleteSync()
    // console.timeEnd('GaussianDensityTexture2d')
    var _b = fieldFromTexture2d(webgl, texture, gridDim, gridTexDim), field = _b.field, idField = _b.idField;
    return { field: field, idField: idField, transform: getTransform(scale, bbox) };
}
exports.GaussianDensityGPU = GaussianDensityGPU;
function GaussianDensityTexture(webgl, position, box, radius, props, oldTexture) {
    return webgl.isWebGL2 ?
        GaussianDensityTexture3d(webgl, position, box, radius, props, oldTexture) :
        GaussianDensityTexture2d(webgl, position, box, radius, props, oldTexture);
}
exports.GaussianDensityTexture = GaussianDensityTexture;
function GaussianDensityTexture2d(webgl, position, box, radius, props, oldTexture) {
    return finalizeGaussianDensityTexture(calcGaussianDensityTexture2d(webgl, position, box, radius, props, oldTexture));
}
exports.GaussianDensityTexture2d = GaussianDensityTexture2d;
function GaussianDensityTexture3d(webgl, position, box, radius, props, oldTexture) {
    return finalizeGaussianDensityTexture(calcGaussianDensityTexture3d(webgl, position, box, radius, props, oldTexture));
}
exports.GaussianDensityTexture3d = GaussianDensityTexture3d;
function finalizeGaussianDensityTexture(_a) {
    var texture = _a.texture, scale = _a.scale, bbox = _a.bbox, gridDim = _a.gridDim, gridTexDim = _a.gridTexDim, gridTexScale = _a.gridTexScale;
    return { transform: getTransform(scale, bbox), texture: texture, bbox: bbox, gridDim: gridDim, gridTexDim: gridTexDim, gridTexScale: gridTexScale };
}
function getTransform(scale, bbox) {
    var transform = linear_algebra_1.Mat4.identity();
    linear_algebra_1.Mat4.fromScaling(transform, scale);
    linear_algebra_1.Mat4.setTranslation(transform, bbox.min);
    return transform;
}
function calcGaussianDensityTexture2d(webgl, position, box, radius, props, texture) {
    var smoothness = props.smoothness;
    var _a = prepareGaussianDensityData(position, box, radius, props), drawCount = _a.drawCount, positions = _a.positions, radii = _a.radii, groups = _a.groups, scale = _a.scale, expandedBox = _a.expandedBox, dim = _a.dim;
    var dx = dim[0], dy = dim[1], dz = dim[2];
    var _b = getTexture2dSize(dim), texDimX = _b.texDimX, texDimY = _b.texDimY, texCols = _b.texCols, powerOfTwoSize = _b.powerOfTwoSize;
    // console.log({ texDimX, texDimY, texCols, powerOfTwoSize, dim })
    var gridTexDim = linear_algebra_1.Vec3.create(texDimX, texDimY, 0);
    var gridTexScale = linear_algebra_1.Vec2.create(texDimX / powerOfTwoSize, texDimY / powerOfTwoSize);
    var minDistanceTexture = texture_1.createTexture(webgl, 'image-float32', 'rgba', 'float', 'nearest');
    minDistanceTexture.define(powerOfTwoSize, powerOfTwoSize);
    var renderable = getGaussianDensityRenderable(webgl, drawCount, positions, radii, groups, minDistanceTexture, expandedBox, dim, gridTexDim, gridTexScale, smoothness, props.resolution);
    //
    var gl = webgl.gl, framebufferCache = webgl.framebufferCache, state = webgl.state;
    var _c = renderable.values, uCurrentSlice = _c.uCurrentSlice, uCurrentX = _c.uCurrentX, uCurrentY = _c.uCurrentY;
    var framebuffer = framebufferCache.get(FramebufferName).value;
    framebuffer.bind();
    setRenderingDefaults(webgl);
    if (!texture) {
        texture = texture_1.createTexture(webgl, 'image-float32', 'rgba', 'float', 'nearest');
        texture.define(powerOfTwoSize, powerOfTwoSize);
    }
    else if (texture.width !== powerOfTwoSize || texture.height !== powerOfTwoSize) {
        texture.define(powerOfTwoSize, powerOfTwoSize);
    }
    // console.log(renderable)
    function render(fbTex, clear) {
        state.currentRenderItemId = -1;
        fbTex.attachFramebuffer(framebuffer, 0);
        if (clear)
            gl.clear(gl.COLOR_BUFFER_BIT);
        var currCol = 0;
        var currY = 0;
        var currX = 0;
        for (var i = 0; i < dz; ++i) {
            if (currCol >= texCols) {
                currCol -= texCols;
                currY += dy;
                currX = 0;
                mol_util_1.ValueCell.update(uCurrentY, currY);
            }
            // console.log({ i, currX, currY })
            mol_util_1.ValueCell.update(uCurrentX, currX);
            mol_util_1.ValueCell.update(uCurrentSlice, i);
            gl.viewport(currX, currY, dx, dy);
            renderable.render();
            ++currCol;
            currX += dx;
        }
        gl.finish();
    }
    setupDensityRendering(webgl, renderable);
    render(texture, true);
    setupMinDistanceRendering(webgl, renderable);
    render(minDistanceTexture, true);
    setupGroupIdRendering(webgl, renderable);
    render(texture, false);
    // printTexture(webgl, texture, 1)
    return { texture: texture, scale: scale, bbox: expandedBox, gridDim: dim, gridTexDim: gridTexDim, gridTexScale: gridTexScale };
}
function calcGaussianDensityTexture3d(webgl, position, box, radius, props, texture) {
    var smoothness = props.smoothness;
    var _a = prepareGaussianDensityData(position, box, radius, props), drawCount = _a.drawCount, positions = _a.positions, radii = _a.radii, groups = _a.groups, scale = _a.scale, expandedBox = _a.expandedBox, dim = _a.dim;
    var dx = dim[0], dy = dim[1], dz = dim[2];
    var minDistanceTexture = texture_1.createTexture(webgl, 'volume-float32', 'rgba', 'float', 'nearest');
    minDistanceTexture.define(dx, dy, dz);
    var gridTexScale = linear_algebra_1.Vec2.create(1, 1);
    var renderable = getGaussianDensityRenderable(webgl, drawCount, positions, radii, groups, minDistanceTexture, expandedBox, dim, dim, gridTexScale, smoothness, props.resolution);
    //
    var gl = webgl.gl, framebufferCache = webgl.framebufferCache;
    var uCurrentSlice = renderable.values.uCurrentSlice;
    var framebuffer = framebufferCache.get(FramebufferName).value;
    framebuffer.bind();
    setRenderingDefaults(webgl);
    gl.viewport(0, 0, dx, dy);
    if (!texture)
        texture = texture_1.createTexture(webgl, 'volume-float32', 'rgba', 'float', 'nearest');
    texture.define(dx, dy, dz);
    function render(fbTex) {
        for (var i = 0; i < dz; ++i) {
            mol_util_1.ValueCell.update(uCurrentSlice, i);
            fbTex.attachFramebuffer(framebuffer, 0, i);
            renderable.render();
        }
    }
    setupMinDistanceRendering(webgl, renderable);
    render(minDistanceTexture);
    setupDensityRendering(webgl, renderable);
    render(texture);
    setupGroupIdRendering(webgl, renderable);
    render(texture);
    return { texture: texture, scale: scale, bbox: expandedBox, gridDim: dim, gridTexDim: dim, gridTexScale: gridTexScale };
}
//
function prepareGaussianDensityData(position, box, radius, props) {
    var resolution = props.resolution, radiusOffset = props.radiusOffset;
    var scaleFactor = 1 / resolution;
    var indices = position.indices, x = position.x, y = position.y, z = position.z;
    var n = int_1.OrderedSet.size(indices);
    var positions = new Float32Array(n * 3);
    var radii = new Float32Array(n);
    var groups = new Float32Array(n);
    var maxRadius = 0;
    for (var i = 0; i < n; ++i) {
        var j = int_1.OrderedSet.getAt(indices, i);
        positions[i * 3] = x[j];
        positions[i * 3 + 1] = y[j];
        positions[i * 3 + 2] = z[j];
        var r = radius(j) + radiusOffset;
        if (maxRadius < r)
            maxRadius = r;
        radii[i] = r;
        groups[i] = i;
    }
    var pad = maxRadius * 2 + resolution * 4;
    var expandedBox = geometry_1.Box3D.expand(geometry_1.Box3D(), box, linear_algebra_1.Vec3.create(pad, pad, pad));
    var scaledBox = geometry_1.Box3D.scale(geometry_1.Box3D(), expandedBox, scaleFactor);
    var dim = geometry_1.Box3D.size(linear_algebra_1.Vec3(), scaledBox);
    linear_algebra_1.Vec3.ceil(dim, dim);
    var scale = linear_algebra_1.Vec3.create(resolution, resolution, resolution);
    return { drawCount: n, positions: positions, radii: radii, groups: groups, scale: scale, expandedBox: expandedBox, dim: dim };
}
function getGaussianDensityRenderable(webgl, drawCount, positions, radii, groups, minDistanceTexture, box, gridDim, gridTexDim, gridTexScale, smoothness, resolution) {
    var extent = linear_algebra_1.Vec3.sub(linear_algebra_1.Vec3.zero(), box.max, box.min);
    var values = {
        drawCount: mol_util_1.ValueCell.create(drawCount),
        instanceCount: mol_util_1.ValueCell.create(1),
        aRadius: mol_util_1.ValueCell.create(radii),
        aPosition: mol_util_1.ValueCell.create(positions),
        aGroup: mol_util_1.ValueCell.create(groups),
        uCurrentSlice: mol_util_1.ValueCell.create(0),
        uCurrentX: mol_util_1.ValueCell.create(0),
        uCurrentY: mol_util_1.ValueCell.create(0),
        uBboxMin: mol_util_1.ValueCell.create(box.min),
        uBboxSize: mol_util_1.ValueCell.create(extent),
        uGridDim: mol_util_1.ValueCell.create(gridDim),
        uGridTexDim: mol_util_1.ValueCell.create(gridTexDim),
        uGridTexScale: mol_util_1.ValueCell.create(gridTexScale),
        uAlpha: mol_util_1.ValueCell.create(smoothness),
        uResolution: mol_util_1.ValueCell.create(resolution),
        tMinDistanceTex: mol_util_1.ValueCell.create(minDistanceTexture),
        dGridTexType: mol_util_1.ValueCell.create(minDistanceTexture.depth > 0 ? '3d' : '2d'),
        dCalcType: mol_util_1.ValueCell.create('minDistance'),
    };
    var schema = __assign({}, exports.GaussianDensitySchema);
    var shaderCode = exports.GaussianDensityShaderCode;
    var renderItem = render_item_1.createComputeRenderItem(webgl, 'points', shaderCode, schema, values);
    return renderable_1.createComputeRenderable(renderItem, values);
}
function setRenderingDefaults(ctx) {
    var gl = ctx.gl, state = ctx.state;
    state.disable(gl.CULL_FACE);
    state.enable(gl.BLEND);
    state.disable(gl.DEPTH_TEST);
    state.disable(gl.SCISSOR_TEST);
    state.depthMask(false);
    state.clearColor(0, 0, 0, 0);
}
function setupMinDistanceRendering(webgl, renderable) {
    var gl = webgl.gl, state = webgl.state;
    mol_util_1.ValueCell.update(renderable.values.dCalcType, 'minDistance');
    renderable.update();
    state.colorMask(false, false, false, true);
    state.blendFunc(gl.ONE, gl.ONE);
    // the shader writes 1 - dist so we set blending to MAX
    if (!webgl.extensions.blendMinMax) {
        throw new Error('GPU gaussian surface calculation requires EXT_blend_minmax');
    }
    state.blendEquation(webgl.extensions.blendMinMax.MAX);
}
function setupDensityRendering(webgl, renderable) {
    var gl = webgl.gl, state = webgl.state;
    mol_util_1.ValueCell.update(renderable.values.dCalcType, 'density');
    renderable.update();
    state.colorMask(false, false, false, true);
    state.blendFunc(gl.ONE, gl.ONE);
    // state.colorMask(true, true, true, true)
    // state.blendFuncSeparate(gl.ONE, gl.ZERO, gl.ONE, gl.ONE)
    state.blendEquation(gl.FUNC_ADD);
}
function setupGroupIdRendering(webgl, renderable) {
    var gl = webgl.gl, state = webgl.state;
    mol_util_1.ValueCell.update(renderable.values.dCalcType, 'groupId');
    renderable.update();
    // overwrite color, don't change alpha
    state.colorMask(true, true, true, false);
    state.blendFunc(gl.ONE, gl.ZERO);
    state.blendEquation(gl.FUNC_ADD);
}
function getTexture2dSize(gridDim) {
    var area = gridDim[0] * gridDim[1] * gridDim[2];
    var squareDim = Math.sqrt(area);
    var powerOfTwoSize = Math.pow(2, Math.ceil(Math.log(squareDim) / Math.log(2)));
    var texDimX = 0;
    var texDimY = gridDim[1];
    var texRows = 1;
    var texCols = gridDim[2];
    if (powerOfTwoSize < gridDim[0] * gridDim[2]) {
        texCols = Math.floor(powerOfTwoSize / gridDim[0]);
        texRows = Math.ceil(gridDim[2] / texCols);
        texDimX = texCols * gridDim[0];
        texDimY *= texRows;
    }
    else {
        texDimX = gridDim[0] * gridDim[2];
    }
    return { texDimX: texDimX, texDimY: texDimY, texRows: texRows, texCols: texCols, powerOfTwoSize: texDimY < powerOfTwoSize ? powerOfTwoSize : powerOfTwoSize * 2 };
}
function fieldFromTexture2d(ctx, texture, dim, texDim) {
    // console.time('fieldFromTexture2d')
    var framebufferCache = ctx.framebufferCache;
    var dx = dim[0], dy = dim[1], dz = dim[2];
    // const { width, height } = texture
    var width = texDim[0], height = texDim[1];
    var fboTexCols = Math.floor(width / dx);
    var space = linear_algebra_1.Tensor.Space(dim, [2, 1, 0], Float32Array);
    var data = space.create();
    var field = linear_algebra_1.Tensor.create(space, data);
    var idData = space.create();
    var idField = linear_algebra_1.Tensor.create(space, idData);
    // const image = new Uint8Array(width * height * 4)
    var image = new Float32Array(width * height * 4);
    var framebuffer = framebufferCache.get(FramebufferName).value;
    framebuffer.bind();
    texture.attachFramebuffer(framebuffer, 0);
    ctx.readPixels(0, 0, width, height, image);
    // printImageData(createImageData(image, width, height), 1/3)
    var j = 0;
    var tmpCol = 0;
    var tmpRow = 0;
    for (var iz = 0; iz < dz; ++iz) {
        if (tmpCol >= fboTexCols) {
            tmpCol = 0;
            tmpRow += dy;
        }
        for (var iy = 0; iy < dy; ++iy) {
            for (var ix = 0; ix < dx; ++ix) {
                var idx = 4 * (tmpCol * dx + (iy + tmpRow) * width + ix);
                data[j] = image[idx + 3]; // / 255
                idData[j] = float_packing_1.decodeFloatRGB(image[idx] * 255, image[idx + 1] * 255, image[idx + 2] * 255);
                j++;
            }
        }
        tmpCol++;
    }
    // console.timeEnd('fieldFromTexture2d')
    return { field: field, idField: idField };
}
exports.fieldFromTexture2d = fieldFromTexture2d;
