"use strict";
/**
 * Copyright (c) 2018-2019 mol* contributors, licensed under MIT, See LICENSE file for more info.
 *
 * @author David Sehnal <david.sehnal@gmail.com>
 * @author Alexander Rose <alexander.rose@weirdbyte.de>
 */
Object.defineProperty(exports, "__esModule", { value: true });
var linear_algebra_1 = require("../../linear-algebra");
var int_1 = require("../../../mol-data/int");
function Sphere3D() {
    return Sphere3D.zero();
}
exports.Sphere3D = Sphere3D;
(function (Sphere3D) {
    function create(center, radius) { return { center: center, radius: radius }; }
    Sphere3D.create = create;
    function zero() { return { center: linear_algebra_1.Vec3.zero(), radius: 0 }; }
    Sphere3D.zero = zero;
    function clone(a) {
        var out = zero();
        linear_algebra_1.Vec3.copy(out.center, a.center);
        out.radius = a.radius;
        return out;
    }
    Sphere3D.clone = clone;
    function copy(out, a) {
        linear_algebra_1.Vec3.copy(out.center, a.center);
        out.radius = a.radius;
        return out;
    }
    Sphere3D.copy = copy;
    function computeBounding(data) {
        var x = data.x, y = data.y, z = data.z, indices = data.indices;
        var cx = 0, cy = 0, cz = 0;
        var radiusSq = 0;
        var size = int_1.OrderedSet.size(indices);
        for (var t = 0; t < size; t++) {
            var i = int_1.OrderedSet.getAt(indices, t);
            cx += x[i];
            cy += y[i];
            cz += z[i];
        }
        if (size > 0) {
            cx /= size;
            cy /= size;
            cz /= size;
        }
        for (var t = 0; t < size; t++) {
            var i = int_1.OrderedSet.getAt(indices, t);
            var dx = x[i] - cx, dy = y[i] - cy, dz = z[i] - cz;
            var d = dx * dx + dy * dy + dz * dz;
            if (d > radiusSq)
                radiusSq = d;
        }
        return { center: linear_algebra_1.Vec3.create(cx, cy, cz), radius: Math.sqrt(radiusSq) };
    }
    Sphere3D.computeBounding = computeBounding;
    /** Transform sphere with a Mat4 */
    function transform(out, sphere, m) {
        linear_algebra_1.Vec3.transformMat4(out.center, sphere.center, m);
        out.radius = sphere.radius * linear_algebra_1.Mat4.getMaxScaleOnAxis(m);
        return out;
    }
    Sphere3D.transform = transform;
    function toArray(s, out, offset) {
        linear_algebra_1.Vec3.toArray(s.center, out, offset);
        out[offset + 3] = s.radius;
    }
    Sphere3D.toArray = toArray;
    function fromArray(out, array, offset) {
        linear_algebra_1.Vec3.fromArray(out.center, array, offset);
        out.radius = array[offset + 3];
        return out;
    }
    Sphere3D.fromArray = fromArray;
    function fromBox3D(out, box) {
        linear_algebra_1.Vec3.scale(out.center, linear_algebra_1.Vec3.add(out.center, box.max, box.min), 0.5);
        out.radius = linear_algebra_1.Vec3.distance(out.center, box.max);
        return out;
    }
    Sphere3D.fromBox3D = fromBox3D;
    var tmpAddVec3 = linear_algebra_1.Vec3();
    function addVec3(out, s, v) {
        var d = linear_algebra_1.Vec3.distance(s.center, v);
        if (d < s.radius)
            return Sphere3D.copy(out, s);
        linear_algebra_1.Vec3.sub(tmpAddVec3, s.center, v);
        linear_algebra_1.Vec3.sub(tmpAddVec3, s.center, tmpAddVec3);
        linear_algebra_1.Vec3.setMagnitude(tmpAddVec3, tmpAddVec3, s.radius);
        linear_algebra_1.Vec3.scale(out.center, linear_algebra_1.Vec3.add(tmpAddVec3, tmpAddVec3, v), 0.5);
        out.radius = linear_algebra_1.Vec3.distance(out.center, v);
        return out;
    }
    Sphere3D.addVec3 = addVec3;
    /** Expand sphere radius by another sphere */
    function expandBySphere(out, sphere) {
        out.radius = Math.max(out.radius, linear_algebra_1.Vec3.distance(out.center, sphere.center) + sphere.radius);
        return out;
    }
    Sphere3D.expandBySphere = expandBySphere;
    /** Expand sphere radius by delta */
    function expand(out, sphere, delta) {
        out.radius = sphere.radius + delta;
        return out;
    }
    Sphere3D.expand = expand;
    /**
     * Returns whether or not the spheres have exactly the same center and radius (when compared with ===)
     */
    function exactEquals(a, b) {
        return a.radius === b.radius && linear_algebra_1.Vec3.exactEquals(a.center, b.center);
    }
    Sphere3D.exactEquals = exactEquals;
    /**
     * Returns whether or not the spheres have approximately the same center and radius.
     */
    function equals(a, b) {
        var ar = a.radius;
        var br = b.radius;
        return (Math.abs(ar - br) <= 0.000001 /* Value */ * Math.max(1.0, Math.abs(ar), Math.abs(br)) &&
            linear_algebra_1.Vec3.equals(a.center, b.center));
    }
    Sphere3D.equals = equals;
})(Sphere3D || (Sphere3D = {}));
exports.Sphere3D = Sphere3D;
