"use strict";
/**
 * Copyright (c) 2018 mol* contributors, licensed under MIT, See LICENSE file for more info.
 *
 * @author David Sehnal <david.sehnal@gmail.com>
 */
Object.defineProperty(exports, "__esModule", { value: true });
var linear_algebra_1 = require("../../linear-algebra");
var tables_1 = require("./tables");
var geometry_1 = require("../../geometry");
var SpacegroupCell;
(function (SpacegroupCell) {
    /** Create a 'P 1' with cellsize [1, 1, 1] */
    SpacegroupCell.Zero = create('P 1', linear_algebra_1.Vec3.create(1, 1, 1), linear_algebra_1.Vec3.create(Math.PI / 2, Math.PI / 2, Math.PI / 2));
    /** True if 'P 1' with cellsize [1, 1, 1] */
    function isZero(cell) {
        return cell.index === 0 && cell.size[0] === 1 && cell.size[1] === 1 && cell.size[1] === 1;
    }
    SpacegroupCell.isZero = isZero;
    // returns Zero cell if the spacegroup does not exist
    function create(nameOrNumber, size, anglesInRadians) {
        var index = tables_1.getSpacegroupIndex(nameOrNumber);
        if (index < 0) {
            console.warn("Unknown spacegroup '" + nameOrNumber + "', returning a 'P 1' with cellsize [1, 1, 1]");
            return SpacegroupCell.Zero;
        }
        var alpha = anglesInRadians[0];
        var beta = anglesInRadians[1];
        var gamma = anglesInRadians[2];
        var xScale = size[0], yScale = size[1], zScale = size[2];
        var z1 = Math.cos(beta);
        var z2 = (Math.cos(alpha) - Math.cos(beta) * Math.cos(gamma)) / Math.sin(gamma);
        var z3 = Math.sqrt(1.0 - z1 * z1 - z2 * z2);
        var x = [xScale, 0.0, 0.0];
        var y = [Math.cos(gamma) * yScale, Math.sin(gamma) * yScale, 0.0];
        var z = [z1 * zScale, z2 * zScale, z3 * zScale];
        var fromFractional = linear_algebra_1.Mat4.ofRows([
            [x[0], y[0], z[0], 0],
            [0, y[1], z[1], 0],
            [0, 0, z[2], 0],
            [0, 0, 0, 1.0]
        ]);
        var toFractional = linear_algebra_1.Mat4.invert(linear_algebra_1.Mat4.zero(), fromFractional);
        return { index: index, size: size, anglesInRadians: anglesInRadians, toFractional: toFractional, fromFractional: fromFractional };
    }
    SpacegroupCell.create = create;
})(SpacegroupCell || (SpacegroupCell = {}));
exports.SpacegroupCell = SpacegroupCell;
var Spacegroup;
(function (Spacegroup) {
    // P1 with [1, 1, 1] cell.
    Spacegroup.ZeroP1 = create(SpacegroupCell.Zero);
    function create(cell) {
        var operators = tables_1.GroupData[cell.index].map(function (i) { return getOperatorMatrix(tables_1.OperatorData[i]); });
        return { name: tables_1.SpacegroupNames[cell.index], cell: cell, operators: operators };
    }
    Spacegroup.create = create;
    var _tempVec = linear_algebra_1.Vec3.zero(), _tempMat = linear_algebra_1.Mat4.zero();
    function updateOperatorMatrix(spacegroup, index, i, j, k, target) {
        _tempVec[0] = i;
        _tempVec[1] = j;
        _tempVec[2] = k;
        linear_algebra_1.Mat4.fromTranslation(_tempMat, _tempVec);
        return linear_algebra_1.Mat4.mul(target, linear_algebra_1.Mat4.mul(target, linear_algebra_1.Mat4.mul(target, spacegroup.cell.fromFractional, _tempMat), spacegroup.operators[index]), spacegroup.cell.toFractional);
    }
    Spacegroup.updateOperatorMatrix = updateOperatorMatrix;
    function getSymmetryOperator(spacegroup, index, i, j, k) {
        var operator = updateOperatorMatrix(spacegroup, index, i, j, k, linear_algebra_1.Mat4.zero());
        return geometry_1.SymmetryOperator.create(index + 1 + "_" + (5 + i) + (5 + j) + (5 + k), operator, { id: '', operList: [] }, '', linear_algebra_1.Vec3.create(i, j, k), index);
    }
    Spacegroup.getSymmetryOperator = getSymmetryOperator;
    function getOperatorMatrix(ids) {
        var r1 = tables_1.TransformData[ids[0]];
        var r2 = tables_1.TransformData[ids[1]];
        var r3 = tables_1.TransformData[ids[2]];
        return linear_algebra_1.Mat4.ofRows([r1, r2, r3, [0, 0, 0, 1]]);
    }
})(Spacegroup || (Spacegroup = {}));
exports.Spacegroup = Spacegroup;
