/**
 * Copyright (c) 2018 mol* contributors, licensed under MIT, See LICENSE file for more info.
 *
 * @author David Sehnal <david.sehnal@gmail.com>
 * @author Alexander Rose <alexander.rose@weirdbyte.de>
 */
import { SortedArray } from '../../mol-data/int';
/**
 * Represent a graph using vertex adjacency list.
 *
 * Edges of the i-th vertex are stored in the arrays a and b
 * for indices in the range [offset[i], offset[i+1]).
 *
 * Edge properties are indexed same as in the arrays a and b.
 */
export interface IntAdjacencyGraph<EdgeProps extends IntAdjacencyGraph.EdgePropsBase = {}> {
    readonly offset: ArrayLike<number>;
    readonly a: ArrayLike<number>;
    readonly b: ArrayLike<number>;
    readonly vertexCount: number;
    readonly edgeCount: number;
    readonly edgeProps: Readonly<EdgeProps>;
    /**
     * Get the edge index between i-th and j-th vertex.
     * -1 if the edge does not exist.
     *
     * Because the a and b arrays contains each edge twice,
     * this always returns the smaller of the indices.
     *
     * `getEdgeIndex(i, j) === getEdgeIndex(j, i)`
     */
    getEdgeIndex(i: number, j: number): number;
    /**
     * Get the edge index between i-th and j-th vertex.
     * -1 if the edge does not exist.
     *
     * `getEdgeIndex(i, j) !== getEdgeIndex(j, i)`
     */
    getDirectedEdgeIndex(i: number, j: number): number;
    getVertexEdgeCount(i: number): number;
}
export declare namespace IntAdjacencyGraph {
    type EdgePropsBase = {
        [name: string]: ArrayLike<any>;
    };
    function create<EdgeProps extends IntAdjacencyGraph.EdgePropsBase = {}>(offset: ArrayLike<number>, a: ArrayLike<number>, b: ArrayLike<number>, edgeCount: number, edgeProps?: EdgeProps): IntAdjacencyGraph<EdgeProps>;
    class EdgeBuilder {
        vertexCount: number;
        xs: ArrayLike<number>;
        ys: ArrayLike<number>;
        private bucketFill;
        private current;
        private curA;
        private curB;
        offsets: Int32Array;
        edgeCount: number;
        /** the size of the A and B arrays */
        slotCount: number;
        a: Int32Array;
        b: Int32Array;
        createGraph<EdgeProps extends IntAdjacencyGraph.EdgePropsBase = {}>(edgeProps?: EdgeProps): IntAdjacencyGraph<EdgeProps>;
        /**
         * @example
         *   const property = new Int32Array(builder.slotCount);
         *   for (let i = 0; i < builder.edgeCount; i++) {
         *     builder.addNextEdge();
         *     builder.assignProperty(property, srcProp[i]);
         *   }
         *   return builder.createGraph({ property });
         */
        addNextEdge(): void;
        /** Builds property-less graph */
        addAllEdges(): void;
        assignProperty<T>(prop: {
            [i: number]: T;
        }, value: T): void;
        constructor(vertexCount: number, xs: ArrayLike<number>, ys: ArrayLike<number>);
    }
    class DirectedEdgeBuilder {
        vertexCount: number;
        xs: ArrayLike<number>;
        ys: ArrayLike<number>;
        private bucketFill;
        private current;
        private curA;
        offsets: Int32Array;
        edgeCount: number;
        /** the size of the A and B arrays */
        slotCount: number;
        a: Int32Array;
        b: Int32Array;
        createGraph<EdgeProps extends IntAdjacencyGraph.EdgePropsBase = {}>(edgeProps?: EdgeProps): IntAdjacencyGraph<EdgeProps>;
        /**
         * @example
         *   const property = new Int32Array(builder.slotCount);
         *   for (let i = 0; i < builder.edgeCount; i++) {
         *     builder.addNextEdge();
         *     builder.assignProperty(property, srcProp[i]);
         *   }
         *   return builder.createGraph({ property });
         */
        addNextEdge(): void;
        /** Builds property-less graph */
        addAllEdges(): void;
        assignProperty<T>(prop: {
            [i: number]: T;
        }, value: T): void;
        constructor(vertexCount: number, xs: ArrayLike<number>, ys: ArrayLike<number>);
    }
    class UniqueEdgeBuilder {
        vertexCount: number;
        private xs;
        private ys;
        private included;
        addEdge(i: number, j: number): boolean;
        getGraph(): IntAdjacencyGraph;
        getEdgeBuiler(): EdgeBuilder;
        constructor(vertexCount: number);
    }
    function fromVertexPairs(vertexCount: number, xs: number[], ys: number[]): IntAdjacencyGraph<{}>;
    function induceByVertices<P extends IntAdjacencyGraph.EdgePropsBase>(graph: IntAdjacencyGraph<P>, vertexIndices: ArrayLike<number>): IntAdjacencyGraph<P>;
    function connectedComponents(graph: IntAdjacencyGraph): {
        componentCount: number;
        componentIndex: Int32Array;
    };
    /**
     * Check if any vertex in `verticesA` is connected to any vertex in `verticesB`
     * via at most `maxDistance` edges.
     *
     * Returns true if verticesA and verticesB are intersecting.
     */
    function areVertexSetsConnected(graph: IntAdjacencyGraph, verticesA: SortedArray<number>, verticesB: SortedArray<number>, maxDistance: number): boolean;
}
