"use strict";
/**
 * Copyright (c) 2017-2018 mol* contributors, licensed under MIT, See LICENSE file for more info.
 *
 * @author David Sehnal <david.sehnal@gmail.com>
 */
var __assign = (this && this.__assign) || function () {
    __assign = Object.assign || function(t) {
        for (var s, i = 1, n = arguments.length; i < n; i++) {
            s = arguments[i];
            for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p))
                t[p] = s[p];
        }
        return t;
    };
    return __assign.apply(this, arguments);
};
Object.defineProperty(exports, "__esModule", { value: true });
var db_1 = require("../../../mol-data/db");
var int_1 = require("../../../mol-data/int");
var uuid_1 = require("../../../mol-util/uuid");
var atomic_1 = require("../../../mol-model/structure/model/properties/atomic");
var atomic_index_1 = require("../../../mol-model/structure/model/properties/utils/atomic-index");
var types_1 = require("../../../mol-model/structure/model/types");
var atomic_ranges_1 = require("../../../mol-model/structure/model/properties/utils/atomic-ranges");
var atomic_derived_1 = require("../../../mol-model/structure/model/properties/utils/atomic-derived");
function findHierarchyOffsets(atom_site) {
    if (atom_site._rowCount === 0)
        return { residues: [], chains: [] };
    var start = 0, end = atom_site._rowCount;
    var residues = [start], chains = [start];
    var label_entity_id = atom_site.label_entity_id, label_asym_id = atom_site.label_asym_id, label_seq_id = atom_site.label_seq_id, auth_seq_id = atom_site.auth_seq_id, pdbx_PDB_ins_code = atom_site.pdbx_PDB_ins_code, label_comp_id = atom_site.label_comp_id;
    for (var i = start + 1; i < end; i++) {
        var newChain = !label_entity_id.areValuesEqual(i - 1, i) || !label_asym_id.areValuesEqual(i - 1, i);
        var newResidue = newChain
            || !label_seq_id.areValuesEqual(i - 1, i)
            || !auth_seq_id.areValuesEqual(i - 1, i)
            || !pdbx_PDB_ins_code.areValuesEqual(i - 1, i)
            || !label_comp_id.areValuesEqual(i - 1, i);
        if (newResidue)
            residues[residues.length] = i;
        if (newChain)
            chains[chains.length] = i;
    }
    return { residues: residues, chains: chains };
}
function createHierarchyData(atom_site, sourceIndex, offsets) {
    var atoms = db_1.Table.ofColumns(atomic_1.AtomsSchema, {
        type_symbol: db_1.Column.ofArray({ array: db_1.Column.mapToArray(atom_site.type_symbol, types_1.ElementSymbol), schema: db_1.Column.Schema.Aliased(db_1.Column.Schema.str) }),
        label_atom_id: atom_site.label_atom_id,
        auth_atom_id: atom_site.auth_atom_id,
        label_alt_id: atom_site.label_alt_id,
        pdbx_formal_charge: atom_site.pdbx_formal_charge,
        sourceIndex: sourceIndex
    });
    var residues = db_1.Table.view(atom_site, atomic_1.ResiduesSchema, offsets.residues);
    // Optimize the numeric columns
    db_1.Table.columnToArray(residues, 'label_seq_id', Int32Array);
    db_1.Table.columnToArray(residues, 'auth_seq_id', Int32Array);
    var chains = db_1.Table.view(atom_site, atomic_1.ChainsSchema, offsets.chains);
    return { atoms: atoms, residues: residues, chains: chains };
}
function getConformation(atom_site) {
    return {
        id: uuid_1.default.create22(),
        atomId: atom_site.id,
        occupancy: atom_site.occupancy,
        B_iso_or_equiv: atom_site.B_iso_or_equiv,
        x: atom_site.Cartn_x.toArray({ array: Float32Array }),
        y: atom_site.Cartn_y.toArray({ array: Float32Array }),
        z: atom_site.Cartn_z.toArray({ array: Float32Array }),
    };
}
function isHierarchyDataEqual(a, b) {
    // TODO need to cast because of how TS handles type resolution for interfaces https://github.com/Microsoft/TypeScript/issues/15300
    return db_1.Table.areEqual(a.chains, b.chains)
        && db_1.Table.areEqual(a.residues, b.residues)
        && db_1.Table.areEqual(a.atoms, b.atoms);
}
function getAtomicHierarchyAndConformation(atom_site, sourceIndex, entities, formatData, previous) {
    var hierarchyOffsets = findHierarchyOffsets(atom_site);
    var hierarchyData = createHierarchyData(atom_site, sourceIndex, hierarchyOffsets);
    if (previous && isHierarchyDataEqual(previous.atomicHierarchy, hierarchyData)) {
        return {
            sameAsPrevious: true,
            hierarchy: previous.atomicHierarchy,
            conformation: getConformation(atom_site)
        };
    }
    var conformation = getConformation(atom_site);
    var hierarchySegments = {
        residueAtomSegments: int_1.Segmentation.ofOffsets(hierarchyOffsets.residues, int_1.Interval.ofBounds(0, atom_site._rowCount)),
        chainAtomSegments: int_1.Segmentation.ofOffsets(hierarchyOffsets.chains, int_1.Interval.ofBounds(0, atom_site._rowCount)),
    };
    var index = atomic_index_1.getAtomicIndex(hierarchyData, entities, hierarchySegments);
    var derived = atomic_derived_1.getAtomicDerivedData(hierarchyData, index, formatData.chemicalComponentMap);
    var hierarchyRanges = atomic_ranges_1.getAtomicRanges(hierarchyData, hierarchySegments, conformation, index, derived.residue.moleculeType);
    var hierarchy = __assign({}, hierarchyData, hierarchySegments, hierarchyRanges, { index: index, derived: derived });
    return { sameAsPrevious: false, hierarchy: hierarchy, conformation: conformation };
}
exports.getAtomicHierarchyAndConformation = getAtomicHierarchyAndConformation;
