"use strict";
/**
 * Copyright (c) 2017-2018 Mol* contributors, licensed under MIT, See LICENSE file for more info.
 *
 * @author David Sehnal <david.sehnal@gmail.com>
 * @author Alexander Rose <alexander.rose@weirdbyte.de>
 */
Object.defineProperty(exports, "__esModule", { value: true });
var structure_1 = require("../../../../mol-model/structure");
var int_1 = require("../../../../mol-data/int");
var cif_1 = require("../../../../mol-io/writer/cif");
var ComponentBond;
(function (ComponentBond) {
    ComponentBond.Descriptor = {
        isStatic: true,
        name: 'chem_comp_bond',
        cifExport: {
            prefix: '',
            categories: [{
                    name: 'chem_comp_bond',
                    instance: function (ctx) {
                        var chem_comp_bond = getChemCompBond(ctx.structures[0].model);
                        if (!chem_comp_bond)
                            return cif_1.CifWriter.Category.Empty;
                        var comp_names = getUniqueResidueNames(ctx.structures[0]);
                        var comp_id = chem_comp_bond.comp_id, _rowCount = chem_comp_bond._rowCount;
                        var indices = [];
                        for (var i = 0; i < _rowCount; i++) {
                            if (comp_names.has(comp_id.value(i)))
                                indices[indices.length] = i;
                        }
                        return cif_1.CifWriter.Category.ofTable(chem_comp_bond, indices);
                    }
                }]
        }
    };
    function attachFromMmCif(model) {
        if (model.customProperties.has(ComponentBond.Descriptor))
            return true;
        if (model.sourceData.kind !== 'mmCIF')
            return false;
        var chem_comp_bond = model.sourceData.data.chem_comp_bond;
        if (chem_comp_bond._rowCount === 0)
            return false;
        model.customProperties.add(ComponentBond.Descriptor);
        model._staticPropertyData.__ComponentBondData__ = chem_comp_bond;
        return true;
    }
    ComponentBond.attachFromMmCif = attachFromMmCif;
    function attachFromExternalData(model, bonds, force) {
        if (force === void 0) { force = false; }
        if (!force && model.customProperties.has(ComponentBond.Descriptor))
            return true;
        if (model._staticPropertyData.__ComponentBondData__)
            delete model._staticPropertyData.__ComponentBondData__;
        model.customProperties.add(ComponentBond.Descriptor);
        model._staticPropertyData[ComponentBond.PropName] = bonds;
        return true;
    }
    ComponentBond.attachFromExternalData = attachFromExternalData;
    var ComponentBondImpl = /** @class */ (function () {
        function ComponentBondImpl() {
            this.entries = new Map();
        }
        ComponentBondImpl.prototype.addEntry = function (id) {
            var e = new Entry(id);
            this.entries.set(id, e);
            return e;
        };
        return ComponentBondImpl;
    }());
    ComponentBond.ComponentBondImpl = ComponentBondImpl;
    var Entry = /** @class */ (function () {
        function Entry(id) {
            this.id = id;
            this.map = new Map();
        }
        Entry.prototype.add = function (a, b, order, flags, swap) {
            if (swap === void 0) { swap = true; }
            var e = this.map.get(a);
            if (e !== void 0) {
                var f = e.get(b);
                if (f === void 0) {
                    e.set(b, { order: order, flags: flags });
                }
            }
            else {
                var map = new Map();
                map.set(b, { order: order, flags: flags });
                this.map.set(a, map);
            }
            if (swap)
                this.add(b, a, order, flags, false);
        };
        return Entry;
    }());
    ComponentBond.Entry = Entry;
    function parseChemCompBond(data) {
        var comp_id = data.comp_id, atom_id_1 = data.atom_id_1, atom_id_2 = data.atom_id_2, value_order = data.value_order, pdbx_aromatic_flag = data.pdbx_aromatic_flag, rowCount = data._rowCount;
        var compBond = new ComponentBondImpl();
        var entry = compBond.addEntry(comp_id.value(0));
        for (var i = 0; i < rowCount; i++) {
            var id = comp_id.value(i);
            var nameA = atom_id_1.value(i);
            var nameB = atom_id_2.value(i);
            var order = value_order.value(i);
            var aromatic = pdbx_aromatic_flag.value(i) === 'Y';
            if (entry.id !== id) {
                entry = compBond.addEntry(id);
            }
            var flags = 1 /* Covalent */;
            var ord = 1;
            if (aromatic)
                flags |= 32 /* Aromatic */;
            switch (order.toLowerCase()) {
                case 'doub':
                case 'delo':
                    ord = 2;
                    break;
                case 'trip':
                    ord = 3;
                    break;
                case 'quad':
                    ord = 4;
                    break;
            }
            entry.add(nameA, nameB, ord, flags);
        }
        return compBond;
    }
    ComponentBond.parseChemCompBond = parseChemCompBond;
    function getChemCompBond(model) {
        return model._staticPropertyData.__ComponentBondData__;
    }
    ComponentBond.PropName = '__ComponentBond__';
    function get(model) {
        if (model._staticPropertyData[ComponentBond.PropName])
            return model._staticPropertyData[ComponentBond.PropName];
        if (!model.customProperties.has(ComponentBond.Descriptor))
            return void 0;
        var chem_comp_bond = getChemCompBond(model);
        if (!chem_comp_bond)
            return void 0;
        var chemComp = parseChemCompBond(chem_comp_bond);
        model._staticPropertyData[ComponentBond.PropName] = chemComp;
        return chemComp;
    }
    ComponentBond.get = get;
    function getUniqueResidueNames(s) {
        var prop = structure_1.StructureProperties.residue.label_comp_id;
        var names = new Set();
        var loc = structure_1.StructureElement.create();
        for (var _i = 0, _a = s.units; _i < _a.length; _i++) {
            var unit = _a[_i];
            if (!structure_1.Unit.isAtomic(unit))
                continue;
            var residues = int_1.Segmentation.transientSegments(unit.model.atomicHierarchy.residueAtomSegments, unit.elements);
            loc.unit = unit;
            while (residues.hasNext) {
                var seg = residues.move();
                loc.element = unit.elements[seg.start];
                names.add(prop(loc));
            }
        }
        return names;
    }
})(ComponentBond = exports.ComponentBond || (exports.ComponentBond = {}));
