"use strict";
/**
 * Copyright (c) 2017-2018 Mol* contributors, licensed under MIT, See LICENSE file for more info.
 *
 * @author David Sehnal <david.sehnal@gmail.com>
 * @author Alexander Rose <alexander.rose@weirdbyte.de>
 */
Object.defineProperty(exports, "__esModule", { value: true });
var util_1 = require("../util");
var int_1 = require("../../../../mol-data/int");
var cif_1 = require("../../../../mol-io/writer/cif");
var StructConn;
(function (StructConn) {
    StructConn.Descriptor = {
        isStatic: true,
        name: 'struct_conn',
        cifExport: {
            prefix: '',
            categories: [{
                    name: 'struct_conn',
                    instance: function (ctx) {
                        var structure = ctx.structures[0], model = structure.model;
                        var struct_conn = getStructConn(model);
                        if (!struct_conn)
                            return cif_1.CifWriter.Category.Empty;
                        var strConn = get(model);
                        if (!strConn || strConn.entries.length === 0)
                            return cif_1.CifWriter.Category.Empty;
                        var foundAtoms = new Set();
                        var indices = [];
                        for (var _a = 0, _b = strConn.entries; _a < _b.length; _a++) {
                            var entry = _b[_a];
                            var partners = entry.partners;
                            var hasAll = true;
                            for (var i = 0, _i = partners.length; i < _i; i++) {
                                var atom = partners[i].atomIndex;
                                if (foundAtoms.has(atom))
                                    continue;
                                if (hasAtom(structure, atom)) {
                                    foundAtoms.add(atom);
                                }
                                else {
                                    hasAll = false;
                                    break;
                                }
                            }
                            if (hasAll) {
                                indices[indices.length] = entry.rowIndex;
                            }
                        }
                        return cif_1.CifWriter.Category.ofTable(struct_conn, indices);
                    }
                }]
        }
    };
    function hasAtom(_a, element) {
        var units = _a.units;
        for (var i = 0, _i = units.length; i < _i; i++) {
            if (int_1.SortedArray.indexOf(units[i].elements, element) >= 0)
                return true;
        }
        return false;
    }
    function _resKey(rA, rB) {
        if (rA < rB)
            return rA + "-" + rB;
        return rB + "-" + rA;
    }
    var _emptyEntry = [];
    var StructConnImpl = /** @class */ (function () {
        function StructConnImpl(entries) {
            this.entries = entries;
            this._residuePairIndex = void 0;
            this._atomIndex = void 0;
        }
        StructConnImpl.prototype.getResiduePairIndex = function () {
            if (this._residuePairIndex)
                return this._residuePairIndex;
            this._residuePairIndex = new Map();
            for (var _a = 0, _b = this.entries; _a < _b.length; _a++) {
                var e = _b[_a];
                var ps = e.partners;
                var l = ps.length;
                for (var i = 0; i < l - 1; i++) {
                    for (var j = i + i; j < l; j++) {
                        var key = _resKey(ps[i].residueIndex, ps[j].residueIndex);
                        if (this._residuePairIndex.has(key)) {
                            this._residuePairIndex.get(key).push(e);
                        }
                        else {
                            this._residuePairIndex.set(key, [e]);
                        }
                    }
                }
            }
            return this._residuePairIndex;
        };
        StructConnImpl.prototype.getAtomIndex = function () {
            if (this._atomIndex)
                return this._atomIndex;
            this._atomIndex = new Map();
            for (var _a = 0, _b = this.entries; _a < _b.length; _a++) {
                var e = _b[_a];
                for (var _c = 0, _d = e.partners; _c < _d.length; _c++) {
                    var p = _d[_c];
                    var key = p.atomIndex;
                    if (this._atomIndex.has(key)) {
                        this._atomIndex.get(key).push(e);
                    }
                    else {
                        this._atomIndex.set(key, [e]);
                    }
                }
            }
            return this._atomIndex;
        };
        StructConnImpl.prototype.getResidueEntries = function (residueAIndex, residueBIndex) {
            return this.getResiduePairIndex().get(_resKey(residueAIndex, residueBIndex)) || _emptyEntry;
        };
        StructConnImpl.prototype.getAtomEntries = function (atomIndex) {
            return this.getAtomIndex().get(atomIndex) || _emptyEntry;
        };
        return StructConnImpl;
    }());
    function attachFromMmCif(model) {
        if (model.customProperties.has(StructConn.Descriptor))
            return true;
        if (model.sourceData.kind !== 'mmCIF')
            return false;
        var struct_conn = model.sourceData.data.struct_conn;
        if (struct_conn._rowCount === 0)
            return false;
        model.customProperties.add(StructConn.Descriptor);
        model._staticPropertyData.__StructConnData__ = struct_conn;
        return true;
    }
    StructConn.attachFromMmCif = attachFromMmCif;
    function getStructConn(model) {
        return model._staticPropertyData.__StructConnData__;
    }
    StructConn.PropName = '__StructConn__';
    function get(model) {
        if (model._staticPropertyData[StructConn.PropName])
            return model._staticPropertyData[StructConn.PropName];
        if (!model.customProperties.has(StructConn.Descriptor))
            return void 0;
        var struct_conn = getStructConn(model);
        var conn_type_id = struct_conn.conn_type_id, pdbx_dist_value = struct_conn.pdbx_dist_value, pdbx_value_order = struct_conn.pdbx_value_order;
        var p1 = {
            label_asym_id: struct_conn.ptnr1_label_asym_id,
            label_seq_id: struct_conn.ptnr1_label_seq_id,
            auth_seq_id: struct_conn.ptnr1_auth_seq_id,
            label_atom_id: struct_conn.ptnr1_label_atom_id,
            label_alt_id: struct_conn.pdbx_ptnr1_label_alt_id,
            ins_code: struct_conn.pdbx_ptnr1_PDB_ins_code,
            symmetry: struct_conn.ptnr1_symmetry
        };
        var p2 = {
            label_asym_id: struct_conn.ptnr2_label_asym_id,
            label_seq_id: struct_conn.ptnr2_label_seq_id,
            auth_seq_id: struct_conn.ptnr2_auth_seq_id,
            label_atom_id: struct_conn.ptnr2_label_atom_id,
            label_alt_id: struct_conn.pdbx_ptnr2_label_alt_id,
            ins_code: struct_conn.pdbx_ptnr2_PDB_ins_code,
            symmetry: struct_conn.ptnr2_symmetry
        };
        var _p = function (row, ps) {
            if (ps.label_asym_id.valueKind(row) !== 0 /* Present */)
                return void 0;
            var asymId = ps.label_asym_id.value(row);
            var residueIndex = model.atomicHierarchy.index.findResidue(util_1.findEntityIdByAsymId(model, asymId), asymId, ps.auth_seq_id.value(row), ps.ins_code.value(row));
            if (residueIndex < 0)
                return void 0;
            var atomName = ps.label_atom_id.value(row);
            // turns out "mismat" records might not have atom name value
            if (!atomName)
                return void 0;
            var atomIndex = util_1.findAtomIndexByLabelName(model, residueIndex, atomName, ps.label_alt_id.value(row));
            if (atomIndex < 0)
                return void 0;
            return { residueIndex: residueIndex, atomIndex: atomIndex, symmetry: ps.symmetry.value(row) || '1_555' };
        };
        var _ps = function (row) {
            var ret = [];
            var p = _p(row, p1);
            if (p)
                ret.push(p);
            p = _p(row, p2);
            if (p)
                ret.push(p);
            return ret;
        };
        var entries = [];
        for (var i = 0; i < struct_conn._rowCount; i++) {
            var partners = _ps(i);
            if (partners.length < 2)
                continue;
            var type = conn_type_id.value(i); // TODO workaround for dictionary inconsistency
            var orderType = (pdbx_value_order.value(i) || '').toLowerCase();
            var flags = 0 /* None */;
            var order = 1;
            switch (orderType) {
                case 'sing':
                    order = 1;
                    break;
                case 'doub':
                    order = 2;
                    break;
                case 'trip':
                    order = 3;
                    break;
                case 'quad':
                    order = 4;
                    break;
            }
            switch (type) {
                case 'covale':
                case 'covale_base':
                case 'covale_phosphate':
                case 'covale_sugar':
                case 'modres':
                    flags = 1 /* Covalent */;
                    break;
                case 'disulf':
                    flags = 1 /* Covalent */ | 16 /* Sulfide */;
                    break;
                case 'hydrog':
                case 'mismat':
                    flags = 4 /* Hydrogen */;
                    break;
                case 'metalc':
                    flags = 2 /* MetallicCoordination */;
                    break;
                case 'saltbr':
                    flags = 8 /* Ionic */;
                    break;
            }
            entries.push({ rowIndex: i, flags: flags, order: order, distance: pdbx_dist_value.value(i), partners: partners });
        }
        var ret = new StructConnImpl(entries);
        model._staticPropertyData[StructConn.PropName] = ret;
        return ret;
    }
    StructConn.get = get;
})(StructConn = exports.StructConn || (exports.StructConn = {}));
