"use strict";
/**
 * Copyright (c) 2018 mol* contributors, licensed under MIT, See LICENSE file for more info.
 *
 * @author David Sehnal <david.sehnal@gmail.com>
 */
Object.defineProperty(exports, "__esModule", { value: true });
var types_1 = require("../../../mol-model/structure/model/types");
var atomic_1 = require("../../../mol-model/structure/model/properties/atomic");
var seconday_structure_1 = require("../../../mol-model/structure/model/properties/seconday-structure");
function getSecondaryStructure(data, hierarchy) {
    var map = new Map();
    var elements = [{ kind: 'none' }];
    addHelices(data.struct_conf, map, elements);
    // must add Helices 1st because of 'key' value assignment.
    addSheets(data.struct_sheet_range, map, data.struct_conf._rowCount, elements);
    var n = hierarchy.residues._rowCount;
    var getIndex = function (rI) { return rI; };
    var secStruct = {
        type: new Int32Array(n),
        key: new Int32Array(n),
        elements: elements
    };
    if (map.size > 0)
        assignSecondaryStructureRanges(hierarchy, map, secStruct);
    return seconday_structure_1.SecondaryStructure(secStruct.type, secStruct.key, secStruct.elements, getIndex);
}
exports.getSecondaryStructure = getSecondaryStructure;
function addHelices(cat, map, elements) {
    if (!cat._rowCount)
        return;
    var beg_label_asym_id = cat.beg_label_asym_id, beg_label_seq_id = cat.beg_label_seq_id, pdbx_beg_PDB_ins_code = cat.pdbx_beg_PDB_ins_code;
    var end_label_seq_id = cat.end_label_seq_id, pdbx_end_PDB_ins_code = cat.pdbx_end_PDB_ins_code;
    var pdbx_PDB_helix_class = cat.pdbx_PDB_helix_class, conf_type_id = cat.conf_type_id, details = cat.details;
    for (var i = 0, _i = cat._rowCount; i < _i; i++) {
        var type = types_1.SecondaryStructureType.create(pdbx_PDB_helix_class.valueKind(i) === 0 /* Present */
            ? types_1.SecondaryStructureType.SecondaryStructurePdb[pdbx_PDB_helix_class.value(i)]
            : conf_type_id.valueKind(i) === 0 /* Present */
                ? types_1.SecondaryStructureType.SecondaryStructureMmcif[conf_type_id.value(i)]
                : 536870912 /* NA */);
        var element = {
            kind: 'helix',
            flags: type,
            type_id: conf_type_id.valueKind(i) === 0 /* Present */ ? conf_type_id.value(i) : 'HELIX_P',
            helix_class: pdbx_PDB_helix_class.value(i),
            details: details.valueKind(i) === 0 /* Present */ ? details.value(i) : void 0
        };
        var entry = {
            startSeqNumber: beg_label_seq_id.value(i),
            startInsCode: pdbx_beg_PDB_ins_code.value(i),
            endSeqNumber: end_label_seq_id.value(i),
            endInsCode: pdbx_end_PDB_ins_code.value(i),
            type: type,
            key: elements.length
        };
        elements[elements.length] = element;
        var asymId = beg_label_asym_id.value(i);
        if (map.has(asymId)) {
            map.get(asymId).set(entry.startSeqNumber, entry);
        }
        else {
            map.set(asymId, new Map([[entry.startSeqNumber, entry]]));
        }
    }
}
function addSheets(cat, map, sheetCount, elements) {
    if (!cat._rowCount)
        return;
    var beg_label_asym_id = cat.beg_label_asym_id, beg_label_seq_id = cat.beg_label_seq_id, pdbx_beg_PDB_ins_code = cat.pdbx_beg_PDB_ins_code;
    var end_label_seq_id = cat.end_label_seq_id, pdbx_end_PDB_ins_code = cat.pdbx_end_PDB_ins_code;
    var sheet_id = cat.sheet_id;
    var sheet_id_key = new Map();
    var currentKey = sheetCount + 1;
    for (var i = 0, _i = cat._rowCount; i < _i; i++) {
        var id = sheet_id.value(i);
        var key = void 0;
        if (sheet_id_key.has(id))
            key = sheet_id_key.get(id);
        else {
            key = currentKey++;
            sheet_id_key.set(id, key);
        }
        var type = types_1.SecondaryStructureType.create(4 /* Beta */ | 8388608 /* BetaSheet */);
        var element = {
            kind: 'sheet',
            flags: type,
            sheet_id: id,
            symmetry: void 0
        };
        var entry = {
            startSeqNumber: beg_label_seq_id.value(i),
            startInsCode: pdbx_beg_PDB_ins_code.value(i),
            endSeqNumber: end_label_seq_id.value(i),
            endInsCode: pdbx_end_PDB_ins_code.value(i),
            type: type,
            key: elements.length
        };
        elements[elements.length] = element;
        var asymId = beg_label_asym_id.value(i);
        if (map.has(asymId)) {
            map.get(asymId).set(entry.startSeqNumber, entry);
        }
        else {
            map.set(asymId, new Map([[entry.startSeqNumber, entry]]));
        }
    }
    return;
}
function assignSecondaryStructureEntry(hierarchy, entry, resStart, resEnd, data) {
    var _a = hierarchy.residues, label_seq_id = _a.label_seq_id, pdbx_PDB_ins_code = _a.pdbx_PDB_ins_code;
    var endSeqNumber = entry.endSeqNumber, endInsCode = entry.endInsCode, key = entry.key, type = entry.type;
    var rI = resStart;
    while (rI < resEnd) {
        var seqNumber = label_seq_id.value(rI);
        data.type[rI] = type;
        data.key[rI] = key;
        if ((seqNumber > endSeqNumber) ||
            (seqNumber === endSeqNumber && pdbx_PDB_ins_code.value(rI) === endInsCode)) {
            break;
        }
        rI++;
    }
}
function assignSecondaryStructureRanges(hierarchy, map, data) {
    var chainCount = hierarchy.chainAtomSegments.count;
    var label_asym_id = hierarchy.chains.label_asym_id;
    var _a = hierarchy.residues, label_seq_id = _a.label_seq_id, pdbx_PDB_ins_code = _a.pdbx_PDB_ins_code;
    for (var cI = 0; cI < chainCount; cI++) {
        var resStart = atomic_1.AtomicHierarchy.chainStartResidueIndex(hierarchy, cI), resEnd = atomic_1.AtomicHierarchy.chainEndResidueIndexExcl(hierarchy, cI);
        var asymId = label_asym_id.value(cI);
        if (map.has(asymId)) {
            var entries = map.get(asymId);
            for (var rI = resStart; rI < resEnd; rI++) {
                var seqNumber = label_seq_id.value(rI);
                if (entries.has(seqNumber)) {
                    var entry = entries.get(seqNumber);
                    var insCode = pdbx_PDB_ins_code.value(rI);
                    if (entry.startInsCode !== insCode)
                        continue;
                    assignSecondaryStructureEntry(hierarchy, entry, rI, resEnd, data);
                }
            }
        }
    }
}
