/**
 * Copyright (c) 2018-2019 mol* contributors, licensed under MIT, See LICENSE file for more info.
 *
 * @author Alexander Rose <alexander.rose@weirdbyte.de>
 */
import { Model, CustomPropertyDescriptor } from '../../mol-model/structure';
import { Database as _Database, Column, Table } from '../../mol-data/db';
import { CifExportContext } from '../../mol-model/structure/export/mmcif';
import { Task, RuntimeContext } from '../../mol-task';
import { GraphQLClient } from '../../mol-util/graphql-client';
export interface AssemblySymmetry {
    '@type': 'rcsb_assembly_symmetry';
    db: AssemblySymmetry.Database;
    getSymmetries(assemblyIds: string[]): Table<AssemblySymmetry.Schema['rcsb_assembly_symmetry']>;
    getClusters(symmetryId: number): Table<AssemblySymmetry.Schema['rcsb_assembly_symmetry_cluster']>;
    getClusterMembers(clusterId: number): Table<AssemblySymmetry.Schema['rcsb_assembly_symmetry_cluster_member']>;
    getAxes(symmetryId: number): Table<AssemblySymmetry.Schema['rcsb_assembly_symmetry_axis']>;
}
export declare function AssemblySymmetry(db: AssemblySymmetry.Database): AssemblySymmetry;
declare type SymmetryKind = 'GLOBAL' | 'LOCAL' | 'PSEUDO';
declare type SymmetryType = 'ASYMMETRIC' | 'CYCLIC' | 'DIHEDRAL' | 'HELICAL' | 'ICOSAHEDRAL' | 'OCTAHEDRAL' | 'TETRAHEDRAL';
export declare namespace AssemblySymmetry {
    function is(x: any): x is AssemblySymmetry;
    const GraphQLEndpointURL = "//rest-staging.rcsb.org/graphql";
    const Schema: {
        rcsb_assembly_symmetry_info: {
            updated_datetime_utc: Column.Schema.Str;
        };
        rcsb_assembly_symmetry: {
            /** Uniquely identifies a record in `rcsb_assembly_symmetry` */
            id: Column.Schema.Int;
            /**
             * A pointer to `pdbx_struct_assembly.id`.
             * The value 'deposited' refers to the coordinates as given in the file.
             * */
            assembly_id: Column.Schema.Str;
            /** Type of protein symmetry */
            kind: Column.Schema.Aliased<SymmetryKind>;
            /** Quantitative description of every individual subunit in a given protein */
            stoichiometry: Column.Schema.List<string>;
            /**
             * Symmetry symbol refers to point group or helical symmetry of identical subunits.
             * Contains point group symbol (e.g., C2, C5, D2, T, O, I) or H for helical symmetry.
             */
            symbol: Column.Schema.Str;
            /** Point group or helical symmetry */
            type: Column.Schema.Aliased<SymmetryType>;
            /**
             * Oligomeric state refers to a composition of subunits in quaternary structure.
             * Quaternary structure may be composed either exclusively of several copies of identical
             * subunits, in which case they are termed homo-oligomers, or alternatively by at least
             * one copy of different subunits (hetero-oligomers). Quaternary structure composed of
             * a single subunit is demoted as 'Monomer'.
             */
            oligomeric_state: Column.Schema.Str;
        };
        rcsb_assembly_symmetry_cluster: {
            /** Uniquely identifies a record in `rcsb_assembly_symmetry_cluster` */
            id: Column.Schema.Int;
            /** A pointer to `rcsb_assembly_symmetry.id` */
            symmetry_id: Column.Schema.Int;
            /** Average RMSD between members of a given cluster */
            avg_rmsd: Column.Schema.Float;
        };
        rcsb_assembly_symmetry_cluster_member: {
            /** A pointer to `rcsb_assembly_symmetry_cluster.id` */
            cluster_id: Column.Schema.Int;
            /** The `label_asym_id` value of the member */
            asym_id: Column.Schema.Str;
            /** List of `pdbx_struct_oper_list_id` values of the member */
            pdbx_struct_oper_list_ids: Column.Schema.List<string>;
        };
        rcsb_assembly_symmetry_axis: {
            /** A pointer to `rcsb_assembly_symmetry.id` */
            symmetry_id: Column.Schema.Int;
            /**
             * The number of times (order of rotation) that a subunit can be repeated by a rotation
             * operation, being transformed into a new state indistinguishable from its starting state.
             */
            order: Column.Schema.Int;
            /** The x,y,z coordinate of the start point of a symmetry axis */
            start: Column.Schema.Tensor;
            /** The x,y,z coordinate of the end point of a symmetry axis */
            end: Column.Schema.Tensor;
        };
    };
    type Schema = typeof Schema;
    interface Database extends _Database<Schema> {
    }
    const Descriptor: CustomPropertyDescriptor<CifExportContext, {}>;
    function attachFromCifOrAPI(model: Model, client: GraphQLClient, ctx?: RuntimeContext): Promise<boolean>;
    function createAttachTask(fetch: import('../../mol-util/data-source').AjaxTask): (model: Model) => Task<boolean>;
    function get(model: Model): AssemblySymmetry | undefined;
}
export {};
