"use strict";
/**
 * Copyright (c) 2018-2019 mol* contributors, licensed under MIT, See LICENSE file for more info.
 *
 * @author Alexander Rose <alexander.rose@weirdbyte.de>
 */
Object.defineProperty(exports, "__esModule", { value: true });
var structure_1 = require("./structure");
var links_1 = require("./structure/structure/unit/links");
var shape_1 = require("./shape");
var geometry_1 = require("../mol-math/geometry");
var centroid_helper_1 = require("../mol-math/geometry/centroid-helper");
var linear_algebra_1 = require("../mol-math/linear-algebra");
var int_1 = require("../mol-data/int");
var structure_2 = require("./structure/structure");
/** A Loci that includes every loci */
exports.EveryLoci = { kind: 'every-loci' };
function isEveryLoci(x) {
    return !!x && x.kind === 'every-loci';
}
exports.isEveryLoci = isEveryLoci;
/** A Loci that is empty */
exports.EmptyLoci = { kind: 'empty-loci' };
function isEmptyLoci(x) {
    return !!x && x.kind === 'empty-loci';
}
exports.isEmptyLoci = isEmptyLoci;
function isDataLoci(x) {
    return !!x && x.kind === 'data-loci';
}
exports.isDataLoci = isDataLoci;
function areDataLociEqual(a, b) {
    return a.data === b.data && a.tag === b.tag && int_1.OrderedSet.areEqual(a.indices, b.indices);
}
exports.areDataLociEqual = areDataLociEqual;
function createDataLoci(data, tag, indices) {
    return { kind: 'data-loci', data: data, tag: tag, indices: indices };
}
exports.createDataLoci = createDataLoci;
var Loci;
(function (Loci) {
    function areEqual(lociA, lociB) {
        if (isEveryLoci(lociA) && isEveryLoci(lociB))
            return true;
        if (isEmptyLoci(lociA) && isEmptyLoci(lociB))
            return true;
        if (isDataLoci(lociA) && isDataLoci(lociB)) {
            return areDataLociEqual(lociA, lociB);
        }
        if (structure_2.Structure.isLoci(lociA) && structure_2.Structure.isLoci(lociB)) {
            return structure_2.Structure.areLociEqual(lociA, lociB);
        }
        if (structure_1.StructureElement.isLoci(lociA) && structure_1.StructureElement.isLoci(lociB)) {
            return structure_1.StructureElement.areLociEqual(lociA, lociB);
        }
        if (links_1.Link.isLoci(lociA) && links_1.Link.isLoci(lociB)) {
            return links_1.Link.areLociEqual(lociA, lociB);
        }
        if (shape_1.Shape.isLoci(lociA) && shape_1.Shape.isLoci(lociB)) {
            return shape_1.Shape.areLociEqual(lociA, lociB);
        }
        if (shape_1.ShapeGroup.isLoci(lociA) && shape_1.ShapeGroup.isLoci(lociB)) {
            return shape_1.ShapeGroup.areLociEqual(lociA, lociB);
        }
        return false;
    }
    Loci.areEqual = areEqual;
    var sphereHelper = new centroid_helper_1.CentroidHelper(), tempPos = linear_algebra_1.Vec3.zero();
    function getBoundingSphere(loci, boundingSphere) {
        if (loci.kind === 'every-loci' || loci.kind === 'empty-loci')
            return void 0;
        if (!boundingSphere)
            boundingSphere = geometry_1.Sphere3D.zero();
        sphereHelper.reset();
        if (loci.kind === 'structure-loci') {
            return geometry_1.Sphere3D.copy(boundingSphere, loci.structure.boundary.sphere);
        }
        else if (loci.kind === 'element-loci') {
            return structure_1.StructureElement.Loci.getBoundary(loci).sphere;
        }
        else if (loci.kind === 'link-loci') {
            for (var _i = 0, _a = loci.links; _i < _a.length; _i++) {
                var e = _a[_i];
                e.aUnit.conformation.position(e.aUnit.elements[e.aIndex], tempPos);
                sphereHelper.includeStep(tempPos);
                e.bUnit.conformation.position(e.bUnit.elements[e.bIndex], tempPos);
                sphereHelper.includeStep(tempPos);
            }
            sphereHelper.finishedIncludeStep();
            for (var _b = 0, _c = loci.links; _b < _c.length; _b++) {
                var e = _c[_b];
                e.aUnit.conformation.position(e.aUnit.elements[e.aIndex], tempPos);
                sphereHelper.radiusStep(tempPos);
                e.aUnit.conformation.position(e.bUnit.elements[e.bIndex], tempPos);
                sphereHelper.radiusStep(tempPos);
            }
        }
        else if (loci.kind === 'shape-loci') {
            // TODO
            return void 0;
        }
        else if (loci.kind === 'group-loci') {
            // TODO
            return void 0;
        }
        else if (loci.kind === 'data-loci') {
            // TODO maybe add loci.getBoundingSphere()???
            return void 0;
        }
        linear_algebra_1.Vec3.copy(boundingSphere.center, sphereHelper.center);
        boundingSphere.radius = Math.sqrt(sphereHelper.radiusSq);
        return boundingSphere;
    }
    Loci.getBoundingSphere = getBoundingSphere;
    var tmpSphere3D = geometry_1.Sphere3D.zero();
    function getCenter(loci, center) {
        var boundingSphere = getBoundingSphere(loci, tmpSphere3D);
        return boundingSphere ? linear_algebra_1.Vec3.copy(center || linear_algebra_1.Vec3.zero(), boundingSphere.center) : undefined;
    }
    Loci.getCenter = getCenter;
})(Loci || (Loci = {}));
exports.Loci = Loci;
