"use strict";
/**
 * Copyright (c) 2017-2018 mol* contributors, licensed under MIT, See LICENSE file for more info.
 *
 * @author David Sehnal <david.sehnal@gmail.com>
 * @author Alexander Rose <alexander.rose@weirdbyte.de>
 */
Object.defineProperty(exports, "__esModule", { value: true });
var cif_1 = require("../../../mol-io/writer/cif");
var atom_site_1 = require("./categories/atom_site");
var CifCategory = cif_1.CifWriter.Category;
var secondary_structure_1 = require("./categories/secondary-structure");
var modified_residues_1 = require("./categories/modified-residues");
var misc_1 = require("./categories/misc");
var utils_1 = require("./categories/utils");
var sequence_1 = require("./categories/sequence");
var custom_property_1 = require("../common/custom-property");
var CifExportContext;
(function (CifExportContext) {
    function create(structures) {
        var structureArray = Array.isArray(structures) ? structures : [structures];
        return {
            structures: structureArray,
            firstModel: structureArray[0].model,
            cache: Object.create(null)
        };
    }
    CifExportContext.create = create;
})(CifExportContext = exports.CifExportContext || (exports.CifExportContext = {}));
var _entity = {
    name: 'entity',
    instance: function (_a) {
        var structures = _a.structures;
        var indices = utils_1.getUniqueEntityIndicesFromStructures(structures);
        return CifCategory.ofTable(structures[0].model.entities.data, indices);
    }
};
function isWithoutSymmetry(structure) {
    return structure.units.every(function (u) { return u.conformation.operator.isIdentity; });
}
var Categories = [
    // Basics
    utils_1.copy_mmCif_category('entry'),
    utils_1.copy_mmCif_category('exptl'),
    _entity,
    // Symmetry
    utils_1.copy_mmCif_category('cell', isWithoutSymmetry),
    utils_1.copy_mmCif_category('symmetry', isWithoutSymmetry),
    // Assemblies
    utils_1.copy_mmCif_category('pdbx_struct_assembly', isWithoutSymmetry),
    utils_1.copy_mmCif_category('pdbx_struct_assembly_gen', isWithoutSymmetry),
    utils_1.copy_mmCif_category('pdbx_struct_oper_list', isWithoutSymmetry),
    // Secondary structure
    secondary_structure_1._struct_conf,
    secondary_structure_1._struct_sheet_range,
    // Sequence
    sequence_1._struct_asym,
    sequence_1._entity_poly,
    sequence_1._entity_poly_seq,
    // Branch
    utils_1.copy_mmCif_category('pdbx_entity_branch'),
    utils_1.copy_mmCif_category('pdbx_entity_branch_link'),
    utils_1.copy_mmCif_category('pdbx_branch_scheme'),
    // Misc
    // TODO: filter for actual present residues?
    misc_1._chem_comp,
    misc_1._pdbx_chem_comp_identifier,
    utils_1.copy_mmCif_category('atom_sites'),
    misc_1._pdbx_nonpoly_scheme,
    modified_residues_1._pdbx_struct_mod_residue,
    // Atoms
    atom_site_1._atom_site
];
var _Filters;
(function (_Filters) {
    _Filters.AtomSitePositionsFieldNames = new Set(['id', 'Cartn_x', 'Cartn_y', 'Cartn_z']);
})(_Filters || (_Filters = {}));
exports.mmCIF_Export_Filters = {
    onlyPositions: {
        includeCategory: function (name) { return name === 'atom_site'; },
        includeField: function (cat, field) { return _Filters.AtomSitePositionsFieldNames.has(field); }
    }
};
function encodeCustomProp(customProp, ctx, encoder, params) {
    if (!customProp.cifExport || customProp.cifExport.categories.length === 0)
        return;
    var prefix = customProp.cifExport.prefix;
    var cats = customProp.cifExport.categories;
    var propCtx = ctx;
    if (customProp.cifExport.context) {
        var propId = custom_property_1.CustomPropertyDescriptor.getUUID(customProp);
        if (ctx.cache[propId + '__ctx'])
            propCtx = ctx.cache[propId + '__ctx'];
        else {
            propCtx = customProp.cifExport.context(ctx) || ctx;
            ctx.cache[propId + '__ctx'] = propCtx;
        }
    }
    for (var _i = 0, cats_1 = cats; _i < cats_1.length; _i++) {
        var cat = cats_1[_i];
        if (params.skipCategoryNames && params.skipCategoryNames.has(cat.name))
            continue;
        if (cat.name.indexOf(prefix) !== 0)
            throw new Error("Custom category '" + cat.name + "' name must start with prefix '" + prefix + ".'");
        encoder.writeCategory(cat, propCtx);
    }
}
/** Doesn't start a data block */
function encode_mmCIF_categories(encoder, structures, params) {
    var first = Array.isArray(structures) ? structures[0] : structures;
    var models = first.models;
    if (models.length !== 1)
        throw 'Can\'t export stucture composed from multiple models.';
    var _params = params || {};
    var ctx = params && params.exportCtx ? params.exportCtx : CifExportContext.create(structures);
    for (var _i = 0, Categories_1 = Categories; _i < Categories_1.length; _i++) {
        var cat = Categories_1[_i];
        if (_params.skipCategoryNames && _params.skipCategoryNames.has(cat.name))
            continue;
        encoder.writeCategory(cat, ctx);
    }
    for (var _a = 0, _b = models[0].customProperties.all; _a < _b.length; _a++) {
        var customProp = _b[_a];
        encodeCustomProp(customProp, ctx, encoder, _params);
    }
    var structureCustomProps = new Set();
    for (var _c = 0, _d = ctx.structures; _c < _d.length; _c++) {
        var s = _d[_c];
        if (!s.hasCustomProperties)
            continue;
        for (var _e = 0, _f = s.customPropertyDescriptors.all; _e < _f.length; _e++) {
            var p = _f[_e];
            structureCustomProps.add(p);
        }
    }
    structureCustomProps.forEach(function (customProp) { return encodeCustomProp(customProp, ctx, encoder, _params); });
}
exports.encode_mmCIF_categories = encode_mmCIF_categories;
function to_mmCIF(name, structure, asBinary) {
    if (asBinary === void 0) { asBinary = false; }
    var enc = cif_1.CifWriter.createEncoder({ binary: asBinary });
    enc.startDataBlock(name);
    encode_mmCIF_categories(enc, structure);
    return enc.getData();
}
exports.default = to_mmCIF;
