/**
 * Copyright (c) 2017-2019 mol* contributors, licensed under MIT, See LICENSE file for more info.
 *
 * @author Alexander Rose <alexander.rose@weirdbyte.de>
 * @author David Sehnal <david.sehnal@gmail.com>
 */
import BitFlags from '../../../mol-util/bit-flags';
import { mmCIF_Schema } from '../../../mol-io/reader/cif/schema/mmcif';
export declare type ElementSymbol = string & {
    '@type': 'element-symbol';
};
export declare function ElementSymbol(s: string): ElementSymbol;
/** Entity types as defined in the mmCIF dictionary */
export declare const enum EntityType {
    'unknown' = 0,
    'polymer' = 1,
    'non-polymer' = 2,
    'macrolide' = 3,
    'water' = 4,
    'branched' = 5
}
export declare const enum MoleculeType {
    /** the molecule type is not known */
    unknown = 0,
    /** a known, but here not listed molecule type */
    other = 1,
    /** water molecule */
    water = 2,
    /** small ionic molecule */
    ion = 3,
    /** protein, e.g. component type included in `ProteinComponentTypeNames` */
    protein = 4,
    /** RNA, e.g. component type included in `RNAComponentTypeNames` */
    RNA = 5,
    /** DNA, e.g. component type included in `DNAComponentTypeNames` */
    DNA = 6,
    /** PNA, peptide nucleic acid, comp id included in `PeptideBaseNames` */
    PNA = 7,
    /** sacharide, e.g. component type included in `SaccharideComponentTypeNames` */
    saccharide = 8
}
export declare type AtomRole = 'trace' | 'directionFrom' | 'directionTo' | 'backboneStart' | 'backboneEnd' | 'coarseBackbone';
export declare const MoleculeTypeAtomRoleId: {
    [k: number]: {
        [k in AtomRole]: Set<string>;
    };
};
export declare const ProteinBackboneAtoms: Set<string>;
export declare const NucleicBackboneAtoms: Set<string>;
/** Chemical component type names for protein */
export declare const ProteinComponentTypeNames: Set<string>;
/** Chemical component type names for DNA */
export declare const DNAComponentTypeNames: string[];
/** Chemical component type names for RNA */
export declare const RNAComponentTypeNames: Set<string>;
/** Chemical component type names for saccharide */
export declare const SaccharideComponentTypeNames: Set<string>;
/** Chemical component type names for other */
export declare const OtherComponentTypeNames: Set<string>;
/** Common names for water molecules */
export declare const WaterNames: Set<string>;
export declare const AminoAcidNames: Set<string>;
export declare const RnaBaseNames: Set<string>;
export declare const DnaBaseNames: Set<string>;
export declare const PeptideBaseNames: Set<string>;
export declare const PurinBaseNames: Set<string>;
export declare const PyrimidineBaseNames: Set<string>;
export declare const BaseNames: Set<string>;
export declare const isPurinBase: (compId: string) => boolean;
export declare const isPyrimidineBase: (compId: string) => boolean;
/** get the molecule type from component type and id */
export declare function getMoleculeType(compType: string, compId: string): MoleculeType;
export declare function getComponentType(compId: string): mmCIF_Schema['chem_comp']['type']['T'];
export declare function getEntityType(compId: string): mmCIF_Schema['entity']['type']['T'];
export declare function isPolymer(moleculeType: MoleculeType): boolean;
export declare function isNucleic(moleculeType: MoleculeType): boolean;
export declare function isProtein(moleculeType: MoleculeType): boolean;
/**
 * TODO write script that read CCD and outputs list of ion names
 *
 * all chemical components with the word "ion" in their name, Sep 2016
 *
 * SET SESSION group_concat_max_len = 1000000;
 * SELECT GROUP_CONCAT(id_ ORDER BY id_ ASC SEPARATOR '", "') from
 * (
 *     SELECT count(obj_id) as c, id_
 *     FROM pdb.chem_comp WHERE name LIKE "% ION%"
 *     GROUP BY id_
 * ) AS t1;
 */
export declare const IonNames: Set<string>;
export declare type SecondaryStructureType = BitFlags<SecondaryStructureType.Flag>;
export declare namespace SecondaryStructureType {
    const is: (ss: SecondaryStructureType, f: Flag) => boolean;
    const create: (fs: Flag) => SecondaryStructureType;
    const enum Flag {
        None = 0,
        DoubleHelix = 1,
        Helix = 2,
        Beta = 4,
        Bend = 8,
        Turn = 16,
        LeftHanded = 32,
        RightHanded = 64,
        ClassicTurn = 128,
        InverseTurn = 256,
        HelixOther = 512,
        Helix27 = 1024,
        Helix3Ten = 2048,
        HelixAlpha = 4096,
        HelixGamma = 8192,
        HelixOmega = 16384,
        HelixPi = 32768,
        HelixPolyproline = 65536,
        DoubleHelixOther = 131072,
        DoubleHelixZ = 262144,
        DoubleHelixA = 524288,
        DoubleHelixB = 1048576,
        BetaOther = 2097152,
        BetaStrand = 4194304,
        BetaSheet = 8388608,
        BetaBarell = 16777216,
        TurnOther = 33554432,
        Turn1 = 67108864,
        Turn2 = 134217728,
        Turn3 = 268435456,
        NA = 536870912
    }
    const SecondaryStructureMmcif: {
        [value: string]: number;
    };
    const SecondaryStructurePdb: {
        [value: string]: number;
    };
    const SecondaryStructureStride: {
        [value: string]: number;
    };
    const SecondaryStructureDssp: {
        [value: string]: number;
    };
}
/** Maximum accessible surface area observed for amino acids. Taken from: http://dx.doi.org/10.1371/journal.pone.0080635 */
export declare const MaxAsa: {
    'ALA': number;
    'ARG': number;
    'ASN': number;
    'ASP': number;
    'CYS': number;
    'GLU': number;
    'GLN': number;
    'GLY': number;
    'HIS': number;
    'ILE': number;
    'LEU': number;
    'LYS': number;
    'MET': number;
    'PHE': number;
    'PRO': number;
    'SER': number;
    'THR': number;
    'TRP': number;
    'TYR': number;
    'VAL': number;
};
export declare const DefaultMaxAsa = 121;
export declare const VdwRadii: {
    'H': number;
    'HE': number;
    'LI': number;
    'BE': number;
    'B': number;
    'C': number;
    'N': number;
    'O': number;
    'F': number;
    'NE': number;
    'NA': number;
    'MG': number;
    'AL': number;
    'SI': number;
    'P': number;
    'S': number;
    'CL': number;
    'AR': number;
    'K': number;
    'CA': number;
    'SC': number;
    'TI': number;
    'V': number;
    'CR': number;
    'MN': number;
    'FE': number;
    'CO': number;
    'NI': number;
    'CU': number;
    'ZN': number;
    'GA': number;
    'GE': number;
    'AS': number;
    'SE': number;
    'BR': number;
    'KR': number;
    'RB': number;
    'SR': number;
    'Y': number;
    'ZR': number;
    'NB': number;
    'MO': number;
    'TC': number;
    'RU': number;
    'RH': number;
    'PD': number;
    'AG': number;
    'CD': number;
    'IN': number;
    'SN': number;
    'SB': number;
    'TE': number;
    'I': number;
    'XE': number;
    'CS': number;
    'BA': number;
    'LA': number;
    'CE': number;
    'PR': number;
    'ND': number;
    'PM': number;
    'SM': number;
    'EU': number;
    'GD': number;
    'TB': number;
    'DY': number;
    'HO': number;
    'ER': number;
    'TM': number;
    'YB': number;
    'LU': number;
    'HF': number;
    'TA': number;
    'W': number;
    'RE': number;
    'OS': number;
    'IR': number;
    'PT': number;
    'AU': number;
    'HG': number;
    'TL': number;
    'PB': number;
    'BI': number;
    'PO': number;
    'AT': number;
    'RN': number;
    'FR': number;
    'RA': number;
    'AC': number;
    'TH': number;
    'PA': number;
    'U': number;
    'NP': number;
    'PU': number;
    'AM': number;
    'CM': number;
    'BK': number;
    'CF': number;
    'ES': number;
    'FM': number;
    'MD': number;
    'NO': number;
    'LR': number;
    'RF': number;
    'DB': number;
    'SG': number;
    'BH': number;
    'HS': number;
    'MT': number;
    'DS': number;
    'RG': number;
    'CN': number;
    'UUT': number;
    'FL': number;
    'UUP': number;
    'LV': number;
    'UUH': number;
};
export declare const DefaultVdwRadius = 2;
export declare type LinkType = BitFlags<LinkType.Flag>;
export declare namespace LinkType {
    const is: (b: LinkType, f: Flag) => boolean;
    const enum Flag {
        None = 0,
        Covalent = 1,
        MetallicCoordination = 2,
        Hydrogen = 4,
        Ionic = 8,
        Sulfide = 16,
        Aromatic = 32,
        Computed = 64
    }
    function create(flags: Flag): LinkType;
    function isCovalent(flags: LinkType.Flag): boolean;
}
/**
 * "Experimentally determined hydrophobicity scale for proteins at membrane interfaces"
 * by Wimely and White (doi:10.1038/nsb1096-842)
 * http://blanco.biomol.uci.edu/Whole_residue_HFscales.txt
 * https://www.nature.com/articles/nsb1096-842
 */
export declare const ResidueHydrophobicity: {
    'ALA': number[];
    'ARG': number[];
    'ASN': number[];
    'ASP': number[];
    'ASH': number[];
    'CYS': number[];
    'GLN': number[];
    'GLU': number[];
    'GLH': number[];
    'GLY': number[];
    'HIS': number[];
    'ILE': number[];
    'LEU': number[];
    'LYS': number[];
    'MET': number[];
    'PHE': number[];
    'PRO': number[];
    'SER': number[];
    'THR': number[];
    'TRP': number[];
    'TYR': number[];
    'VAL': number[];
};
export declare const DefaultResidueHydrophobicity: number[];
