"use strict";
/**
 * Copyright (c) 2017 mol* contributors, licensed under MIT, See LICENSE file for more info.
 *
 * @author David Sehnal <david.sehnal@gmail.com>
 */
var __assign = (this && this.__assign) || function () {
    __assign = Object.assign || function(t) {
        for (var s, i = 1, n = arguments.length; i < n; i++) {
            s = arguments[i];
            for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p))
                t[p] = s[p];
        }
        return t;
    };
    return __assign.apply(this, arguments);
};
Object.defineProperty(exports, "__esModule", { value: true });
var int_1 = require("../../../../mol-data/int");
var selection_1 = require("../selection");
var builders_1 = require("../utils/builders");
var unique_subset_builder_1 = require("../../structure/util/unique-subset-builder");
var structure_set_1 = require("../utils/structure-set");
var generic_1 = require("../../../../mol-data/generic");
function getWholeResidues(ctx, source, structure) {
    var builder = source.subsetBuilder(true);
    for (var _a = 0, _b = structure.units; _a < _b.length; _a++) {
        var unit = _b[_a];
        if (unit.kind !== 0 /* Atomic */) {
            // just copy non-atomic units.
            builder.setUnit(unit.id, unit.elements);
            continue;
        }
        var residueAtomSegments = unit.model.atomicHierarchy.residueAtomSegments;
        var elements = unit.elements;
        builder.beginUnit(unit.id);
        var residuesIt = int_1.Segmentation.transientSegments(residueAtomSegments, elements);
        while (residuesIt.hasNext) {
            var rI = residuesIt.move().index;
            for (var j = residueAtomSegments.offsets[rI], _j = residueAtomSegments.offsets[rI + 1]; j < _j; j++) {
                builder.addElement(j);
            }
        }
        builder.commitUnit();
        ctx.throwIfTimedOut();
    }
    return builder.getStructure();
}
function wholeResidues(query) {
    return function (ctx) {
        var inner = query(ctx);
        if (selection_1.StructureSelection.isSingleton(inner)) {
            return selection_1.StructureSelection.Singletons(ctx.inputStructure, getWholeResidues(ctx, ctx.inputStructure, inner.structure));
        }
        else {
            var builder = new builders_1.UniqueStructuresBuilder(ctx.inputStructure);
            for (var _a = 0, _b = inner.structures; _a < _b.length; _a++) {
                var s = _b[_a];
                builder.add(getWholeResidues(ctx, ctx.inputStructure, s));
            }
            return builder.getSelection();
        }
    };
}
exports.wholeResidues = wholeResidues;
function getIncludeSurroundings(ctx, source, structure, params) {
    var builder = new unique_subset_builder_1.StructureUniqueSubsetBuilder(source);
    var lookup = source.lookup3d;
    var r = params.radius;
    for (var _a = 0, _b = structure.units; _a < _b.length; _a++) {
        var unit = _b[_a];
        var _c = unit.conformation, x = _c.x, y = _c.y, z = _c.z;
        var elements = unit.elements;
        for (var i = 0, _i = elements.length; i < _i; i++) {
            var e = elements[i];
            lookup.findIntoBuilder(x(e), y(e), z(e), r, builder);
        }
        ctx.throwIfTimedOut();
    }
    return !!params.wholeResidues ? getWholeResidues(ctx, source, builder.getStructure()) : builder.getStructure();
}
function getIncludeSurroundingsWithRadius(ctx, source, structure, params) {
    var builder = new unique_subset_builder_1.StructureUniqueSubsetBuilder(source);
    var lookup = source.lookup3d;
    var elementRadius = params.elementRadius, elementRadiusClosure = params.elementRadiusClosure, sourceMaxRadius = params.sourceMaxRadius, radius = params.radius;
    ctx.pushCurrentElement();
    for (var _a = 0, _b = structure.units; _a < _b.length; _a++) {
        var unit = _b[_a];
        ctx.element.unit = unit;
        var _c = unit.conformation, x = _c.x, y = _c.y, z = _c.z;
        var elements = unit.elements;
        for (var i = 0, _i = elements.length; i < _i; i++) {
            var e = elements[i];
            ctx.element.element = e;
            var eRadius = elementRadius(ctx);
            lookup.findIntoBuilderWithRadius(x(e), y(e), z(e), eRadius, sourceMaxRadius, radius, elementRadiusClosure, builder);
        }
        ctx.throwIfTimedOut();
    }
    ctx.popCurrentElement();
    return !!params.wholeResidues ? getWholeResidues(ctx, source, builder.getStructure()) : builder.getStructure();
}
function createElementRadiusFn(ctx, eRadius) {
    return function (e) {
        ctx.element.unit = e.unit;
        ctx.element.element = e.element;
        return eRadius(ctx);
    };
}
function findStructureRadius(ctx, eRadius) {
    var r = 0;
    for (var _a = 0, _b = ctx.inputStructure.units; _a < _b.length; _a++) {
        var unit = _b[_a];
        ctx.element.unit = unit;
        var elements = unit.elements;
        for (var i = 0, _i = elements.length; i < _i; i++) {
            var e = elements[i];
            ctx.element.element = e;
            var eR = eRadius(ctx);
            if (eR > r)
                r = eR;
        }
    }
    ctx.throwIfTimedOut();
    return r;
}
function includeSurroundings(query, params) {
    return function (ctx) {
        var inner = query(ctx);
        if (params.elementRadius) {
            var prms = __assign({}, params, { elementRadius: params.elementRadius, elementRadiusClosure: createElementRadiusFn(ctx, params.elementRadius), sourceMaxRadius: findStructureRadius(ctx, params.elementRadius) });
            if (selection_1.StructureSelection.isSingleton(inner)) {
                var surr = getIncludeSurroundingsWithRadius(ctx, ctx.inputStructure, inner.structure, prms);
                var ret = selection_1.StructureSelection.Singletons(ctx.inputStructure, surr);
                return ret;
            }
            else {
                var builder = new builders_1.UniqueStructuresBuilder(ctx.inputStructure);
                for (var _a = 0, _b = inner.structures; _a < _b.length; _a++) {
                    var s = _b[_a];
                    builder.add(getIncludeSurroundingsWithRadius(ctx, ctx.inputStructure, s, prms));
                }
                return builder.getSelection();
            }
        }
        if (selection_1.StructureSelection.isSingleton(inner)) {
            var surr = getIncludeSurroundings(ctx, ctx.inputStructure, inner.structure, params);
            var ret = selection_1.StructureSelection.Singletons(ctx.inputStructure, surr);
            return ret;
        }
        else {
            var builder = new builders_1.UniqueStructuresBuilder(ctx.inputStructure);
            for (var _c = 0, _d = inner.structures; _c < _d.length; _c++) {
                var s = _d[_c];
                builder.add(getIncludeSurroundings(ctx, ctx.inputStructure, s, params));
            }
            return builder.getSelection();
        }
    };
}
exports.includeSurroundings = includeSurroundings;
function querySelection(selection, query) {
    return function (ctx) {
        var targetSel = selection(ctx);
        if (selection_1.StructureSelection.structureCount(targetSel) === 0)
            return targetSel;
        var ret = selection_1.StructureSelection.UniqueBuilder(ctx.inputStructure);
        var add = function (s) { return ret.add(s); };
        selection_1.StructureSelection.forEach(targetSel, function (s, sI) {
            ctx.pushInputStructure(s);
            selection_1.StructureSelection.forEach(query(ctx), add);
            ctx.popInputStructure();
            if (sI % 10 === 0)
                ctx.throwIfTimedOut();
        });
        return ret.getSelection();
    };
}
exports.querySelection = querySelection;
function intersectBy(query, by) {
    return function (ctx) {
        var selection = query(ctx);
        if (selection_1.StructureSelection.structureCount(selection) === 0)
            return selection;
        var bySel = by(ctx);
        if (selection_1.StructureSelection.structureCount(bySel) === 0)
            return selection_1.StructureSelection.Empty(ctx.inputStructure);
        var unionBy = selection_1.StructureSelection.unionStructure(bySel);
        var ret = selection_1.StructureSelection.UniqueBuilder(ctx.inputStructure);
        selection_1.StructureSelection.forEach(selection, function (s, sI) {
            var ii = structure_set_1.structureIntersect(unionBy, s);
            if (ii.elementCount !== 0)
                ret.add(ii);
            if (sI % 50 === 0)
                ctx.throwIfTimedOut();
        });
        return ret.getSelection();
    };
}
exports.intersectBy = intersectBy;
function exceptBy(query, by) {
    return function (ctx) {
        var selection = query(ctx);
        if (selection_1.StructureSelection.structureCount(selection) === 0)
            return selection;
        var bySel = by(ctx);
        if (selection_1.StructureSelection.structureCount(bySel) === 0)
            return selection_1.StructureSelection.Empty(ctx.inputStructure);
        var subtractBy = selection_1.StructureSelection.unionStructure(bySel);
        var ret = selection_1.StructureSelection.UniqueBuilder(ctx.inputStructure);
        selection_1.StructureSelection.forEach(selection, function (s, sI) {
            var diff = structure_set_1.structureSubtract(s, subtractBy);
            if (diff.elementCount !== 0)
                ret.add(diff);
            if (sI % 50 === 0)
                ctx.throwIfTimedOut();
        });
        return ret.getSelection();
    };
}
exports.exceptBy = exceptBy;
function union(query) {
    return function (ctx) {
        var ret = selection_1.StructureSelection.LinearBuilder(ctx.inputStructure);
        ret.add(selection_1.StructureSelection.unionStructure(query(ctx)));
        return ret.getSelection();
    };
}
exports.union = union;
function expandProperty(query, property) {
    return function (ctx) {
        var src = query(ctx);
        var propertyToStructureIndexMap = new Map();
        var builders = [];
        ctx.pushCurrentElement();
        selection_1.StructureSelection.forEach(src, function (s, sI) {
            for (var _a = 0, _b = s.units; _a < _b.length; _a++) {
                var unit = _b[_a];
                ctx.element.unit = unit;
                var elements = unit.elements;
                for (var i = 0, _i = elements.length; i < _i; i++) {
                    ctx.element.element = elements[i];
                    var p = property(ctx);
                    var arr = void 0;
                    if (propertyToStructureIndexMap.has(p))
                        arr = propertyToStructureIndexMap.get(p);
                    else {
                        arr = generic_1.UniqueArray.create();
                        propertyToStructureIndexMap.set(p, arr);
                    }
                    generic_1.UniqueArray.add(arr, sI, sI);
                }
            }
            builders[sI] = ctx.inputStructure.subsetBuilder(true);
            if (sI % 10 === 0)
                ctx.throwIfTimedOut();
        });
        for (var _a = 0, _b = ctx.inputStructure.units; _a < _b.length; _a++) {
            var unit = _b[_a];
            ctx.element.unit = unit;
            var elements = unit.elements;
            for (var i = 0, _i = elements.length; i < _i; i++) {
                ctx.element.element = elements[i];
                var p = property(ctx);
                if (!propertyToStructureIndexMap.has(p))
                    continue;
                var indices = propertyToStructureIndexMap.get(p).array;
                for (var _sI = 0, __sI = indices.length; _sI < __sI; _sI++) {
                    builders[indices[_sI]].addToUnit(unit.id, elements[i]);
                }
            }
        }
        ctx.popCurrentElement();
        var ret = selection_1.StructureSelection.UniqueBuilder(ctx.inputStructure);
        for (var _c = 0, builders_2 = builders; _c < builders_2.length; _c++) {
            var b = builders_2[_c];
            ret.add(b.getStructure());
        }
        return ret.getSelection();
    };
}
exports.expandProperty = expandProperty;
function includeConnected(_a) {
    var query = _a.query, layerCount = _a.layerCount, wholeResidues = _a.wholeResidues, bondTest = _a.bondTest;
    return function (ctx) {
        return 0;
    };
}
exports.includeConnected = includeConnected;
// function defaultBondTest(ctx: QueryContext) {
//     return true;
// }
// interface IncludeConnectedCtx {
//     queryCtx: QueryContext,
//     input: Structure,
//     bondTest: QueryFn<boolean>,
//     wholeResidues: boolean
// }
// type FrontierSet = UniqueArray<StructureElement.UnitIndex, StructureElement.UnitIndex>
// type Frontier = { unitIds: UniqueArray<number>, elements: Map<number /* unit id */, FrontierSet> }
// namespace Frontier {
//     export function has({ elements }: Frontier, unitId: number, element: StructureElement.UnitIndex) {
//         if (!elements.has(unitId)) return false;
//         const xs = elements.get(unitId)!;
//         return xs.keys.has(element);
//     }
//     export function create(pivot: Structure, input: Structure) {
//         const unitIds = UniqueArray.create<number>();
//         const elements: Frontier['elements'] = new Map();
//         for (const unit of pivot.units) {
//             if (!Unit.isAtomic(unit)) continue;
//             UniqueArray.add(unitIds, unit.id, unit.id);
//             const xs: FrontierSet = UniqueArray.create();
//             elements.set(unit.id, xs);
//             const pivotElements = unit.elements;
//             const inputElements = input.unitMap.get(unit.id).elements;
//             for (let i = 0, _i = pivotElements.length; i < _i; i++) {
//                 const idx = SortedArray.indexOf(inputElements, pivotElements[i]) as StructureElement.UnitIndex;
//                 UniqueArray.add(xs, idx, idx);
//             }
//         }
//         return { unitIds, elements };
//     }
//     export function addFrontier(target: Frontier, from: Frontier) {
//         for (const unitId of from.unitIds.array) {
//             let xs: FrontierSet;
//             if (target.elements.has(unitId)) {
//                 xs = target.elements.get(unitId)!;
//             } else {
//                 xs = UniqueArray.create();
//                 target.elements.set(unitId, xs);
//                 UniqueArray.add(target.unitIds, unitId, unitId);
//             }
//             for (const e of from.elements.get(unitId)!.array) {
//                 UniqueArray.add(xs, e, e);
//             }
//         }
//         return target;
//     }
//     export function includeWholeResidues(structure: Structure, frontier: Frontier) {
//         // ...
//     }
// }
// function expandFrontier(ctx: IncludeConnectedCtx, currentFrontier: Frontier, result: Frontier): Frontier {
//     return 0 as any;
// }
// TODO: unionBy (skip this one?), cluster
