"use strict";
/**
 * Copyright (c) 2018 mol* contributors, licensed under MIT, See LICENSE file for more info.
 *
 * @author Alexander Rose <alexander.rose@weirdbyte.de>
 * @author David Sehnal <david.sehnal@gmail.com>
 */
var __assign = (this && this.__assign) || function () {
    __assign = Object.assign || function(t) {
        for (var s, i = 1, n = arguments.length; i < n; i++) {
            s = arguments[i];
            for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p))
                t[p] = s[p];
        }
        return t;
    };
    return __assign.apply(this, arguments);
};
Object.defineProperty(exports, "__esModule", { value: true });
var int_1 = require("../../../../mol-data/int");
var combination_1 = require("../../../../mol-data/util/combination");
var graph_1 = require("../../../../mol-math/graph");
var linear_algebra_1 = require("../../../../mol-math/linear-algebra");
var principal_axes_1 = require("../../../../mol-math/linear-algebra/matrix/principal-axes");
var array_1 = require("../../../../mol-util/array");
var types_1 = require("../../model/types");
var util_1 = require("../../util");
var unit_1 = require("../unit");
var data_1 = require("./data");
var rings_1 = require("../unit/rings");
var C = types_1.ElementSymbol('C'), O = types_1.ElementSymbol('O');
var SugarRingFps = [rings_1.UnitRing.elementFingerprint([C, C, C, C, C, O]), rings_1.UnitRing.elementFingerprint([C, C, C, C, O])];
function getAnomericCarbon(unit, ringAtoms) {
    var indexHasTwoOxygen = -1, indexHasOxygenAndCarbon = -1, indexHasC1Name = -1, indexIsCarbon = -1;
    var elements = unit.elements;
    var _a = unit.model.atomicHierarchy.atoms, type_symbol = _a.type_symbol, label_atom_id = _a.label_atom_id;
    var _b = unit.links, neighbor = _b.b, offset = _b.offset;
    for (var i = 0, il = ringAtoms.length; i < il; ++i) {
        var ei = elements[ringAtoms[i]];
        if (type_symbol.value(ei) !== C)
            continue;
        var linkedOxygenCount = 0;
        var linkedCarbonCount = 0;
        for (var j = offset[ringAtoms[i]], jl = offset[ringAtoms[i] + 1]; j < jl; ++j) {
            var ej = elements[neighbor[j]];
            var typeSymbol = type_symbol.value(ej);
            if (typeSymbol === O)
                ++linkedOxygenCount;
            else if (typeSymbol === C)
                ++linkedCarbonCount;
        }
        if (linkedOxygenCount === 2) {
            // found anomeric carbon
            indexHasTwoOxygen = ei;
            break;
        }
        else if (linkedOxygenCount === 1 && linkedCarbonCount === 1) {
            // possibly an anomeric carbon if this is a mono-saccharide without a glycosidic bond
            indexHasOxygenAndCarbon = ei;
        }
        else if (label_atom_id.value(ei).startsWith('C1')) {
            // likely the anomeric carbon as it is named C1 by convention
            indexHasC1Name = ei;
        }
        else {
            // use any carbon as a fallback
            indexIsCarbon = ei;
        }
    }
    return (indexHasTwoOxygen !== -1 ? indexHasTwoOxygen
        : indexHasOxygenAndCarbon !== -1 ? indexHasOxygenAndCarbon
            : indexHasC1Name !== -1 ? indexHasC1Name
                : indexIsCarbon !== -1 ? indexIsCarbon
                    : elements[ringAtoms[0]]);
}
/** Return first non-empty label_alt_id or an empty string */
function getRingAltId(unit, ringAtoms) {
    var elements = unit.elements;
    var label_alt_id = unit.model.atomicHierarchy.atoms.label_alt_id;
    for (var i = 0, il = ringAtoms.length; i < il; ++i) {
        var ei = elements[ringAtoms[i]];
        var altId = label_alt_id.value(ei);
        if (altId)
            return altId;
    }
    return '';
}
function getAltId(unit, index) {
    var elements = unit.elements;
    var label_alt_id = unit.model.atomicHierarchy.atoms.label_alt_id;
    return label_alt_id.value(elements[index]);
}
function getDirection(direction, unit, index, center) {
    var position = unit.conformation.position;
    linear_algebra_1.Vec3.normalize(direction, linear_algebra_1.Vec3.sub(direction, center, position(index, direction)));
    return direction;
}
function getAtomId(unit, index) {
    var elements = unit.elements;
    var label_atom_id = unit.model.atomicHierarchy.atoms.label_atom_id;
    return label_atom_id.value(elements[index]);
}
function filterFusedRings(unitRings, rings) {
    if (!rings || !rings.length)
        return;
    var fusedRings = new Set();
    var ringCombinations = combination_1.combinations(array_1.fillSerial(new Array(rings.length)), 2);
    for (var i = 0, il = ringCombinations.length; i < il; ++i) {
        var rc = ringCombinations[i];
        var r0 = unitRings.all[rings[rc[0]]], r1 = unitRings.all[rings[rc[1]]];
        if (int_1.SortedArray.areIntersecting(r0, r1)) {
            fusedRings.add(rings[rc[0]]);
            fusedRings.add(rings[rc[1]]);
        }
    }
    if (fusedRings.size) {
        var filteredRings = [];
        for (var i = 0, il = rings.length; i < il; ++i) {
            if (!fusedRings.has(rings[i]))
                filteredRings.push(rings[i]);
        }
        return filteredRings;
    }
    else {
        return rings;
    }
}
function getSaccharideComp(compId, model) {
    return model.properties.saccharideComponentMap.get(compId);
}
function computeCarbohydrates(structure) {
    // skip computation if there are no saccharide components in any model
    if (structure.models.reduce(function (a, v) { return a + v.properties.saccharideComponentMap.size; }, 0) === 0)
        return data_1.EmptyCarbohydrates;
    var links = [];
    var terminalLinks = [];
    var elements = [];
    var partialElements = [];
    var elementsWithRingMap = new Map();
    function ringElementKey(residueIndex, unitId, altId) {
        return residueIndex + "|" + unitId + "|" + altId;
    }
    function fixLinkDirection(iA, iB) {
        linear_algebra_1.Vec3.sub(elements[iA].geometry.direction, elements[iB].geometry.center, elements[iA].geometry.center);
        linear_algebra_1.Vec3.normalize(elements[iA].geometry.direction, elements[iA].geometry.direction);
    }
    var tmpV = linear_algebra_1.Vec3.zero();
    function fixTerminalLinkDirection(iA, indexB, unitB) {
        var pos = unitB.conformation.position, geo = elements[iA].geometry;
        linear_algebra_1.Vec3.sub(geo.direction, pos(unitB.elements[indexB], tmpV), geo.center);
        linear_algebra_1.Vec3.normalize(geo.direction, geo.direction);
    }
    // get carbohydrate elements and carbohydrate links induced by intra-residue bonds
    for (var i = 0, il = structure.units.length; i < il; ++i) {
        var unit = structure.units[i];
        if (!unit_1.default.isAtomic(unit))
            continue;
        var model = unit.model;
        var _a = model.atomicHierarchy, chainAtomSegments = _a.chainAtomSegments, residueAtomSegments = _a.residueAtomSegments, residues = _a.residues;
        var label_comp_id = residues.label_comp_id;
        var chainIt = int_1.Segmentation.transientSegments(chainAtomSegments, unit.elements);
        var residueIt = int_1.Segmentation.transientSegments(residueAtomSegments, unit.elements);
        var sugarResidueMap = void 0;
        while (chainIt.hasNext) {
            residueIt.setSegment(chainIt.move());
            while (residueIt.hasNext) {
                var residueIndex = residueIt.move().index;
                var saccharideComp = getSaccharideComp(label_comp_id.value(residueIndex), model);
                if (!saccharideComp)
                    continue;
                if (!sugarResidueMap) {
                    sugarResidueMap = rings_1.UnitRings.byFingerprintAndResidue(unit.rings, SugarRingFps);
                }
                var sugarRings = filterFusedRings(unit.rings, sugarResidueMap.get(residueIndex));
                if (!sugarRings || !sugarRings.length) {
                    partialElements.push({ unit: unit, residueIndex: residueIndex, component: saccharideComp });
                    continue;
                }
                var rings = unit.rings;
                var ringElements = [];
                for (var j = 0, jl = sugarRings.length; j < jl; ++j) {
                    var ringAtoms = rings.all[sugarRings[j]];
                    var anomericCarbon = getAnomericCarbon(unit, ringAtoms);
                    var pa = new principal_axes_1.default(util_1.getPositionMatrix(unit, ringAtoms));
                    var center = linear_algebra_1.Vec3.copy(linear_algebra_1.Vec3.zero(), pa.center);
                    var normal = linear_algebra_1.Vec3.copy(linear_algebra_1.Vec3.zero(), pa.normVecC);
                    var direction = getDirection(linear_algebra_1.Vec3.zero(), unit, anomericCarbon, center);
                    linear_algebra_1.Vec3.orthogonalize(direction, normal, direction);
                    var ringAltId = getRingAltId(unit, ringAtoms);
                    var elementIndex = elements.length;
                    ringElements.push(elementIndex);
                    elementsWithRingMap.set(ringElementKey(residueIndex, unit.id, ringAltId), elementIndex);
                    elements.push({
                        geometry: { center: center, normal: normal, direction: direction },
                        component: saccharideComp,
                        unit: unit, residueIndex: residueIndex, anomericCarbon: anomericCarbon, ringAltId: ringAltId
                    });
                }
                // add carbohydrate links induced by intra-residue bonds
                // (e.g. for structures from the PDB archive __before__ carbohydrate remediation)
                var ringCombinations = combination_1.combinations(array_1.fillSerial(new Array(sugarRings.length)), 2);
                for (var j = 0, jl = ringCombinations.length; j < jl; ++j) {
                    var rc = ringCombinations[j];
                    var r0 = rings.all[sugarRings[rc[0]]], r1 = rings.all[sugarRings[rc[1]]];
                    // 1,6 glycosidic links are distance 3 and 1,4 glycosidic links are distance 2
                    if (graph_1.IntAdjacencyGraph.areVertexSetsConnected(unit.links, r0, r1, 3)) {
                        var re0 = ringElements[rc[0]];
                        var re1 = ringElements[rc[1]];
                        if (elements[re0].ringAltId === elements[re1].ringAltId) {
                            // TODO handle better, for now fix both directions as it is unclear where the C1 atom is
                            //  would need to know the path connecting the two rings
                            fixLinkDirection(re0, re1);
                            fixLinkDirection(re1, re0);
                            links.push({ carbohydrateIndexA: re0, carbohydrateIndexB: re1 });
                            links.push({ carbohydrateIndexA: re1, carbohydrateIndexB: re0 });
                        }
                    }
                }
            }
        }
    }
    function getRingElementIndex(unit, index) {
        return elementsWithRingMap.get(ringElementKey(unit.getResidueIndex(index), unit.id, getAltId(unit, index)));
    }
    // add carbohydrate links induced by intra-unit bonds
    // (e.g. for structures from the PDB archive __after__ carbohydrate remediation)
    for (var i = 0, il = elements.length; i < il; ++i) {
        var carbohydrate = elements[i];
        var unit = carbohydrate.unit, residueIndex = carbohydrate.residueIndex, anomericCarbon = carbohydrate.anomericCarbon;
        var _b = unit.links, offset = _b.offset, b = _b.b;
        var ac = int_1.SortedArray.indexOf(unit.elements, anomericCarbon);
        for (var j = offset[ac], jl = offset[ac + 1]; j < jl; ++j) {
            var bj = b[j];
            if (residueIndex !== unit.getResidueIndex(bj)) {
                var ringElementIndex = getRingElementIndex(unit, bj);
                if (ringElementIndex !== undefined && ringElementIndex !== i) {
                    fixLinkDirection(i, ringElementIndex);
                    links.push({
                        carbohydrateIndexA: i,
                        carbohydrateIndexB: ringElementIndex
                    });
                    links.push({
                        carbohydrateIndexA: ringElementIndex,
                        carbohydrateIndexB: i
                    });
                }
            }
        }
    }
    // get carbohydrate links induced by inter-unit bonds, that is
    // terminal links plus inter monosaccharide links for structures from the
    // PDB archive __before__ carbohydrate remediation
    for (var i = 0, il = structure.units.length; i < il; ++i) {
        var unit = structure.units[i];
        if (!unit_1.default.isAtomic(unit))
            continue;
        structure.links.getLinkedUnits(unit).forEach(function (pairBonds) {
            pairBonds.linkedElementIndices.forEach(function (indexA) {
                pairBonds.getBonds(indexA).forEach(function (bondInfo) {
                    var unitA = pairBonds.unitA, unitB = pairBonds.unitB;
                    var indexB = bondInfo.indexB;
                    var ringElementIndexA = getRingElementIndex(unitA, indexA);
                    var ringElementIndexB = getRingElementIndex(unitB, indexB);
                    if (ringElementIndexA !== undefined && ringElementIndexB !== undefined) {
                        var atomIdA = getAtomId(unitA, indexA);
                        if (atomIdA.startsWith('O1') || atomIdA.startsWith('C1')) {
                            fixLinkDirection(ringElementIndexA, ringElementIndexB);
                        }
                        links.push({
                            carbohydrateIndexA: ringElementIndexA,
                            carbohydrateIndexB: ringElementIndexB
                        });
                    }
                    else if (ringElementIndexA !== undefined) {
                        var atomIdA = getAtomId(unitA, indexA);
                        if (atomIdA.startsWith('O1') || atomIdA.startsWith('C1')) {
                            fixTerminalLinkDirection(ringElementIndexA, indexB, unitB);
                        }
                        terminalLinks.push({
                            carbohydrateIndex: ringElementIndexA,
                            elementIndex: indexB,
                            elementUnit: unitB,
                            fromCarbohydrate: true
                        });
                    }
                    else if (ringElementIndexB !== undefined) {
                        terminalLinks.push({
                            carbohydrateIndex: ringElementIndexB,
                            elementIndex: indexA,
                            elementUnit: unitA,
                            fromCarbohydrate: false
                        });
                    }
                });
            });
        });
    }
    return __assign({ links: links, terminalLinks: terminalLinks, elements: elements, partialElements: partialElements }, buildLookups(elements, links, terminalLinks));
}
exports.computeCarbohydrates = computeCarbohydrates;
function buildLookups(elements, links, terminalLinks) {
    // element lookup
    function elementKey(unit, anomericCarbon) {
        return unit.id + "|" + anomericCarbon;
    }
    var elementMap = new Map();
    for (var i = 0, il = elements.length; i < il; ++i) {
        var _a = elements[i], unit = _a.unit, anomericCarbon = _a.anomericCarbon;
        elementMap.set(elementKey(unit, anomericCarbon), i);
    }
    function getElementIndex(unit, anomericCarbon) {
        return elementMap.get(elementKey(unit, anomericCarbon));
    }
    // link lookup
    function linkKey(unitA, anomericCarbonA, unitB, anomericCarbonB) {
        return unitA.id + "|" + anomericCarbonA + "|" + unitB.id + "|" + anomericCarbonB;
    }
    var linkMap = new Map();
    for (var i = 0, il = links.length; i < il; ++i) {
        var l = links[i];
        var _b = elements[l.carbohydrateIndexA], unitA = _b.unit, anomericCarbonA = _b.anomericCarbon;
        var _c = elements[l.carbohydrateIndexB], unitB = _c.unit, anomericCarbonB = _c.anomericCarbon;
        linkMap.set(linkKey(unitA, anomericCarbonA, unitB, anomericCarbonB), i);
    }
    function getLinkIndex(unitA, anomericCarbonA, unitB, anomericCarbonB) {
        return linkMap.get(linkKey(unitA, anomericCarbonA, unitB, anomericCarbonB));
    }
    // links lookup
    function linksKey(unit, anomericCarbon) {
        return unit.id + "|" + anomericCarbon;
    }
    var linksMap = new Map();
    for (var i = 0, il = links.length; i < il; ++i) {
        var l = links[i];
        var _d = elements[l.carbohydrateIndexA], unit = _d.unit, anomericCarbon = _d.anomericCarbon;
        var k = linksKey(unit, anomericCarbon);
        var e = linksMap.get(k);
        if (e === undefined)
            linksMap.set(k, [i]);
        else
            e.push(i);
    }
    function getLinkIndices(unit, anomericCarbon) {
        return linksMap.get(linksKey(unit, anomericCarbon)) || [];
    }
    // terminal link lookup
    function terminalLinkKey(unitA, elementA, unitB, elementB) {
        return unitA.id + "|" + elementA + "|" + unitB.id + "|" + elementB;
    }
    var terminalLinkMap = new Map();
    for (var i = 0, il = terminalLinks.length; i < il; ++i) {
        var _e = terminalLinks[i], fromCarbohydrate = _e.fromCarbohydrate, carbohydrateIndex = _e.carbohydrateIndex, elementUnit = _e.elementUnit, elementIndex = _e.elementIndex;
        var _f = elements[carbohydrateIndex], unit = _f.unit, anomericCarbon = _f.anomericCarbon;
        if (fromCarbohydrate) {
            terminalLinkMap.set(terminalLinkKey(unit, anomericCarbon, elementUnit, elementUnit.elements[elementIndex]), i);
        }
        else {
            terminalLinkMap.set(terminalLinkKey(elementUnit, elementUnit.elements[elementIndex], unit, anomericCarbon), i);
        }
    }
    function getTerminalLinkIndex(unitA, elementA, unitB, elementB) {
        return terminalLinkMap.get(terminalLinkKey(unitA, elementA, unitB, elementB));
    }
    // terminal links lookup
    function terminalLinksKey(unit, element) {
        return unit.id + "|" + element;
    }
    var terminalLinksMap = new Map();
    for (var i = 0, il = terminalLinks.length; i < il; ++i) {
        var _g = terminalLinks[i], fromCarbohydrate = _g.fromCarbohydrate, carbohydrateIndex = _g.carbohydrateIndex, elementUnit = _g.elementUnit, elementIndex = _g.elementIndex;
        var _h = elements[carbohydrateIndex], unit = _h.unit, anomericCarbon = _h.anomericCarbon;
        var k = void 0;
        if (fromCarbohydrate) {
            k = terminalLinksKey(unit, anomericCarbon);
        }
        else {
            k = terminalLinksKey(elementUnit, elementUnit.elements[elementIndex]);
        }
        var e = terminalLinksMap.get(k);
        if (e === undefined)
            terminalLinksMap.set(k, [i]);
        else
            e.push(i);
    }
    function getTerminalLinkIndices(unit, element) {
        return terminalLinksMap.get(terminalLinksKey(unit, element)) || [];
    }
    // anomeric carbon lookup
    function anomericCarbonKey(unit, residueIndex) {
        return unit.id + "|" + residueIndex;
    }
    var anomericCarbonMap = new Map();
    for (var i = 0, il = elements.length; i < il; ++i) {
        var _j = elements[i], unit = _j.unit, anomericCarbon = _j.anomericCarbon;
        var residueIndex = unit.model.atomicHierarchy.residueAtomSegments.index[anomericCarbon];
        var k = anomericCarbonKey(unit, residueIndex);
        if (anomericCarbonMap.has(k)) {
            anomericCarbonMap.get(k).push(anomericCarbon);
        }
        else {
            anomericCarbonMap.set(k, [anomericCarbon]);
        }
    }
    var EmptyArray = [];
    function getAnomericCarbons(unit, residueIndex) {
        return anomericCarbonMap.get(anomericCarbonKey(unit, residueIndex)) || EmptyArray;
    }
    return { getElementIndex: getElementIndex, getLinkIndex: getLinkIndex, getLinkIndices: getLinkIndices, getTerminalLinkIndex: getTerminalLinkIndex, getTerminalLinkIndices: getTerminalLinkIndices, getAnomericCarbons: getAnomericCarbons };
}
