"use strict";
/**
 * Copyright (c) 2017-2019 mol* contributors, licensed under MIT, See LICENSE file for more info.
 *
 * @author David Sehnal <david.sehnal@gmail.com>
 * @author Alexander Rose <alexander.rose@weirdbyte.de>
 */
Object.defineProperty(exports, "__esModule", { value: true });
var generic_1 = require("../../../mol-data/generic");
var int_1 = require("../../../mol-data/int");
var boundary_helper_1 = require("../../../mol-math/geometry/boundary-helper");
var linear_algebra_1 = require("../../../mol-math/linear-algebra");
var builder_1 = require("../../../mol-script/language/builder");
var unit_1 = require("./unit");
var structure_1 = require("../structure");
var util_1 = require("../../../mol-data/util");
var StructureElement;
(function (StructureElement) {
    function create(unit, element) {
        return { kind: 'element-location', unit: unit, element: element || 0 };
    }
    StructureElement.create = create;
    function set(a, unit, element) {
        if (unit)
            a.unit = unit;
        if (element !== undefined)
            a.element = element;
        return a;
    }
    StructureElement.set = set;
    function property(p) { return p; }
    StructureElement.property = property;
    function _wrongUnitKind(kind) { throw new Error("Property only available for " + kind + " models."); }
    function atomicProperty(p) {
        return property(function (l) { return unit_1.default.isAtomic(l.unit) ? p(l) : _wrongUnitKind('atomic'); });
    }
    StructureElement.atomicProperty = atomicProperty;
    function coarseProperty(p) {
        return property(function (l) { return unit_1.default.isCoarse(l.unit) ? p(l) : _wrongUnitKind('coarse'); });
    }
    StructureElement.coarseProperty = coarseProperty;
    function Loci(structure, elements) {
        return { kind: 'element-loci', structure: structure, elements: elements };
    }
    StructureElement.Loci = Loci;
    function isLoci(x) {
        return !!x && x.kind === 'element-loci';
    }
    StructureElement.isLoci = isLoci;
    function areLociEqual(a, b) {
        if (a.elements.length !== b.elements.length)
            return false;
        for (var i = 0, il = a.elements.length; i < il; ++i) {
            var elementA = a.elements[i];
            var elementB = b.elements[i];
            if (elementA.unit.id !== elementB.unit.id)
                return false;
            if (!int_1.OrderedSet.areEqual(elementA.indices, elementB.indices))
                return false;
        }
        return true;
    }
    StructureElement.areLociEqual = areLociEqual;
    function isLocation(x) {
        return !!x && x.kind === 'element-location';
    }
    StructureElement.isLocation = isLocation;
    function residueIndex(e) {
        if (unit_1.default.isAtomic(e.unit)) {
            return e.unit.residueIndex[e.element];
        }
        else {
            // TODO: throw error instead?
            return -1;
        }
    }
    StructureElement.residueIndex = residueIndex;
    function chainIndex(e) {
        if (unit_1.default.isAtomic(e.unit)) {
            return e.unit.chainIndex[e.element];
        }
        else {
            // TODO: throw error instead?
            return -1;
        }
    }
    StructureElement.chainIndex = chainIndex;
    function entityIndex(l) {
        switch (l.unit.kind) {
            case 0 /* Atomic */:
                return l.unit.model.atomicHierarchy.index.getEntityFromChain(l.unit.chainIndex[l.element]);
            case 1 /* Spheres */:
                return l.unit.model.coarseHierarchy.spheres.entityKey[l.element];
            case 2 /* Gaussians */:
                return l.unit.model.coarseHierarchy.gaussians.entityKey[l.element];
        }
    }
    StructureElement.entityIndex = entityIndex;
    (function (Loci) {
        function size(loci) {
            var s = 0;
            for (var _a = 0, _b = loci.elements; _a < _b.length; _a++) {
                var u = _b[_a];
                s += int_1.OrderedSet.size(u.indices);
            }
            return s;
        }
        Loci.size = size;
        function all(structure) {
            return Loci(structure, structure.units.map(function (unit) { return ({
                unit: unit,
                indices: int_1.OrderedSet.ofRange(0, unit.elements.length)
            }); }));
        }
        Loci.all = all;
        function remap(loci, structure) {
            return Loci(structure, loci.elements.map(function (e) { return ({
                unit: structure.unitMap.get(e.unit.id),
                indices: e.indices
            }); }));
        }
        Loci.remap = remap;
        function union(xs, ys) {
            if (xs.elements.length > ys.elements.length)
                return union(ys, xs);
            if (xs.elements.length === 0)
                return ys;
            var map = new Map();
            for (var _a = 0, _b = xs.elements; _a < _b.length; _a++) {
                var e = _b[_a];
                map.set(e.unit.id, e.indices);
            }
            var elements = [];
            for (var _c = 0, _d = ys.elements; _c < _d.length; _c++) {
                var e = _d[_c];
                if (map.has(e.unit.id)) {
                    elements[elements.length] = { unit: e.unit, indices: int_1.OrderedSet.union(map.get(e.unit.id), e.indices) };
                    map.delete(e.unit.id);
                }
                else {
                    elements[elements.length] = e;
                }
            }
            map.forEach(function (indices, id) {
                elements[elements.length] = { unit: xs.structure.unitMap.get(id), indices: indices };
            });
            return Loci(xs.structure, elements);
        }
        Loci.union = union;
        function subtract(xs, ys) {
            var map = new Map();
            for (var _a = 0, _b = ys.elements; _a < _b.length; _a++) {
                var e = _b[_a];
                map.set(e.unit.id, e.indices);
            }
            var elements = [];
            for (var _c = 0, _d = xs.elements; _c < _d.length; _c++) {
                var e = _d[_c];
                if (map.has(e.unit.id)) {
                    var indices = int_1.OrderedSet.subtract(e.indices, map.get(e.unit.id));
                    if (int_1.OrderedSet.size(indices) === 0)
                        continue;
                    elements[elements.length] = { unit: e.unit, indices: indices };
                }
                else {
                    elements[elements.length] = e;
                }
            }
            return Loci(xs.structure, elements);
        }
        Loci.subtract = subtract;
        function areIntersecting(xs, ys) {
            if (xs.elements.length > ys.elements.length)
                return areIntersecting(ys, xs);
            if (xs.elements.length === 0)
                return ys.elements.length === 0;
            var map = new Map();
            for (var _a = 0, _b = xs.elements; _a < _b.length; _a++) {
                var e = _b[_a];
                map.set(e.unit.id, e.indices);
            }
            for (var _c = 0, _d = ys.elements; _c < _d.length; _c++) {
                var e = _d[_c];
                if (!map.has(e.unit.id))
                    continue;
                if (int_1.OrderedSet.areIntersecting(map.get(e.unit.id), e.indices))
                    return true;
            }
            return false;
        }
        Loci.areIntersecting = areIntersecting;
        function extendToWholeResidues(loci) {
            var elements = [];
            for (var _a = 0, _b = loci.elements; _a < _b.length; _a++) {
                var lociElement = _b[_a];
                if (lociElement.unit.kind === 0 /* Atomic */) {
                    var unitElements = lociElement.unit.elements;
                    var h = lociElement.unit.model.atomicHierarchy;
                    var _c = h.residueAtomSegments, residueIndex_1 = _c.index, residueOffsets = _c.offsets;
                    var newIndices = [];
                    var indices = lociElement.indices, len = int_1.OrderedSet.size(indices);
                    var i = 0;
                    while (i < len) {
                        var rI = residueIndex_1[unitElements[int_1.OrderedSet.getAt(indices, i)]];
                        i++;
                        while (i < len && residueIndex_1[unitElements[int_1.OrderedSet.getAt(indices, i)]] === rI) {
                            i++;
                        }
                        for (var j = residueOffsets[rI], _j = residueOffsets[rI + 1]; j < _j; j++) {
                            var idx = int_1.OrderedSet.indexOf(unitElements, j);
                            if (idx >= 0)
                                newIndices[newIndices.length] = idx;
                        }
                    }
                    elements[elements.length] = { unit: lociElement.unit, indices: int_1.SortedArray.ofSortedArray(newIndices) };
                }
                else {
                    // coarse elements are already by-residue
                    elements[elements.length] = lociElement;
                }
            }
            return Loci(loci.structure, elements);
        }
        Loci.extendToWholeResidues = extendToWholeResidues;
        function getChainSegments(unit) {
            switch (unit.kind) {
                case 0 /* Atomic */: return unit.model.atomicHierarchy.chainAtomSegments;
                case 1 /* Spheres */: return unit.model.coarseHierarchy.spheres.chainElementSegments;
                case 2 /* Gaussians */: return unit.model.coarseHierarchy.gaussians.chainElementSegments;
            }
        }
        function extendToWholeChains(loci) {
            var elements = [];
            for (var _a = 0, _b = loci.elements; _a < _b.length; _a++) {
                var lociElement = _b[_a];
                var newIndices = [];
                var unitElements = lociElement.unit.elements;
                var _c = getChainSegments(lociElement.unit), chainIndex_1 = _c.index, chainOffsets = _c.offsets;
                var indices = lociElement.indices, len = int_1.OrderedSet.size(indices);
                var i = 0;
                while (i < len) {
                    var cI = chainIndex_1[unitElements[int_1.OrderedSet.getAt(indices, i)]];
                    i++;
                    while (i < len && chainIndex_1[unitElements[int_1.OrderedSet.getAt(indices, i)]] === cI) {
                        i++;
                    }
                    for (var j = chainOffsets[cI], _j = chainOffsets[cI + 1]; j < _j; j++) {
                        var idx = int_1.OrderedSet.indexOf(unitElements, j);
                        if (idx >= 0)
                            newIndices[newIndices.length] = idx;
                    }
                }
                elements[elements.length] = { unit: lociElement.unit, indices: int_1.SortedArray.ofSortedArray(newIndices) };
            }
            return Loci(loci.structure, elements);
        }
        Loci.extendToWholeChains = extendToWholeChains;
        var boundaryHelper = new boundary_helper_1.BoundaryHelper(), tempPos = linear_algebra_1.Vec3.zero();
        function getBoundary(loci) {
            boundaryHelper.reset(0);
            for (var _a = 0, _b = loci.elements; _a < _b.length; _a++) {
                var e = _b[_a];
                var indices = e.indices;
                var pos = e.unit.conformation.position, r = e.unit.conformation.r;
                var elements = e.unit.elements;
                for (var i = 0, _i = int_1.OrderedSet.size(indices); i < _i; i++) {
                    var eI = elements[int_1.OrderedSet.getAt(indices, i)];
                    pos(eI, tempPos);
                    boundaryHelper.boundaryStep(tempPos, r(eI));
                }
            }
            boundaryHelper.finishBoundaryStep();
            for (var _c = 0, _d = loci.elements; _c < _d.length; _c++) {
                var e = _d[_c];
                var indices = e.indices;
                var pos = e.unit.conformation.position, r = e.unit.conformation.r;
                var elements = e.unit.elements;
                for (var i = 0, _i = int_1.OrderedSet.size(indices); i < _i; i++) {
                    var eI = elements[int_1.OrderedSet.getAt(indices, i)];
                    pos(eI, tempPos);
                    boundaryHelper.extendStep(tempPos, r(eI));
                }
            }
            return { box: boundaryHelper.getBox(), sphere: boundaryHelper.getSphere() };
        }
        Loci.getBoundary = getBoundary;
        function toScriptExpression(loci) {
            if (loci.structure.models.length > 1) {
                console.warn('toScriptExpression is only supported for Structure with single model, returning empty expression.');
                return builder_1.MolScriptBuilder.struct.generator.empty();
            }
            if (loci.elements.length === 0)
                return builder_1.MolScriptBuilder.struct.generator.empty();
            var sourceIndexMap = new Map();
            var el = StructureElement.create(), p = structure_1.StructureProperties.atom.sourceIndex;
            for (var _a = 0, _b = loci.elements; _a < _b.length; _a++) {
                var e = _b[_a];
                var indices = e.indices;
                var elements = e.unit.elements;
                var opName = e.unit.conformation.operator.name;
                var sourceIndices = void 0;
                if (sourceIndexMap.has(opName))
                    sourceIndices = sourceIndexMap.get(opName);
                else {
                    sourceIndices = generic_1.UniqueArray.create();
                    sourceIndexMap.set(opName, sourceIndices);
                }
                el.unit = e.unit;
                for (var i = 0, _i = int_1.OrderedSet.size(indices); i < _i; i++) {
                    el.element = elements[int_1.OrderedSet.getAt(indices, i)];
                    var idx = p(el);
                    generic_1.UniqueArray.add(sourceIndices, idx, idx);
                }
            }
            var byOpName = [];
            var keys = sourceIndexMap.keys();
            while (true) {
                var k = keys.next();
                if (k.done)
                    break;
                byOpName.push(getOpNameQuery(k.value, sourceIndexMap.get(k.value).array));
            }
            return builder_1.MolScriptBuilder.struct.modifier.union([
                byOpName.length === 1 ? byOpName[0] : builder_1.MolScriptBuilder.struct.combinator.merge.apply(null, byOpName)
            ]);
        }
        Loci.toScriptExpression = toScriptExpression;
        function getOpNameQuery(opName, xs) {
            util_1.sortArray(xs);
            var ranges = [];
            var set = [];
            var i = 0, len = xs.length;
            while (i < len) {
                var start = i;
                i++;
                while (i < len && xs[i - 1] + 1 === xs[i])
                    i++;
                var end = i;
                // TODO: is this a good value?
                if (end - start > 12) {
                    ranges[ranges.length] = xs[start];
                    ranges[ranges.length] = xs[end - 1];
                }
                else {
                    for (var j = start; j < end; j++) {
                        set[set.length] = xs[j];
                    }
                }
            }
            var siProp = builder_1.MolScriptBuilder.struct.atomProperty.core.sourceIndex();
            var tests = [];
            // TODO: add set.ofRanges constructor to MolQL???
            if (set.length > 0) {
                tests[tests.length] = builder_1.MolScriptBuilder.core.set.has([builder_1.MolScriptBuilder.set.apply(null, set), siProp]);
            }
            for (var rI = 0, _rI = ranges.length / 2; rI < _rI; rI++) {
                tests[tests.length] = builder_1.MolScriptBuilder.core.rel.inRange([siProp, ranges[2 * rI], ranges[2 * rI + 1]]);
            }
            return builder_1.MolScriptBuilder.struct.generator.atomGroups({
                'atom-test': tests.length > 1 ? builder_1.MolScriptBuilder.core.logic.or.apply(null, tests) : tests[0],
                'chain-test': builder_1.MolScriptBuilder.core.rel.eq([builder_1.MolScriptBuilder.struct.atomProperty.core.operatorName(), opName])
            });
        }
    })(Loci = StructureElement.Loci || (StructureElement.Loci = {}));
})(StructureElement || (StructureElement = {}));
exports.default = StructureElement;
