/**
 * Copyright (c) 2017-2018 mol* contributors, licensed under MIT, See LICENSE file for more info.
 *
 * @author David Sehnal <david.sehnal@gmail.com>
 * @author Alexander Rose <alexander.rose@weirdbyte.de>
 */
import { IntMap, Iterator } from '../../../mol-data/int';
import { SymmetryOperator } from '../../../mol-math/geometry/symmetry-operator';
import { Model } from '../model';
import StructureElement from './element';
import Unit from './unit';
import { StructureLookup3D } from './util/lookup3d';
import { StructureSubsetBuilder } from './util/subset-builder';
import { InterUnitBonds } from './unit/links';
import { PairRestraints, CrossLinkRestraint } from './unit/pair-restraints';
import { ResidueIndex, EntityIndex } from '../model/indexing';
import { Carbohydrates } from './carbohydrates/data';
import { Vec3, Mat4 } from '../../../mol-math/linear-algebra';
import { UUID } from '../../../mol-util';
import { CustomProperties } from '../common/custom-property';
declare class Structure {
    /** Maps unit.id to unit */
    readonly unitMap: IntMap<Unit>;
    /** Array of all units in the structure, sorted by unit.id */
    readonly units: ReadonlyArray<Unit>;
    private _props;
    subsetBuilder(isSorted: boolean): StructureSubsetBuilder;
    /** Count of all elements in the structure, i.e. the sum of the elements in the units */
    readonly elementCount: number;
    readonly hasCustomProperties: boolean;
    readonly customPropertyDescriptors: CustomProperties;
    /**
     * Property data unique to this instance of the structure.
     */
    readonly currentPropertyData: any;
    /**
     * Property data of the parent structure if it exists, currentPropertyData otherwise.
     */
    readonly inheritedPropertyData: any;
    /** Count of all polymer residues in the structure */
    readonly polymerResidueCount: number;
    /** Coarse structure, defined as Containing less than twice as many elements as polymer residues */
    readonly isCoarse: boolean;
    readonly isEmpty: boolean;
    readonly hashCode: number;
    readonly transformHash: number;
    private computeHash;
    /** Returns a new element location iterator */
    elementLocations(): Iterator<StructureElement>;
    /** the root/top-most parent or `undefined` in case this is the root */
    readonly parent: Structure | undefined;
    readonly coordinateSystem: SymmetryOperator;
    readonly boundary: {
        box: import("../../../mol-math/geometry/primitives/box3d").Box3D;
        sphere: import("../../../mol-math/geometry/primitives/sphere3d").Sphere3D;
    };
    readonly lookup3d: StructureLookup3D;
    readonly links: InterUnitBonds;
    readonly crossLinkRestraints: PairRestraints<CrossLinkRestraint>;
    readonly unitSymmetryGroups: ReadonlyArray<Unit.SymmetryGroup>;
    readonly carbohydrates: Carbohydrates;
    readonly models: ReadonlyArray<Model>;
    readonly uniqueResidueNames: Set<string>;
    readonly entityIndices: readonly EntityIndex[];
    readonly uniqueAtomicResidueIndices: ReadonlyMap<UUID, readonly ResidueIndex[]>;
    /** If the structure is based on a single model, return it. Otherwise throw an exception. */
    readonly model: Model;
    hasElement(e: StructureElement): boolean;
    private initUnits;
    constructor(units: ArrayLike<Unit>, parent: Structure | undefined, coordinateSystem?: SymmetryOperator);
}
declare namespace Structure {
    const Empty: Structure;
    /** Represents a single structure */
    interface Loci {
        readonly kind: 'structure-loci';
        readonly structure: Structure;
    }
    function Loci(structure: Structure): Loci;
    function isLoci(x: any): x is Loci;
    function areLociEqual(a: Loci, b: Loci): boolean;
    function create(units: ReadonlyArray<Unit>, parent: Structure | undefined, coordinateSystem?: SymmetryOperator): Structure;
    /**
     * Construct a Structure from a model.
     *
     * Generally, a single unit corresponds to a single chain, with the exception
     * of consecutive "single atom chains".
     */
    function ofModel(model: Model): Structure;
    function transform(s: Structure, transform: Mat4): Structure;
    class StructureBuilder {
        private parent;
        private coordinateSystem;
        private units;
        private invariantId;
        addUnit(kind: Unit.Kind, model: Model, operator: SymmetryOperator, elements: StructureElement.Set): Unit;
        addWithOperator(unit: Unit, operator: SymmetryOperator): Unit;
        getStructure(): Structure;
        readonly isEmpty: boolean;
        constructor(parent: Structure | undefined, coordinateSystem: SymmetryOperator | undefined);
    }
    function Builder(parent: Structure | undefined, coordinateSystem: SymmetryOperator | undefined): StructureBuilder;
    function hashCode(s: Structure): number;
    /** Hash based on all unit.model conformation values in the structure */
    function conformationHash(s: Structure): number;
    function areUnitAndIndicesEqual(a: Structure, b: Structure): boolean;
    function areEquivalent(a: Structure, b: Structure): boolean;
    /** Check if the structures or their parents are equivalent */
    function areParentsEquivalent(a: Structure, b: Structure): boolean;
    /** Check if the structures or their parents are equal */
    function areParentsEqual(a: Structure, b: Structure): boolean;
    class ElementLocationIterator implements Iterator<StructureElement> {
        private structure;
        private current;
        private unitIndex;
        private elements;
        private maxIdx;
        private idx;
        hasNext: boolean;
        move(): StructureElement;
        private advance;
        constructor(structure: Structure);
    }
    function minDistanceToPoint(s: Structure, point: Vec3, radius: number): number;
    function distance(a: Structure, b: Structure): number;
}
export default Structure;
