"use strict";
/**
 * Copyright (c) 2017-2018 mol* contributors, licensed under MIT, See LICENSE file for more info.
 *
 * @author David Sehnal <david.sehnal@gmail.com>
 * @author Alexander Rose <alexander.rose@weirdbyte.de>
 */
Object.defineProperty(exports, "__esModule", { value: true });
var int_1 = require("../../../mol-data/int");
var generic_1 = require("../../../mol-data/generic");
var symmetry_operator_1 = require("../../../mol-math/geometry/symmetry-operator");
var util_1 = require("../../../mol-data/util");
var element_1 = require("./element");
var unit_1 = require("./unit");
var lookup3d_1 = require("./util/lookup3d");
var subset_builder_1 = require("./util/subset-builder");
var links_1 = require("./unit/links");
var pair_restraints_1 = require("./unit/pair-restraints");
var symmetry_1 = require("./symmetry");
var properties_1 = require("./properties");
var compute_1 = require("./carbohydrates/compute");
var linear_algebra_1 = require("../../../mol-math/linear-algebra");
var id_factory_1 = require("../../../mol-util/id-factory");
var geometry_1 = require("../../../mol-math/geometry");
var custom_property_1 = require("../common/custom-property");
var Structure = /** @class */ (function () {
    function Structure(units, parent, coordinateSystem) {
        this._props = {
            hashCode: -1,
            transformHash: -1,
            elementCount: 0,
            polymerResidueCount: 0,
            coordinateSystem: symmetry_operator_1.SymmetryOperator.Default
        };
        this.unitMap = this.initUnits(units);
        this.units = units;
        if (parent)
            this._props.parent = parent.parent || parent;
        if (coordinateSystem)
            this._props.coordinateSystem = coordinateSystem;
        else if (parent)
            this._props.coordinateSystem = parent.coordinateSystem;
    }
    Structure.prototype.subsetBuilder = function (isSorted) {
        return new subset_builder_1.StructureSubsetBuilder(this, isSorted);
    };
    Object.defineProperty(Structure.prototype, "elementCount", {
        /** Count of all elements in the structure, i.e. the sum of the elements in the units */
        get: function () {
            return this._props.elementCount;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(Structure.prototype, "hasCustomProperties", {
        get: function () {
            return !!this._props.customProps && this._props.customProps.all.length > 0;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(Structure.prototype, "customPropertyDescriptors", {
        get: function () {
            if (!this._props.customProps)
                this._props.customProps = new custom_property_1.CustomProperties();
            return this._props.customProps;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(Structure.prototype, "currentPropertyData", {
        /**
         * Property data unique to this instance of the structure.
         */
        get: function () {
            if (!this._props.propertyData)
                this._props.propertyData = Object.create(null);
            return this._props.propertyData;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(Structure.prototype, "inheritedPropertyData", {
        /**
         * Property data of the parent structure if it exists, currentPropertyData otherwise.
         */
        get: function () {
            return this.parent ? this.parent.currentPropertyData : this.currentPropertyData;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(Structure.prototype, "polymerResidueCount", {
        /** Count of all polymer residues in the structure */
        get: function () {
            return this._props.polymerResidueCount;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(Structure.prototype, "isCoarse", {
        /** Coarse structure, defined as Containing less than twice as many elements as polymer residues */
        get: function () {
            var ec = this.elementCount;
            var prc = this.polymerResidueCount;
            return prc && ec ? ec / prc < 2 : false;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(Structure.prototype, "isEmpty", {
        get: function () {
            return this.units.length === 0;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(Structure.prototype, "hashCode", {
        get: function () {
            if (this._props.hashCode !== -1)
                return this._props.hashCode;
            return this.computeHash();
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(Structure.prototype, "transformHash", {
        get: function () {
            if (this._props.transformHash !== -1)
                return this._props.transformHash;
            this._props.transformHash = util_1.hashFnv32a(this.units.map(function (u) { return u.id; }));
            return this._props.transformHash;
        },
        enumerable: true,
        configurable: true
    });
    Structure.prototype.computeHash = function () {
        var hash = 23;
        for (var i = 0, _i = this.units.length; i < _i; i++) {
            var u = this.units[i];
            hash = (31 * hash + u.id) | 0;
            hash = (31 * hash + int_1.SortedArray.hashCode(u.elements)) | 0;
        }
        hash = (31 * hash + this.elementCount) | 0;
        hash = util_1.hash1(hash);
        if (hash === -1)
            hash = 0;
        this._props.hashCode = hash;
        return hash;
    };
    /** Returns a new element location iterator */
    Structure.prototype.elementLocations = function () {
        return new Structure.ElementLocationIterator(this);
    };
    Object.defineProperty(Structure.prototype, "parent", {
        /** the root/top-most parent or `undefined` in case this is the root */
        get: function () {
            return this._props.parent;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(Structure.prototype, "coordinateSystem", {
        get: function () {
            return this._props.coordinateSystem;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(Structure.prototype, "boundary", {
        get: function () {
            return this.lookup3d.boundary;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(Structure.prototype, "lookup3d", {
        get: function () {
            if (this._props.lookup3d)
                return this._props.lookup3d;
            this._props.lookup3d = new lookup3d_1.StructureLookup3D(this);
            return this._props.lookup3d;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(Structure.prototype, "links", {
        get: function () {
            if (this._props.links)
                return this._props.links;
            this._props.links = links_1.computeInterUnitBonds(this);
            return this._props.links;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(Structure.prototype, "crossLinkRestraints", {
        get: function () {
            if (this._props.crossLinkRestraints)
                return this._props.crossLinkRestraints;
            this._props.crossLinkRestraints = pair_restraints_1.extractCrossLinkRestraints(this);
            return this._props.crossLinkRestraints;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(Structure.prototype, "unitSymmetryGroups", {
        get: function () {
            if (this._props.unitSymmetryGroups)
                return this._props.unitSymmetryGroups;
            this._props.unitSymmetryGroups = symmetry_1.default.computeTransformGroups(this);
            return this._props.unitSymmetryGroups;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(Structure.prototype, "carbohydrates", {
        get: function () {
            if (this._props.carbohydrates)
                return this._props.carbohydrates;
            this._props.carbohydrates = compute_1.computeCarbohydrates(this);
            return this._props.carbohydrates;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(Structure.prototype, "models", {
        get: function () {
            if (this._props.models)
                return this._props.models;
            this._props.models = getModels(this);
            return this._props.models;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(Structure.prototype, "uniqueResidueNames", {
        get: function () {
            return this._props.uniqueResidueNames
                || (this._props.uniqueResidueNames = getUniqueResidueNames(this));
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(Structure.prototype, "entityIndices", {
        get: function () {
            return this._props.entityIndices || (this._props.entityIndices = getEntityIndices(this));
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(Structure.prototype, "uniqueAtomicResidueIndices", {
        get: function () {
            return this._props.uniqueAtomicResidueIndices
                || (this._props.uniqueAtomicResidueIndices = getUniqueAtomicResidueIndices(this));
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(Structure.prototype, "model", {
        /** If the structure is based on a single model, return it. Otherwise throw an exception. */
        get: function () {
            if (this._props.model)
                return this._props.model;
            var models = this.models;
            if (models.length > 1)
                throw new Error('The structre is based on multiple models.');
            this._props.model = models[0];
            return this._props.model;
        },
        enumerable: true,
        configurable: true
    });
    Structure.prototype.hasElement = function (e) {
        if (!this.unitMap.has(e.unit.id))
            return false;
        return int_1.SortedArray.has(this.unitMap.get(e.unit.id).elements, e.element);
    };
    Structure.prototype.initUnits = function (units) {
        var map = int_1.IntMap.Mutable();
        var elementCount = 0;
        var polymerResidueCount = 0;
        var isSorted = true;
        var lastId = units.length > 0 ? units[0].id : 0;
        for (var i = 0, _i = units.length; i < _i; i++) {
            var u = units[i];
            map.set(u.id, u);
            elementCount += u.elements.length;
            polymerResidueCount += u.polymerElements.length;
            if (u.id < lastId)
                isSorted = false;
            lastId = u.id;
        }
        if (!isSorted)
            util_1.sort(units, 0, units.length, cmpUnits, util_1.arraySwap);
        this._props.elementCount = elementCount;
        this._props.polymerResidueCount = polymerResidueCount;
        return map;
    };
    return Structure;
}());
function cmpUnits(units, i, j) { return units[i].id - units[j].id; }
function getModels(s) {
    var units = s.units;
    var arr = generic_1.UniqueArray.create();
    for (var _a = 0, units_1 = units; _a < units_1.length; _a++) {
        var u = units_1[_a];
        generic_1.UniqueArray.add(arr, u.model.id, u.model);
    }
    return arr.array;
}
function getUniqueResidueNames(s) {
    var prop = properties_1.default.residue.label_comp_id;
    var names = new Set();
    var loc = element_1.default.create();
    for (var _a = 0, _b = s.units; _a < _b.length; _a++) {
        var unit = _b[_a];
        // TODO: support coarse unit?
        if (!unit_1.default.isAtomic(unit))
            continue;
        var residues = int_1.Segmentation.transientSegments(unit.model.atomicHierarchy.residueAtomSegments, unit.elements);
        loc.unit = unit;
        while (residues.hasNext) {
            var seg = residues.move();
            loc.element = unit.elements[seg.start];
            names.add(prop(loc));
        }
    }
    return names;
}
function getEntityIndices(structure) {
    var units = structure.units;
    var l = element_1.default.create();
    var keys = generic_1.UniqueArray.create();
    for (var _a = 0, units_2 = units; _a < units_2.length; _a++) {
        var unit = units_2[_a];
        var prop = unit.kind === 0 /* Atomic */ ? properties_1.default.entity.key : properties_1.default.coarse.entityKey;
        l.unit = unit;
        var elements = unit.elements;
        var chainsIt = int_1.Segmentation.transientSegments(unit.model.atomicHierarchy.chainAtomSegments, elements);
        while (chainsIt.hasNext) {
            var chainSegment = chainsIt.move();
            l.element = elements[chainSegment.start];
            var key = prop(l);
            generic_1.UniqueArray.add(keys, key, key);
        }
    }
    util_1.sortArray(keys.array);
    return keys.array;
}
function getUniqueAtomicResidueIndices(structure) {
    var map = new Map();
    var modelIds = [];
    var unitGroups = structure.unitSymmetryGroups;
    for (var _a = 0, unitGroups_1 = unitGroups; _a < unitGroups_1.length; _a++) {
        var unitGroup = unitGroups_1[_a];
        var unit = unitGroup.units[0];
        if (!unit_1.default.isAtomic(unit))
            continue;
        var uniqueResidues = void 0;
        if (map.has(unit.model.id))
            uniqueResidues = map.get(unit.model.id);
        else {
            uniqueResidues = generic_1.UniqueArray.create();
            modelIds.push(unit.model.id);
            map.set(unit.model.id, uniqueResidues);
        }
        var residues = int_1.Segmentation.transientSegments(unit.model.atomicHierarchy.residueAtomSegments, unit.elements);
        while (residues.hasNext) {
            var seg = residues.move();
            generic_1.UniqueArray.add(uniqueResidues, seg.index, seg.index);
        }
    }
    var ret = new Map();
    for (var _b = 0, modelIds_1 = modelIds; _b < modelIds_1.length; _b++) {
        var id = modelIds_1[_b];
        var array = map.get(id).array;
        util_1.sortArray(array);
        ret.set(id, array);
    }
    return ret;
}
(function (Structure) {
    Structure.Empty = new Structure([], void 0, void 0);
    function Loci(structure) {
        return { kind: 'structure-loci', structure: structure };
    }
    Structure.Loci = Loci;
    function isLoci(x) {
        return !!x && x.kind === 'structure-loci';
    }
    Structure.isLoci = isLoci;
    function areLociEqual(a, b) {
        return a.structure === b.structure;
    }
    Structure.areLociEqual = areLociEqual;
    function create(units, parent, coordinateSystem) {
        return new Structure(units, parent, coordinateSystem);
    }
    Structure.create = create;
    /**
     * Construct a Structure from a model.
     *
     * Generally, a single unit corresponds to a single chain, with the exception
     * of consecutive "single atom chains".
     */
    function ofModel(model) {
        var chains = model.atomicHierarchy.chainAtomSegments;
        var builder = new StructureBuilder(void 0, void 0);
        for (var c = 0; c < chains.count; c++) {
            var start = chains.offsets[c];
            // merge all consecutive "single atom chains"
            while (c + 1 < chains.count
                && chains.offsets[c + 1] - chains.offsets[c] === 1
                && chains.offsets[c + 2] - chains.offsets[c + 1] === 1) {
                c++;
            }
            var elements = int_1.SortedArray.ofBounds(start, chains.offsets[c + 1]);
            if (isWaterChain(model, c, elements)) {
                partitionAtomicUnit(model, elements, builder);
            }
            else {
                builder.addUnit(0 /* Atomic */, model, symmetry_operator_1.SymmetryOperator.Default, elements);
            }
        }
        var cs = model.coarseHierarchy;
        if (cs.isDefined) {
            if (cs.spheres.count > 0) {
                addCoarseUnits(builder, model, model.coarseHierarchy.spheres, 1 /* Spheres */);
            }
            if (cs.gaussians.count > 0) {
                addCoarseUnits(builder, model, model.coarseHierarchy.gaussians, 2 /* Gaussians */);
            }
        }
        return builder.getStructure();
    }
    Structure.ofModel = ofModel;
    function isWaterChain(model, chainIndex, indices) {
        var e = model.atomicHierarchy.index.getEntityFromChain(chainIndex);
        return model.entities.data.type.value(e) === 'water';
    }
    function partitionAtomicUnit(model, indices, builder) {
        var _a = model.atomicConformation, x = _a.x, y = _a.y, z = _a.z;
        var lookup = geometry_1.GridLookup3D({ x: x, y: y, z: z, indices: indices }, linear_algebra_1.Vec3.create(64, 64, 64));
        var _b = lookup.buckets, offset = _b.offset, count = _b.count, array = _b.array;
        for (var i = 0, _i = offset.length; i < _i; i++) {
            var start = offset[i];
            var set = new Int32Array(count[i]);
            for (var j = 0, _j = count[i]; j < _j; j++) {
                set[j] = indices[array[start + j]];
            }
            builder.addUnit(0 /* Atomic */, model, symmetry_operator_1.SymmetryOperator.Default, int_1.SortedArray.ofSortedArray(set));
        }
    }
    function addCoarseUnits(builder, model, elements, kind) {
        var chainElementSegments = elements.chainElementSegments;
        for (var cI = 0; cI < chainElementSegments.count; cI++) {
            var elements_1 = int_1.SortedArray.ofBounds(chainElementSegments.offsets[cI], chainElementSegments.offsets[cI + 1]);
            builder.addUnit(kind, model, symmetry_operator_1.SymmetryOperator.Default, elements_1);
        }
    }
    function transform(s, transform) {
        if (linear_algebra_1.Mat4.isIdentity(transform))
            return s;
        if (!linear_algebra_1.Mat4.isRotationAndTranslation(transform, symmetry_operator_1.SymmetryOperator.RotationTranslationEpsilon))
            throw new Error('Only rotation/translation combination can be applied.');
        var units = [];
        for (var _a = 0, _b = s.units; _a < _b.length; _a++) {
            var u = _b[_a];
            var old = u.conformation.operator;
            var op = symmetry_operator_1.SymmetryOperator.create(old.name, transform, old.assembly, old.ncsId, old.hkl);
            units.push(u.applyOperator(u.id, op));
        }
        var cs = s.coordinateSystem;
        var newCS = symmetry_operator_1.SymmetryOperator.compose(symmetry_operator_1.SymmetryOperator.create(cs.name, transform, cs.assembly, cs.ncsId, cs.hkl), cs);
        return new Structure(units, s, newCS);
    }
    Structure.transform = transform;
    var StructureBuilder = /** @class */ (function () {
        function StructureBuilder(parent, coordinateSystem) {
            this.parent = parent;
            this.coordinateSystem = coordinateSystem;
            this.units = [];
            this.invariantId = id_factory_1.idFactory();
        }
        StructureBuilder.prototype.addUnit = function (kind, model, operator, elements) {
            var unit = unit_1.default.create(this.units.length, this.invariantId(), kind, model, operator, elements);
            this.units.push(unit);
            return unit;
        };
        StructureBuilder.prototype.addWithOperator = function (unit, operator) {
            var newUnit = unit.applyOperator(this.units.length, operator);
            this.units.push(newUnit);
            return newUnit;
        };
        StructureBuilder.prototype.getStructure = function () {
            return create(this.units, this.parent, this.coordinateSystem);
        };
        Object.defineProperty(StructureBuilder.prototype, "isEmpty", {
            get: function () {
                return this.units.length === 0;
            },
            enumerable: true,
            configurable: true
        });
        return StructureBuilder;
    }());
    Structure.StructureBuilder = StructureBuilder;
    function Builder(parent, coordinateSystem) {
        return new StructureBuilder(parent, coordinateSystem);
    }
    Structure.Builder = Builder;
    function hashCode(s) {
        return s.hashCode;
    }
    Structure.hashCode = hashCode;
    /** Hash based on all unit.model conformation values in the structure */
    function conformationHash(s) {
        return util_1.hashString(s.units.map(function (u) { return unit_1.default.conformationId(u); }).join('|'));
    }
    Structure.conformationHash = conformationHash;
    function areUnitAndIndicesEqual(a, b) {
        if (a.elementCount !== b.elementCount)
            return false;
        var len = a.units.length;
        if (len !== b.units.length)
            return false;
        for (var i = 0; i < len; i++) {
            if (a.units[i].id !== b.units[i].id)
                return false;
        }
        for (var i = 0; i < len; i++) {
            if (!int_1.SortedArray.areEqual(a.units[i].elements, b.units[i].elements))
                return false;
        }
        return true;
    }
    Structure.areUnitAndIndicesEqual = areUnitAndIndicesEqual;
    function areEquivalent(a, b) {
        return a === b || (a.hashCode === b.hashCode &&
            symmetry_1.default.areTransformGroupsEquivalent(a.unitSymmetryGroups, b.unitSymmetryGroups));
    }
    Structure.areEquivalent = areEquivalent;
    /** Check if the structures or their parents are equivalent */
    function areParentsEquivalent(a, b) {
        return areEquivalent(a.parent || a, b.parent || b);
    }
    Structure.areParentsEquivalent = areParentsEquivalent;
    /** Check if the structures or their parents are equal */
    function areParentsEqual(a, b) {
        return (a.parent || a) === (b.parent || b);
    }
    Structure.areParentsEqual = areParentsEqual;
    var ElementLocationIterator = /** @class */ (function () {
        function ElementLocationIterator(structure) {
            this.structure = structure;
            this.current = element_1.default.create();
            this.unitIndex = 0;
            this.maxIdx = 0;
            this.idx = -1;
            this.hasNext = structure.elementCount > 0;
            if (this.hasNext) {
                this.elements = structure.units[0].elements;
                this.maxIdx = this.elements.length - 1;
                this.current.unit = structure.units[0];
            }
        }
        ElementLocationIterator.prototype.move = function () {
            this.advance();
            this.current.element = this.elements[this.idx];
            return this.current;
        };
        ElementLocationIterator.prototype.advance = function () {
            if (this.idx < this.maxIdx) {
                this.idx++;
                if (this.idx === this.maxIdx)
                    this.hasNext = this.unitIndex + 1 < this.structure.units.length;
                return;
            }
            this.idx = 0;
            this.unitIndex++;
            if (this.unitIndex >= this.structure.units.length) {
                this.hasNext = false;
                return;
            }
            this.current.unit = this.structure.units[this.unitIndex];
            this.elements = this.current.unit.elements;
            this.maxIdx = this.elements.length - 1;
        };
        return ElementLocationIterator;
    }());
    Structure.ElementLocationIterator = ElementLocationIterator;
    var distVec = linear_algebra_1.Vec3.zero();
    function unitElementMinDistance(unit, p, eRadius) {
        var elements = unit.elements, _a = unit.conformation, position = _a.position, r = _a.r, dV = distVec;
        var minD = Number.MAX_VALUE;
        for (var i = 0, _i = elements.length; i < _i; i++) {
            var e = elements[i];
            var d = linear_algebra_1.Vec3.distance(p, position(e, dV)) - eRadius - r(e);
            if (d < minD)
                minD = d;
        }
        return minD;
    }
    function minDistanceToPoint(s, point, radius) {
        var units = s.units;
        var minD = Number.MAX_VALUE;
        for (var i = 0, _i = units.length; i < _i; i++) {
            var unit = units[i];
            var d = unitElementMinDistance(unit, point, radius);
            if (d < minD)
                minD = d;
        }
        return minD;
    }
    Structure.minDistanceToPoint = minDistanceToPoint;
    var distPivot = linear_algebra_1.Vec3.zero();
    function distance(a, b) {
        if (a.elementCount === 0 || b.elementCount === 0)
            return 0;
        var units = a.units;
        var minD = Number.MAX_VALUE;
        for (var i = 0, _i = units.length; i < _i; i++) {
            var unit = units[i];
            var elements = unit.elements, _a = unit.conformation, position = _a.position, r = _a.r;
            for (var i_1 = 0, _i_1 = elements.length; i_1 < _i_1; i_1++) {
                var e = elements[i_1];
                var d = minDistanceToPoint(b, position(e, distPivot), r(e));
                if (d < minD)
                    minD = d;
            }
        }
        return minD;
    }
    Structure.distance = distance;
})(Structure || (Structure = {}));
exports.default = Structure;
