/**
 * Copyright (c) 2017-2019 mol* contributors, licensed under MIT, See LICENSE file for more info.
 *
 * @author David Sehnal <david.sehnal@gmail.com>
 * @author Alexander Rose <alexander.rose@weirdbyte.de>
 */
import { SymmetryOperator } from '../../../mol-math/geometry/symmetry-operator';
import { Model } from '../model';
import { Lookup3D } from '../../../mol-math/geometry';
import { IntraUnitLinks } from './unit/links';
import { CoarseElements, CoarseSphereConformation, CoarseGaussianConformation } from '../model/properties/coarse';
import { ValueRef } from '../../../mol-util';
import { UnitRings } from './unit/rings';
import StructureElement from './element';
import { ChainIndex, ResidueIndex, ElementIndex } from '../model/indexing';
import { IntMap, SortedArray } from '../../../mol-data/int';
/**
 * A building block of a structure that corresponds to an atomic or
 * a coarse grained representation 'conveniently grouped together'.
 */
declare type Unit = Unit.Atomic | Unit.Spheres | Unit.Gaussians;
declare namespace Unit {
    const enum Kind {
        Atomic = 0,
        Spheres = 1,
        Gaussians = 2
    }
    function isAtomic(u: Unit): u is Atomic;
    function isCoarse(u: Unit): u is Spheres | Gaussians;
    function isSpheres(u: Unit): u is Spheres;
    function isGaussians(u: Unit): u is Gaussians;
    function create(id: number, invariantId: number, kind: Kind, model: Model, operator: SymmetryOperator, elements: StructureElement.Set): Unit;
    /** A group of units that differ only by symmetry operators. */
    type SymmetryGroup = {
        readonly elements: StructureElement.Set;
        readonly units: ReadonlyArray<Unit>;
        /** Maps unit.id to index of unit in units array */
        readonly unitIndexMap: IntMap<number>;
        /** Hash based on unit.invariantId which is the same for all units in the group */
        readonly hashCode: number;
        /** Hash based on all unit.id values in the group, reflecting the units transformation*/
        readonly transformHash: number;
    };
    function SymmetryGroup(units: Unit[]): {
        elements: SortedArray<ElementIndex>;
        units: Unit[];
        readonly unitIndexMap: IntMap<number>;
        hashCode: number;
        transformHash: number;
    };
    namespace SymmetryGroup {
        function areInvariantElementsEqual(a: SymmetryGroup, b: SymmetryGroup): boolean;
    }
    function conformationId(unit: Unit): import("../../../mol-util/uuid").default;
    function hashUnit(u: Unit): number;
    interface Base {
        readonly id: number;
        readonly invariantId: number;
        readonly elements: StructureElement.Set;
        readonly model: Model;
        readonly conformation: SymmetryOperator.ArrayMapping<ElementIndex>;
        getChild(elements: StructureElement.Set): Unit;
        applyOperator(id: number, operator: SymmetryOperator, dontCompose?: boolean): Unit;
        readonly lookup3d: Lookup3D;
        readonly polymerElements: SortedArray<ElementIndex>;
        readonly gapElements: SortedArray<ElementIndex>;
    }
    /**
     * A bulding block of a structure that corresponds
     * to a "natural group of atoms" (most often a "chain")
     * together with a tranformation (rotation and translation)
     * that is dynamically applied to the underlying atom set.
     *
     * An atom set can be referenced by multiple different units which
     * makes construction of assemblies and spacegroups very efficient.
     */
    class Atomic implements Base {
        readonly kind = Kind.Atomic;
        readonly id: number;
        readonly invariantId: number;
        readonly elements: StructureElement.Set;
        readonly model: Model;
        readonly conformation: SymmetryOperator.ArrayMapping<ElementIndex>;
        /** Reference some commonly accessed things for faster access. */
        readonly residueIndex: ArrayLike<ResidueIndex>;
        readonly chainIndex: ArrayLike<ChainIndex>;
        private props;
        getChild(elements: StructureElement.Set): Unit;
        applyOperator(id: number, operator: SymmetryOperator, dontCompose?: boolean): Unit;
        readonly lookup3d: Lookup3D<number>;
        readonly links: import("../../../mol-math/graph").IntAdjacencyGraph<{
            readonly order: ArrayLike<number>;
            readonly flags: ArrayLike<import("../model/types").LinkType.Flag>;
        }>;
        readonly rings: UnitRings;
        readonly polymerElements: SortedArray<ElementIndex>;
        readonly gapElements: SortedArray<ElementIndex>;
        readonly nucleotideElements: SortedArray<ElementIndex>;
        readonly proteinElements: SortedArray<ElementIndex>;
        getResidueIndex(elementIndex: StructureElement.UnitIndex): ResidueIndex;
        constructor(id: number, invariantId: number, model: Model, elements: StructureElement.Set, conformation: SymmetryOperator.ArrayMapping<ElementIndex>, props: AtomicProperties);
    }
    interface AtomicProperties {
        lookup3d: ValueRef<Lookup3D | undefined>;
        links: ValueRef<IntraUnitLinks | undefined>;
        rings: ValueRef<UnitRings | undefined>;
        polymerElements: ValueRef<SortedArray<ElementIndex> | undefined>;
        gapElements: ValueRef<SortedArray<ElementIndex> | undefined>;
        nucleotideElements: ValueRef<SortedArray<ElementIndex> | undefined>;
        proteinElements: ValueRef<SortedArray<ElementIndex> | undefined>;
    }
    function AtomicProperties(): AtomicProperties;
    class Coarse<K extends Kind.Gaussians | Kind.Spheres, C extends CoarseSphereConformation | CoarseGaussianConformation> implements Base {
        readonly kind: K;
        readonly id: number;
        readonly invariantId: number;
        readonly elements: StructureElement.Set;
        readonly model: Model;
        readonly conformation: SymmetryOperator.ArrayMapping<ElementIndex>;
        readonly coarseElements: CoarseElements;
        readonly coarseConformation: C;
        private props;
        getChild(elements: StructureElement.Set): Unit;
        applyOperator(id: number, operator: SymmetryOperator, dontCompose?: boolean): Unit;
        readonly lookup3d: Lookup3D<number>;
        readonly polymerElements: SortedArray<ElementIndex>;
        readonly gapElements: SortedArray<ElementIndex>;
        private getCoarseConformation;
        constructor(id: number, invariantId: number, model: Model, kind: K, elements: StructureElement.Set, conformation: SymmetryOperator.ArrayMapping<ElementIndex>, props: CoarseProperties);
    }
    interface CoarseProperties {
        lookup3d: ValueRef<Lookup3D | undefined>;
        polymerElements: ValueRef<SortedArray<ElementIndex> | undefined>;
        gapElements: ValueRef<SortedArray<ElementIndex> | undefined>;
    }
    function CoarseProperties(): CoarseProperties;
    class Spheres extends Coarse<Kind.Spheres, CoarseSphereConformation> {
    }
    class Gaussians extends Coarse<Kind.Gaussians, CoarseGaussianConformation> {
    }
}
export default Unit;
