"use strict";
/**
 * Copyright (c) 2017-2019 mol* contributors, licensed under MIT, See LICENSE file for more info.
 *
 * @author David Sehnal <david.sehnal@gmail.com>
 * @author Alexander Rose <alexander.rose@weirdbyte.de>
 */
var __extends = (this && this.__extends) || (function () {
    var extendStatics = function (d, b) {
        extendStatics = Object.setPrototypeOf ||
            ({ __proto__: [] } instanceof Array && function (d, b) { d.__proto__ = b; }) ||
            function (d, b) { for (var p in b) if (b.hasOwnProperty(p)) d[p] = b[p]; };
        return extendStatics(d, b);
    };
    return function (d, b) {
        extendStatics(d, b);
        function __() { this.constructor = d; }
        d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
    };
})();
Object.defineProperty(exports, "__esModule", { value: true });
var symmetry_operator_1 = require("../../../mol-math/geometry/symmetry-operator");
var geometry_1 = require("../../../mol-math/geometry");
var links_1 = require("./unit/links");
var mol_util_1 = require("../../../mol-util");
var rings_1 = require("./unit/rings");
var int_1 = require("../../../mol-data/int");
var util_1 = require("../../../mol-data/util");
var polymer_1 = require("./util/polymer");
var Unit;
(function (Unit) {
    function isAtomic(u) { return u.kind === 0 /* Atomic */; }
    Unit.isAtomic = isAtomic;
    function isCoarse(u) { return u.kind === 1 /* Spheres */ || u.kind === 2 /* Gaussians */; }
    Unit.isCoarse = isCoarse;
    function isSpheres(u) { return u.kind === 1 /* Spheres */; }
    Unit.isSpheres = isSpheres;
    function isGaussians(u) { return u.kind === 2 /* Gaussians */; }
    Unit.isGaussians = isGaussians;
    function create(id, invariantId, kind, model, operator, elements) {
        switch (kind) {
            case 0 /* Atomic */: return new Atomic(id, invariantId, model, elements, symmetry_operator_1.SymmetryOperator.createMapping(operator, model.atomicConformation, void 0), AtomicProperties());
            case 1 /* Spheres */: return createCoarse(id, invariantId, model, 1 /* Spheres */, elements, symmetry_operator_1.SymmetryOperator.createMapping(operator, model.coarseConformation.spheres, getSphereRadiusFunc(model)), CoarseProperties());
            case 2 /* Gaussians */: return createCoarse(id, invariantId, model, 2 /* Gaussians */, elements, symmetry_operator_1.SymmetryOperator.createMapping(operator, model.coarseConformation.gaussians, getGaussianRadiusFunc(model)), CoarseProperties());
        }
    }
    Unit.create = create;
    function getUnitIndexMap(units) {
        var unitIndexMap = int_1.IntMap.Mutable();
        for (var i = 0, _i = units.length; i < _i; i++) {
            unitIndexMap.set(units[i].id, i);
        }
        return unitIndexMap;
    }
    function SymmetryGroup(units) {
        var props = {};
        return {
            elements: units[0].elements,
            units: units,
            get unitIndexMap() {
                if (props.unitIndexMap)
                    return props.unitIndexMap;
                props.unitIndexMap = getUnitIndexMap(units);
                return props.unitIndexMap;
            },
            hashCode: hashUnit(units[0]),
            transformHash: util_1.hashFnv32a(units.map(function (u) { return u.id; }))
        };
    }
    Unit.SymmetryGroup = SymmetryGroup;
    (function (SymmetryGroup) {
        function areInvariantElementsEqual(a, b) {
            if (a.hashCode !== b.hashCode)
                return false;
            return int_1.SortedArray.areEqual(a.elements, b.elements);
        }
        SymmetryGroup.areInvariantElementsEqual = areInvariantElementsEqual;
    })(SymmetryGroup = Unit.SymmetryGroup || (Unit.SymmetryGroup = {}));
    function conformationId(unit) {
        return Unit.isAtomic(unit) ? unit.model.atomicConformation.id : unit.model.coarseConformation.id;
    }
    Unit.conformationId = conformationId;
    function hashUnit(u) {
        return util_1.hash2(u.invariantId, int_1.SortedArray.hashCode(u.elements));
    }
    Unit.hashUnit = hashUnit;
    function getSphereRadiusFunc(model) {
        var r = model.coarseConformation.spheres.radius;
        return function (i) { return r[i]; };
    }
    function getGaussianRadiusFunc(model) {
        // TODO: compute radius for gaussians
        return function (i) { return 0; };
    }
    /**
     * A bulding block of a structure that corresponds
     * to a "natural group of atoms" (most often a "chain")
     * together with a tranformation (rotation and translation)
     * that is dynamically applied to the underlying atom set.
     *
     * An atom set can be referenced by multiple different units which
     * makes construction of assemblies and spacegroups very efficient.
     */
    var Atomic = /** @class */ (function () {
        function Atomic(id, invariantId, model, elements, conformation, props) {
            this.kind = 0 /* Atomic */;
            this.id = id;
            this.invariantId = invariantId;
            this.model = model;
            this.elements = elements;
            this.conformation = conformation;
            this.residueIndex = model.atomicHierarchy.residueAtomSegments.index;
            this.chainIndex = model.atomicHierarchy.chainAtomSegments.index;
            this.props = props;
        }
        Atomic.prototype.getChild = function (elements) {
            if (elements.length === this.elements.length)
                return this;
            return new Atomic(this.id, this.invariantId, this.model, elements, this.conformation, AtomicProperties());
        };
        Atomic.prototype.applyOperator = function (id, operator, dontCompose) {
            if (dontCompose === void 0) { dontCompose = false; }
            var op = dontCompose ? operator : symmetry_operator_1.SymmetryOperator.compose(this.conformation.operator, operator);
            return new Atomic(id, this.invariantId, this.model, this.elements, symmetry_operator_1.SymmetryOperator.createMapping(op, this.model.atomicConformation, this.conformation.r), this.props);
        };
        Object.defineProperty(Atomic.prototype, "lookup3d", {
            get: function () {
                if (this.props.lookup3d.ref)
                    return this.props.lookup3d.ref;
                var _a = this.model.atomicConformation, x = _a.x, y = _a.y, z = _a.z;
                this.props.lookup3d.ref = geometry_1.GridLookup3D({ x: x, y: y, z: z, indices: this.elements });
                return this.props.lookup3d.ref;
            },
            enumerable: true,
            configurable: true
        });
        Object.defineProperty(Atomic.prototype, "links", {
            get: function () {
                if (this.props.links.ref)
                    return this.props.links.ref;
                this.props.links.ref = links_1.computeIntraUnitBonds(this);
                return this.props.links.ref;
            },
            enumerable: true,
            configurable: true
        });
        Object.defineProperty(Atomic.prototype, "rings", {
            get: function () {
                if (this.props.rings.ref)
                    return this.props.rings.ref;
                this.props.rings.ref = rings_1.UnitRings.create(this);
                return this.props.rings.ref;
            },
            enumerable: true,
            configurable: true
        });
        Object.defineProperty(Atomic.prototype, "polymerElements", {
            get: function () {
                if (this.props.polymerElements.ref)
                    return this.props.polymerElements.ref;
                this.props.polymerElements.ref = polymer_1.getAtomicPolymerElements(this);
                return this.props.polymerElements.ref;
            },
            enumerable: true,
            configurable: true
        });
        Object.defineProperty(Atomic.prototype, "gapElements", {
            get: function () {
                if (this.props.gapElements.ref)
                    return this.props.gapElements.ref;
                this.props.gapElements.ref = polymer_1.getAtomicGapElements(this);
                return this.props.gapElements.ref;
            },
            enumerable: true,
            configurable: true
        });
        Object.defineProperty(Atomic.prototype, "nucleotideElements", {
            get: function () {
                if (this.props.nucleotideElements.ref)
                    return this.props.nucleotideElements.ref;
                this.props.nucleotideElements.ref = polymer_1.getNucleotideElements(this);
                return this.props.nucleotideElements.ref;
            },
            enumerable: true,
            configurable: true
        });
        Object.defineProperty(Atomic.prototype, "proteinElements", {
            get: function () {
                if (this.props.proteinElements.ref)
                    return this.props.proteinElements.ref;
                this.props.proteinElements.ref = polymer_1.getProteinElements(this);
                return this.props.proteinElements.ref;
            },
            enumerable: true,
            configurable: true
        });
        Atomic.prototype.getResidueIndex = function (elementIndex) {
            return this.model.atomicHierarchy.residueAtomSegments.index[this.elements[elementIndex]];
        };
        return Atomic;
    }());
    Unit.Atomic = Atomic;
    function AtomicProperties() {
        return {
            lookup3d: mol_util_1.ValueRef.create(void 0),
            links: mol_util_1.ValueRef.create(void 0),
            rings: mol_util_1.ValueRef.create(void 0),
            polymerElements: mol_util_1.ValueRef.create(void 0),
            gapElements: mol_util_1.ValueRef.create(void 0),
            nucleotideElements: mol_util_1.ValueRef.create(void 0),
            proteinElements: mol_util_1.ValueRef.create(void 0),
        };
    }
    var Coarse = /** @class */ (function () {
        function Coarse(id, invariantId, model, kind, elements, conformation, props) {
            this.kind = kind;
            this.id = id;
            this.invariantId = invariantId;
            this.model = model;
            this.elements = elements;
            this.conformation = conformation;
            this.coarseElements = kind === 1 /* Spheres */ ? model.coarseHierarchy.spheres : model.coarseHierarchy.gaussians;
            this.coarseConformation = (kind === 1 /* Spheres */ ? model.coarseConformation.spheres : model.coarseConformation.gaussians);
            this.props = props;
        }
        Coarse.prototype.getChild = function (elements) {
            if (elements.length === this.elements.length)
                return this /** lets call this an ugly temporary hack */;
            return createCoarse(this.id, this.invariantId, this.model, this.kind, elements, this.conformation, CoarseProperties());
        };
        Coarse.prototype.applyOperator = function (id, operator, dontCompose) {
            if (dontCompose === void 0) { dontCompose = false; }
            var op = dontCompose ? operator : symmetry_operator_1.SymmetryOperator.compose(this.conformation.operator, operator);
            var ret = createCoarse(id, this.invariantId, this.model, this.kind, this.elements, symmetry_operator_1.SymmetryOperator.createMapping(op, this.getCoarseConformation(), this.conformation.r), this.props);
            // (ret as Coarse<K, C>)._lookup3d = this._lookup3d;
            return ret;
        };
        Object.defineProperty(Coarse.prototype, "lookup3d", {
            get: function () {
                if (this.props.lookup3d.ref)
                    return this.props.lookup3d.ref;
                // TODO: support sphere radius?
                var _a = this.getCoarseConformation(), x = _a.x, y = _a.y, z = _a.z;
                this.props.lookup3d.ref = geometry_1.GridLookup3D({ x: x, y: y, z: z, indices: this.elements });
                return this.props.lookup3d.ref;
            },
            enumerable: true,
            configurable: true
        });
        Object.defineProperty(Coarse.prototype, "polymerElements", {
            get: function () {
                if (this.props.polymerElements.ref)
                    return this.props.polymerElements.ref;
                this.props.polymerElements.ref = polymer_1.getCoarsePolymerElements(this); // TODO get rid of casting
                return this.props.polymerElements.ref;
            },
            enumerable: true,
            configurable: true
        });
        Object.defineProperty(Coarse.prototype, "gapElements", {
            get: function () {
                if (this.props.gapElements.ref)
                    return this.props.gapElements.ref;
                this.props.gapElements.ref = polymer_1.getCoarseGapElements(this); // TODO get rid of casting
                return this.props.gapElements.ref;
            },
            enumerable: true,
            configurable: true
        });
        Coarse.prototype.getCoarseConformation = function () {
            return this.kind === 1 /* Spheres */ ? this.model.coarseConformation.spheres : this.model.coarseConformation.gaussians;
        };
        return Coarse;
    }());
    function CoarseProperties() {
        return {
            lookup3d: mol_util_1.ValueRef.create(void 0),
            polymerElements: mol_util_1.ValueRef.create(void 0),
            gapElements: mol_util_1.ValueRef.create(void 0),
        };
    }
    function createCoarse(id, invariantId, model, kind, elements, conformation, props) {
        return new Coarse(id, invariantId, model, kind, elements, conformation, props) /** lets call this an ugly temporary hack */;
    }
    var Spheres = /** @class */ (function (_super) {
        __extends(Spheres, _super);
        function Spheres() {
            return _super !== null && _super.apply(this, arguments) || this;
        }
        return Spheres;
    }(Coarse));
    Unit.Spheres = Spheres;
    var Gaussians = /** @class */ (function (_super) {
        __extends(Gaussians, _super);
        function Gaussians() {
            return _super !== null && _super.apply(this, arguments) || this;
        }
        return Gaussians;
    }(Coarse));
    Unit.Gaussians = Gaussians;
})(Unit || (Unit = {}));
exports.default = Unit;
