"use strict";
/**
 * Copyright (c) 2017 Mol* contributors, licensed under MIT, See LICENSE file for more info.
 *
 * @author David Sehnal <david.sehnal@gmail.com>
 */
Object.defineProperty(exports, "__esModule", { value: true });
var graph_1 = require("../../../../../mol-math/graph");
var common_1 = require("./common");
var int_1 = require("../../../../../mol-data/int");
var bonds_1 = require("../../../../../mol-model-formats/structure/mmcif/bonds");
function getGraph(atomA, atomB, _order, _flags, atomCount) {
    var builder = new graph_1.IntAdjacencyGraph.EdgeBuilder(atomCount, atomA, atomB);
    var flags = new Uint16Array(builder.slotCount);
    var order = new Int8Array(builder.slotCount);
    for (var i = 0, _i = builder.edgeCount; i < _i; i++) {
        builder.addNextEdge();
        builder.assignProperty(flags, _flags[i]);
        builder.assignProperty(order, _order[i]);
    }
    return builder.createGraph({ flags: flags, order: order });
}
function _computeBonds(unit, params) {
    var MAX_RADIUS = 4;
    var _a = unit.model.atomicConformation, x = _a.x, y = _a.y, z = _a.z;
    var atomCount = unit.elements.length;
    var atoms = unit.elements, residueIndex = unit.residueIndex;
    var _b = unit.model.atomicHierarchy.atoms, type_symbol = _b.type_symbol, label_atom_id = _b.label_atom_id, label_alt_id = _b.label_alt_id;
    var label_comp_id = unit.model.atomicHierarchy.residues.label_comp_id;
    var query3d = unit.lookup3d;
    var structConn = unit.model.sourceData.kind === 'mmCIF' ? bonds_1.StructConn.get(unit.model) : void 0;
    var component = unit.model.sourceData.kind === 'mmCIF' ? bonds_1.ComponentBond.get(unit.model) : void 0;
    var atomA = [];
    var atomB = [];
    var flags = [];
    var order = [];
    var lastResidue = -1;
    var componentMap = void 0;
    for (var _aI = 0; _aI < atomCount; _aI++) {
        var aI = atoms[_aI];
        var raI = residueIndex[aI];
        if (!params.forceCompute && raI !== lastResidue) {
            var resn = label_comp_id.value(raI);
            if (!!component && component.entries.has(resn)) {
                componentMap = component.entries.get(resn).map;
            }
            else {
                componentMap = void 0;
            }
        }
        lastResidue = raI;
        var componentPairs = componentMap ? componentMap.get(label_atom_id.value(aI)) : void 0;
        var aeI = common_1.getElementIdx(type_symbol.value(aI));
        var _c = query3d.find(x[aI], y[aI], z[aI], MAX_RADIUS), indices = _c.indices, count = _c.count, squaredDistances = _c.squaredDistances;
        var isHa = common_1.isHydrogen(aeI);
        var thresholdA = common_1.getElementThreshold(aeI);
        var altA = label_alt_id.value(aI);
        var metalA = common_1.MetalsSet.has(aeI);
        var structConnEntries = params.forceCompute ? void 0 : structConn && structConn.getAtomEntries(aI);
        if (structConnEntries) {
            for (var _d = 0, structConnEntries_1 = structConnEntries; _d < structConnEntries_1.length; _d++) {
                var se = structConnEntries_1[_d];
                if (se.distance < MAX_RADIUS)
                    continue;
                for (var _e = 0, _f = se.partners; _e < _f.length; _e++) {
                    var p = _f[_e];
                    var _bI = int_1.SortedArray.indexOf(unit.elements, p.atomIndex);
                    if (_bI < 0)
                        continue;
                    atomA[atomA.length] = _aI;
                    atomB[atomB.length] = _bI;
                    flags[flags.length] = se.flags;
                    order[order.length] = se.order;
                }
            }
        }
        for (var ni = 0; ni < count; ni++) {
            var _bI = indices[ni];
            var bI = atoms[_bI];
            if (bI <= aI)
                continue;
            var altB = label_alt_id.value(bI);
            if (altA && altB && altA !== altB)
                continue;
            var beI = common_1.getElementIdx(type_symbol.value(bI));
            var isMetal = metalA || common_1.MetalsSet.has(beI);
            var rbI = residueIndex[bI];
            // handle "component dictionary" bonds.
            if (raI === rbI && componentPairs) {
                var e = componentPairs.get(label_atom_id.value(bI));
                if (e) {
                    atomA[atomA.length] = _aI;
                    atomB[atomB.length] = _bI;
                    order[order.length] = e.order;
                    var flag = e.flags;
                    if (isMetal) {
                        if (flag | 1 /* Covalent */)
                            flag ^= 1 /* Covalent */;
                        flag |= 2 /* MetallicCoordination */;
                    }
                    flags[flags.length] = flag;
                }
                continue;
            }
            var isHb = common_1.isHydrogen(beI);
            if (isHa && isHb)
                continue;
            var dist = Math.sqrt(squaredDistances[ni]);
            if (dist === 0)
                continue;
            // handle "struct conn" bonds.
            if (structConnEntries && structConnEntries.length) {
                var added = false;
                for (var _g = 0, structConnEntries_2 = structConnEntries; _g < structConnEntries_2.length; _g++) {
                    var se = structConnEntries_2[_g];
                    for (var _h = 0, _j = se.partners; _h < _j.length; _h++) {
                        var p = _j[_h];
                        if (p.atomIndex === bI) {
                            atomA[atomA.length] = _aI;
                            atomB[atomB.length] = _bI;
                            flags[flags.length] = se.flags;
                            order[order.length] = se.order;
                            added = true;
                            break;
                        }
                    }
                    if (added)
                        break;
                }
                if (added)
                    continue;
            }
            if (isHa || isHb) {
                if (dist < params.maxHbondLength) {
                    atomA[atomA.length] = _aI;
                    atomB[atomB.length] = _bI;
                    order[order.length] = 1;
                    flags[flags.length] = 1 /* Covalent */ | 64 /* Computed */; // TODO: check if correct
                }
                continue;
            }
            var thresholdAB = common_1.getElementPairThreshold(aeI, beI);
            var pairingThreshold = thresholdAB > 0
                ? thresholdAB
                : beI < 0 ? thresholdA : Math.max(thresholdA, common_1.getElementThreshold(beI));
            if (dist <= pairingThreshold) {
                atomA[atomA.length] = _aI;
                atomB[atomB.length] = _bI;
                order[order.length] = 1;
                flags[flags.length] = (isMetal ? 2 /* MetallicCoordination */ : 1 /* Covalent */) | 64 /* Computed */;
            }
        }
    }
    return getGraph(atomA, atomB, order, flags, atomCount);
}
function computeIntraUnitBonds(unit, params) {
    return _computeBonds(unit, {
        maxHbondLength: (params && params.maxHbondLength) || 1.15,
        forceCompute: !!(params && params.forceCompute),
    });
}
exports.computeIntraUnitBonds = computeIntraUnitBonds;
