/**
 * Copyright (c) 2018-2019 mol* contributors, licensed under MIT, See LICENSE file for more info.
 *
 * @author David Sehnal <david.sehnal@gmail.com>
 * @author Alexander Rose <alexander.rose@weirdbyte.de>
 */
/// <reference types="react" />
import { List } from 'immutable';
import { Canvas3D } from '../mol-canvas3d/canvas3d';
import { CustomPropertyRegistry } from '../mol-model-props/common/custom-property-registry';
import { StructureRepresentationRegistry } from '../mol-repr/structure/registry';
import { VolumeRepresentationRegistry } from '../mol-repr/volume/registry';
import { State, StateTransform, StateTransformer } from '../mol-state';
import { Task } from '../mol-task';
import { ThemeRegistryContext } from '../mol-theme/theme';
import { ajaxGet } from '../mol-util/data-source';
import { LogEntry } from '../mol-util/log-entry';
import { PluginCommand } from './command';
import { PluginLayout } from './layout';
import { PluginSpec } from './spec';
import { PluginState } from './state';
import { DataFormatRegistry } from './state/actions/data-format';
import { StateTransformParameters } from './ui/state/common';
import { LociLabelEntry, LociLabelManager } from './util/loci-label-manager';
import { TaskManager } from './util/task-manager';
import { StructureElementSelectionManager } from './util/structure-element-selection';
import { SubstructureParentHelper } from './util/substructure-parent-helper';
import { Model, Structure } from '../mol-model/structure';
import { Interactivity } from './util/interactivity';
interface Log {
    entries: List<LogEntry>;
    readonly entry: (e: LogEntry) => void;
    readonly error: (msg: string) => void;
    readonly message: (msg: string) => void;
    readonly info: (msg: string) => void;
    readonly warn: (msg: string) => void;
}
export declare class PluginContext {
    spec: PluginSpec;
    private disposed;
    private ev;
    private tasks;
    readonly state: PluginState;
    readonly commands: PluginCommand.Manager;
    readonly events: {
        state: {
            cell: {
                readonly stateUpdated: import("rxjs").Observable<State.ObjectEvent & {
                    cell: import("../mol-state/object").StateObjectCell<import("../mol-state/object").StateObject<any, import("../mol-state/object").StateObject.Type<any>>, StateTransform<StateTransformer<any, import("../mol-state/object").StateObject<any, import("../mol-state/object").StateObject.Type<any>>, any>>>;
                }>;
                readonly created: import("rxjs").Observable<State.ObjectEvent & {
                    cell: import("../mol-state/object").StateObjectCell<import("../mol-state/object").StateObject<any, import("../mol-state/object").StateObject.Type<any>>, StateTransform<StateTransformer<any, import("../mol-state/object").StateObject<any, import("../mol-state/object").StateObject.Type<any>>, any>>>;
                }>;
                readonly removed: import("rxjs").Observable<State.ObjectEvent & {
                    parent: string;
                }>;
            };
            object: {
                readonly created: import("rxjs").Observable<State.ObjectEvent & {
                    obj: import("../mol-state/object").StateObject<any, import("../mol-state/object").StateObject.Type<any>>;
                }>;
                readonly removed: import("rxjs").Observable<State.ObjectEvent & {
                    obj?: import("../mol-state/object").StateObject<any, import("../mol-state/object").StateObject.Type<any>> | undefined;
                }>;
                readonly updated: import("rxjs").Observable<State.ObjectEvent & {
                    action: "in-place" | "recreate";
                    obj: import("../mol-state/object").StateObject<any, import("../mol-state/object").StateObject.Type<any>>;
                    oldObj?: import("../mol-state/object").StateObject<any, import("../mol-state/object").StateObject.Type<any>> | undefined;
                }>;
            };
            readonly cameraSnapshots: {
                changed: import("rxjs").Subject<unknown>;
            };
            readonly snapshots: {
                changed: import("rxjs").Subject<unknown>;
            };
        };
        readonly log: import("rxjs").Subject<LogEntry>;
        readonly task: {
            progress: import("rxjs").Subject<TaskManager.ProgressEvent>;
            finished: import("rxjs").Subject<{
                id: number;
            }>;
        };
        canvas3d: {
            readonly settingsUpdated: import("rxjs").Subject<unknown>;
        };
        interactivity: {
            readonly propsUpdated: import("rxjs").Subject<unknown>;
        };
    };
    readonly behaviors: {
        state: {
            readonly isAnimating: import("rxjs").BehaviorSubject<boolean>;
            readonly isUpdating: import("rxjs").BehaviorSubject<boolean>;
        };
        interaction: {
            readonly highlight: import("rxjs").BehaviorSubject<Interactivity.HighlightEvent>;
            readonly click: import("rxjs").BehaviorSubject<Interactivity.ClickEvent>;
        };
        labels: {
            readonly highlight: import("rxjs").BehaviorSubject<{
                entries: readonly LociLabelEntry[];
            }>;
        };
    };
    readonly canvas3d: Canvas3D;
    readonly layout: PluginLayout;
    readonly interactivity: Interactivity;
    readonly lociLabels: LociLabelManager;
    readonly structureRepresentation: {
        readonly registry: StructureRepresentationRegistry;
        readonly themeCtx: ThemeRegistryContext;
    };
    readonly volumeRepresentation: {
        readonly registry: VolumeRepresentationRegistry;
        readonly themeCtx: ThemeRegistryContext;
    };
    readonly dataFormat: {
        readonly registry: DataFormatRegistry<import("./state/objects").PluginStateObject.Data.String | import("./state/objects").PluginStateObject.Data.Binary>;
    };
    readonly customModelProperties: CustomPropertyRegistry<Model>;
    readonly customStructureProperties: CustomPropertyRegistry<Structure>;
    readonly customParamEditors: Map<string, import("react").ComponentClass<StateTransformParameters.Props, any>>;
    readonly helpers: {
        readonly structureSelection: StructureElementSelectionManager;
        readonly substructureParent: SubstructureParentHelper;
    };
    initViewer(canvas: HTMLCanvasElement, container: HTMLDivElement): boolean;
    readonly log: Log;
    /**
     * This should be used in all transform related request so that it could be "spoofed" to allow
     * "static" access to resources.
     */
    readonly fetch: typeof ajaxGet;
    runTask<T>(task: Task<T>): Promise<T>;
    dispose(): void;
    applyTransform(state: State, a: StateTransform.Ref, transformer: StateTransformer, params: any): Promise<void>;
    updateTransform(state: State, a: StateTransform.Ref, params: any): Promise<void>;
    private initBehaviorEvents;
    private initBuiltInBehavior;
    private initBehaviors;
    private initDataActions;
    private initAnimations;
    private initCustomParamEditors;
    constructor(spec: PluginSpec);
}
export {};
