"use strict";
/**
 * Copyright (c) 2019 mol* contributors, licensed under MIT, See LICENSE file for more info.
 *
 * @author David Sehnal <david.sehnal@gmail.com>
 * @author Alexander Rose <alexander.rose@weirdbyte.de>
 */
Object.defineProperty(exports, "__esModule", { value: true });
var int_1 = require("../../mol-data/int");
var loci_1 = require("../../mol-model/loci");
var structure_1 = require("../../mol-model/structure");
var objects_1 = require("../state/objects");
var StructureElementSelectionManager = /** @class */ (function () {
    function StructureElementSelectionManager(plugin) {
        var _this = this;
        this.plugin = plugin;
        this.entries = new Map();
        this.prevHighlight = void 0;
        plugin.state.dataState.events.object.removed.subscribe(function (e) { return _this.onRemove(e.ref); });
        plugin.state.dataState.events.object.updated.subscribe(function (e) { return _this.onUpdate(e.ref, e.oldObj, e.obj); });
    }
    StructureElementSelectionManager.prototype.getEntry = function (s) {
        var cell = this.plugin.helpers.substructureParent.get(s);
        if (!cell)
            return;
        var ref = cell.transform.ref;
        if (!this.entries.has(ref)) {
            var entry = SelectionEntry(s);
            this.entries.set(ref, entry);
            return entry;
        }
        return this.entries.get(ref);
    };
    StructureElementSelectionManager.prototype.add = function (loci) {
        if (structure_1.StructureElement.isLoci(loci)) {
            var entry = this.getEntry(loci.structure);
            if (entry) {
                entry.selection = structure_1.StructureElement.Loci.union(entry.selection, loci);
                return entry.selection;
            }
        }
        return loci_1.EmptyLoci;
    };
    StructureElementSelectionManager.prototype.remove = function (loci) {
        if (structure_1.StructureElement.isLoci(loci)) {
            var entry = this.getEntry(loci.structure);
            if (entry) {
                entry.selection = structure_1.StructureElement.Loci.subtract(entry.selection, loci);
                return entry.selection.elements.length === 0 ? loci_1.EmptyLoci : entry.selection;
            }
        }
        return loci_1.EmptyLoci;
    };
    StructureElementSelectionManager.prototype.set = function (loci) {
        if (structure_1.StructureElement.isLoci(loci)) {
            var entry = this.getEntry(loci.structure);
            if (entry) {
                entry.selection = loci;
                return entry.selection.elements.length === 0 ? loci_1.EmptyLoci : entry.selection;
            }
        }
        return loci_1.EmptyLoci;
    };
    StructureElementSelectionManager.prototype.clear = function () {
        var keys = this.entries.keys();
        var selections = [];
        while (true) {
            var k = keys.next();
            if (k.done)
                break;
            var s = this.entries.get(k.value);
            if (s.selection.elements.length > 0)
                selections.push(s.selection);
            s.selection = structure_1.StructureElement.Loci(s.selection.structure, []);
        }
        return selections;
    };
    StructureElementSelectionManager.prototype.get = function (structure) {
        var entry = this.getEntry(structure);
        if (!entry)
            return loci_1.EmptyLoci;
        return entry.selection;
    };
    StructureElementSelectionManager.prototype.has = function (loci) {
        if (structure_1.StructureElement.isLoci(loci)) {
            var entry = this.getEntry(loci.structure);
            if (entry) {
                return structure_1.StructureElement.Loci.areIntersecting(loci, entry.selection);
            }
        }
        return false;
    };
    StructureElementSelectionManager.prototype.tryGetRange = function (loci) {
        if (!structure_1.StructureElement.isLoci(loci))
            return;
        if (loci.elements.length !== 1)
            return;
        var entry = this.getEntry(loci.structure);
        if (!entry)
            return;
        var xs = loci.elements[0];
        var e;
        for (var _i = 0, _a = entry.selection.elements; _i < _a.length; _i++) {
            var _e = _a[_i];
            if (xs.unit === _e.unit) {
                e = _e;
                break;
            }
        }
        if (!e)
            return;
        return tryGetElementRange(entry.selection.structure, e, xs);
    };
    StructureElementSelectionManager.prototype.accumulateInteractiveHighlight = function (loci) {
        if (structure_1.StructureElement.isLoci(loci)) {
            if (this.prevHighlight) {
                this.prevHighlight = structure_1.StructureElement.Loci.union(this.prevHighlight, loci);
            }
            else {
                this.prevHighlight = loci;
            }
        }
        return this.prevHighlight;
    };
    StructureElementSelectionManager.prototype.clearInteractiveHighlight = function () {
        var ret = this.prevHighlight;
        this.prevHighlight = void 0;
        return ret || loci_1.EmptyLoci;
    };
    StructureElementSelectionManager.prototype.onRemove = function (ref) {
        if (this.entries.has(ref))
            this.entries.delete(ref);
    };
    StructureElementSelectionManager.prototype.onUpdate = function (ref, oldObj, obj) {
        if (!objects_1.PluginStateObject.Molecule.Structure.is(obj))
            return;
        if (this.entries.has(ref)) {
            if (!objects_1.PluginStateObject.Molecule.Structure.is(oldObj) || oldObj === obj || oldObj.data === obj.data)
                return;
            // remap the old selection to be related to the new object if possible.
            if (structure_1.Structure.areUnitAndIndicesEqual(oldObj.data, obj.data)) {
                this.entries.set(ref, remapSelectionEntry(this.entries.get(ref), obj.data));
                return;
            }
            // clear the selection
            this.entries.set(ref, SelectionEntry(obj.data));
        }
    };
    return StructureElementSelectionManager;
}());
exports.StructureElementSelectionManager = StructureElementSelectionManager;
function SelectionEntry(s) {
    return {
        selection: structure_1.StructureElement.Loci(s, [])
    };
}
/** remap `selection-entry` to be related to `structure` if possible */
function remapSelectionEntry(e, s) {
    return {
        selection: structure_1.StructureElement.Loci.remap(e.selection, s)
    };
}
/**
 * Assumes `ref` and `ext` belong to the same unit in the same structure
 */
function tryGetElementRange(structure, ref, ext) {
    var refMin = int_1.OrderedSet.min(ref.indices);
    var refMax = int_1.OrderedSet.max(ref.indices);
    var extMin = int_1.OrderedSet.min(ext.indices);
    var extMax = int_1.OrderedSet.max(ext.indices);
    var min;
    var max;
    if (refMax < extMin) {
        min = refMax + 1;
        max = extMax;
    }
    else if (extMax < refMin) {
        min = extMin;
        max = refMin - 1;
    }
    else {
        // TODO handle range overlap cases
        return;
    }
    return structure_1.StructureElement.Loci(structure, [{
            unit: ref.unit,
            indices: int_1.OrderedSet.ofRange(min, max)
        }]);
}
