"use strict";
/**
 * Copyright (c) 2018 mol* contributors, licensed under MIT, See LICENSE file for more info.
 *
 * @author Alexander Rose <alexander.rose@weirdbyte.de>
 */
var __assign = (this && this.__assign) || function () {
    __assign = Object.assign || function(t) {
        for (var s, i = 1, n = arguments.length; i < n; i++) {
            s = arguments[i];
            for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p))
                t[p] = s[p];
        }
        return t;
    };
    return __assign.apply(this, arguments);
};
Object.defineProperty(exports, "__esModule", { value: true });
var structure_1 = require("../../../mol-model/structure");
var loci_1 = require("../../../mol-model/loci");
var linear_algebra_1 = require("../../../mol-math/linear-algebra");
var link_1 = require("./util/link");
var int_1 = require("../../../mol-data/int");
var complex_visual_1 = require("../complex-visual");
var mol_util_1 = require("../../../mol-util");
var units_visual_1 = require("../units-visual");
var param_definition_1 = require("../../../mol-util/param-definition");
var location_iterator_1 = require("../../../mol-geo/util/location-iterator");
function createCarbohydrateLinkCylinderMesh(ctx, structure, theme, props, mesh) {
    var _a = structure.carbohydrates, links = _a.links, elements = _a.elements;
    var linkSizeFactor = props.linkSizeFactor;
    var location = structure_1.StructureElement.create();
    var builderProps = {
        linkCount: links.length,
        referencePosition: function (edgeIndex) { return null; },
        position: function (posA, posB, edgeIndex) {
            var l = links[edgeIndex];
            linear_algebra_1.Vec3.copy(posA, elements[l.carbohydrateIndexA].geometry.center);
            linear_algebra_1.Vec3.copy(posB, elements[l.carbohydrateIndexB].geometry.center);
        },
        order: function (edgeIndex) { return 1; },
        flags: function (edgeIndex) { return mol_util_1.BitFlags.create(0 /* None */); },
        radius: function (edgeIndex) {
            var l = links[edgeIndex];
            location.unit = elements[l.carbohydrateIndexA].unit;
            location.element = elements[l.carbohydrateIndexA].anomericCarbon;
            return theme.size.size(location) * linkSizeFactor;
        }
    };
    return link_1.createLinkCylinderMesh(ctx, builderProps, props, mesh);
}
exports.CarbohydrateLinkParams = __assign({}, units_visual_1.UnitsMeshParams, link_1.LinkCylinderParams, { linkSizeFactor: param_definition_1.ParamDefinition.Numeric(0.3, { min: 0, max: 3, step: 0.01 }) });
function CarbohydrateLinkVisual(materialId) {
    return complex_visual_1.ComplexMeshVisual({
        defaultProps: param_definition_1.ParamDefinition.getDefaultValues(exports.CarbohydrateLinkParams),
        createGeometry: createCarbohydrateLinkCylinderMesh,
        createLocationIterator: CarbohydrateLinkIterator,
        getLoci: getLinkLoci,
        eachLocation: eachCarbohydrateLink,
        setUpdateState: function (state, newProps, currentProps) {
            state.createGeometry = (newProps.linkSizeFactor !== currentProps.linkSizeFactor ||
                newProps.radialSegments !== currentProps.radialSegments);
        }
    }, materialId);
}
exports.CarbohydrateLinkVisual = CarbohydrateLinkVisual;
function CarbohydrateLinkIterator(structure) {
    var _a = structure.carbohydrates, elements = _a.elements, links = _a.links;
    var groupCount = links.length;
    var instanceCount = 1;
    var location = structure_1.Link.Location();
    var getLocation = function (groupIndex) {
        var link = links[groupIndex];
        var carbA = elements[link.carbohydrateIndexA];
        var carbB = elements[link.carbohydrateIndexB];
        var indexA = int_1.OrderedSet.indexOf(carbA.unit.elements, carbA.anomericCarbon);
        var indexB = int_1.OrderedSet.indexOf(carbB.unit.elements, carbB.anomericCarbon);
        location.aUnit = carbA.unit;
        location.aIndex = indexA;
        location.bUnit = carbB.unit;
        location.bIndex = indexB;
        return location;
    };
    return location_iterator_1.LocationIterator(groupCount, instanceCount, getLocation, true);
}
function getLinkLoci(pickingId, structure, id) {
    var objectId = pickingId.objectId, groupId = pickingId.groupId;
    if (id === objectId) {
        var _a = structure.carbohydrates, links = _a.links, elements = _a.elements;
        var l = links[groupId];
        var carbA = elements[l.carbohydrateIndexA];
        var carbB = elements[l.carbohydrateIndexB];
        var indexA = int_1.OrderedSet.indexOf(carbA.unit.elements, carbA.anomericCarbon);
        var indexB = int_1.OrderedSet.indexOf(carbB.unit.elements, carbB.anomericCarbon);
        if (indexA !== -1 && indexB !== -1) {
            return structure_1.Link.Loci(structure, [
                structure_1.Link.Location(carbA.unit, indexA, carbB.unit, indexB),
                structure_1.Link.Location(carbB.unit, indexB, carbA.unit, indexA)
            ]);
        }
    }
    return loci_1.EmptyLoci;
}
function eachCarbohydrateLink(loci, structure, apply) {
    var changed = false;
    if (structure_1.Link.isLoci(loci)) {
        if (!structure_1.Structure.areParentsEquivalent(loci.structure, structure))
            return false;
        var getLinkIndex = structure.carbohydrates.getLinkIndex;
        for (var _i = 0, _a = loci.links; _i < _a.length; _i++) {
            var l = _a[_i];
            var idx = getLinkIndex(l.aUnit, l.aUnit.elements[l.aIndex], l.bUnit, l.bUnit.elements[l.bIndex]);
            if (idx !== undefined) {
                if (apply(int_1.Interval.ofSingleton(idx)))
                    changed = true;
            }
        }
    }
    else if (structure_1.StructureElement.isLoci(loci)) {
        if (!structure_1.Structure.areParentsEquivalent(loci.structure, structure))
            return false;
        // TODO mark link only when both of the link elements are in a StructureElement.Loci
        var _b = structure.carbohydrates, getElementIndex_1 = _b.getElementIndex, getLinkIndices_1 = _b.getLinkIndices, elements_1 = _b.elements;
        var _loop_1 = function (e) {
            int_1.OrderedSet.forEach(e.indices, function (v) {
                var carbI = getElementIndex_1(e.unit, e.unit.elements[v]);
                if (carbI !== undefined) {
                    var carb = elements_1[carbI];
                    var indices = getLinkIndices_1(carb.unit, carb.anomericCarbon);
                    for (var i = 0, il = indices.length; i < il; ++i) {
                        if (apply(int_1.Interval.ofSingleton(indices[i])))
                            changed = true;
                    }
                }
            });
        };
        for (var _c = 0, _d = loci.elements; _c < _d.length; _c++) {
            var e = _d[_c];
            _loop_1(e);
        }
    }
    return changed;
}
