"use strict";
/**
 * Copyright (c) 2018 mol* contributors, licensed under MIT, See LICENSE file for more info.
 *
 * @author Alexander Rose <alexander.rose@weirdbyte.de>
 */
var __assign = (this && this.__assign) || function () {
    __assign = Object.assign || function(t) {
        for (var s, i = 1, n = arguments.length; i < n; i++) {
            s = arguments[i];
            for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p))
                t[p] = s[p];
        }
        return t;
    };
    return __assign.apply(this, arguments);
};
Object.defineProperty(exports, "__esModule", { value: true });
var param_definition_1 = require("../../../mol-util/param-definition");
var structure_1 = require("../../../mol-model/structure");
var linear_algebra_1 = require("../../../mol-math/linear-algebra");
var mol_util_1 = require("../../../mol-util");
var link_1 = require("./util/link");
var units_visual_1 = require("../units-visual");
var complex_visual_1 = require("../complex-visual");
var location_iterator_1 = require("../../../mol-geo/util/location-iterator");
var int_1 = require("../../../mol-data/int");
var loci_1 = require("../../../mol-model/loci");
function createCarbohydrateTerminalLinkCylinderMesh(ctx, structure, theme, props, mesh) {
    var _a = structure.carbohydrates, terminalLinks = _a.terminalLinks, elements = _a.elements;
    var linkSizeFactor = props.linkSizeFactor;
    var location = structure_1.StructureElement.create();
    var builderProps = {
        linkCount: terminalLinks.length,
        referencePosition: function (edgeIndex) { return null; },
        position: function (posA, posB, edgeIndex) {
            var l = terminalLinks[edgeIndex];
            if (l.fromCarbohydrate) {
                linear_algebra_1.Vec3.copy(posA, elements[l.carbohydrateIndex].geometry.center);
                l.elementUnit.conformation.position(l.elementUnit.elements[l.elementIndex], posB);
            }
            else {
                l.elementUnit.conformation.position(l.elementUnit.elements[l.elementIndex], posA);
                linear_algebra_1.Vec3.copy(posB, elements[l.carbohydrateIndex].geometry.center);
            }
        },
        order: function (edgeIndex) { return 1; },
        flags: function (edgeIndex) { return mol_util_1.BitFlags.create(0 /* None */); },
        radius: function (edgeIndex) {
            var l = terminalLinks[edgeIndex];
            if (l.fromCarbohydrate) {
                location.unit = elements[l.carbohydrateIndex].unit;
                location.element = elements[l.carbohydrateIndex].anomericCarbon;
            }
            else {
                location.unit = l.elementUnit;
                location.element = l.elementUnit.elements[l.elementIndex];
            }
            return theme.size.size(location) * linkSizeFactor;
        }
    };
    return link_1.createLinkCylinderMesh(ctx, builderProps, props, mesh);
}
exports.CarbohydrateTerminalLinkParams = __assign({}, units_visual_1.UnitsMeshParams, link_1.LinkCylinderParams, { linkSizeFactor: param_definition_1.ParamDefinition.Numeric(0.3, { min: 0, max: 3, step: 0.01 }) });
function CarbohydrateTerminalLinkVisual(materialId) {
    return complex_visual_1.ComplexMeshVisual({
        defaultProps: param_definition_1.ParamDefinition.getDefaultValues(exports.CarbohydrateTerminalLinkParams),
        createGeometry: createCarbohydrateTerminalLinkCylinderMesh,
        createLocationIterator: CarbohydrateTerminalLinkIterator,
        getLoci: getTerminalLinkLoci,
        eachLocation: eachTerminalLink,
        setUpdateState: function (state, newProps, currentProps) {
            state.createGeometry = (newProps.linkSizeFactor !== currentProps.linkSizeFactor ||
                newProps.radialSegments !== currentProps.radialSegments);
        }
    }, materialId);
}
exports.CarbohydrateTerminalLinkVisual = CarbohydrateTerminalLinkVisual;
function CarbohydrateTerminalLinkIterator(structure) {
    var _a = structure.carbohydrates, elements = _a.elements, terminalLinks = _a.terminalLinks;
    var groupCount = terminalLinks.length;
    var instanceCount = 1;
    var location = structure_1.Link.Location();
    var getLocation = function (groupIndex) {
        var terminalLink = terminalLinks[groupIndex];
        var carb = elements[terminalLink.carbohydrateIndex];
        var indexCarb = int_1.OrderedSet.indexOf(carb.unit.elements, carb.anomericCarbon);
        if (terminalLink.fromCarbohydrate) {
            location.aUnit = carb.unit;
            location.aIndex = indexCarb;
            location.bUnit = terminalLink.elementUnit;
            location.bIndex = terminalLink.elementIndex;
        }
        else {
            location.aUnit = terminalLink.elementUnit;
            location.aIndex = terminalLink.elementIndex;
            location.bUnit = carb.unit;
            location.bIndex = indexCarb;
        }
        return location;
    };
    return location_iterator_1.LocationIterator(groupCount, instanceCount, getLocation, true);
}
function getTerminalLinkLoci(pickingId, structure, id) {
    var objectId = pickingId.objectId, groupId = pickingId.groupId;
    if (id === objectId) {
        var _a = structure.carbohydrates, terminalLinks = _a.terminalLinks, elements = _a.elements;
        var l = terminalLinks[groupId];
        var carb = elements[l.carbohydrateIndex];
        var carbIndex = int_1.OrderedSet.indexOf(carb.unit.elements, carb.anomericCarbon);
        return structure_1.Link.Loci(structure, [
            structure_1.Link.Location(carb.unit, carbIndex, l.elementUnit, l.elementIndex),
            structure_1.Link.Location(l.elementUnit, l.elementIndex, carb.unit, carbIndex)
        ]);
    }
    return loci_1.EmptyLoci;
}
// TODO for each link when both of the link elements are in a StructureElement.Loci
function eachTerminalLink(loci, structure, apply) {
    var getTerminalLinkIndex = structure.carbohydrates.getTerminalLinkIndex;
    var changed = false;
    if (structure_1.Link.isLoci(loci)) {
        if (!structure_1.Structure.areParentsEquivalent(loci.structure, structure))
            return false;
        for (var _i = 0, _a = loci.links; _i < _a.length; _i++) {
            var l = _a[_i];
            var idx = getTerminalLinkIndex(l.aUnit, l.aUnit.elements[l.aIndex], l.bUnit, l.bUnit.elements[l.bIndex]);
            if (idx !== undefined) {
                if (apply(int_1.Interval.ofSingleton(idx)))
                    changed = true;
            }
        }
    }
    else if (structure_1.StructureElement.isLoci(loci)) {
        if (!structure_1.Structure.areParentsEquivalent(loci.structure, structure))
            return false;
        // TODO mark link only when both of the link elements are in a StructureElement.Loci
        var _b = structure.carbohydrates, getElementIndex_1 = _b.getElementIndex, getTerminalLinkIndices_1 = _b.getTerminalLinkIndices, elements_1 = _b.elements;
        var _loop_1 = function (e) {
            int_1.OrderedSet.forEach(e.indices, function (v) {
                var carbI = getElementIndex_1(e.unit, e.unit.elements[v]);
                if (carbI !== undefined) {
                    var carb = elements_1[carbI];
                    var indices = getTerminalLinkIndices_1(carb.unit, carb.anomericCarbon);
                    for (var i = 0, il = indices.length; i < il; ++i) {
                        if (apply(int_1.Interval.ofSingleton(indices[i])))
                            changed = true;
                    }
                }
            });
        };
        for (var _c = 0, _d = loci.elements; _c < _d.length; _c++) {
            var e = _d[_c];
            _loop_1(e);
        }
    }
    return changed;
}
