"use strict";
/**
 * Copyright (c) 2018-2019 mol* contributors, licensed under MIT, See LICENSE file for more info.
 *
 * @author Alexander Rose <alexander.rose@weirdbyte.de>
 */
var __assign = (this && this.__assign) || function () {
    __assign = Object.assign || function(t) {
        for (var s, i = 1, n = arguments.length; i < n; i++) {
            s = arguments[i];
            for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p))
                t[p] = s[p];
        }
        return t;
    };
    return __assign.apply(this, arguments);
};
Object.defineProperty(exports, "__esModule", { value: true });
var param_definition_1 = require("../../../mol-util/param-definition");
var structure_1 = require("../../../mol-model/structure");
var mesh_1 = require("../../../mol-geo/geometry/mesh/mesh");
var mesh_builder_1 = require("../../../mol-geo/geometry/mesh/mesh-builder");
var polymer_1 = require("./util/polymer");
var types_1 = require("../../../mol-model/structure/model/types");
var sheet_1 = require("../../../mol-geo/geometry/mesh/builder/sheet");
var tube_1 = require("../../../mol-geo/geometry/mesh/builder/tube");
var units_visual_1 = require("../units-visual");
var secondary_structure_1 = require("../../../mol-model-props/computed/secondary-structure");
exports.PolymerTraceMeshParams = {
    sizeFactor: param_definition_1.ParamDefinition.Numeric(0.2, { min: 0, max: 10, step: 0.01 }),
    linearSegments: param_definition_1.ParamDefinition.Numeric(8, { min: 1, max: 48, step: 1 }),
    radialSegments: param_definition_1.ParamDefinition.Numeric(16, { min: 3, max: 56, step: 1 }),
    aspectRatio: param_definition_1.ParamDefinition.Numeric(5, { min: 0.1, max: 10, step: 0.1 }),
    arrowFactor: param_definition_1.ParamDefinition.Numeric(1.5, { min: 0.1, max: 5, step: 0.1 }),
};
exports.DefaultPolymerTraceMeshProps = param_definition_1.ParamDefinition.getDefaultValues(exports.PolymerTraceMeshParams);
// TODO handle polymer ends properly
function createPolymerTraceMesh(ctx, unit, structure, theme, props, mesh) {
    var _a, _b, _c;
    var polymerElementCount = unit.polymerElements.length;
    if (!polymerElementCount)
        return mesh_1.Mesh.createEmpty(mesh);
    var sizeFactor = props.sizeFactor, linearSegments = props.linearSegments, radialSegments = props.radialSegments, aspectRatio = props.aspectRatio, arrowFactor = props.arrowFactor;
    var vertexCount = linearSegments * radialSegments * polymerElementCount + (radialSegments + 1) * polymerElementCount * 2;
    var builderState = mesh_builder_1.MeshBuilder.createState(vertexCount, vertexCount / 10, mesh);
    var isCoarse = structure_1.Unit.isCoarse(unit);
    var state = polymer_1.createCurveSegmentState(linearSegments);
    var curvePoints = state.curvePoints, normalVectors = state.normalVectors, binormalVectors = state.binormalVectors, widthValues = state.widthValues, heightValues = state.heightValues;
    var i = 0;
    var polymerTraceIt = polymer_1.PolymerTraceIterator(unit, structure);
    while (polymerTraceIt.hasNext) {
        var v = polymerTraceIt.move();
        builderState.currentGroup = i;
        var isNucleicType = types_1.isNucleic(v.moleculeType);
        var isSheet = types_1.SecondaryStructureType.is(v.secStrucType, 4 /* Beta */);
        var isHelix = types_1.SecondaryStructureType.is(v.secStrucType, 2 /* Helix */);
        var tension = isHelix ? 0.9 : 0.5;
        var shift = isNucleicType ? 0.3 : 0.5;
        polymer_1.interpolateCurveSegment(state, v, tension, shift);
        var w0 = theme.size.size(v.centerPrev) * sizeFactor;
        var w1 = theme.size.size(v.center) * sizeFactor;
        var w2 = theme.size.size(v.centerNext) * sizeFactor;
        if (isCoarse) {
            w0 *= aspectRatio / 2;
            w1 *= aspectRatio / 2;
            w2 *= aspectRatio / 2;
        }
        if (isSheet) {
            var h1 = w1 * aspectRatio;
            var arrowHeight = v.secStrucLast ? h1 * arrowFactor : 0;
            sheet_1.addSheet(builderState, curvePoints, normalVectors, binormalVectors, linearSegments, w1, h1, arrowHeight, v.secStrucFirst, v.secStrucLast);
        }
        else {
            var h0 = void 0, h1 = void 0, h2 = void 0;
            if (isHelix && !v.isCoarseBackbone) {
                h0 = w0 * aspectRatio;
                h1 = w1 * aspectRatio;
                h2 = w2 * aspectRatio;
            }
            else if (isNucleicType && !v.isCoarseBackbone) {
                h0 = w0 * aspectRatio;
                _a = [h0, w0], w0 = _a[0], h0 = _a[1];
                h1 = w1 * aspectRatio;
                _b = [h1, w1], w1 = _b[0], h1 = _b[1];
                h2 = w2 * aspectRatio;
                _c = [h2, w2], w2 = _c[0], h2 = _c[1];
            }
            else {
                h0 = w0;
                h1 = w1;
                h2 = w2;
            }
            polymer_1.interpolateSizes(state, w0, w1, w2, h0, h1, h2, shift);
            tube_1.addTube(builderState, curvePoints, normalVectors, binormalVectors, linearSegments, radialSegments, widthValues, heightValues, 1, v.secStrucFirst || v.coarseBackboneFirst, v.secStrucLast || v.coarseBackboneLast);
        }
        ++i;
    }
    return mesh_builder_1.MeshBuilder.getMesh(builderState);
}
exports.PolymerTraceParams = __assign({}, units_visual_1.UnitsMeshParams, exports.PolymerTraceMeshParams);
function PolymerTraceVisual(materialId) {
    return units_visual_1.UnitsMeshVisual({
        defaultProps: param_definition_1.ParamDefinition.getDefaultValues(exports.PolymerTraceParams),
        createGeometry: createPolymerTraceMesh,
        createLocationIterator: polymer_1.PolymerLocationIterator.fromGroup,
        getLoci: polymer_1.getPolymerElementLoci,
        eachLocation: polymer_1.eachPolymerElement,
        setUpdateState: function (state, newProps, currentProps, newTheme, currentTheme, newStructureGroup, currentStructureGroup) {
            state.createGeometry = (newProps.sizeFactor !== currentProps.sizeFactor ||
                newProps.linearSegments !== currentProps.linearSegments ||
                newProps.radialSegments !== currentProps.radialSegments ||
                newProps.aspectRatio !== currentProps.aspectRatio ||
                newProps.arrowFactor !== currentProps.arrowFactor);
            var computedSecondaryStructure = secondary_structure_1.ComputedSecondaryStructure.get(newStructureGroup.structure);
            if (state.info.computedSecondaryStructure !== computedSecondaryStructure) {
                state.createGeometry = true;
                state.info.computedSecondaryStructure = computedSecondaryStructure;
            }
        }
    }, materialId);
}
exports.PolymerTraceVisual = PolymerTraceVisual;
