"use strict";
/**
 * Copyright (c) 2018-2019 mol* contributors, licensed under MIT, See LICENSE file for more info.
 *
 * @author Alexander Rose <alexander.rose@weirdbyte.de>
 */
Object.defineProperty(exports, "__esModule", { value: true });
var structure_1 = require("../../../../mol-model/structure");
var linear_algebra_1 = require("../../../../mol-math/linear-algebra");
var transform_data_1 = require("../../../../mol-geo/geometry/transform-data");
var int_1 = require("../../../../mol-data/int");
var loci_1 = require("../../../../mol-model/loci");
var physical_1 = require("../../../../mol-theme/size/physical");
/** Return a Loci for the elements of a whole residue the elementIndex belongs to. */
function getResidueLoci(structure, unit, elementIndex) {
    var elements = unit.elements, model = unit.model;
    if (int_1.OrderedSet.indexOf(elements, elementIndex) !== -1) {
        var _a = model.atomicHierarchy.residueAtomSegments, index = _a.index, offsets = _a.offsets;
        var rI = index[elementIndex];
        var _indices = [];
        for (var i = offsets[rI], il = offsets[rI + 1]; i < il; ++i) {
            var unitIndex = int_1.OrderedSet.indexOf(elements, i);
            if (unitIndex !== -1)
                _indices.push(unitIndex);
        }
        var indices = int_1.OrderedSet.ofSortedArray(int_1.SortedArray.ofSortedArray(_indices));
        return structure_1.StructureElement.Loci(structure, [{ unit: unit, indices: indices }]);
    }
    return loci_1.EmptyLoci;
}
exports.getResidueLoci = getResidueLoci;
/**
 * Return a Loci for the elements of a whole residue the elementIndex belongs to but
 * restrict to elements that have the same label_alt_id or none
 */
function getAltResidueLoci(structure, unit, elementIndex) {
    var elements = unit.elements, model = unit.model;
    var label_alt_id = model.atomicHierarchy.atoms.label_alt_id;
    var elementAltId = label_alt_id.value(elementIndex);
    if (int_1.OrderedSet.indexOf(elements, elementIndex) !== -1) {
        var _a = model.atomicHierarchy.residueAtomSegments, index = _a.index, offsets = _a.offsets;
        var rI = index[elementIndex];
        var _indices = [];
        for (var i = offsets[rI], il = offsets[rI + 1]; i < il; ++i) {
            var unitIndex = int_1.OrderedSet.indexOf(elements, i);
            if (unitIndex !== -1) {
                var altId = label_alt_id.value(i);
                if (elementAltId === altId || altId === '') {
                    _indices.push(unitIndex);
                }
            }
        }
        var indices = int_1.OrderedSet.ofSortedArray(int_1.SortedArray.ofSortedArray(_indices));
        return structure_1.StructureElement.Loci(structure, [{ unit: unit, indices: indices }]);
    }
    return loci_1.EmptyLoci;
}
exports.getAltResidueLoci = getAltResidueLoci;
//
function createUnitsTransform(_a, transformData) {
    var units = _a.units;
    var unitCount = units.length;
    var n = unitCount * 16;
    var array = transformData && transformData.aTransform.ref.value.length >= n ? transformData.aTransform.ref.value : new Float32Array(n);
    for (var i = 0; i < unitCount; i++) {
        linear_algebra_1.Mat4.toArray(units[i].conformation.operator.matrix, array, i * 16);
    }
    return transform_data_1.createTransform(array, unitCount, transformData);
}
exports.createUnitsTransform = createUnitsTransform;
exports.UnitKindInfo = {
    'atomic': {},
    'spheres': {},
    'gaussians': {},
};
exports.UnitKindNames = Object.keys(exports.UnitKindInfo);
exports.UnitKindOptions = exports.UnitKindNames.map(function (n) { return [n, n]; });
function includesUnitKind(unitKinds, unit) {
    for (var i = 0, il = unitKinds.length; i < il; ++i) {
        if (structure_1.Unit.isAtomic(unit) && unitKinds[i] === 'atomic')
            return true;
        if (structure_1.Unit.isSpheres(unit) && unitKinds[i] === 'spheres')
            return true;
        if (structure_1.Unit.isGaussians(unit) && unitKinds[i] === 'gaussians')
            return true;
    }
    return false;
}
exports.includesUnitKind = includesUnitKind;
//
function getConformation(unit) {
    switch (unit.kind) {
        case 0 /* Atomic */: return unit.model.atomicConformation;
        case 1 /* Spheres */: return unit.model.coarseConformation.spheres;
        case 2 /* Gaussians */: return unit.model.coarseConformation.gaussians;
    }
}
exports.getConformation = getConformation;
function getUnitConformationAndRadius(unit) {
    var conformation = getConformation(unit);
    var elements = unit.elements;
    var position = {
        indices: elements,
        x: conformation.x,
        y: conformation.y,
        z: conformation.z
    };
    var l = structure_1.StructureElement.create(unit);
    var sizeTheme = physical_1.PhysicalSizeTheme({}, {});
    var radius = function (index) {
        l.element = index;
        return sizeTheme.size(l);
    };
    return { position: position, radius: radius };
}
exports.getUnitConformationAndRadius = getUnitConformationAndRadius;
function getStructureConformationAndRadius(structure) {
    var n = structure.elementCount;
    var xs = new Float32Array(n);
    var ys = new Float32Array(n);
    var zs = new Float32Array(n);
    var rs = new Float32Array(n);
    var l = structure_1.StructureElement.create();
    var sizeTheme = physical_1.PhysicalSizeTheme({}, {});
    var m = 0;
    for (var i = 0, il = structure.units.length; i < il; ++i) {
        var unit = structure.units[i];
        var elements = unit.elements;
        var _a = unit.conformation, x = _a.x, y = _a.y, z = _a.z;
        l.unit = unit;
        for (var j = 0, jl = elements.length; j < jl; ++j) {
            var eI = elements[j];
            xs[m + j] = x(eI);
            ys[m + j] = y(eI);
            zs[m + j] = z(eI);
            l.element = eI;
            rs[m + j] = sizeTheme.size(l);
        }
        m += elements.length;
    }
    var position = { indices: int_1.OrderedSet.ofRange(0, n), x: xs, y: ys, z: zs };
    var radius = function (index) { return rs[index]; };
    return { position: position, radius: radius };
}
exports.getStructureConformationAndRadius = getStructureConformationAndRadius;
