"use strict";
/**
 * Copyright (c) 2018 mol* contributors, licensed under MIT, See LICENSE file for more info.
 *
 * @author Alexander Rose <alexander.rose@weirdbyte.de>
 */
Object.defineProperty(exports, "__esModule", { value: true });
var linear_algebra_1 = require("../../../../mol-math/linear-algebra");
var types_1 = require("../../../../mol-model/structure/model/types");
var structure_1 = require("../../../../mol-model/structure");
var param_definition_1 = require("../../../../mol-util/param-definition");
var mesh_1 = require("../../../../mol-geo/geometry/mesh/mesh");
var mesh_builder_1 = require("../../../../mol-geo/geometry/mesh/mesh-builder");
var cylinder_1 = require("../../../../mol-geo/geometry/mesh/builder/cylinder");
var location_iterator_1 = require("../../../../mol-geo/util/location-iterator");
exports.LinkCylinderParams = {
    linkScale: param_definition_1.ParamDefinition.Numeric(0.4, { min: 0, max: 1, step: 0.1 }),
    linkSpacing: param_definition_1.ParamDefinition.Numeric(1, { min: 0, max: 2, step: 0.01 }),
    radialSegments: param_definition_1.ParamDefinition.Numeric(16, { min: 3, max: 56, step: 1 }),
};
exports.DefaultLinkCylinderProps = param_definition_1.ParamDefinition.getDefaultValues(exports.LinkCylinderParams);
var tmpShiftV12 = linear_algebra_1.Vec3.zero();
var tmpShiftV13 = linear_algebra_1.Vec3.zero();
/** Calculate 'shift' direction that is perpendiculat to v1 - v2 and goes through v3 */
function calculateShiftDir(out, v1, v2, v3) {
    linear_algebra_1.Vec3.normalize(tmpShiftV12, linear_algebra_1.Vec3.sub(tmpShiftV12, v1, v2));
    if (v3 !== null) {
        linear_algebra_1.Vec3.sub(tmpShiftV13, v1, v3);
    }
    else {
        linear_algebra_1.Vec3.copy(tmpShiftV13, v1); // no reference point, use v1
    }
    linear_algebra_1.Vec3.normalize(tmpShiftV13, tmpShiftV13);
    // ensure v13 and v12 are not colinear
    var dp = linear_algebra_1.Vec3.dot(tmpShiftV12, tmpShiftV13);
    if (1 - Math.abs(dp) < 1e-5) {
        linear_algebra_1.Vec3.set(tmpShiftV13, 1, 0, 0);
        dp = linear_algebra_1.Vec3.dot(tmpShiftV12, tmpShiftV13);
        if (1 - Math.abs(dp) < 1e-5) {
            linear_algebra_1.Vec3.set(tmpShiftV13, 0, 1, 0);
            dp = linear_algebra_1.Vec3.dot(tmpShiftV12, tmpShiftV13);
        }
    }
    linear_algebra_1.Vec3.setMagnitude(tmpShiftV12, tmpShiftV12, dp);
    linear_algebra_1.Vec3.sub(tmpShiftV13, tmpShiftV13, tmpShiftV12);
    return linear_algebra_1.Vec3.normalize(out, tmpShiftV13);
}
exports.calculateShiftDir = calculateShiftDir;
/**
 * Each edge is included twice to allow for coloring/picking
 * the half closer to the first vertex, i.e. vertex a.
 */
function createLinkCylinderMesh(ctx, linkBuilder, props, mesh) {
    var linkCount = linkBuilder.linkCount, referencePosition = linkBuilder.referencePosition, position = linkBuilder.position, order = linkBuilder.order, flags = linkBuilder.flags, radius = linkBuilder.radius;
    if (!linkCount)
        return mesh_1.Mesh.createEmpty(mesh);
    var linkScale = props.linkScale, linkSpacing = props.linkSpacing, radialSegments = props.radialSegments;
    var vertexCountEstimate = radialSegments * 2 * linkCount * 2;
    var builderState = mesh_builder_1.MeshBuilder.createState(vertexCountEstimate, vertexCountEstimate / 4, mesh);
    var va = linear_algebra_1.Vec3.zero();
    var vb = linear_algebra_1.Vec3.zero();
    var vShift = linear_algebra_1.Vec3.zero();
    var cylinderProps = {
        radiusTop: 1,
        radiusBottom: 1,
        radialSegments: radialSegments,
        topCap: false,
        bottomCap: false
    };
    for (var edgeIndex = 0, _eI = linkCount; edgeIndex < _eI; ++edgeIndex) {
        position(va, vb, edgeIndex);
        var linkRadius = radius(edgeIndex);
        var o = order(edgeIndex);
        var f = flags(edgeIndex);
        builderState.currentGroup = edgeIndex;
        if (types_1.LinkType.is(f, 2 /* MetallicCoordination */) || types_1.LinkType.is(f, 4 /* Hydrogen */)) {
            // show metall coordinations and hydrogen bonds with dashed cylinders
            cylinderProps.radiusTop = cylinderProps.radiusBottom = linkRadius / 3;
            cylinderProps.topCap = cylinderProps.bottomCap = true;
            cylinder_1.addFixedCountDashedCylinder(builderState, va, vb, 0.5, 7, cylinderProps);
        }
        else if (o === 2 || o === 3) {
            // show bonds with order 2 or 3 using 2 or 3 parallel cylinders
            var multiRadius = linkRadius * (linkScale / (0.5 * o));
            var absOffset = (linkRadius - multiRadius) * linkSpacing;
            calculateShiftDir(vShift, va, vb, referencePosition(edgeIndex));
            linear_algebra_1.Vec3.setMagnitude(vShift, vShift, absOffset);
            cylinderProps.radiusTop = cylinderProps.radiusBottom = multiRadius;
            cylinderProps.topCap = cylinderProps.bottomCap = false;
            if (o === 3)
                cylinder_1.addCylinder(builderState, va, vb, 0.5, cylinderProps);
            cylinder_1.addDoubleCylinder(builderState, va, vb, 0.5, vShift, cylinderProps);
        }
        else {
            cylinderProps.radiusTop = cylinderProps.radiusBottom = linkRadius;
            cylinderProps.topCap = cylinderProps.bottomCap = false;
            cylinder_1.addCylinder(builderState, va, vb, 0.5, cylinderProps);
        }
    }
    return mesh_builder_1.MeshBuilder.getMesh(builderState);
}
exports.createLinkCylinderMesh = createLinkCylinderMesh;
var LinkIterator;
(function (LinkIterator) {
    function fromGroup(group) {
        var unit = group.units[0];
        var groupCount = structure_1.Unit.isAtomic(unit) ? unit.links.edgeCount * 2 : 0;
        var instanceCount = group.units.length;
        var location = structure_1.StructureElement.create();
        var getLocation = function (groupIndex, instanceIndex) {
            var unit = group.units[instanceIndex];
            location.unit = unit;
            location.element = unit.elements[unit.links.a[groupIndex]];
            return location;
        };
        return location_iterator_1.LocationIterator(groupCount, instanceCount, getLocation);
    }
    LinkIterator.fromGroup = fromGroup;
    function fromStructure(structure) {
        var groupCount = structure.links.bondCount;
        var instanceCount = 1;
        var location = structure_1.Link.Location();
        var getLocation = function (groupIndex) {
            var bond = structure.links.bonds[groupIndex];
            location.aUnit = bond.unitA;
            location.aIndex = bond.indexA;
            location.bUnit = bond.unitB;
            location.bIndex = bond.indexB;
            return location;
        };
        return location_iterator_1.LocationIterator(groupCount, instanceCount, getLocation, true);
    }
    LinkIterator.fromStructure = fromStructure;
})(LinkIterator = exports.LinkIterator || (exports.LinkIterator = {}));
