"use strict";
/**
 * Copyright (c) 2018 mol* contributors, licensed under MIT, See LICENSE file for more info.
 *
 * @author Alexander Rose <alexander.rose@weirdbyte.de>
 */
function __export(m) {
    for (var p in m) if (!exports.hasOwnProperty(p)) exports[p] = m[p];
}
Object.defineProperty(exports, "__esModule", { value: true });
var structure_1 = require("../../../../mol-model/structure");
var int_1 = require("../../../../mol-data/int");
var loci_1 = require("../../../../mol-model/loci");
var location_iterator_1 = require("../../../../mol-geo/util/location-iterator");
var common_1 = require("./common");
__export(require("./polymer/backbone-iterator"));
__export(require("./polymer/gap-iterator"));
__export(require("./polymer/trace-iterator"));
__export(require("./polymer/curve-segment"));
function getPolymerRanges(unit) {
    switch (unit.kind) {
        case 0 /* Atomic */: return unit.model.atomicHierarchy.polymerRanges;
        case 1 /* Spheres */: return unit.model.coarseHierarchy.spheres.polymerRanges;
        case 2 /* Gaussians */: return unit.model.coarseHierarchy.gaussians.polymerRanges;
    }
}
exports.getPolymerRanges = getPolymerRanges;
function getGapRanges(unit) {
    switch (unit.kind) {
        case 0 /* Atomic */: return unit.model.atomicHierarchy.gapRanges;
        case 1 /* Spheres */: return unit.model.coarseHierarchy.spheres.gapRanges;
        case 2 /* Gaussians */: return unit.model.coarseHierarchy.gaussians.gapRanges;
    }
}
exports.getGapRanges = getGapRanges;
var PolymerLocationIterator;
(function (PolymerLocationIterator) {
    function fromGroup(group) {
        var polymerElements = group.units[0].polymerElements;
        var groupCount = polymerElements.length;
        var instanceCount = group.units.length;
        var location = structure_1.StructureElement.create();
        var getLocation = function (groupIndex, instanceIndex) {
            var unit = group.units[instanceIndex];
            location.unit = unit;
            location.element = polymerElements[groupIndex];
            return location;
        };
        return location_iterator_1.LocationIterator(groupCount, instanceCount, getLocation);
    }
    PolymerLocationIterator.fromGroup = fromGroup;
})(PolymerLocationIterator = exports.PolymerLocationIterator || (exports.PolymerLocationIterator = {}));
var PolymerGapLocationIterator;
(function (PolymerGapLocationIterator) {
    function fromGroup(group) {
        var gapElements = group.units[0].gapElements;
        var groupCount = gapElements.length;
        var instanceCount = group.units.length;
        var location = structure_1.StructureElement.create();
        var getLocation = function (groupIndex, instanceIndex) {
            var unit = group.units[instanceIndex];
            location.unit = unit;
            location.element = gapElements[groupIndex];
            return location;
        };
        return location_iterator_1.LocationIterator(groupCount, instanceCount, getLocation);
    }
    PolymerGapLocationIterator.fromGroup = fromGroup;
})(PolymerGapLocationIterator = exports.PolymerGapLocationIterator || (exports.PolymerGapLocationIterator = {}));
/** Return a Loci for the elements of the whole residue of a polymer element. */
function getPolymerElementLoci(pickingId, structureGroup, id) {
    var objectId = pickingId.objectId, instanceId = pickingId.instanceId, groupId = pickingId.groupId;
    if (id === objectId) {
        var structure = structureGroup.structure, group = structureGroup.group;
        var unit = group.units[instanceId];
        if (structure_1.Unit.isAtomic(unit)) {
            return common_1.getResidueLoci(structure, unit, unit.polymerElements[groupId]);
        }
        else {
            var elements = unit.elements;
            var elementIndex = unit.polymerElements[groupId];
            var unitIndex = int_1.OrderedSet.indexOf(elements, elementIndex);
            if (unitIndex !== -1) {
                var indices = int_1.OrderedSet.ofSingleton(unitIndex);
                return structure_1.StructureElement.Loci(structure, [{ unit: unit, indices: indices }]);
            }
        }
    }
    return loci_1.EmptyLoci;
}
exports.getPolymerElementLoci = getPolymerElementLoci;
/**
 * Mark a polymer element (e.g. part of a cartoon trace)
 * - for atomic units mark only when all its residue's elements are in a loci
 */
function eachPolymerElement(loci, structureGroup, apply) {
    var changed = false;
    if (!structure_1.StructureElement.isLoci(loci))
        return false;
    var structure = structureGroup.structure, group = structureGroup.group;
    if (!structure_1.Structure.areParentsEquivalent(loci.structure, structure))
        return false;
    var _a = group.units[0], polymerElements = _a.polymerElements, model = _a.model, elements = _a.elements;
    var _b = model.atomicHierarchy.residueAtomSegments, index = _b.index, offsets = _b.offsets;
    var traceElementIndex = model.atomicHierarchy.derived.residue.traceElementIndex;
    var groupCount = polymerElements.length;
    var _loop_1 = function (e) {
        var unitIdx = group.unitIndexMap.get(e.unit.id);
        if (unitIdx !== undefined) {
            if (structure_1.Unit.isAtomic(e.unit)) {
                // TODO optimized implementation for intervals
                int_1.OrderedSet.forEach(e.indices, function (v) {
                    var rI = index[elements[v]];
                    var unitIndexMin = int_1.OrderedSet.findPredecessorIndex(elements, offsets[rI]);
                    var unitIndexMax = int_1.OrderedSet.findPredecessorIndex(elements, offsets[rI + 1] - 1);
                    var unitIndexInterval = int_1.Interval.ofRange(unitIndexMin, unitIndexMax);
                    if (!int_1.OrderedSet.isSubset(e.indices, unitIndexInterval))
                        return;
                    var eI = traceElementIndex[rI];
                    var idx = int_1.OrderedSet.indexOf(e.unit.polymerElements, eI);
                    if (idx !== -1) {
                        if (apply(int_1.Interval.ofSingleton(unitIdx * groupCount + idx)))
                            changed = true;
                    }
                });
            }
            else {
                if (int_1.Interval.is(e.indices)) {
                    var start = unitIdx * groupCount + int_1.Interval.start(e.indices);
                    var end = unitIdx * groupCount + int_1.Interval.end(e.indices);
                    if (apply(int_1.Interval.ofBounds(start, end)))
                        changed = true;
                }
                else {
                    for (var i = 0, _i = e.indices.length; i < _i; i++) {
                        var idx = unitIdx * groupCount + e.indices[i];
                        if (apply(int_1.Interval.ofSingleton(idx)))
                            changed = true;
                    }
                }
            }
        }
    };
    for (var _c = 0, _d = loci.elements; _c < _d.length; _c++) {
        var e = _d[_c];
        _loop_1(e);
    }
    return changed;
}
exports.eachPolymerElement = eachPolymerElement;
/** Return a Loci for both directions of the polymer gap element. */
function getPolymerGapElementLoci(pickingId, structureGroup, id) {
    var objectId = pickingId.objectId, instanceId = pickingId.instanceId, groupId = pickingId.groupId;
    if (id === objectId) {
        var structure = structureGroup.structure, group = structureGroup.group;
        var unit = group.units[instanceId];
        var unitIndexA = int_1.OrderedSet.indexOf(unit.elements, unit.gapElements[groupId]);
        var unitIndexB = int_1.OrderedSet.indexOf(unit.elements, unit.gapElements[groupId % 2 ? groupId - 1 : groupId + 1]);
        if (unitIndexA !== -1 && unitIndexB !== -1) {
            return structure_1.Link.Loci(structure, [
                structure_1.Link.Location(unit, unitIndexA, unit, unitIndexB),
                structure_1.Link.Location(unit, unitIndexB, unit, unitIndexA)
            ]);
        }
    }
    return loci_1.EmptyLoci;
}
exports.getPolymerGapElementLoci = getPolymerGapElementLoci;
function eachPolymerGapElement(loci, structureGroup, apply) {
    var changed = false;
    if (structure_1.Link.isLoci(loci)) {
        var structure = structureGroup.structure, group = structureGroup.group;
        if (!structure_1.Structure.areParentsEquivalent(loci.structure, structure))
            return false;
        var groupCount = group.units[0].gapElements.length;
        for (var _a = 0, _b = loci.links; _a < _b.length; _a++) {
            var b = _b[_a];
            var unitIdx = group.unitIndexMap.get(b.aUnit.id);
            if (unitIdx !== undefined) {
                var idxA = int_1.OrderedSet.indexOf(b.aUnit.gapElements, b.aUnit.elements[b.aIndex]);
                var idxB = int_1.OrderedSet.indexOf(b.bUnit.gapElements, b.bUnit.elements[b.bIndex]);
                if (idxA !== -1 && idxB !== -1) {
                    if (apply(int_1.Interval.ofSingleton(unitIdx * groupCount + idxA)))
                        changed = true;
                }
            }
        }
    }
    else if (structure_1.StructureElement.isLoci(loci)) {
        var structure = structureGroup.structure, group = structureGroup.group;
        if (!structure_1.Structure.areParentsEquivalent(loci.structure, structure))
            return false;
        var groupCount_1 = group.units[0].gapElements.length;
        var _loop_2 = function (e) {
            var unitIdx = group.unitIndexMap.get(e.unit.id);
            if (unitIdx !== undefined) {
                int_1.OrderedSet.forEach(e.indices, function (v) {
                    var idx = int_1.OrderedSet.indexOf(e.unit.gapElements, e.unit.elements[v]);
                    if (idx !== -1) {
                        if (apply(int_1.Interval.ofSingleton(unitIdx * groupCount_1 + idx)))
                            changed = true;
                    }
                });
            }
        };
        for (var _c = 0, _d = loci.elements; _c < _d.length; _c++) {
            var e = _d[_c];
            _loop_2(e);
        }
    }
    return changed;
}
exports.eachPolymerGapElement = eachPolymerGapElement;
