"use strict";
/**
 * Copyright (c) 2018-2019 mol* contributors, licensed under MIT, See LICENSE file for more info.
 *
 * @author Alexander Rose <alexander.rose@weirdbyte.de>
 */
Object.defineProperty(exports, "__esModule", { value: true });
var linear_algebra_1 = require("../../../../../mol-math/linear-algebra");
var interpolate_1 = require("../../../../../mol-math/interpolate");
function createCurveSegmentState(linearSegments) {
    var n = linearSegments + 1;
    var pn = n * 3;
    return {
        curvePoints: new Float32Array(pn),
        tangentVectors: new Float32Array(pn),
        normalVectors: new Float32Array(pn),
        binormalVectors: new Float32Array(pn),
        widthValues: new Float32Array(n),
        heightValues: new Float32Array(n),
        linearSegments: linearSegments
    };
}
exports.createCurveSegmentState = createCurveSegmentState;
function interpolateCurveSegment(state, controls, tension, shift) {
    interpolatePointsAndTangents(state, controls, tension, shift);
    interpolateNormals(state, controls);
}
exports.interpolateCurveSegment = interpolateCurveSegment;
var tanA = linear_algebra_1.Vec3.zero();
var tanB = linear_algebra_1.Vec3.zero();
var curvePoint = linear_algebra_1.Vec3.zero();
function interpolatePointsAndTangents(state, controls, tension, shift) {
    var curvePoints = state.curvePoints, tangentVectors = state.tangentVectors, linearSegments = state.linearSegments;
    var p0 = controls.p0, p1 = controls.p1, p2 = controls.p2, p3 = controls.p3, p4 = controls.p4, secStrucFirst = controls.secStrucFirst, secStrucLast = controls.secStrucLast;
    var shift1 = 1 - shift;
    var tensionBeg = secStrucFirst ? 0.5 : tension;
    var tensionEnd = secStrucLast ? 0.5 : tension;
    for (var j = 0; j <= linearSegments; ++j) {
        var t = j * 1.0 / linearSegments;
        if (t < shift1) {
            var te = interpolate_1.lerp(tensionBeg, tension, t);
            linear_algebra_1.Vec3.spline(curvePoint, p0, p1, p2, p3, t + shift, te);
            linear_algebra_1.Vec3.spline(tanA, p0, p1, p2, p3, t + shift + 0.01, tensionBeg);
            linear_algebra_1.Vec3.spline(tanB, p0, p1, p2, p3, t + shift - 0.01, tensionBeg);
        }
        else {
            var te = interpolate_1.lerp(tension, tensionEnd, t);
            linear_algebra_1.Vec3.spline(curvePoint, p1, p2, p3, p4, t - shift1, te);
            linear_algebra_1.Vec3.spline(tanA, p1, p2, p3, p4, t - shift1 + 0.01, te);
            linear_algebra_1.Vec3.spline(tanB, p1, p2, p3, p4, t - shift1 - 0.01, te);
        }
        linear_algebra_1.Vec3.toArray(curvePoint, curvePoints, j * 3);
        linear_algebra_1.Vec3.normalize(tangentVec, linear_algebra_1.Vec3.sub(tangentVec, tanA, tanB));
        linear_algebra_1.Vec3.toArray(tangentVec, tangentVectors, j * 3);
    }
}
exports.interpolatePointsAndTangents = interpolatePointsAndTangents;
var tmpNormal = linear_algebra_1.Vec3();
var tangentVec = linear_algebra_1.Vec3();
var normalVec = linear_algebra_1.Vec3();
var binormalVec = linear_algebra_1.Vec3();
var prevNormal = linear_algebra_1.Vec3();
var firstTangentVec = linear_algebra_1.Vec3();
var lastTangentVec = linear_algebra_1.Vec3();
var firstNormalVec = linear_algebra_1.Vec3();
var lastNormalVec = linear_algebra_1.Vec3();
/**
 * Populate normalVectors by interpolating from firstDirection to lastDirection with
 * resulting vector perpendicular to tangentVectors and binormalVectors
 */
function interpolateNormals(state, controls) {
    var curvePoints = state.curvePoints, tangentVectors = state.tangentVectors, normalVectors = state.normalVectors, binormalVectors = state.binormalVectors;
    var firstDirection = controls.d12, lastDirection = controls.d23;
    var n = curvePoints.length / 3;
    linear_algebra_1.Vec3.fromArray(firstTangentVec, tangentVectors, 0);
    linear_algebra_1.Vec3.fromArray(lastTangentVec, tangentVectors, (n - 1) * 3);
    linear_algebra_1.Vec3.orthogonalize(firstNormalVec, firstTangentVec, firstDirection);
    linear_algebra_1.Vec3.orthogonalize(lastNormalVec, lastTangentVec, lastDirection);
    if (linear_algebra_1.Vec3.dot(firstNormalVec, lastNormalVec) < 0) {
        linear_algebra_1.Vec3.scale(lastNormalVec, lastNormalVec, -1);
    }
    linear_algebra_1.Vec3.copy(prevNormal, firstNormalVec);
    for (var i = 0; i < n; ++i) {
        var t = i === 0 ? 0 : 1 / (n - i);
        linear_algebra_1.Vec3.fromArray(tangentVec, tangentVectors, i * 3);
        linear_algebra_1.Vec3.orthogonalize(normalVec, tangentVec, linear_algebra_1.Vec3.slerp(tmpNormal, prevNormal, lastNormalVec, t));
        linear_algebra_1.Vec3.toArray(normalVec, normalVectors, i * 3);
        linear_algebra_1.Vec3.copy(prevNormal, normalVec);
        linear_algebra_1.Vec3.normalize(binormalVec, linear_algebra_1.Vec3.cross(binormalVec, tangentVec, normalVec));
        linear_algebra_1.Vec3.toArray(binormalVec, binormalVectors, i * 3);
    }
}
exports.interpolateNormals = interpolateNormals;
function interpolateSizes(state, w0, w1, w2, h0, h1, h2, shift) {
    var widthValues = state.widthValues, heightValues = state.heightValues, linearSegments = state.linearSegments;
    var shift1 = 1 - shift;
    for (var i = 0; i <= linearSegments; ++i) {
        var t = i * 1.0 / linearSegments;
        if (t < shift1) {
            widthValues[i] = interpolate_1.lerp(w0, w1, t + shift);
            heightValues[i] = interpolate_1.lerp(h0, h1, t + shift);
        }
        else {
            widthValues[i] = interpolate_1.lerp(w1, w2, t - shift1);
            heightValues[i] = interpolate_1.lerp(h1, h2, t - shift1);
        }
    }
}
exports.interpolateSizes = interpolateSizes;
