"use strict";
/**
 * Copyright (c) 2018-2019 mol* contributors, licensed under MIT, See LICENSE file for more info.
 *
 * @author Alexander Rose <alexander.rose@weirdbyte.de>
 */
Object.defineProperty(exports, "__esModule", { value: true });
var structure_1 = require("../../../../../mol-model/structure");
var int_1 = require("../../../../../mol-data/int");
var types_1 = require("../../../../../mol-model/structure/model/types");
var linear_algebra_1 = require("../../../../../mol-math/linear-algebra");
var sorted_ranges_1 = require("../../../../../mol-data/int/sorted-ranges");
var polymer_1 = require("../polymer");
var secondary_structure_1 = require("../../../../../mol-model-props/computed/secondary-structure");
/**
 * Iterates over individual residues/coarse elements in polymers of a unit while
 * providing information about the neighbourhood in the underlying model for drawing splines
 */
function PolymerTraceIterator(unit, structure) {
    switch (unit.kind) {
        case 0 /* Atomic */: return new AtomicPolymerTraceIterator(unit, structure);
        case 1 /* Spheres */:
        case 2 /* Gaussians */:
            return new CoarsePolymerTraceIterator(unit, structure);
    }
}
exports.PolymerTraceIterator = PolymerTraceIterator;
var SecStrucTypeNA = types_1.SecondaryStructureType.create(536870912 /* NA */);
function createPolymerTraceElement(unit) {
    return {
        center: structure_1.StructureElement.create(unit),
        centerPrev: structure_1.StructureElement.create(unit),
        centerNext: structure_1.StructureElement.create(unit),
        first: false, last: false,
        secStrucFirst: false, secStrucLast: false,
        secStrucType: SecStrucTypeNA,
        moleculeType: 0 /* unknown */,
        coarseBackboneFirst: false, coarseBackboneLast: false,
        isCoarseBackbone: false,
        p0: linear_algebra_1.Vec3(), p1: linear_algebra_1.Vec3(), p2: linear_algebra_1.Vec3(), p3: linear_algebra_1.Vec3(), p4: linear_algebra_1.Vec3(),
        d12: linear_algebra_1.Vec3(), d23: linear_algebra_1.Vec3()
    };
}
var tmpVecA = linear_algebra_1.Vec3();
var tmpVecB = linear_algebra_1.Vec3();
var AtomicPolymerTraceIterator = /** @class */ (function () {
    function AtomicPolymerTraceIterator(unit, structure) {
        this.unit = unit;
        this.state = 0 /* nextPolymer */;
        this.p0 = linear_algebra_1.Vec3();
        this.p1 = linear_algebra_1.Vec3();
        this.p2 = linear_algebra_1.Vec3();
        this.p3 = linear_algebra_1.Vec3();
        this.p4 = linear_algebra_1.Vec3();
        this.p5 = linear_algebra_1.Vec3();
        this.p6 = linear_algebra_1.Vec3();
        this.d01 = linear_algebra_1.Vec3();
        this.d12 = linear_algebra_1.Vec3();
        this.d23 = linear_algebra_1.Vec3();
        this.d34 = linear_algebra_1.Vec3();
        this.hasNext = false;
        this.atomicConformation = unit.model.atomicConformation;
        this.residueAtomSegments = unit.model.atomicHierarchy.residueAtomSegments;
        this.traceElementIndex = unit.model.atomicHierarchy.derived.residue.traceElementIndex; // can assume it won't be -1 for polymer residues
        this.directionFromElementIndex = unit.model.atomicHierarchy.derived.residue.directionFromElementIndex;
        this.directionToElementIndex = unit.model.atomicHierarchy.derived.residue.directionToElementIndex;
        this.moleculeType = unit.model.atomicHierarchy.derived.residue.moleculeType;
        this.cyclicPolymerMap = unit.model.atomicHierarchy.cyclicPolymerMap;
        this.polymerIt = sorted_ranges_1.default.transientSegments(polymer_1.getPolymerRanges(unit), unit.elements);
        this.residueIt = int_1.Segmentation.transientSegments(this.residueAtomSegments, unit.elements);
        this.value = createPolymerTraceElement(unit);
        this.hasNext = this.residueIt.hasNext && this.polymerIt.hasNext;
        this.secondaryStructureType = unit.model.properties.secondaryStructure.type;
        this.secondaryStructureGetIndex = unit.model.properties.secondaryStructure.getIndex;
        var computedSecondaryStructure = secondary_structure_1.ComputedSecondaryStructure.get(structure);
        if (computedSecondaryStructure) {
            var secondaryStructure = computedSecondaryStructure.map.get(unit.invariantId);
            if (secondaryStructure) {
                this.secondaryStructureType = secondaryStructure.type;
                this.secondaryStructureGetIndex = secondaryStructure.getIndex;
            }
        }
    }
    AtomicPolymerTraceIterator.prototype.pos = function (target, index) {
        if (index !== -1) {
            target[0] = this.atomicConformation.x[index];
            target[1] = this.atomicConformation.y[index];
            target[2] = this.atomicConformation.z[index];
        }
    };
    AtomicPolymerTraceIterator.prototype.updateResidueSegmentRange = function (polymerSegment) {
        var index = this.residueAtomSegments.index;
        this.residueSegmentMin = index[this.unit.elements[polymerSegment.start]];
        this.residueSegmentMax = index[this.unit.elements[polymerSegment.end - 1]];
    };
    AtomicPolymerTraceIterator.prototype.getResidueIndex = function (residueIndex) {
        if (residueIndex < this.residueSegmentMin) {
            var cyclicIndex = this.cyclicPolymerMap.get(this.residueSegmentMin);
            if (cyclicIndex !== undefined) {
                residueIndex = cyclicIndex - (this.residueSegmentMin - residueIndex - 1);
            }
            else {
                residueIndex = this.residueSegmentMin;
            }
        }
        else if (residueIndex > this.residueSegmentMax) {
            var cyclicIndex = this.cyclicPolymerMap.get(this.residueSegmentMax);
            if (cyclicIndex !== undefined) {
                residueIndex = cyclicIndex + (residueIndex - this.residueSegmentMax - 1);
            }
            else {
                residueIndex = this.residueSegmentMax;
            }
        }
        return residueIndex;
    };
    AtomicPolymerTraceIterator.prototype.getSecStruc = function (residueIndex) {
        return this.secondaryStructureType[this.secondaryStructureGetIndex(residueIndex)];
    };
    AtomicPolymerTraceIterator.prototype.setControlPoint = function (out, p1, p2, p3, residueIndex) {
        var ss = this.getSecStruc(residueIndex);
        if (types_1.SecondaryStructureType.is(ss, 4 /* Beta */)) {
            linear_algebra_1.Vec3.scale(out, linear_algebra_1.Vec3.add(out, p1, linear_algebra_1.Vec3.add(out, p3, linear_algebra_1.Vec3.add(out, p2, p2))), 1 / 4);
        }
        else {
            linear_algebra_1.Vec3.copy(out, p2);
        }
    };
    AtomicPolymerTraceIterator.prototype.setFromToVector = function (out, residueIndex) {
        this.pos(tmpVecA, this.directionFromElementIndex[residueIndex]);
        this.pos(tmpVecB, this.directionToElementIndex[residueIndex]);
        linear_algebra_1.Vec3.sub(out, tmpVecB, tmpVecA);
    };
    AtomicPolymerTraceIterator.prototype.setDirection = function (out, v1, v2, v3) {
        linear_algebra_1.Vec3.copy(tmpVecA, v1);
        linear_algebra_1.Vec3.copy(tmpVecB, v3);
        if (linear_algebra_1.Vec3.dot(v2, tmpVecA) < 0)
            linear_algebra_1.Vec3.scale(tmpVecA, tmpVecA, -1);
        if (linear_algebra_1.Vec3.dot(v2, tmpVecB) < 0)
            linear_algebra_1.Vec3.scale(tmpVecB, tmpVecB, -1);
        linear_algebra_1.Vec3.scale(out, linear_algebra_1.Vec3.add(out, tmpVecA, linear_algebra_1.Vec3.add(out, tmpVecB, linear_algebra_1.Vec3.add(out, v2, v2))), 1 / 4);
    };
    AtomicPolymerTraceIterator.prototype.move = function () {
        var _a = this, residueIt = _a.residueIt, polymerIt = _a.polymerIt, value = _a.value;
        if (this.state === 0 /* nextPolymer */) {
            while (polymerIt.hasNext) {
                this.polymerSegment = polymerIt.move();
                residueIt.setSegment(this.polymerSegment);
                this.updateResidueSegmentRange(this.polymerSegment);
                if (residueIt.hasNext) {
                    this.state = 1 /* nextResidue */;
                    this.currSecStrucType = SecStrucTypeNA;
                    this.nextSecStrucType = this.getSecStruc(this.residueSegmentMin);
                    this.currCoarseBackbone = false;
                    this.nextCoarseBackbone = this.directionFromElementIndex[this.residueSegmentMin] === -1 || this.directionToElementIndex[this.residueSegmentMin] === -1;
                    break;
                }
            }
        }
        if (this.state === 1 /* nextResidue */) {
            var residueIndex = residueIt.move().index;
            this.prevSecStrucType = this.currSecStrucType;
            this.currSecStrucType = this.nextSecStrucType;
            this.nextSecStrucType = residueIt.hasNext ? this.getSecStruc(residueIndex + 1) : SecStrucTypeNA;
            this.prevCoarseBackbone = this.currCoarseBackbone;
            this.currCoarseBackbone = this.nextCoarseBackbone;
            this.nextCoarseBackbone = residueIt.hasNext ? (this.directionFromElementIndex[residueIndex + 1] === -1 || this.directionToElementIndex[residueIndex + 1] === -1) : false;
            value.secStrucType = this.currSecStrucType;
            value.secStrucFirst = this.prevSecStrucType !== this.currSecStrucType;
            value.secStrucLast = this.currSecStrucType !== this.nextSecStrucType;
            value.isCoarseBackbone = this.currCoarseBackbone;
            value.coarseBackboneFirst = this.prevCoarseBackbone !== this.currCoarseBackbone;
            value.coarseBackboneLast = this.currCoarseBackbone !== this.nextCoarseBackbone;
            value.first = residueIndex === this.residueSegmentMin;
            value.last = residueIndex === this.residueSegmentMax;
            value.moleculeType = this.moleculeType[residueIndex];
            var residueIndexPrev3 = this.getResidueIndex(residueIndex - 3);
            var residueIndexPrev2 = this.getResidueIndex(residueIndex - 2);
            var residueIndexPrev1 = this.getResidueIndex(residueIndex - 1);
            var residueIndexNext1 = this.getResidueIndex(residueIndex + 1);
            var residueIndexNext2 = this.getResidueIndex(residueIndex + 2);
            var residueIndexNext3 = this.getResidueIndex(residueIndex + 3);
            if (value.first) {
                value.centerPrev.element = this.traceElementIndex[residueIndexPrev1];
                value.center.element = this.traceElementIndex[residueIndex];
                this.pos(this.p0, this.traceElementIndex[residueIndexPrev3]);
                this.pos(this.p1, this.traceElementIndex[residueIndexPrev2]);
                this.pos(this.p2, this.traceElementIndex[residueIndexPrev1]);
                this.pos(this.p3, this.traceElementIndex[residueIndex]);
                this.pos(this.p4, this.traceElementIndex[residueIndexNext1]);
                this.pos(this.p5, this.traceElementIndex[residueIndexNext2]);
                this.setFromToVector(this.d01, residueIndexPrev1);
                this.setFromToVector(this.d12, residueIndex);
                this.setFromToVector(this.d23, residueIndexNext1);
            }
            else {
                value.centerPrev.element = value.center.element;
                value.center.element = value.centerNext.element;
                linear_algebra_1.Vec3.copy(this.p0, this.p1);
                linear_algebra_1.Vec3.copy(this.p1, this.p2);
                linear_algebra_1.Vec3.copy(this.p2, this.p3);
                linear_algebra_1.Vec3.copy(this.p3, this.p4);
                linear_algebra_1.Vec3.copy(this.p4, this.p5);
                linear_algebra_1.Vec3.copy(this.p5, this.p6);
                linear_algebra_1.Vec3.copy(this.d01, this.d12);
                linear_algebra_1.Vec3.copy(this.d12, this.d23);
                linear_algebra_1.Vec3.copy(this.d23, this.d34);
            }
            value.centerNext.element = this.traceElementIndex[residueIndexNext1];
            this.pos(this.p6, this.traceElementIndex[residueIndexNext3]);
            this.setFromToVector(this.d34, residueIndexNext2);
            value.isCoarseBackbone = this.directionFromElementIndex[residueIndex] === -1 || this.directionToElementIndex[residueIndex] === -1;
            this.setControlPoint(value.p0, this.p0, this.p1, this.p2, residueIndexPrev2);
            this.setControlPoint(value.p1, this.p1, this.p2, this.p3, residueIndexPrev1);
            this.setControlPoint(value.p2, this.p2, this.p3, this.p4, residueIndex);
            this.setControlPoint(value.p3, this.p3, this.p4, this.p5, residueIndexNext1);
            this.setControlPoint(value.p4, this.p4, this.p5, this.p6, residueIndexNext2);
            this.setDirection(value.d12, this.d01, this.d12, this.d23);
            this.setDirection(value.d23, this.d12, this.d23, this.d34);
            if (!residueIt.hasNext) {
                this.state = 0 /* nextPolymer */;
            }
        }
        this.hasNext = residueIt.hasNext || polymerIt.hasNext;
        return this.value;
    };
    return AtomicPolymerTraceIterator;
}());
exports.AtomicPolymerTraceIterator = AtomicPolymerTraceIterator;
var CoarsePolymerTraceIterator = /** @class */ (function () {
    function CoarsePolymerTraceIterator(unit, structure) {
        this.unit = unit;
        this.state = 0 /* nextPolymer */;
        this.hasNext = false;
        this.polymerIt = sorted_ranges_1.default.transientSegments(polymer_1.getPolymerRanges(unit), unit.elements);
        this.value = createPolymerTraceElement(unit);
        switch (unit.kind) {
            case 1 /* Spheres */:
                this.conformation = unit.model.coarseConformation.spheres;
                break;
            case 2 /* Gaussians */:
                this.conformation = unit.model.coarseConformation.gaussians;
                break;
        }
        this.hasNext = this.polymerIt.hasNext;
    }
    CoarsePolymerTraceIterator.prototype.getElementIndex = function (elementIndex) {
        return Math.min(Math.max(this.polymerSegment.start, elementIndex), this.polymerSegment.end - 1);
    };
    CoarsePolymerTraceIterator.prototype.pos = function (target, elementIndex) {
        var index = this.unit.elements[elementIndex];
        target[0] = this.conformation.x[index];
        target[1] = this.conformation.y[index];
        target[2] = this.conformation.z[index];
    };
    CoarsePolymerTraceIterator.prototype.move = function () {
        if (this.state === 0 /* nextPolymer */) {
            while (this.polymerIt.hasNext) {
                this.polymerSegment = this.polymerIt.move();
                this.elementIndex = this.polymerSegment.start;
                if (this.elementIndex + 1 < this.polymerSegment.end) {
                    this.state = 1 /* nextElement */;
                    break;
                }
            }
        }
        if (this.state === 1 /* nextElement */) {
            this.elementIndex += 1;
            var elementIndexPrev2 = this.getElementIndex(this.elementIndex - 2);
            var elementIndexPrev1 = this.getElementIndex(this.elementIndex - 1);
            var elementIndexNext1 = this.getElementIndex(this.elementIndex + 1);
            var elementIndexNext2 = this.getElementIndex(this.elementIndex + 2);
            this.value.centerPrev.element = this.value.center.unit.elements[elementIndexPrev1];
            this.value.center.element = this.value.center.unit.elements[this.elementIndex];
            this.value.centerNext.element = this.value.center.unit.elements[elementIndexNext1];
            this.pos(this.value.p0, elementIndexPrev2);
            this.pos(this.value.p1, elementIndexPrev1);
            this.pos(this.value.p2, this.elementIndex);
            this.pos(this.value.p3, elementIndexNext1);
            this.pos(this.value.p4, elementIndexNext2);
            this.value.first = this.elementIndex === this.polymerSegment.start;
            this.value.last = this.elementIndex === this.polymerSegment.end - 1;
            if (this.elementIndex + 1 >= this.polymerSegment.end) {
                this.state = 0 /* nextPolymer */;
            }
        }
        this.hasNext = this.elementIndex + 1 < this.polymerSegment.end || this.polymerIt.hasNext;
        return this.value;
    };
    return CoarsePolymerTraceIterator;
}());
exports.CoarsePolymerTraceIterator = CoarsePolymerTraceIterator;
