"use strict";
/**
 * Copyright (c) 2018 mol* contributors, licensed under MIT, See LICENSE file for more info.
 *
 * @author Alexander Rose <alexander.rose@weirdbyte.de>
 */
Object.defineProperty(exports, "__esModule", { value: true });
var loci_1 = require("../mol-model/loci");
var marker_action_1 = require("../mol-util/marker-action");
var linear_algebra_1 = require("../mol-math/linear-algebra");
var transform_data_1 = require("../mol-geo/geometry/transform-data");
var util_1 = require("../mol-gl/renderable/util");
var mol_util_1 = require("../mol-util");
var overpaint_data_1 = require("../mol-geo/geometry/overpaint-data");
var int_1 = require("../mol-data/int");
var transparency_data_1 = require("../mol-geo/geometry/transparency-data");
var Visual;
(function (Visual) {
    function setVisibility(renderObject, visible) {
        if (renderObject)
            renderObject.state.visible = visible;
    }
    Visual.setVisibility = setVisibility;
    function setAlphaFactor(renderObject, alphaFactor) {
        if (renderObject)
            renderObject.state.alphaFactor = alphaFactor;
    }
    Visual.setAlphaFactor = setAlphaFactor;
    function setPickable(renderObject, pickable) {
        if (renderObject)
            renderObject.state.pickable = pickable;
    }
    Visual.setPickable = setPickable;
    function mark(renderObject, loci, action, lociApply) {
        if (!renderObject)
            return false;
        var tMarker = renderObject.values.tMarker;
        function apply(interval) {
            return marker_action_1.applyMarkerAction(tMarker.ref.value.array, interval, action);
        }
        var changed = lociApply(loci, apply);
        if (changed)
            mol_util_1.ValueCell.update(tMarker, tMarker.ref.value);
        return changed;
    }
    Visual.mark = mark;
    function setOverpaint(renderObject, overpaint, lociApply, clear) {
        if (!renderObject)
            return;
        var _a = renderObject.values, tOverpaint = _a.tOverpaint, uGroupCount = _a.uGroupCount, instanceCount = _a.instanceCount;
        var count = uGroupCount.ref.value * instanceCount.ref.value;
        // ensure texture has right size
        overpaint_data_1.createOverpaint(overpaint.layers.length ? count : 0, renderObject.values);
        // clear if requested
        if (clear)
            overpaint_data_1.clearOverpaint(tOverpaint.ref.value.array, 0, count);
        var _loop_1 = function (i, il) {
            var _a = overpaint.layers[i], loci = _a.loci, color = _a.color;
            var apply = function (interval) {
                var start = int_1.Interval.start(interval);
                var end = int_1.Interval.end(interval);
                return overpaint_data_1.applyOverpaintColor(tOverpaint.ref.value.array, start, end, color, overpaint.alpha);
            };
            lociApply(loci, apply);
        };
        for (var i = 0, il = overpaint.layers.length; i < il; ++i) {
            _loop_1(i, il);
        }
        mol_util_1.ValueCell.update(tOverpaint, tOverpaint.ref.value);
    }
    Visual.setOverpaint = setOverpaint;
    function setTransparency(renderObject, transparency, lociApply, clear) {
        if (!renderObject)
            return;
        var _a = renderObject.values, tTransparency = _a.tTransparency, uGroupCount = _a.uGroupCount, instanceCount = _a.instanceCount;
        var count = uGroupCount.ref.value * instanceCount.ref.value;
        var loci = transparency.loci, value = transparency.value, variant = transparency.variant;
        // ensure texture has right size and variant
        transparency_data_1.createTransparency(value && !loci_1.isEmptyLoci(loci) ? count : 0, variant, renderObject.values);
        // clear if requested
        if (clear)
            transparency_data_1.clearTransparency(tTransparency.ref.value.array, 0, count);
        var apply = function (interval) {
            var start = int_1.Interval.start(interval);
            var end = int_1.Interval.end(interval);
            return transparency_data_1.applyTransparencyValue(tTransparency.ref.value.array, start, end, value);
        };
        lociApply(loci, apply);
        mol_util_1.ValueCell.update(tTransparency, tTransparency.ref.value);
    }
    Visual.setTransparency = setTransparency;
    function setTransform(renderObject, transform, instanceTransforms) {
        if (!renderObject || (!transform && !instanceTransforms))
            return;
        var values = renderObject.values;
        if (transform) {
            linear_algebra_1.Mat4.copy(values.matrix.ref.value, transform);
            mol_util_1.ValueCell.update(values.matrix, values.matrix.ref.value);
        }
        if (instanceTransforms) {
            values.extraTransform.ref.value.set(instanceTransforms);
            mol_util_1.ValueCell.update(values.extraTransform, values.extraTransform.ref.value);
        }
        else if (instanceTransforms === null) {
            transform_data_1.fillIdentityTransform(values.extraTransform.ref.value, values.instanceCount.ref.value);
            mol_util_1.ValueCell.update(values.extraTransform, values.extraTransform.ref.value);
        }
        transform_data_1.updateTransformData(values);
        var boundingSphere = util_1.calculateTransformBoundingSphere(values.invariantBoundingSphere.ref.value, values.aTransform.ref.value, values.instanceCount.ref.value);
        mol_util_1.ValueCell.update(values.boundingSphere, boundingSphere);
    }
    Visual.setTransform = setTransform;
})(Visual || (Visual = {}));
exports.Visual = Visual;
