/**
 * Copyright (c) 2018-2019 mol* contributors, licensed under MIT, See LICENSE file for more info.
 *
 * @author David Sehnal <david.sehnal@gmail.com>
 */
import { StateObject, StateObjectCell } from './object';
import { StateTree } from './tree';
import { StateTransform } from './transform';
import { StateTransformer } from './transformer';
import { Task } from '../mol-task';
import { StateSelection } from './state/selection';
import { StateBuilder } from './state/builder';
import { StateAction } from './action';
import { StateActionManager } from './action/manager';
import { LogEntry } from '../mol-util/log-entry';
export { State };
declare class State {
    private _tree;
    protected errorFree: boolean;
    private ev;
    readonly globalContext: unknown;
    readonly events: {
        cell: {
            stateUpdated: import("rxjs").Subject<State.ObjectEvent & {
                cell: StateObjectCell<StateObject<any, StateObject.Type<any>>, StateTransform<StateTransformer<any, StateObject<any, StateObject.Type<any>>, any>>>;
            }>;
            created: import("rxjs").Subject<State.ObjectEvent & {
                cell: StateObjectCell<StateObject<any, StateObject.Type<any>>, StateTransform<StateTransformer<any, StateObject<any, StateObject.Type<any>>, any>>>;
            }>;
            removed: import("rxjs").Subject<State.ObjectEvent & {
                parent: string;
            }>;
        };
        object: {
            updated: import("rxjs").Subject<State.ObjectEvent & {
                action: "in-place" | "recreate";
                obj: StateObject<any, StateObject.Type<any>>;
                oldObj?: StateObject<any, StateObject.Type<any>> | undefined;
            }>;
            created: import("rxjs").Subject<State.ObjectEvent & {
                obj: StateObject<any, StateObject.Type<any>>;
            }>;
            removed: import("rxjs").Subject<State.ObjectEvent & {
                obj?: StateObject<any, StateObject.Type<any>> | undefined;
            }>;
        };
        log: import("rxjs").Subject<LogEntry>;
        changed: import("rxjs").Subject<void>;
        isUpdating: import("rxjs").Subject<boolean>;
    };
    readonly behaviors: {
        currentObject: import("rxjs").BehaviorSubject<State.ObjectEvent>;
    };
    readonly actions: StateActionManager;
    readonly tree: StateTree;
    readonly transforms: StateTree.Transforms;
    readonly current: string;
    build(): StateBuilder.Root;
    readonly cells: State.Cells;
    private spine;
    getSnapshot(): State.Snapshot;
    setSnapshot(snapshot: State.Snapshot): Task<void>;
    setCurrent(ref: StateTransform.Ref): void;
    updateCellState(ref: StateTransform.Ref, stateOrProvider: ((old: StateTransform.State) => Partial<StateTransform.State>) | Partial<StateTransform.State>): void;
    dispose(): void;
    /**
     * Select Cells using the provided selector.
     * @example state.query(StateSelection.Generators.byRef('test').ancestorOfType([type]))
     * @example state.query('test')
     */
    select<C extends StateObjectCell>(selector: StateSelection.Selector<C>): C[];
    /**
     * Select Cells by building a query generated on the fly.
     * @example state.select(q => q.byRef('test').subtree())
     */
    selectQ<C extends StateObjectCell>(selector: (q: typeof StateSelection.Generators) => StateSelection.Selector<C>): C[];
    /**
     * Creates a Task that applies the specified StateAction (i.e. must use run* on the result)
     * If no ref is specified, apply to root.
     */
    applyAction<A extends StateAction>(action: A, params: StateAction.Params<A>, ref?: StateTransform.Ref): Task<void>;
    /**
     * Queues up a reconciliation of the existing state tree.
     *
     * If the tree is StateBuilder.To<T>, the corresponding StateObject is returned by the task.
     * @param tree Tree instance or a tree builder instance
     * @param doNotReportTiming Indicates whether to log timing of the individual transforms
     */
    updateTree<T extends StateObject>(tree: StateBuilder.To<T>, options?: Partial<State.UpdateOptions>): Task<T>;
    updateTree(tree: StateTree | StateBuilder, options?: Partial<State.UpdateOptions>): Task<void>;
    private updateQueue;
    private _updateTree;
    private updateTreeAndCreateCtx;
    constructor(rootObject: StateObject, params?: {
        globalContext?: unknown;
        rootState?: StateTransform.State;
    });
}
declare namespace State {
    type Cells = ReadonlyMap<StateTransform.Ref, StateObjectCell>;
    type Tree = StateTree;
    type Builder = StateBuilder;
    interface ObjectEvent {
        state: State;
        ref: Ref;
    }
    interface Snapshot {
        readonly tree: StateTree.Serialized;
    }
    interface UpdateOptions {
        doNotLogTiming: boolean;
        doNotUpdateCurrent: boolean;
    }
    function create(rootObject: StateObject, params?: {
        globalContext?: unknown;
        rootState?: StateTransform.State;
    }): State;
}
declare type Ref = StateTransform.Ref;
