"use strict";
/**
 * Copyright (c) 2019 mol* contributors, licensed under MIT, See LICENSE file for more info.
 *
 * @author Alexander Rose <alexander.rose@weirdbyte.de>
 */
var __assign = (this && this.__assign) || function () {
    __assign = Object.assign || function(t) {
        for (var s, i = 1, n = arguments.length; i < n; i++) {
            s = arguments[i];
            for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p))
                t[p] = s[p];
        }
        return t;
    };
    return __assign.apply(this, arguments);
};
Object.defineProperty(exports, "__esModule", { value: true });
var structure_1 = require("../../mol-model/structure");
var color_1 = require("../../mol-util/color");
var param_definition_1 = require("../../mol-util/param-definition");
var util_1 = require("./util");
var DefaultColor = color_1.Color(0xCCCCCC);
var Description = 'Gives ranges of a polymer chain a color based on the entity source it originates from. Genes get the same color per entity';
exports.EntitySourceColorThemeParams = __assign({}, util_1.getPaletteParams({ scaleList: 'RedYellowBlue' }));
function getEntitySourceColorThemeParams(ctx) {
    return exports.EntitySourceColorThemeParams; // TODO return copy
}
exports.getEntitySourceColorThemeParams = getEntitySourceColorThemeParams;
function modelEntityKey(modelIndex, entityId) {
    return modelIndex + "|" + entityId;
}
function srcKey(modelIndex, entityId, srcId, gene) {
    return modelIndex + "|" + entityId + "|" + (gene ? gene : srcId);
}
function addSrc(seqToSrcByModelEntity, srcKeySerialMap, modelIndex, model, entity_src, gene_src_gene) {
    var entity_id = entity_src.entity_id, pdbx_src_id = entity_src.pdbx_src_id, pdbx_beg_seq_num = entity_src.pdbx_beg_seq_num, pdbx_end_seq_num = entity_src.pdbx_end_seq_num;
    for (var j = 0, jl = entity_src._rowCount; j < jl; ++j) {
        var entityId = entity_id.value(j);
        var mK = modelEntityKey(modelIndex, entityId);
        var seqToSrc = void 0;
        if (!seqToSrcByModelEntity.has(mK)) {
            var entityIndex = model.entities.getEntityIndex(entityId);
            var seq = model.sequence.sequences[entityIndex].sequence;
            seqToSrc = new Int16Array(seq.sequence.length);
            seqToSrcByModelEntity.set(mK, seqToSrc);
        }
        else {
            seqToSrc = seqToSrcByModelEntity.get(mK);
        }
        var sK = srcKey(modelIndex, entityId, pdbx_src_id.value(j), gene_src_gene ? gene_src_gene.value(j).join(',') : '');
        // may not be given (= 0) indicating src is for the whole seq
        var beg = pdbx_beg_seq_num.valueKind(j) === 0 /* Present */ ? pdbx_beg_seq_num.value(j) : 1;
        var end = pdbx_end_seq_num.valueKind(j) === 0 /* Present */ ? pdbx_end_seq_num.value(j) : seqToSrc.length;
        var srcIndex = void 0; // serial no starting from 1
        if (srcKeySerialMap.has(sK)) {
            srcIndex = srcKeySerialMap.get(sK);
        }
        else {
            srcIndex = srcKeySerialMap.size + 1;
            srcKeySerialMap.set(sK, srcIndex);
        }
        // set src index
        for (var i = beg, il = end; i <= il; ++i) {
            seqToSrc[i - 1] = srcIndex;
        }
    }
}
function EntitySourceColorTheme(ctx, props) {
    var color;
    var legend;
    var structure = ctx.structure;
    if (structure) {
        var l_1 = structure_1.StructureElement.create();
        var models = structure.models;
        var seqToSrcByModelEntity_1 = new Map();
        var srcKeySerialMap = new Map(); // serial no starting from 1
        for (var i = 0, il = models.length; i < il; ++i) {
            var m = models[i];
            if (m.sourceData.kind !== 'mmCIF')
                continue;
            var _a = m.sourceData.data, entity_src_gen = _a.entity_src_gen, entity_src_nat = _a.entity_src_nat, pdbx_entity_src_syn = _a.pdbx_entity_src_syn;
            addSrc(seqToSrcByModelEntity_1, srcKeySerialMap, i, m, entity_src_gen, entity_src_gen.pdbx_gene_src_gene);
            addSrc(seqToSrcByModelEntity_1, srcKeySerialMap, i, m, entity_src_nat);
            addSrc(seqToSrcByModelEntity_1, srcKeySerialMap, i, m, pdbx_entity_src_syn);
        }
        var palette_1 = util_1.getPalette(srcKeySerialMap.size + 1, props);
        legend = palette_1.legend;
        var getSrcColor_1 = function (location) {
            var modelIndex = structure.models.indexOf(location.unit.model);
            var entityId = structure_1.StructureProperties.entity.id(location);
            var mK = modelEntityKey(modelIndex, entityId);
            var seqToSrc = seqToSrcByModelEntity_1.get(mK);
            if (seqToSrc) {
                // minus 1 to convert seqId to array index
                return palette_1.color(seqToSrc[structure_1.StructureProperties.residue.label_seq_id(location) - 1]);
            }
            else {
                return DefaultColor;
            }
        };
        color = function (location) {
            if (structure_1.StructureElement.isLocation(location)) {
                return getSrcColor_1(location);
            }
            else if (structure_1.Link.isLocation(location)) {
                l_1.unit = location.aUnit;
                l_1.element = location.aUnit.elements[location.aIndex];
                return getSrcColor_1(l_1);
            }
            return DefaultColor;
        };
    }
    else {
        color = function () { return DefaultColor; };
    }
    return {
        factory: EntitySourceColorTheme,
        granularity: 'group',
        color: color,
        props: props,
        description: Description,
        legend: legend
    };
}
exports.EntitySourceColorTheme = EntitySourceColorTheme;
exports.EntitySourceColorThemeProvider = {
    label: 'Entity Source',
    factory: EntitySourceColorTheme,
    getParams: getEntitySourceColorThemeParams,
    defaultValues: param_definition_1.ParamDefinition.getDefaultValues(exports.EntitySourceColorThemeParams),
    isApplicable: function (ctx) { return !!ctx.structure; }
};
