"use strict";
/**
 * Copyright (c) 2018-2019 mol* contributors, licensed under MIT, See LICENSE file for more info.
 *
 * @author Alexander Rose <alexander.rose@weirdbyte.de>
 */
Object.defineProperty(exports, "__esModule", { value: true });
var color_1 = require("../../mol-util/color");
var structure_1 = require("../../mol-model/structure");
var types_1 = require("../../mol-model/structure/model/types");
var util_1 = require("../../mol-model/structure/util");
var param_definition_1 = require("../../mol-util/param-definition");
var tables_1 = require("../../mol-util/color/tables");
var secondary_structure_1 = require("../../mol-model-props/computed/secondary-structure");
var color_2 = require("../../mol-util/color/color");
// from Jmol http://jmol.sourceforge.net/jscolors/ (shapely)
var SecondaryStructureColors = color_1.ColorMap({
    'alphaHelix': 0xFF0080,
    'threeTenHelix': 0xA00080,
    'piHelix': 0x600080,
    'betaTurn': 0x6080FF,
    'betaStrand': 0xFFC800,
    'coil': 0xFFFFFF,
    'bend': 0x66D8C9 /* biting original color used 0x00FF00 */,
    'turn': 0x00B266,
    'dna': 0xAE00FE,
    'rna': 0xFD0162,
    'carbohydrate': 0xA6A6FA
});
var DefaultSecondaryStructureColor = color_1.Color(0x808080);
var Description = 'Assigns a color based on the type of secondary structure and basic molecule type.';
exports.SecondaryStructureColorThemeParams = {
    saturation: param_definition_1.ParamDefinition.Numeric(-1, { min: -6, max: 6, step: 0.1 }),
    lightness: param_definition_1.ParamDefinition.Numeric(0, { min: -6, max: 6, step: 0.1 })
};
function getSecondaryStructureColorThemeParams(ctx) {
    return exports.SecondaryStructureColorThemeParams; // TODO return copy
}
exports.getSecondaryStructureColorThemeParams = getSecondaryStructureColorThemeParams;
function secondaryStructureColor(colorMap, unit, element, computedSecondaryStructure) {
    var secStrucType = types_1.SecondaryStructureType.create(0 /* None */);
    if (structure_1.Unit.isAtomic(unit)) {
        secStrucType = unit.model.properties.secondaryStructure.type[unit.residueIndex[element]];
        if (computedSecondaryStructure) {
            var secondaryStructure = computedSecondaryStructure.map.get(unit.invariantId);
            if (secondaryStructure)
                secStrucType = secondaryStructure.type[secondaryStructure.getIndex(unit.residueIndex[element])];
        }
    }
    if (types_1.SecondaryStructureType.is(secStrucType, 2 /* Helix */)) {
        if (types_1.SecondaryStructureType.is(secStrucType, 2048 /* Helix3Ten */)) {
            return colorMap.threeTenHelix;
        }
        else if (types_1.SecondaryStructureType.is(secStrucType, 32768 /* HelixPi */)) {
            return colorMap.piHelix;
        }
        return colorMap.alphaHelix;
    }
    else if (types_1.SecondaryStructureType.is(secStrucType, 4 /* Beta */)) {
        return colorMap.betaStrand;
    }
    else if (types_1.SecondaryStructureType.is(secStrucType, 8 /* Bend */)) {
        return colorMap.bend;
    }
    else if (types_1.SecondaryStructureType.is(secStrucType, 16 /* Turn */)) {
        return colorMap.turn;
    }
    else {
        var moleculeType = util_1.getElementMoleculeType(unit, element);
        if (moleculeType === 6 /* DNA */) {
            return colorMap.dna;
        }
        else if (moleculeType === 5 /* RNA */) {
            return colorMap.rna;
        }
        else if (moleculeType === 8 /* saccharide */) {
            return colorMap.carbohydrate;
        }
        else if (moleculeType === 4 /* protein */) {
            return colorMap.coil;
        }
    }
    return DefaultSecondaryStructureColor;
}
exports.secondaryStructureColor = secondaryStructureColor;
function SecondaryStructureColorTheme(ctx, props) {
    var computedSecondaryStructure = ctx.structure && secondary_structure_1.ComputedSecondaryStructure.get(ctx.structure);
    // use `computedSecondaryStructure.id` as context hash, when available
    var contextHash = computedSecondaryStructure && computedSecondaryStructure.id;
    var colorMap = color_2.getAdjustedColorMap(SecondaryStructureColors, props.saturation, props.lightness);
    function color(location) {
        if (structure_1.StructureElement.isLocation(location)) {
            return secondaryStructureColor(colorMap, location.unit, location.element, computedSecondaryStructure);
        }
        else if (structure_1.Link.isLocation(location)) {
            return secondaryStructureColor(colorMap, location.aUnit, location.aUnit.elements[location.aIndex], computedSecondaryStructure);
        }
        return DefaultSecondaryStructureColor;
    }
    return {
        factory: SecondaryStructureColorTheme,
        granularity: 'group',
        color: color,
        props: props,
        contextHash: contextHash,
        description: Description,
        legend: tables_1.TableLegend(Object.keys(SecondaryStructureColors).map(function (name) {
            return [name, SecondaryStructureColors[name]];
        }).concat([['Other', DefaultSecondaryStructureColor]]))
    };
}
exports.SecondaryStructureColorTheme = SecondaryStructureColorTheme;
exports.SecondaryStructureColorThemeProvider = {
    label: 'Secondary Structure',
    factory: SecondaryStructureColorTheme,
    getParams: getSecondaryStructureColorThemeParams,
    defaultValues: param_definition_1.ParamDefinition.getDefaultValues(exports.SecondaryStructureColorThemeParams),
    isApplicable: function (ctx) { return !!ctx.structure; }
};
