"use strict";
/**
 * Copyright (c) 2019 mol* contributors, licensed under MIT, See LICENSE file for more info.
 *
 * @author Alexander Rose <alexander.rose@weirdbyte.de>
 */
Object.defineProperty(exports, "__esModule", { value: true });
var color_1 = require("../../mol-util/color");
var structure_1 = require("../../mol-model/structure");
var param_definition_1 = require("../../mol-util/param-definition");
var scale_1 = require("../../mol-util/color/scale");
var DefaultUncertaintyColor = color_1.Color(0xffff99);
var Description = "Assigns a color based on the uncertainty of an element's position, , e.g. B-factor or RMSF, depending on the data availability and experimental technique.";
exports.UncertaintyColorThemeParams = {
    domain: param_definition_1.ParamDefinition.Interval([0, 100]),
    list: param_definition_1.ParamDefinition.ColorScale('RedWhiteBlue', scale_1.ColorListOptions),
};
function getUncertaintyColorThemeParams(ctx) {
    return exports.UncertaintyColorThemeParams; // TODO return copy
}
exports.getUncertaintyColorThemeParams = getUncertaintyColorThemeParams;
function getUncertainty(unit, element) {
    if (structure_1.Unit.isAtomic(unit)) {
        return unit.model.atomicConformation.B_iso_or_equiv.value(element);
    }
    else if (structure_1.Unit.isSpheres(unit)) {
        return unit.model.coarseConformation.spheres.rmsf[element];
    }
    else {
        return 0;
    }
}
exports.getUncertainty = getUncertainty;
function UncertaintyColorTheme(ctx, props) {
    var scale = color_1.ColorScale.create({
        reverse: true,
        domain: props.domain,
        listOrName: props.list,
    });
    // TODO calc domain based on data, set min/max as 10/90 percentile to be robust against outliers
    function color(location) {
        if (structure_1.StructureElement.isLocation(location)) {
            return scale.color(getUncertainty(location.unit, location.element));
        }
        else if (structure_1.Link.isLocation(location)) {
            return scale.color(getUncertainty(location.aUnit, location.aUnit.elements[location.aIndex]));
        }
        return DefaultUncertaintyColor;
    }
    return {
        factory: UncertaintyColorTheme,
        granularity: 'group',
        color: color,
        props: props,
        description: Description,
        legend: scale ? scale.legend : undefined
    };
}
exports.UncertaintyColorTheme = UncertaintyColorTheme;
exports.UncertaintyColorThemeProvider = {
    label: 'Uncertainty',
    factory: UncertaintyColorTheme,
    getParams: getUncertaintyColorThemeParams,
    defaultValues: param_definition_1.ParamDefinition.getDefaultValues(exports.UncertaintyColorThemeParams),
    isApplicable: function (ctx) { return !!ctx.structure; }
};
