/**
 * Copyright (c) 2018 mol* contributors, licensed under MIT, See LICENSE file for more info.
 *
 * @author Alexander Rose <alexander.rose@weirdbyte.de>
 * @author David Sehnal <david.sehnal@gmail.com>
 */
import { Color as ColorData } from './color';
import { Vec2 as Vec2Data, Vec3 as Vec3Data } from '../mol-math/linear-algebra';
export declare namespace ParamDefinition {
    interface Info {
        label?: string;
        description?: string;
        fieldLabels?: {
            [name: string]: string;
        };
        isHidden?: boolean;
    }
    interface Base<T> extends Info {
        isOptional?: boolean;
        defaultValue: T;
    }
    interface Optional<T extends Any = Any> extends Base<T['defaultValue'] | undefined> {
        type: T['type'];
    }
    function Optional<T>(p: Base<T>): Base<T | undefined>;
    interface Value<T> extends Base<T> {
        type: 'value';
    }
    function Value<T>(defaultValue: T, info?: Info): Value<T>;
    interface Select<T extends string | number> extends Base<T> {
        type: 'select';
        /** array of (value, label) tuples */
        options: [T, string][];
    }
    function Select<T extends string | number>(defaultValue: T, options: [T, string][], info?: Info): Select<T>;
    interface ColorScale<T extends string> extends Base<T> {
        type: 'color-scale';
        /** array of (value, label) tuples */
        options: [T, string][];
    }
    function ColorScale<T extends string>(defaultValue: T, options: [T, string][], info?: Info): ColorScale<T>;
    interface MultiSelect<E extends string, T = E[]> extends Base<T> {
        type: 'multi-select';
        /** array of (value, label) tuples */
        options: [E, string][];
    }
    function MultiSelect<E extends string, T = E[]>(defaultValue: T, options: [E, string][], info?: Info): MultiSelect<E, T>;
    interface BooleanParam extends Base<boolean> {
        type: 'boolean';
    }
    function Boolean(defaultValue: boolean, info?: Info): BooleanParam;
    interface Text<T extends string = string> extends Base<T> {
        type: 'text';
    }
    function Text<T extends string = string>(defaultValue?: string, info?: Info): Text<T>;
    interface Color extends Base<ColorData> {
        type: 'color';
    }
    function Color(defaultValue: ColorData, info?: Info): Color;
    interface Vec3 extends Base<Vec3Data> {
        type: 'vec3';
    }
    function Vec3(defaultValue: Vec3Data, info?: Info): Vec3;
    interface FileParam extends Base<File> {
        type: 'file';
        accept?: string;
    }
    function File(info?: Info & {
        accept?: string;
    }): FileParam;
    interface Range {
        /** If given treat as a range. */
        min?: number;
        /** If given treat as a range. */
        max?: number;
        /**
         * If given treat as a range.
         * If an `integer` parse value with parseInt, otherwise use parseFloat.
         */
        step?: number;
    }
    interface Numeric extends Base<number>, Range {
        type: 'number';
    }
    function Numeric(defaultValue: number, range?: {
        min?: number;
        max?: number;
        step?: number;
    }, info?: Info): Numeric;
    interface Interval extends Base<[number, number]>, Range {
        type: 'interval';
    }
    function Interval(defaultValue: [number, number], range?: {
        min?: number;
        max?: number;
        step?: number;
    }, info?: Info): Interval;
    interface LineGraph extends Base<Vec2Data[]> {
        type: 'line-graph';
    }
    function LineGraph(defaultValue: Vec2Data[], info?: Info): LineGraph;
    interface Group<T> extends Base<T> {
        type: 'group';
        params: Params;
        isExpanded?: boolean;
        isFlat?: boolean;
    }
    function Group<T>(params: For<T>, info?: Info & {
        isExpanded?: boolean;
        isFlat?: boolean;
    }): Group<Normalize<T>>;
    interface NamedParams<T = any, K = string> {
        name: K;
        params: T;
    }
    type NamedParamUnion<P extends Params, K extends keyof P = keyof P> = K extends any ? NamedParams<P[K]['defaultValue'], K> : never;
    interface Mapped<T extends NamedParams<any, any>> extends Base<T> {
        type: 'mapped';
        select: Select<string>;
        map(name: string): Any;
    }
    function Mapped<T>(defaultKey: string, names: [string, string][], map: (name: string) => Any, info?: Info): Mapped<NamedParams<T>>;
    function MappedStatic<C extends Params>(defaultKey: keyof C, map: C, info?: Info & {
        options?: [keyof C, string][];
    }): Mapped<NamedParamUnion<C>>;
    interface ObjectList<T = any> extends Base<T[]> {
        type: 'object-list';
        element: Params;
        ctor(): T;
        getLabel(t: T): string;
    }
    function ObjectList<T>(element: For<T>, getLabel: (e: T) => string, info?: Info & {
        defaultValue?: T[];
        ctor?: () => T;
    }): ObjectList<Normalize<T>>;
    interface Converted<T, C> extends Base<T> {
        type: 'converted';
        converted: Any;
        /** converts from prop value to display value */
        fromValue(v: T): C;
        /** converts from display value to prop value */
        toValue(v: C): T;
    }
    function Converted<T, C extends Any>(fromValue: (v: T) => C['defaultValue'], toValue: (v: C['defaultValue']) => T, converted: C): Converted<T, C['defaultValue']>;
    interface Conditioned<T, P extends Base<T>, C = {
        [k: string]: P;
    }> extends Base<T> {
        type: 'conditioned';
        select: Select<string>;
        conditionParams: C;
        conditionForValue(v: T): keyof C;
        conditionedValue(v: T, condition: keyof C): T;
    }
    function Conditioned<T, P extends Base<T>, C = {
        [k: string]: P;
    }>(defaultValue: T, conditionParams: C, conditionForValue: (v: T) => keyof C, conditionedValue: (v: T, condition: keyof C) => T): Conditioned<T, P, C>;
    interface ScriptExpression extends Base<{
        language: 'mol-script';
        expression: string;
    }> {
        type: 'script-expression';
    }
    function ScriptExpression(defaultValue: ScriptExpression['defaultValue'], info?: Info): ScriptExpression;
    type Any = Value<any> | Select<any> | MultiSelect<any> | BooleanParam | Text | Color | Vec3 | Numeric | FileParam | Interval | LineGraph | ColorScale<any> | Group<any> | Mapped<any> | Converted<any, any> | Conditioned<any, any, any> | ScriptExpression | ObjectList;
    type Params = {
        [k: string]: Any;
    };
    type Values<T extends Params> = {
        [k in keyof T]: T[k]['defaultValue'];
    };
    type Optionals<P> = {
        [K in keyof P]-?: undefined extends P[K] ? K : never;
    }[keyof P];
    type NonOptionals<P> = {
        [K in keyof P]-?: undefined extends P[K] ? never : K;
    }[keyof P];
    type Normalize<P> = Pick<P, NonOptionals<P>> & Partial<Pick<P, Optionals<P>>>;
    type For<P> = {
        [K in keyof P]-?: Base<P[K]>;
    };
    function getDefaultValues<T extends Params>(params: T): Values<T>;
    function clone<P extends Params>(params: P): P;
    /**
     * List of [error text, pathToValue]
     * i.e. ['Missing Nested Id', ['group1', 'id']]
     */
    type ParamErrors = [string, string | string[]][];
    function validate(params: Params, values: any): ParamErrors | undefined;
    function areEqual(params: Params, a: any, b: any): boolean;
    function isParamEqual(p: Any, a: any, b: any): boolean;
}
