"use strict";
/**
 * Copyright (c) 2018 mol* contributors, licensed under MIT, See LICENSE file for more info.
 *
 * @author David Sehnal <david.sehnal@gmail.com>
 */
var __assign = (this && this.__assign) || function () {
    __assign = Object.assign || function(t) {
        for (var s, i = 1, n = arguments.length; i < n; i++) {
            s = arguments[i];
            for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p))
                t[p] = s[p];
        }
        return t;
    };
    return __assign.apply(this, arguments);
};
Object.defineProperty(exports, "__esModule", { value: true });
var structure_1 = require("../../../mol-model/structure");
var atoms_1 = require("../query/atoms");
var QueryParamType;
(function (QueryParamType) {
    QueryParamType[QueryParamType["JSON"] = 0] = "JSON";
    QueryParamType[QueryParamType["String"] = 1] = "String";
    QueryParamType[QueryParamType["Integer"] = 2] = "Integer";
    QueryParamType[QueryParamType["Float"] = 3] = "Float";
})(QueryParamType = exports.QueryParamType || (exports.QueryParamType = {}));
var AtomSiteTestParams = {
    name: 'atom_site',
    type: QueryParamType.JSON,
    description: 'Object or array of objects describing atom properties. Names are same as in wwPDB mmCIF dictionary of the atom_site category.',
    exampleValues: [{ label_comp_id: 'ALA' }, { label_seq_id: 123, label_asym_id: 'A' }]
};
var RadiusParam = {
    name: 'radius',
    type: QueryParamType.Float,
    defaultValue: 5,
    exampleValues: [5],
    description: 'Value in Angstroms.',
    validation: function (v) {
        if (v < 1 || v > 10) {
            throw "Invalid radius for residue interaction query (must be a value between 1 and 10).";
        }
    }
};
function Q(definition) {
    return definition;
}
var QueryMap = {
    'full': Q({ niceName: 'Full Structure', query: function () { return structure_1.Queries.generators.all; }, description: 'The full structure.' }),
    'atoms': Q({
        niceName: 'Atoms',
        description: 'Atoms satisfying the given criteria.',
        query: function (p) { return structure_1.Queries.combinators.merge(atoms_1.getAtomsTests(p.atom_site).map(function (test) { return structure_1.Queries.generators.atoms(test); })); },
        params: [AtomSiteTestParams]
    }),
    'symmetryMates': Q({
        niceName: 'Symmetry Mates',
        description: 'Computes crystal symmetry mates within the specified radius.',
        query: function () { return structure_1.Queries.generators.all; },
        structureTransform: function (p, s) {
            return structure_1.StructureSymmetry.builderSymmetryMates(s, p.radius).run();
        },
        params: [RadiusParam]
    }),
    'assembly': Q({
        niceName: 'Assembly',
        description: 'Computes structural assembly.',
        query: function () { return structure_1.Queries.generators.all; },
        structureTransform: function (p, s) {
            return structure_1.StructureSymmetry.buildAssembly(s, '' + (p.name || '1')).run();
        },
        params: [{
                name: 'name',
                type: QueryParamType.String,
                defaultValue: '1',
                exampleValues: ['1'],
                description: 'Assembly name.'
            }]
    }),
    'residueInteraction': Q({
        niceName: 'Residue Interaction',
        description: 'Identifies all residues within the given radius from the source residue. Takes crystal symmetry into account.',
        query: function (p) {
            var tests = atoms_1.getAtomsTests(p.atom_site);
            var center = structure_1.Queries.combinators.merge(tests.map(function (test) { return structure_1.Queries.generators.atoms(__assign({}, test, { entityTest: test.entityTest
                    ? function (ctx) { return test.entityTest(ctx) && ctx.element.unit.conformation.operator.isIdentity; }
                    : function (ctx) { return ctx.element.unit.conformation.operator.isIdentity; } })); }));
            return structure_1.Queries.modifiers.includeSurroundings(center, { radius: p.radius, wholeResidues: true });
        },
        structureTransform: function (p, s) {
            return structure_1.StructureSymmetry.builderSymmetryMates(s, p.radius).run();
        },
        params: [AtomSiteTestParams, RadiusParam]
    }),
    'residueSurroundings': Q({
        niceName: 'Residue Surroundings',
        description: 'Identifies all residues within the given radius from the source residue.',
        query: function (p) {
            var tests = atoms_1.getAtomsTests(p.atom_site);
            var center = structure_1.Queries.combinators.merge(tests.map(function (test) { return structure_1.Queries.generators.atoms(test); }));
            return structure_1.Queries.modifiers.includeSurroundings(center, { radius: p.radius, wholeResidues: true });
        },
        params: [AtomSiteTestParams, RadiusParam]
    })
};
function getQueryByName(name) {
    return QueryMap[name];
}
exports.getQueryByName = getQueryByName;
exports.QueryList = (function () {
    var list = [];
    for (var _i = 0, _a = Object.keys(QueryMap); _i < _a.length; _i++) {
        var k = _a[_i];
        list.push({ name: k, definition: QueryMap[k] });
    }
    list.sort(function (a, b) { return a.name < b.name ? -1 : a.name > b.name ? 1 : 0; });
    return list;
})();
// normalize the queries
(function () {
    for (var _i = 0, QueryList_1 = exports.QueryList; _i < QueryList_1.length; _i++) {
        var q = QueryList_1[_i];
        var m = q.definition;
        m.name = q.name;
        m.params = m.params || [];
    }
})();
// function _normalizeQueryParams(params: { [p: string]: string }, paramList: QueryParamInfo[]): { [p: string]: string | number | boolean } {
//     const ret: any = {};
//     for (const p of paramList) {
//         const key = p.name;
//         const value = params[key];
//         if (typeof value === 'undefined' || (typeof value !== 'undefined' && value !== null && value['length'] === 0)) {
//             if (p.required) {
//                 throw `The parameter '${key}' is required.`;
//             }
//             if (typeof p.defaultValue !== 'undefined') ret[key] = p.defaultValue;
//         } else {
//             switch (p.type) {
//                 case QueryParamType.JSON: ret[key] = JSON.parse(value); break;
//                 case QueryParamType.String: ret[key] = value; break;
//                 case QueryParamType.Integer: ret[key] = parseInt(value); break;
//                 case QueryParamType.Float: ret[key] = parseFloat(value); break;
//             }
//             if (p.validation) p.validation(ret[key]);
//         }
//     }
//     return ret;
// }
function normalizeQueryParams(query, params) {
    return params;
    // return _normalizeQueryParams(params, query.params);
}
exports.normalizeQueryParams = normalizeQueryParams;
